
---------------------------------------------------------
---------------- Auto Bundled Code Block ----------------
---------------------------------------------------------

do
    local searchers = package.searchers or package.loaders
    local origin_seacher = searchers[2]
    searchers[2] = function(path)
        local files =
        {
------------------------
-- Modules part begin --
------------------------

["insurrection.components"] = function()
--------------------
-- Module: 'insurrection.components'
--------------------
local balltze = Balltze
local engine = Engine
local blam = require "blam"
local getTag = blam.getTag
local uiWidgetDefinition = blam.uiWidgetDefinition
local unicodeStringList = blam.unicodeStringList
local isNull = blam.isNull
local core = require "insurrection.core"

---@class uiComponent
local component = {
    ---@type number
    tagId = nil,
    ---@type tag
    tag = nil,
    ---@type uiWidgetDefinition
    widgetDefinition = nil,
    ---@type uiComponentEvents
    events = {},
    ---@type boolean
    isBackgroundAnimated = false,
    ---@type boolean
    isBackgroundLooped = false,
    ---@type "generic" | "list" | "button" | "checkbox" | "slider" | "dropdown" | "text" | "image" | "spinner" | "progress"
    type = "generic"
    -- @type table<string, widgetAnimation>
    -- animations = {}
}

---@class uiComponentEvents
---@field onClick? fun(value?: string | boolean | number): boolean
---@field onFocus? function
---@field onOpen? fun(previousWidgetTag?: MetaEngineTag)
---@field onClose? fun():boolean
---@field animate? function

---@type table<number, uiComponent>
component.widgets = {}

-- TODO Make this local and port functions to component
VirtualInputValue = {}
---@type MetaEngineTag
local previousWidgetTag

function component.callbacks()
    ---@type MetaEngineTagDataUiWidgetDefinition?
    local editableWidgetTagData
    ---@type MetaEngineTag?
    local editableWidgetTagEntry
    ---@type MetaEngineTag?
    local lastFocusedWidgetTagEntry

    balltze.event.uiWidgetAccept.subscribe(function(event)
        if event.time == "before" then
            log("Accepting widget: {}", event.context.widget.definitionTagHandle.value)
            local isCanceled = false
            local instance = component.widgets[event.context.widget.definitionTagHandle.value]
            if instance then
                if instance.events.onClick then
                    isCanceled = instance.events.onClick() == false
                end
            end
            if isCanceled then
                event:cancel()
            end
        end
    end)

    ---@type BalltzeUIWidgetFocusEventCallback
    local function onWidgetFocus(event)
        if event.time == "before" then
            local definitionTagHandleValue = event.context.widget.definitionTagHandle.value
            local component = component.widgets[definitionTagHandleValue]
            if component and component.events.onFocus then
                component.events.onFocus()
            end
            local focusedWidgetTag = engine.tag.getTag(definitionTagHandleValue,
                                                       engine.tag.classes.uiWidgetDefinition)
            -- local focusedWidgetTag = engine.tag.getTag(definitionTagHandleValue)
            if focusedWidgetTag then
                lastFocusedWidgetTagEntry = focusedWidgetTag
                ---@diagnostic disable-next-line: undefined-field
                if focusedWidgetTag.data.flags1:editable() or
                    focusedWidgetTag.data.flags1:password() then
                    editableWidgetTagData = focusedWidgetTag.data
                    editableWidgetTagEntry = focusedWidgetTag
                else
                    editableWidgetTagData = nil
                    editableWidgetTagEntry = nil
                end
            end
        end
    end
    balltze.event.uiWidgetFocus.subscribe(onWidgetFocus)

    balltze.event.uiWidgetMouseButtonPress.subscribe(function(event)
        if event.time == "before" then
            local button = event.context.button:label()
            local widgetTag = engine.userInterface.findWidget(event.context.widget
                                                                  .definitionTagHandle.value)
            assert(widgetTag, "Invalid widget tag")
            if editableWidgetTagData and editableWidgetTagEntry then
                if widgetTag.definitionTagHandle.value == editableWidgetTagEntry.handle.value then
                    if button == "right" then
                        local inputString = core.getStringFromWidget(
                                                editableWidgetTagEntry.handle.value)
                        local text = inputString .. core.getClipboard()
                        core.setStringToWidget(text, editableWidgetTagEntry.handle.value)
                        local component = component.widgets[editableWidgetTagEntry.handle.value]
                        if component and component.events.onInputText then
                            component.events.onInputText(text)
                        end
                    end
                end
            end
        end
    end)

    local function onMouseScroll(widgetTagHandle)
        local widget = engine.userInterface.findWidget(widgetTagHandle)
        if not widget then
            return
        end
        local parentWidget = widget.parentWidget
        if not parentWidget then
            return
        end
        local component = component.widgets[parentWidget.definitionTagHandle.value] --[[@as uiComponentList]]
        if component and component.type == "list" and component.onScroll then
            local mouse = core.getMouseState()
            component:scroll(mouse.scroll, true)
        end
    end
    balltze.event.frame.subscribe(function(event)
        if event.time == "before" then
            local widget = engine.userInterface.getRootWidget()
            if widget then
                if lastFocusedWidgetTagEntry then
                    local mouse = core.getMouseState()
                    if mouse.scroll ~= 0 then
                        onMouseScroll(lastFocusedWidgetTagEntry.handle.value)
                    end
                    if mouse.rightClick > 0 then
                        -- TODO BALLTZE MIGRATE
                    end
                end
            end
        end
    end)

    balltze.event.uiWidgetCreate.subscribe(function(event)
        if event.time == "after" then
            local tagHandle = event.context.definitionTagHandle.value
            local widget = engine.userInterface.findWidget(tagHandle)
            if widget then
                local widgetTag = engine.tag
                                      .getTag(tagHandle, engine.tag.classes.uiWidgetDefinition)
                assert(widgetTag, "Invalid widget tag")
                log("Opening tag: {}", widgetTag.path)
                local componentInstance = component.widgets[tagHandle]
                if componentInstance and componentInstance.events.onOpen then
                    componentInstance.events.onOpen(previousWidgetTag)
                end
                if previousWidgetTag then
                    local previousComponentInstance = component.widgets[previousWidgetTag.handle.value]
                    if previousComponentInstance and previousComponentInstance.events.onClose then
                        --previousComponentInstance.events.onClose()
                    end
                end
                if previousWidgetTag ~= widgetTag then
                    previousWidgetTag = widgetTag
                end

                if widgetTag then
                    assert(widgetTag, "Invalid widget tag")
                    local widgetTagData = widgetTag.data
                    local widgetCount = widgetTagData.childWidgets.count
                    if widgetTagData and widgetCount > 0 then
                        local optionWidget = widgetTagData.childWidgets.elements[widgetCount]
                        -- log("Option widget: {}", inspect(table.keys(optionWidget.widgetTag)))
                        local optionsWidgetTag = engine.tag.getTag(
                                                     optionWidget.widgetTag.tagHandle.value,
                                                     engine.tag.classes.uiWidgetDefinition)
                        assert(optionsWidgetTag, "Invalid options widget tag")
                        local optionsWidgetTagData = optionsWidgetTag.data
                        -- Auto focus on the first editable widget
                        if optionsWidgetTagData and optionsWidgetTagData.childWidgets[1] then
                            onWidgetFocus(optionsWidget.childWidgets[1].widgetTag)
                        end
                    end
                end
            end
        end
    end)

    -- We might be able to use this in the future to play custom sounds or something
    --balltze.event.uiWidgetSound.subscribe(function(event)
    --    if event.time == "before" then
    --        local sound = event.context.sound
    --    end
    --end)

    balltze.event.uiWidgetBack.subscribe(function(event)
        if event.time == "before" then
            log("Closing tag: {}", event.context.widget.definitionTagHandle.value)
            local widgetTagHandleValue = event.context.widget.definitionTagHandle.value
            local component = component.widgets[widgetTagHandleValue]
            if component and component.events.onClose then
                if component.events.onClose() == false then
                    event:cancel()
                end
            end
            editableWidgetTagData = nil
        end
    end)

    balltze.event.uiWidgetListTab.subscribe(function(event)
        if event.time == "before" then
            local pressedKey = event.context.tab
            log("Pressed key: {}", pressedKey)

            local listWidgetTagId = event.context.widgetList.definitionTagHandle.value
            local previousFocusedWidgetId = event.context.widgetList.focusedChild
                                                .definitionTagHandle.value
            local widgetList = blam.uiWidgetDefinition(listWidgetTagId)
            assert(widgetList, "Invalid widget list tag id")
            -- Handle component spinner scrolling
            -- if pressedKey == "dpad left" or pressedKey == "dpad right" then
            if pressedKey == Balltze.event.uiWidgetListTabTypes.tabThruChildrenNextHorizontal or
                pressedKey == Balltze.event.uiWidgetListTabTypes.tabThruChildrenPrev then
                local component = component.widgets[listWidgetTagId] --[[@as uiComponentSpinner]]
                if component and component.type == "spinner" and component.events.onScroll then
                    -- component:scroll(pressedKey == "dpad left" and -1 or 1)
                    component:scroll(pressedKey ==
                                         Balltze.event.uiWidgetListTabTypes.tabThruChildrenPrev and
                                         -1 or 1)
                    return
                end
            end

            for childIndex, child in pairs(widgetList.childWidgets) do
                if child.widgetTag == previousFocusedWidgetId then
                    local nextChildIndex
                    -- if pressedKey == "dpad up" or pressedKey == "dpad left" then
                    if pressedKey == Balltze.event.uiWidgetListTabTypes.tabThruChildrenPrev then
                        if childIndex - 1 < 1 then
                            nextChildIndex = widgetList.childWidgetsCount
                        else
                            nextChildIndex = childIndex - 1
                        end
                        -- elseif pressedKey == "dpad down" or pressedKey == "dpad right" then
                    elseif Balltze.event.uiWidgetListTabTypes.tabThruChildrenNextHorizontal or
                        Balltze.event.uiWidgetListTabTypes.tabThruChildrenNextVertical then
                        if childIndex + 1 > widgetList.childWidgetsCount then
                            nextChildIndex = 1
                        else
                            nextChildIndex = childIndex + 1
                        end
                    end
                    local widgetTagId = widgetList.childWidgets[nextChildIndex].widgetTag
                    if widgetTagId and not isNull(widgetTagId) then
                        onWidgetFocus({
                            context = {widget = {definitionTagHandle = {value = widgetTagId}}},
                            time = "before"
                        })
                    end
                end
            end
        end
    end)

    balltze.event.keyboardInput.subscribe(function(event)
        if event.time == "before" and not console_is_open() then
            local modifiers = event.context.key.modifier
            local char = event.context.key.character
            local keycode = event.context.key.keycode
            if editableWidgetTagData and editableWidgetTagEntry then
                -- engine.core.consolePrint("Editable widget tag found")
                -- engine.core.consolePrint("Char: " .. char)
                -- engine.core.consolePrint("Keycode: " .. keycode)
                -- Get pressed key from the keyboard
                local pressedKey
                if char ~= -1 then
                    pressedKey = char
                elseif keycode then
                    pressedKey = core.translateKeycode(keycode)
                end
                -- If we pressed a key, update our editable widget
                if pressedKey then
                    -- engine.core.consolePrint("Pressed key: " .. pressedKey)
                    local inputString =
                        core.getStringFromWidget(editableWidgetTagEntry.handle.value)
                    -- engine.core.consolePrint("Input string: " .. inputString)
                    local text = core.mapKeyToText(pressedKey, inputString)
                    if text then
                        -- TODO Use widget text flags from widget tag instead (add support for that in lua-blam)
                        -- if editableWidgetTagData.name:find "password" then
                        if editableWidgetTagData.name:find "password" then
                            core.setStringToWidget(text, editableWidgetTagEntry.handle.value, "*")
                        else
                            core.setStringToWidget(text, editableWidgetTagEntry.handle.value)
                        end
                        component = component.widgets[editableWidgetTagEntry.handle.value]
                        if component and component.events.onInputText then
                            component.events.onInputText(text)
                        end
                    end
                end
            end
        end
    end)

    -- harmony.set_callback("widget list tab", "OnMenuListTab")
    -- harmony.set_callback("widget mouse focus", "OnMouseFocus")
    -- harmony.set_callback("widget mouse button press", "OnMouseButtonPress")
    -- harmony.set_callback("widget close", "OnWidgetClose")
    -- harmony.set_callback("widget open", "OnWidgetOpen")
    -- harmony.set_callback("key press", "OnKeypress")
end

function component.cleanAllEditableWidgets()
    local editableWidgets = blam.findTagsList("input", blam.tagClasses.uiWidgetDefinition) or {}
    for _, widgetTag in pairs(editableWidgets) do
        -- log("Cleaning widget " .. widgetTag.path)
        local widget = blam.uiWidgetDefinition(widgetTag.id)
        assert(widget, "No widget found with tag id " .. widgetTag.id)
        local widgetStrings = blam.unicodeStringList(widget.unicodeStringListTag)
        if widgetStrings then
            local strings = widgetStrings.strings
            strings[1] = ""
            log("Cleaned widget " .. widgetTag.path)
            widgetStrings.strings = strings
        end
    end
end

---@param tagId number
---@return uiComponent
function component.new(tagId)
    local instance = setmetatable({}, {__index = component})
    instance.tagId = tagId
    instance.tag = getTag(instance.tagId) or error("Invalid tagId") --[[@as tag]]
    instance.selectedWidgetTagId = nil
    instance.widgetDefinition = uiWidgetDefinition(tagId) or error("Invalid tagId") --[[@as uiWidgetDefinition]]
    instance.events = {}
    instance.isBackgroundAnimated = false
    -- log("Created component: " .. instance.tag.path, "info")
    component.widgets[tagId] = instance
    return instance
end

---@param tagId number
---@return uiComponent
function component.get(tagId)
    return component.widgets[tagId]
end

---@param self uiComponent
function component.onFocus(self, callback)
    self.events.onFocus = callback
end

---@param self uiComponent
---@return string
function component.getText(self)
    local virtualValue = VirtualInputValue[self.tagId]
    if virtualValue then
        return virtualValue
    end
    local unicodeStrings = blam.unicodeStringList(self.widgetDefinition.unicodeStringListTag)
    if unicodeStrings then
        return unicodeStrings.strings[self.widgetDefinition.stringListIndex + 1]
    end
    error("No unicodeStringList found for widgetDefinition")
end

---@param self uiComponent
---@param text string
---@param mask? string
function component.setText(self, text, mask)
    local childUnicodeStrings
    local childWidgetDefinition
    local widgetDefinition = self.widgetDefinition
    if self.widgetDefinition.childWidgetsCount > 0 then
        local childTagId = self.widgetDefinition.childWidgets[1].widgetTag
        childWidgetDefinition = uiWidgetDefinition(childTagId) --[[@as uiWidgetDefinition]]
        childUnicodeStrings = unicodeStringList(childWidgetDefinition.unicodeStringListTag)
    end
    local unicodeStrings = unicodeStringList(self.widgetDefinition.unicodeStringListTag)
    if not (unicodeStrings and not isNull(unicodeStrings)) then
        unicodeStrings = childUnicodeStrings --[[@as unicodeStringList]]
        widgetDefinition = childWidgetDefinition --[[@as uiWidgetDefinition]]
    end
    if not (unicodeStrings and not isNull(unicodeStrings)) then
        error("No unicodeStringList found for widgetDefinition " .. self.tag.path)
    end
    local stringListIndex = widgetDefinition.stringListIndex
    local newStrings = unicodeStrings.strings
    if mask then
        VirtualInputValue[self.tagId] = text
        newStrings[stringListIndex + 1] = string.rep(mask, #text)
    else
        newStrings[stringListIndex + 1] = text
    end
    unicodeStrings.strings = newStrings
end

---@param self uiComponent
---@param callback fun(previousWidgetTag?: MetaEngineTag)
function component.onOpen(self, callback)
    self.events.onOpen = callback
end

---@param self uiComponent
---@param callback fun(): boolean?
function component.onClose(self, callback)
    self.events.onClose = callback
end

---Animate component background
---@param self uiComponent
---@param isLooped? boolean
function component.animate(self, isLooped)
    self.isBackgroundAnimated = true
    self.isBackgroundLoop = isLooped
end

function component.free()
    component.widgets = {}
end

---@param self uiComponent
---@return tag[]
function component.getChildWidgetTags(self)
    -- TODO Filter this instead of mapping
    return table.map(self.widgetDefinition.childWidgets, function(childWidget)
        if not isNull(childWidget.widgetTag) then
            local tag = getTag(childWidget.widgetTag)
            return tag
        end
        return nil
    end)
end

---@param self uiComponent
---@param name string
function component.findChildWidgetTag(self, name)
    local childWidgetTags = self:getChildWidgetTags()
    for _, childTag in pairs(childWidgetTags) do
        if childTag.path:find(name, 1, true) then
            return childTag
        end
        local widgetDefinition = uiWidgetDefinition(childTag.id)
        if widgetDefinition then
            for _, childWidget in pairs(widgetDefinition.childWidgets) do
                local tag = getTag(childWidget.widgetTag) --[[@as tag]]
                if not isNull(childWidget.widgetTag) then
                    if tag.path:find(name, 1, true) then
                        return tag
                    end
                end
            end
        end
    end
end

---@param self uiComponent
---@param name string
function component.findChildWidgetDefinition(self, name)
    local childWidgetTags = self:getChildWidgetTags()
    for _, childTag in pairs(childWidgetTags) do
        if childTag.path:find(name, 1, true) then
            return uiWidgetDefinition(childTag.id)
        end
        local widgetDefinition = uiWidgetDefinition(childTag.id)
        if widgetDefinition then
            for _, childWidget in pairs(widgetDefinition.childWidgets) do
                local tag = getTag(childWidget.widgetTag) --[[@as tag]]
                if not isNull(childWidget.widgetTag) then
                    if tag.path:find(name, 1, true) then
                        return uiWidgetDefinition(childWidget.widgetTag)
                    end
                end
            end
        end
    end
end

---@param self uiComponent
function component.getType(self)
    return self.type
end

---@param self uiComponent
---@param newWidgetTagId number
function component.replace(self, newWidgetTagId)
    core.replaceWidgetInDom(self.tagId, newWidgetTagId)
end

---@param self uiComponent
---@return uiWidgetValues?
function component.getWidgetValues(self)
    if core.getWidgetHandle(self.tagId) then
        return core.getWidgetValues(self.tagId)
    end
end

---@param self uiComponent
---@param values uiWidgetValues
function component.setWidgetValues(self, values)
    core.setWidgetValues(self.tagId, values)
end

---@param self uiComponent
function component.setBitmapIndex(self, index)
    core.setWidgetValues(self.tagId, {bitmapIndex = index - 1})
end

return component

end,

["blam"] = function()
--------------------
-- Module: 'blam'
--------------------
------------------------------------------------------------------------------
-- Blam! library for Chimera/SAPP Lua scripting
-- Sledmine, JerryBrick
-- Easier memory handle and provides standard functions for scripting
------------------------------------------------------------------------------
local cos = math.cos
local sin = math.sin
local atan = math.atan
local pi = math.pi
math.atan2 = math.atan2 or function(y, x)
    return atan(y / x) + (x < 0 and pi or 0)
end
local atan2 = math.atan2
local sqrt = math.sqrt
local fmod = math.fmod
local rad = math.rad
local deg = math.deg

local blam = {_VERSION = "1.14.0"}

------------------------------------------------------------------------------
-- Useful functions for internal usage
------------------------------------------------------------------------------

-- From legacy glue library!
--- String or number to hex
local function tohex(s, upper)
    if type(s) == "number" then
        return (upper and "%08.8X" or "%08.8x"):format(s)
    end
    if upper then
        return (s:sub(".", function(c)
            return ("%02X"):format(c:byte())
        end))
    else
        return (s:gsub(".", function(c)
            return ("%02x"):format(c:byte())
        end))
    end
end

--- Hex to binary string
local function fromhex(s)
    if #s % 2 == 1 then
        return fromhex("0" .. s)
    end
    return (s:gsub("..", function(cc)
        return string.char(tonumber(cc, 16))
    end))
end

local function split(s, sep)
    if (sep == nil or sep == "") then
        return 1
    end
    local position, array = 0, {}
    for st, sp in function()
        return string.find(s, sep, position, true)
    end do
        table.insert(array, string.sub(s, position, st - 1))
        position = sp + 1
    end
    table.insert(array, string.sub(s, position))
    return array
end

local null = 0xFFFFFFFF

--- Get if given value equals a null value in game engine terms
---@param value any
---@return boolean
function blam.isNull(value)
    if value == 0xFF or value == 0xFFFF or value == null or value == nil then
        return true
    end
    return false
end
local isNull = blam.isNull

---Return if game instance is host
---@return boolean
function blam.isGameHost()
    return server_type == "local"
end

---Return if game instance is single player
---@return boolean
function blam.isGameSinglePlayer()
    return server_type == "none"
end

---Return if the game instance is running on a dedicated server or connected as a "network client"
---@return boolean
function blam.isGameDedicated()
    return server_type == "dedicated"
end

---Return if the game instance is a SAPP server
---@return boolean
function blam.isGameSAPP()
    return register_callback or server_type == "sapp"
end

------------------------------------------------------------------------------
-- Blam! engine data
------------------------------------------------------------------------------

---@alias tagId number

-- Engine address list
local addressList = {
    tagDataHeader = 0x40440000,
    cameraType = 0x00647498, -- from giraffe
    gamePaused = 0x004ACA79,
    gameOnMenus = 0x00622058,
    joystickInput = 0x64D998, -- from aLTis
    firstPerson = 0x40000EB8, -- from aLTis
    objectTable = 0x400506B4,
    deviceGroupsTable = 0x00816110,
    widgetsInstance = 0x6B401C,
    -- syncedNetworkObjects = 0x004F7FA2
    syncedNetworkObjects = 0x006226F0, -- pointer, from Vulpes
    screenResolution = 0x637CF0,
    currentWidgetIdAddress = 0x6B401C,
    cinematicGlobals = 0x0068c83c
}

-- Server side addresses adjustment
if blam.isGameSAPP() then
    addressList.deviceGroupsTable = 0x006E1C50
    addressList.objectTable = 0x4005062C
    addressList.syncedNetworkObjects = 0x00598020 -- not pointer cause cheat engine sucks
    addressList.cinematicGlobals = 0x005f506c
end

-- Tag classes values
---@enum tagClasses
local tagClasses = {
    actorVariant = "actv",
    actor = "actr",
    antenna = "ant!",
    biped = "bipd",
    bitmap = "bitm",
    cameraTrack = "trak",
    colorTable = "colo",
    continuousDamageEffect = "cdmg",
    contrail = "cont",
    damageEffect = "jpt!",
    decal = "deca",
    detailObjectCollection = "dobc",
    deviceControl = "ctrl",
    deviceLightFixture = "lifi",
    deviceMachine = "mach",
    device = "devi",
    dialogue = "udlg",
    effect = "effe",
    equipment = "eqip",
    flag = "flag",
    fog = "fog ",
    font = "font",
    garbage = "garb",
    gbxmodel = "mod2",
    globals = "matg",
    glow = "glw!",
    grenadeHudInterface = "grhi",
    hudGlobals = "hudg",
    hudMessageText = "hmt ",
    hudNumber = "hud#",
    itemCollection = "itmc",
    item = "item",
    lensFlare = "lens",
    lightVolume = "mgs2",
    light = "ligh",
    lightning = "elec",
    materialEffects = "foot",
    meter = "metr",
    modelAnimations = "antr",
    modelCollisiionGeometry = "coll",
    model = "mode",
    multiplayerScenarioDescription = "mply",
    object = "obje",
    particleSystem = "pctl",
    particle = "part",
    physics = "phys",
    placeholder = "plac",
    pointPhysics = "pphy",
    preferencesNetworkGame = "ngpr",
    projectile = "proj",
    scenarioStructureBsp = "sbsp",
    scenario = "scnr",
    scenery = "scen",
    shaderEnvironment = "senv",
    shaderModel = "soso",
    shaderTransparentChicagoExtended = "scex",
    shaderTransparentChicago = "schi",
    shaderTransparentGeneric = "sotr",
    shaderTransparentGlass = "sgla",
    shaderTransparentMeter = "smet",
    shaderTransparentPlasma = "spla",
    shaderTransparentWater = "swat",
    shader = "shdr",
    sky = "sky ",
    soundEnvironment = "snde",
    soundLooping = "lsnd",
    soundScenery = "ssce",
    sound = "snd!",
    spheroid = "boom",
    stringList = "str#",
    tagCollection = "tagc",
    uiWidgetCollection = "Soul",
    uiWidgetDefinition = "DeLa",
    unicodeStringList = "ustr",
    unitHudInterface = "unhi",
    unit = "unit",
    vehicle = "vehi",
    virtualKeyboard = "vcky",
    weaponHudInterface = "wphi",
    weapon = "weap",
    weatherParticleSystem = "rain",
    wind = "wind"
}

-- Blam object classes values
---@enum objectClasses
local objectClasses = {
    biped = 0,
    vehicle = 1,
    weapon = 2,
    equipment = 3,
    garbage = 4,
    projectile = 5,
    scenery = 6,
    machine = 7,
    control = 8,
    lightFixture = 9,
    placeHolder = 10,
    soundScenery = 11
}

-- Camera types
---@enum cameraTypes
local cameraTypes = {
    scripted = 1, -- 22192
    firstPerson = 2, -- 30400
    devcam = 3, -- 30704
    thirdPerson = 4, -- 31952
    deadCamera = 5 -- 23776
}

-- Netgame flag classes
---@enum netgameFlagClasses
local netgameFlagClasses = {
    ctfFlag = 0,
    ctfVehicle = 1,
    ballSpawn = 2,
    raceTrack = 3,
    raceVehicle = 4,
    vegasBank = 5,
    teleportFrom = 6,
    teleportTo = 7,
    hillFlag = 8
}

-- Game type classes
---@enum gameTypeClasses
local gameTypeClasses = {
    none = 0,
    ctf = 1,
    slayer = 2,
    oddball = 3,
    koth = 4,
    race = 5,
    terminator = 6,
    stub = 7,
    ignored1 = 8,
    ignored2 = 9,
    ignored3 = 10,
    ignored4 = 11,
    allGames = 12,
    allExceptCtf = 13,
    allExceptRaceCtf = 14
}

-- Multiplayer team classes
---@enum multiplayerTeamClasses
local multiplayerTeamClasses = {red = 0, blue = 1}

-- Unit team classes
---@enum unitTeamClasses
local unitTeamClasses = {
    defaultByUnit = 0,
    player = 1,
    human = 2,
    covenant = 3,
    flood = 4,
    sentinel = 5,
    unused6 = 6,
    unused7 = 7,
    unused8 = 8,
    unused9 = 9
}

-- Object network role classes
---@enum objectNetworkRoleClasses
local objectNetworkRoleClasses = {
    master = 0,
    puppet = 1,
    locallyControlledPuppet = 2,
    localOnly = 3
}

-- Standard console colors
local consoleColors = {
    success = {1, 0.235, 0.82, 0},
    warning = {1, 0.94, 0.75, 0.098},
    error = {1, 1, 0.2, 0.2},
    unknown = {1, 0.66, 0.66, 0.66}
}

-- Offset input from the joystick game data
local joystickInputs = {
    -- No zero values also pressed time until maxmimum byte size
    button1 = 0, -- Triangle
    button2 = 1, -- Circle
    button3 = 2, -- Cross
    button4 = 3, -- Square
    leftBumper = 4,
    rightBumper = 5,
    leftTrigger = 6,
    rightTrigger = 7,
    backButton = 8,
    startButton = 9,
    leftStick = 10,
    rightStick = 11,
    rightStick2 = 12,
    -- TODO Add joys axis
    leftStickUp = 30,
    leftStickDown = 32,
    rightStickUp = 34,
    rightStickDown = 36,
    triggers = 38,
    -- Multiple values on the same offset, check dPadValues table
    dPad = 96,
    -- Non zero values
    dPadUp = 100,
    dPadDown = 104,
    dPadLeft = 106,
    dPadRight = 102,
    dPadUpRight = 101,
    dPadDownRight = 103,
    dPadUpLeft = 107,
    dPadDownLeft = 105
}

-- Values for the possible dPad values from the joystick inputs
local dPadValues = {
    noButton = 1020,
    upRight = 766,
    downRight = 768,
    upLeft = 772,
    downLeft = 770,
    left = 771,
    right = 767,
    down = 769,
    up = 765
}

local engineConstants = {defaultNetworkObjectsCount = 509}

-- Global variables

---	This is the current gametype that is running. If no gametype is running, this will be set to nil
---, possible values are: ctf, slayer, oddball, king, race.
---@type string | nil
gametype = gametype
---This is the index of the local player. This is a value between 0 and 15, this value does not
---match with player index in the server and is not instantly assigned after joining.
---@type number | nil
local_player_index = local_player_index
---This is the name of the current loaded map.
---@type string
map = map
---Return if the map has protected tags data.
---@type boolean
map_is_protected = map_is_protected
---This is the name of the script. If the script is a global script, it will be defined as the
---filename of the script. Otherwise, it will be the name of the map.
---@type string
script_name = script_name
---This is the script type, possible values are global or map.
---@type string
script_type = script_type
---@type '"none"' | '"local"' | '"dedicated"' | '"sapp"'
server_type = server_type
---Return whether or not the script is sandboxed. See Sandoboxed Scripts for more information.
---@deprecated
---@type boolean
sandboxed = sandboxed ---@diagnostic disable-line: deprecated

local backupFunctions = {}

backupFunctions.console_is_open = _G.console_is_open
backupFunctions.console_out = _G.console_out
backupFunctions.execute_script = _G.execute_script
backupFunctions.get_global = _G.get_global
-- backupFunctions.set_global = _G.set_global
backupFunctions.get_tag = _G.get_tag
backupFunctions.set_callback = _G.set_callback
backupFunctions.set_timer = _G.set_timer
backupFunctions.stop_timer = _G.stop_timer

backupFunctions.spawn_object = _G.spawn_object
backupFunctions.delete_object = _G.delete_object
backupFunctions.get_object = _G.get_object
backupFunctions.get_dynamic_player = _G.get_dynamic_player

backupFunctions.hud_message = _G.hud_message

backupFunctions.create_directory = _G.create_directory
backupFunctions.remove_directory = _G.remove_directory
backupFunctions.directory_exists = _G.directory_exists
backupFunctions.list_directory = _G.list_directory
backupFunctions.write_file = _G.write_file
backupFunctions.read_file = _G.read_file
backupFunctions.delete_file = _G.delete_file
backupFunctions.file_exists = _G.file_exists

------------------------------------------------------------------------------
-- Chimera API auto completion
-- EmmyLua autocompletion for some functions!
-- Functions below do not have a real implementation and are not supossed to be imported
------------------------------------------------------------------------------

---Attempt to spawn an object given tag class, tag path and coordinates.
---Given a tag id is also accepted.
---@overload fun(tagId: number, x: number, y: number, z: number):number
---@param tagClass tagClasses Type of the tag to spawn
---@param tagPath string Path of object to spawn
---@param x number
---@param y number
---@param z number
---@return number? objectId
function spawn_object(tagClass, tagPath, x, y, z)
end

---Attempt to get the address of a player unit object given player index, returning nil on failure.<br>
---If no argument is given, the address to the local player’s unit object is returned, instead.
---@param playerIndex? number
---@return number? objectAddress
function get_dynamic_player(playerIndex)
end

spawn_object = backupFunctions.spawn_object
get_dynamic_player = backupFunctions.get_dynamic_player

------------------------------------------------------------------------------
-- SAPP API bindings
------------------------------------------------------------------------------

---Write content to a text file given file path
---@param path string Path to the file to write
---@param content string Content to write into the file
---@return boolean, string? result True if successful otherwise nil, error
function write_file(path, content)
    local file, error = io.open(path, "w")
    if (not file) then
        return false, error
    end
    local success, err = file:write(content)
    file:close()
    if (not success) then
        os.remove(path)
        return false, err
    else
        return true
    end
end

---Read the contents from a file given file path.
---@param path string Path to the file to read
---@return boolean, string? content string if successful otherwise nil, error
function read_file(path)
    local file, error = io.open(path, "r")
    if (not file) then
        return false, error
    end
    local content, error = file:read("*a")
    if (content == nil) then
        return false, error
    end
    file:close()
    return content
end

---Attempt create a directory with the given path.
---
---An error will occur if the directory can not be created.
---@param path string Path to the directory to create
---@return boolean
function create_directory(path)
    local success, error = os.execute("mkdir " .. path)
    if (not success) then
        return false
    end
    return true
end

---Attempt to remove a directory with the given path.
---
---An error will occur if the directory can not be removed.
---@param path string Path to the directory to remove
---@return boolean
function remove_directory(path)
    local success, error = os.execute("rmdir -r " .. path)
    if (not success) then
        return false
    end
    return true
end

---Verify if a directory exists given directory path
---@param path string
---@return boolean
function directory_exists(path)
    print("directory_exists", path)
    return os.execute("dir \"" .. path .. "\" > nul") == 0
end

---List the contents from a directory given directory path
---@param path string
---@return nil | integer | table
function list_directory(path)
    -- TODO This needs a way to separate folders from files
    if (path) then
        local command = "dir \"" .. path .. "\" /B"
        local pipe = io.popen(command, "r")
        if pipe then
            local output = pipe:read("*a")
            if (output) then
                local items = split(output, "\n")
                for index, item in pairs(items) do
                    if (item and item == "") then
                        items[index] = nil
                    end
                end
                return items
            end
        end
    end
    return nil
end

---Delete a file given file path
---@param path string
---@return boolean
function delete_file(path)
    return os.remove(path)
end

---Return if a file exists given file path.
---@param path string
---@return boolean
function file_exists(path)
    local file = io.open(path, "r")
    if (file) then
        file:close()
        return true
    end
    return false
end

---Return the memory address of a tag given tagId or tagClass and tagPath
---@param tagIdOrTagType string | number
---@param tagPath? string
---@return number?
function get_tag(tagIdOrTagType, tagPath)
    if (not tagPath) then
        return lookup_tag(tagIdOrTagType)
    else
        return lookup_tag(tagIdOrTagType, tagPath)
    end
end

---Execute a custom Halo script.
---
---A script can be either a standalone Halo command or a Lisp-formatted Halo scripting block.
---@param command string
function execute_script(command)
    return execute_command(command)
end

---Return the address of the object memory given object id
---@param objectId number
---@return number?
function get_object(objectId)
    if (objectId) then
        local object_memory = get_object_memory(objectId)
        if (object_memory ~= 0) then
            return object_memory
        end
    end
    return nil
end

---Despawn an object given objectId. An error will occur if the object does not exist.
---@param objectId number
function delete_object(objectId)
    destroy_object(objectId)
end

---Output text to the console, optional text colors in decimal format.<br>
---Avoid sending console messages if console_is_open() is true to avoid annoying the player.
---@param message string | number
---@param red? number
---@param green? number
---@param blue? number
function console_out(message, red, green, blue)
    -- TODO Add color printing to this function on SAPP
    cprint(message)
end

---Output text to console as debug message.
---
---This function will only output text if the debug mode is enabled.
---@param message string
function console_debug(message)
    if DebugMode then
        console_out(message)
    end
end

---Return true if the player has the console open, always returns true on SAPP.
---@return boolean
function console_is_open()
    return true
end

---Get the value of a Halo scripting global.\
---An error will be triggered if the global is not found
---@param name string Name of the global variable to get from hsc
---@return boolean | number
function get_global(name)
    error("SAPP can not retrieve global variables as Chimera does.. yet!")
end

---Print message to player HUD.\
---Messages will be printed to console if SAPP uses this function
---@param message string
function hud_message(message)
    cprint(message)
end

---Set the callback for an event game from the game events available on Chimera
---@param event '"command"' | '"frame"' | '"preframe"' | '"map load"' | '"precamera"' | '"rcon message"' | '"tick"' | '"pretick"' | '"unload"'
---@param callback string Global function name to call when the event is triggered
function set_callback(event, callback)
    if event == "tick" then
        register_callback(cb["EVENT_TICK"], callback)
    elseif event == "pretick" then
        error("SAPP does not support pretick event")
    elseif event == "frame" then
        error("SAPP does not support frame event")
    elseif event == "preframe" then
        error("SAPP does not support preframe event")
    elseif event == "map load" then
        register_callback(cb["EVENT_GAME_START"], callback)
    elseif event == "precamera" then
        error("SAPP does not support precamera event")
    elseif event == "rcon message" then
        _G[callback .. "_rcon_message"] = function(playerIndex,
                                                   command,
                                                   environment,
                                                   password)
            return _G[callback](playerIndex, command, password)
        end
        register_callback(cb["EVENT_COMMAND"], callback .. "_rcon_message")
    elseif event == "command" then
        _G[callback .. "_command"] = function(playerIndex, command, environment)
            return _G[callback](playerIndex, command, environment)
        end
        register_callback(cb["EVENT_COMMAND"], callback .. "_command")
    elseif event == "unload" then
        register_callback(cb["EVENT_GAME_END"], callback)
    else
        error("Unknown event: " .. event)
    end
end

---Register a timer to be called every intervalMilliseconds.<br>
---The callback function will be called with the arguments passed after the callbackName.<br>
---
---**WARNING:** SAPP will not return a timerId, it will return nil instead so timers can not be stopped.
---@param intervalMilliseconds number
---@param globalFunctionCallbackName string
---@vararg any
---@return number?
function set_timer(intervalMilliseconds, globalFunctionCallbackName, ...)
    return timer(intervalMilliseconds, globalFunctionCallbackName, ...)
end

function stop_timer(timerId)
    error("SAPP does not support stopping timers")
end

if register_callback then
    -- Provide global server type variable on SAPP
    server_type = "sapp"
    print("Compatibility with Chimera Lua API has been loaded!")
else
    console_is_open = backupFunctions.console_is_open
    console_out = backupFunctions.console_out
    execute_script = backupFunctions.execute_script
    get_global = backupFunctions.get_global
    -- set_global = -- backupFunctions.set_global
    get_tag = backupFunctions.get_tag
    set_callback = backupFunctions.set_callback
    set_timer = backupFunctions.set_timer
    stop_timer = backupFunctions.stop_timer
    spawn_object = backupFunctions.spawn_object
    delete_object = backupFunctions.delete_object
    get_object = backupFunctions.get_object
    get_dynamic_player = backupFunctions.get_dynamic_player
    hud_message = backupFunctions.hud_message
    create_directory = backupFunctions.create_directory
    remove_directory = backupFunctions.remove_directory
    directory_exists = backupFunctions.directory_exists
    list_directory = backupFunctions.list_directory
    write_file = backupFunctions.write_file
    read_file = backupFunctions.read_file
    delete_file = backupFunctions.delete_file
    file_exists = backupFunctions.file_exists
end

------------------------------------------------------------------------------
-- Generic functions
------------------------------------------------------------------------------

--- Verify if the given variable is a number
---@param var any
---@return boolean
local function isNumber(var)
    return (type(var) == "number")
end

--- Verify if the given variable is a string
---@param var any
---@return boolean
local function isString(var)
    return (type(var) == "string")
end

--- Verify if the given variable is a boolean
---@param var any
---@return boolean
local function isBoolean(var)
    return (type(var) == "boolean")
end

--- Verify if the given variable is a table
---@param var any
---@return boolean
local function isTable(var)
    return (type(var) == "table")
end

--- Remove spaces and tabs from the beginning and the end of a string
---@param str string
---@return string
local function trim(str)
    return str:match("^%s*(.*)"):match("(.-)%s*$")
end

--- Verify if the value is valid
---@param var any
---@return boolean
local function isValid(var)
    return (var and var ~= "" and var ~= 0)
end

------------------------------------------------------------------------------
-- Utilities
------------------------------------------------------------------------------

--- Convert tag class int to string
---@param tagClassInt number
---@return string?
local function tagClassFromInt(tagClassInt)
    if (tagClassInt) then
        local tagClassHex = tohex(tagClassInt)
        local tagClass = ""
        if (tagClassHex) then
            local byte = ""
            for char in string.gmatch(tagClassHex, ".") do
                byte = byte .. char
                if (#byte % 2 == 0) then
                    tagClass = tagClass .. string.char(tonumber(byte, 16))
                    byte = ""
                end
            end
        end
        return tagClass
    end
    return nil
end

--- Return a list of object indexes that are currently spawned, indexed by their object id.
---@return number[]
function blam.getObjects()
    local objects = {}
    for objectIndex = 0, 2047 do
        local object, objectId = blam.getObject(objectIndex)
        if object and objectId then
            objects[objectId] = objectIndex
            -- objects[objectIndex] = objectId
        end
    end
    return objects
end

-- Local reference to the original console_out function
local original_console_out = console_out

--- Print a console message. It also supports multi-line messages!
---@param message string
local function consoleOutput(message, ...)
    -- Put the extra arguments into a table
    local args = {...}

    if (message == nil or #args > 5) then
        consoleOutput(debug.traceback("Wrong number of arguments on console output function", 2),
                      consoleColors.error)
    end

    -- Output color
    local colorARGB = {1, 1, 1, 1}

    -- Get the output color from arguments table
    if (isTable(args[1])) then
        colorARGB = args[1]
    elseif (#args == 3 or #args == 4) then
        colorARGB = args
    end

    -- Set alpha channel if not set
    if (#colorARGB == 3) then
        table.insert(colorARGB, 1, 1)
    end

    if message then
        if (isString(message)) then
            -- Explode the string!!
            for line in message:gmatch("([^\n]+)") do
                -- Trim the line
                local trimmedLine = trim(line)

                -- Print the line
                original_console_out(trimmedLine, table.unpack(colorARGB))
            end
        else
            original_console_out(message, table.unpack(colorARGB))
        end
    end
end

--- Convert booleans to bits and bits to booleans
---@param bitOrBool number
---@return boolean | number
local function b2b(bitOrBool)
    if (bitOrBool == 1) then
        return true
    elseif (bitOrBool == 0) then
        return false
    elseif (bitOrBool == true) then
        return 1
    elseif (bitOrBool == false) then
        return 0
    end
    error("B2B error, expected boolean or bit value, got " .. tostring(bitOrBool) .. " " ..
              type(bitOrBool))
end

------------------------------------------------------------------------------
-- Data manipulation and binding
------------------------------------------------------------------------------

local typesOperations

local function readBit(address, propertyData)
    return b2b(read_bit(address, propertyData.bitLevel))
end

local function writeBit(address, propertyData, propertyValue)
    return write_bit(address, propertyData.bitLevel, b2b(propertyValue))
end

local function readByte(address)
    return read_byte(address)
end

local function writeByte(address, propertyData, propertyValue)
    return write_byte(address, propertyValue)
end

local function readShort(address)
    return read_short(address)
end

local function writeShort(address, propertyData, propertyValue)
    return write_short(address, propertyValue)
end

local function readWord(address)
    return read_word(address)
end

local function writeWord(address, propertyData, propertyValue)
    return write_word(address, propertyValue)
end

local function readInt(address)
    return read_int(address)
end

local function writeInt(address, propertyData, propertyValue)
    return write_int(address, propertyValue)
end

local function readDword(address)
    return read_dword(address)
end

local function writeDword(address, propertyData, propertyValue)
    return write_dword(address, propertyValue)
end

local function readFloat(address)
    return read_float(address)
end

local function writeFloat(address, propertyData, propertyValue)
    return write_float(address, propertyValue)
end

local function readChar(address)
    return read_char(address)
end

local function writeChar(address, propertyData, propertyValue)
    return write_char(address, propertyValue)
end

local function readString(address)
    return read_string(address)
end

local function writeString(address, propertyData, propertyValue)
    return write_string(address, propertyValue)
end

--- Return the string of a unicode string given address
---@param address number
---@param rawRead? boolean
---@return string
function blam.readUnicodeString(address, rawRead)
    local stringAddress
    if rawRead then
        stringAddress = address
    else
        stringAddress = read_dword(address)
    end
    local length = stringAddress / 2
    local output = ""
    -- TODO Refactor this to support full unicode char size
    for i = 1, length do
        local char = read_string(stringAddress + (i - 1) * 0x2)
        if char == "" then
            break
        end
        output = output .. char
    end
    return output
end

--- Writes a unicode string in a given address
---@param address number
---@param newString string
---@param rawWrite? boolean
function blam.writeUnicodeString(address, newString, rawWrite)
    local stringAddress
    if rawWrite then
        stringAddress = address
    else
        stringAddress = read_dword(address)
    end
    -- Allow cancelling writing when the new string is a boolean false value
    if newString == false then
        return
    end
    local newString = tostring(newString)
    -- TODO Refactor this to support writing ASCII and Unicode strings
    for i = 1, #newString do
        write_string(stringAddress + (i - 1) * 0x2, newString:sub(i, i))
        if i == #newString then
            write_byte(stringAddress + #newString * 0x2, 0x0)
        end
    end
    if #newString == 0 then
        write_string(stringAddress, "")
    end
end

local function readPointerUnicodeString(address, propertyData)
    return blam.readUnicodeString(address)
end

local function readUnicodeString(address, propertyData)
    return blam.readUnicodeString(address, true)
end

local function writePointerUnicodeString(address, propertyData, propertyValue)
    return blam.writeUnicodeString(address, propertyValue)
end

local function writeUnicodeString(address, propertyData, propertyValue)
    return blam.writeUnicodeString(address, propertyValue, true)
end

local function readList(address, propertyData)
    local operation = typesOperations[propertyData.elementsType]
    local elementCount = read_word(address - 0x4)
    local addressList = read_dword(address) + 0xC
    if (propertyData.noOffset) then
        addressList = read_dword(address)
    end
    local list = {}
    for currentElement = 1, elementCount do
        list[currentElement] = operation.read(addressList +
                                                  (propertyData.jump * (currentElement - 1)))
    end
    return list
end

local function writeList(address, propertyData, propertyValue)
    local operation = typesOperations[propertyData.elementsType]
    local elementCount = read_word(address - 0x4)
    local addressList
    if (propertyData.noOffset) then
        addressList = read_dword(address)
    else
        addressList = read_dword(address) + 0xC
    end
    for currentElement = 1, elementCount do
        local elementValue = propertyValue[currentElement]
        if (elementValue) then
            -- Check if there are problems at sending property data here due to missing property data
            operation.write(addressList + (propertyData.jump * (currentElement - 1)), propertyData,
                            elementValue)
        else
            if (currentElement > #propertyValue) then
                break
            end
        end
    end
end

local function readTable(address, propertyData)
    local table = {}
    local elementsCount = read_dword(address - 0x4)
    local firstElement = read_dword(address)
    for elementPosition = 1, elementsCount do
        local elementAddress = firstElement + ((elementPosition - 1) * propertyData.jump)
        table[elementPosition] = {}
        for subProperty, subPropertyData in pairs(propertyData.rows) do
            local operation = typesOperations[subPropertyData.type]
            table[elementPosition][subProperty] = operation.read(elementAddress +
                                                                     subPropertyData.offset,
                                                                 subPropertyData)
        end
    end
    return table
end

local function writeTable(address, propertyData, propertyValue)
    local elementCount = read_dword(address - 0x4)
    local firstElement = read_dword(address)
    for currentElement = 1, elementCount do
        local elementAddress = firstElement + (currentElement - 1) * propertyData.jump
        if (propertyValue[currentElement]) then
            for subProperty, subPropertyValue in pairs(propertyValue[currentElement]) do
                local subPropertyData = propertyData.rows[subProperty]
                if (subPropertyData) then
                    local operation = typesOperations[subPropertyData.type]
                    operation.write(elementAddress + subPropertyData.offset, subPropertyData,
                                    subPropertyValue)
                end
            end
        else
            if (currentElement > #propertyValue) then
                break
            end
        end
    end
end

local function readTagReference(address)
    -- local tagClass = read_dword(address)
    -- local tagPathPointer = read_dword(address + 0x4)
    -- local tagPath = read_string(tagPathPointer)
    -- local unknown = read_dword(address + 0x8)
    local tagId = read_dword(address + 0xC)
    return tagId
end

local function writeTagReference(address, propertyData, tagId)
    -- TODO Attempt to validate tag classes and overwrite tag path pointer
    write_dword(address + 0xC, tagId)
end

local function safeReadUnicodeString(address)
    local size = read_dword(address)
    if size == 0 then
        return ""
    end
    return blam.readUnicodeString(address + 0xC)
end

local function safeWriteUnicodeString(address, propertyData, text)
    local size = read_dword(address)
    local text = text
    if #text > size then
        text = text:sub(1, size)
    end
    return blam.writeUnicodeString(address + 0xC, text)
end

-- Data types operations references
typesOperations = {
    bit = {read = readBit, write = writeBit},
    byte = {read = readByte, write = writeByte},
    short = {read = readShort, write = writeShort},
    word = {read = readWord, write = writeWord},
    int = {read = readInt, write = writeInt},
    dword = {read = readDword, write = writeDword},
    float = {read = readFloat, write = writeFloat},
    char = {read = readChar, write = writeChar},
    string = {read = readString, write = writeString},
    -- TODO This is not ok, a pointer type with subtyping should be implemented
    pustring = {read = readPointerUnicodeString, write = writePointerUnicodeString},
    ustring = {read = readUnicodeString, write = writeUnicodeString},
    list = {read = readList, write = writeList},
    table = {read = readTable, write = writeTable},
    tagref = {read = readTagReference, write = writeTagReference},
    sustring = {read = safeReadUnicodeString, write = safeWriteUnicodeString}
}

-- Magic luablam metatable
local dataBindingMetaTable = {
    __newindex = function(object, property, propertyValue)
        -- Get all the data related to property field
        local propertyData = object.structure[property]
        if (propertyData) then
            local operation = typesOperations[propertyData.type]
            local propertyAddress = object.address + propertyData.offset
            operation.write(propertyAddress, propertyData, propertyValue)
        else
            local errorMessage = "Unable to write an invalid property ('" .. property .. "')"
            error(debug.traceback(errorMessage, 2))
        end
    end,
    __index = function(object, property)
        local objectStructure = object.structure
        local propertyData = objectStructure[property]
        if (propertyData) then
            local operation = typesOperations[propertyData.type]
            local propertyAddress = object.address + propertyData.offset
            return operation.read(propertyAddress, propertyData)
        else
            local errorMessage = "Unable to read an invalid property ('" .. property .. "')"
            error(debug.traceback(errorMessage, 2))
        end
    end
}

------------------------------------------------------------------------------
-- Object functions
------------------------------------------------------------------------------

--- Create a bind table for a given address and structure
---@param address number
---@param struct table
---@return table
local function createBindTable(address, struct)
    -- Create object
    local object = {}

    -- Set up legacy values
    object.address = address
    object.structure = struct

    -- Set mechanisim to bind properties to memory
    setmetatable(object, dataBindingMetaTable)

    return object
end

--- Return a dump of a given LuaBlam object
---@param object table
---@return table
local function dumpObject(object)
    local dump = {}
    for k, v in pairs(object.structure) do
        dump[k] = object[k]
    end
    return dump
end

--- Return a extended parent structure with another given structure
---@param parent table
---@param structure table
---@return table
local function extendStructure(parent, structure)
    local extendedStructure = {}
    for k, v in pairs(parent) do
        extendedStructure[k] = v
    end
    for k, v in pairs(structure) do
        extendedStructure[k] = v
    end
    return extendedStructure
end

------------------------------------------------------------------------------
-- Object structures
------------------------------------------------------------------------------

---@class dataTable
---@field name string
---@field maxElements number
---@field elementSize number
---@field capacity number
---@field size number
---@field nextElementId number
---@field firstElementAddress number

local dataTableStructure = {
    name = {type = "string", offset = 0},
    maxElements = {type = "word", offset = 0x20},
    elementSize = {type = "word", offset = 0x22},
    -- padding1 = {size = 0x0A, offset = 0x24},
    capacity = {type = "word", offset = 0x2E},
    size = {type = "word", offset = 0x30},
    nextElementId = {type = "word", offset = 0x32},
    firstElementAddress = {type = "dword", offset = 0x34}
}

local deviceGroupsTableStructure = {
    name = {type = "string", offset = 0},
    maxElements = {type = "word", offset = 0x20},
    elementSize = {type = "word", offset = 0x22},
    firstElementAddress = {type = "dword", offset = 0x34}
}

---@class blamObject
---@field address number
---@field tagId number Object tag ID
---@field networkRoleClass number Object network role class
---@field isGhost boolean Set object in some type of ghost mode
---@field isOnGround boolean Is the object touching ground
---@field isNotAffectedByGravity boolean Enable/disable object gravity
---@field isInWater boolean Is the object touching on water
---@field isStationary boolean Is the object stationary
---@field dynamicShading boolean Enable disable dynamic shading for lightmaps
---@field isNotCastingShadow boolean Enable/disable object shadow casting
---@field isFrozen boolean Freeze/unfreeze object existence
---@field isOutSideMap boolean Is object outside/inside bsp
---@field isCollideable boolean Enable/disable object collision, does not work with bipeds or vehicles
---@field isBeingPickedUp boolean Is the object being picked up
---@field hasNoCollision boolean Enable/disable object collision, causes animation problems
---@field model number Gbxmodel tag ID
---@field scale number Object scale factor
---@field health number Current health of the object
---@field maxHealth number Maximum health of the object
---@field shield number Current shield of the object
---@field maxShield number Maximum shield of the object
---@field colorAUpperRed number Red color channel for A modifier
---@field colorAUpperGreen number Green color channel for A modifier
---@field colorAUpperBlue number Blue color channel for A modifier
---@field colorBUpperRed number Red color channel for B modifier
---@field colorBUpperGreen number Green color channel for B modifier
---@field colorBUpperBlue number Blue color channel for B modifier
---@field colorCUpperRed number Red color channel for C modifier
---@field colorCUpperGreen number Green color channel for C modifier
---@field colorCUpperBlue number Blue color channel for C modifier
---@field colorDUpperRed number Red color channel for D modifier
---@field colorDUpperGreen number Green color channel for D modifier
---@field colorDUpperBlue number Blue color channel for D modifier
---@field colorALowerRed number Red color channel for A modifier
---@field colorALowerGreen number Green color channel for A modifier
---@field colorALowerBlue number Blue color channel for A modifier
---@field colorBLowerRed number Red color channel for B modifier
---@field colorBLowerGreen number Green color channel for B modifier
---@field colorBLowerBlue number Blue color channel for B modifier
---@field colorCLowerRed number Red color channel for C modifier
---@field colorCLowerGreen number Green color channel for C modifier
---@field colorCLowerBlue number Blue color channel for C modifier
---@field colorDLowerRed number Red color channel for D modifier
---@field colorDLowerGreen number Green color channel for D modifier
---@field colorDLowerBlue number Blue color channel for D modifier
---@field x number Current position of the object on X axis
---@field y number Current position of the object on Y axis
---@field z number Current position of the object on Z axis
---@field xVel number Current velocity of the object on X axis
---@field yVel number Current velocity of the object on Y axis
---@field zVel number Current velocity of the object on Z axis
---@field vX number Current x value in first rotation vector
---@field vY number Current y value in first rotation vector
---@field vZ number Current z value in first rotation vector
---@field v2X number Current x value in second rotation vector
---@field v2Y number Current y value in second rotation vector
---@field v2Z number Current z value in second rotation vector
---@field yawVel number Current velocity of the object in yaw
---@field pitchVel number Current velocity of the object in pitch
---@field rollVel number Current velocity of the object in roll
---@field locationId number Current id of the location in the map
---@field boundingRadius number Radius amount of the object in radians
---@field class objectClasses Object type
---@field team number Object multiplayer team
---@field nameIndex number Index of object name in the scenario tag
---@field playerId number Current player id if the object
---@field ownerId number Current owner id of the object any other object id
---@field isApparentlyDead boolean Is the object apparently dead
---@field isSilentlyKilled boolean Is the object really dead
---@field animationTagId number Current animation tag ID
---@field animation number Current animation index
---@field animationFrame number Current animation frame
---@field isNotDamageable boolean Make the object undamageable
---@field shaderPermutationIndex number Current shader permutation index
---@field regionPermutation1 number
---@field regionPermutation2 number
---@field regionPermutation3 number
---@field regionPermutation4 number
---@field regionPermutation5 number
---@field regionPermutation6 number
---@field regionPermutation7 number
---@field regionPermutation8 number
---@field parentObjectId number

-- blamObject structure
local objectStructure = {
    tagId = {type = "dword", offset = 0x0},
    networkRoleClass = {type = "dword", offset = 0x4},
    isNotMoving = {type = "bit", offset = 0x8, bitLevel = 0},
    existanceTime = {type = "dword", offset = 0xC},
    isGhost = {type = "bit", offset = 0x10, bitLevel = 0},
    isOnGround = {type = "bit", offset = 0x10, bitLevel = 1},
    ---@deprecated
    ignoreGravity = {type = "bit", offset = 0x10, bitLevel = 2},
    isNotAffectedByGravity = {type = "bit", offset = 0x10, bitLevel = 2},
    isInWater = {type = "bit", offset = 0x10, bitLevel = 3},
    isStationary = {type = "bit", offset = 0x10, bitLevel = 5},
    hasNoCollision = {type = "bit", offset = 0x10, bitLevel = 7},
    dynamicShading = {type = "bit", offset = 0x10, bitLevel = 14},
    isNotCastingShadow = {type = "bit", offset = 0x10, bitLevel = 18},
    isFrozen = {type = "bit", offset = 0x10, bitLevel = 20},
    -- FIXME Deprecated property, should be erased at a major release later
    frozen = {type = "bit", offset = 0x10, bitLevel = 20},
    isCollideable = {type = "bit", offset = 0x10, bitLevel = 24},
    isBeingPickedUp = {type = "bit", offset = 0x10, bitLevel = 26},
    isOutSideMap = {type = "bit", offset = 0x12, bitLevel = 5},
    model = {type = "dword", offset = 0x34},
    scale = {type = "float", offset = 0xB0},
    health = {type = "float", offset = 0xE0},
    maxHealth = {type = "float", offset = 0xD8},
    shield = {type = "float", offset = 0xE4},
    maxShield = {type = "float", offset = 0xDC},
    ---@deprecated
    redA = {type = "float", offset = 0x1B8},
    ---@deprecated
    greenA = {type = "float", offset = 0x1BC},
    ---@deprecated
    blueA = {type = "float", offset = 0x1C0},
    colorAUpperRed = {type = "float", offset = 0x188},
    colorAUpperGreen = {type = "float", offset = 0x18C},
    colorAUpperBlue = {type = "float", offset = 0x190},
    colorBUpperRed = {type = "float", offset = 0x194},
    colorBUpperGreen = {type = "float", offset = 0x198},
    colorBUpperBlue = {type = "float", offset = 0x19C},
    colorCUpperRed = {type = "float", offset = 0x1A0},
    colorCUpperGreen = {type = "float", offset = 0x1A4},
    colorCUpperBlue = {type = "float", offset = 0x1A8},
    colorDUpperRed = {type = "float", offset = 0x1AC},
    colorDUpperGreen = {type = "float", offset = 0x1B0},
    colorDUpperBlue = {type = "float", offset = 0x1B4},
    colorALowerRed = {type = "float", offset = 0x1B8},
    colorALowerGreen = {type = "float", offset = 0x1BC},
    colorALowerBlue = {type = "float", offset = 0x1C0},
    colorBLowerRed = {type = "float", offset = 0x1C4},
    colorBLowerGreen = {type = "float", offset = 0x1C8},
    colorBLowerBlue = {type = "float", offset = 0x1CC},
    colorCLowerRed = {type = "float", offset = 0x1D0},
    colorCLowerGreen = {type = "float", offset = 0x1D4},
    colorCLowerBlue = {type = "float", offset = 0x1D8},
    colorDLowerRed = {type = "float", offset = 0x1DC},
    colorDLowerGreen = {type = "float", offset = 0x1E0},
    colorDLowerBlue = {type = "float", offset = 0x1E4},
    x = {type = "float", offset = 0x5C},
    y = {type = "float", offset = 0x60},
    z = {type = "float", offset = 0x64},
    xVel = {type = "float", offset = 0x68},
    yVel = {type = "float", offset = 0x6C},
    zVel = {type = "float", offset = 0x70},
    vX = {type = "float", offset = 0x74},
    vY = {type = "float", offset = 0x78},
    vZ = {type = "float", offset = 0x7C},
    v2X = {type = "float", offset = 0x80},
    v2Y = {type = "float", offset = 0x84},
    v2Z = {type = "float", offset = 0x88},
    -- FIXME Some order from this values is probaby wrong, expected order is pitch, yaw, roll
    yawVel = {type = "float", offset = 0x8C},
    pitchVel = {type = "float", offset = 0x90},
    rollVel = {type = "float", offset = 0x94},
    locationId = {type = "dword", offset = 0x98},
    boundingRadius = {type = "float", offset = 0xAC},
    ---@deprecated
    type = {type = "word", offset = 0xB4},
    class = {type = "word", offset = 0xB4},
    team = {type = "word", offset = 0xB8},
    nameIndex = {type = "word", offset = 0xBA},
    playerId = {type = "dword", offset = 0xC0},
    ---@deprecated
    parentId = {type = "dword", offset = 0xC4},
    ownerId = {type = "dword", offset = 0xC4},
    ---@deprecated
    isHealthEmpty = {type = "bit", offset = 0x106, bitLevel = 2},
    isApparentlyDead = {type = "bit", offset = 0x106, bitLevel = 2},
    isSilentlyKilled = {type = "bit", offset = 0x106, bitLevel = 5},
    animationTagId = {type = "dword", offset = 0xCC},
    animation = {type = "word", offset = 0xD0},
    animationFrame = {type = "word", offset = 0xD2},
    isNotDamageable = {type = "bit", offset = 0x106, bitLevel = 11},
    shaderPermutationIndex = {type = "word", offset = 0x176},
    regionPermutation1 = {type = "byte", offset = 0x180},
    regionPermutation2 = {type = "byte", offset = 0x181},
    regionPermutation3 = {type = "byte", offset = 0x182},
    regionPermutation4 = {type = "byte", offset = 0x183},
    regionPermutation5 = {type = "byte", offset = 0x184},
    regionPermutation6 = {type = "byte", offset = 0x185},
    regionPermutation7 = {type = "byte", offset = 0x186},
    regionPermutation8 = {type = "byte", offset = 0x187},
    parentObjectId = {type = "dword", offset = 0x11C}
}

local unitStructure = extendStructure(objectStructure, {
    ---@deprecated
    invisible = {type = "bit", offset = 0x204, bitLevel = 4},
    isCamoActive = {type = "bit", offset = 0x204, bitLevel = 4},
    isControllable = {type = "bit", offset = 0x204, bitLevel = 5},
    isPlayerNotAllowedToEntry = {type = "bit", offset = 0x204, bitLevel = 16},
    parentSeatIndex = {type = "word", offset = 0x2F0},
    weaponAnimationTypeIndex = {type = "byte", offset = 0x2A1},
    weaponSlot = {type = "byte", offset = 0x2F2},
    firstWeaponObjectId = {type = "dword", offset = 0x2F8},
    secondWeaponObjectId = {type = "dword", offset = 0x2FC},
    thirdWeaponObjectId = {type = "dword", offset = 0x300},
    fourthWeaponObjectId = {type = "dword", offset = 0x304},
    camoScale = {type = "float", offset = 0x37C}
})

---@class unit : blamObject
---@field isCamoActive boolean Unit camo state
---@field isControllable boolean Unit controllable state
---@field isPlayerNotAllowedToEntry boolean Unit player not allowed to entry
---@field parentSeatIndex number Unit parent seat index
---@field weaponAnimationTypeIndex number Unit weapon animation type index
---@field weaponSlot number Current unit weapon slot
---@field firstWeaponObjectId number First weapon object id
---@field secondWeaponObjectId number Second weapon object id
---@field thirdWeaponObjectId number Third weapon object id
---@field fourthWeaponObjectId number Fourth weapon object id
---@field camoScale number Unit camo scale

-- Biped structure (extends object structure)
local bipedStructure = extendStructure(unitStructure, {
    noDropItems = {type = "bit", offset = 0x204, bitLevel = 20},
    flashlight = {type = "bit", offset = 0x204, bitLevel = 19},
    cameraX = {type = "float", offset = 0x230},
    cameraY = {type = "float", offset = 0x234},
    cameraZ = {type = "float", offset = 0x238},
    crouchHold = {type = "bit", offset = 0x208, bitLevel = 0},
    jumpHold = {type = "bit", offset = 0x208, bitLevel = 1},
    actionKeyHold = {type = "bit", offset = 0x208, bitLevel = 14},
    actionKey = {type = "bit", offset = 0x208, bitLevel = 6},
    meleeKey = {type = "bit", offset = 0x208, bitLevel = 7},
    reloadKey = {type = "bit", offset = 0x208, bitLevel = 10},
    weaponPTH = {type = "bit", offset = 0x208, bitLevel = 11},
    weaponSTH = {type = "bit", offset = 0x208, bitLevel = 12},
    flashlightKey = {type = "bit", offset = 0x208, bitLevel = 4},
    grenadeHold = {type = "bit", offset = 0x208, bitLevel = 13},
    crouch = {type = "byte", offset = 0x2A0},
    shooting = {type = "float", offset = 0x284},
    zoomLevel = {type = "byte", offset = 0x320},
    ---@deprecated
    invisibleScale = {type = "float", offset = 0x37C},
    primaryNades = {type = "byte", offset = 0x31E},
    secondaryNades = {type = "byte", offset = 0x31F},
    isNotAffectedByGravity = {type = "bit", offset = 0x4CC, bitLevel = 2},
    ignoreCollision = {type = "bit", offset = 0x4CC, bitLevel = 3},
    landing = {type = "byte", offset = 0x508},
    bumpedObjectId = {type = "dword", offset = 0x4FC},
    vehicleObjectId = {type = "dword", offset = 0x11C},
    vehicleSeatIndex = {type = "word", offset = 0x2F0},
    walkingState = {type = "char", offset = 0x503},
    motionState = {type = "byte", offset = 0x4D2},
    mostRecentDamagerPlayer = {type = "dword", offset = 0x43C}
})

---@class biped : unit
---@field invisible boolean Biped invisible state
---@field noDropItems boolean Biped ability to drop items at dead
---@field ignoreCollision boolean Biped ignores collisiion
---@field flashlight boolean Biped has flaslight enabled
---@field cameraX number Current position of the biped  X axis
---@field cameraY number Current position of the biped  Y axis
---@field cameraZ number Current position of the biped  Z axis
---@field crouchHold boolean Biped is holding crouch action
---@field jumpHold boolean Biped is holding jump action
---@field actionKeyHold boolean Biped is holding action key
---@field actionKey boolean Biped pressed action key
---@field meleeKey boolean Biped pressed melee key
---@field reloadKey boolean Biped pressed reload key
---@field weaponPTH boolean Biped is holding primary weapon trigger
---@field weaponSTH boolean Biped is holding secondary weapon trigger
---@field flashlightKey boolean Biped pressed flashlight key
---@field grenadeHold boolean Biped is holding grenade action
---@field crouch number Is biped crouch
---@field shooting number Is biped shooting, 0 when not, 1 when shooting
---@field weaponSlot number Current biped weapon slot
---@field zoomLevel number Current biped weapon zoom level, 0xFF when no zoom, up to 255 when zoomed
---@field invisibleScale number Opacity amount of biped invisiblity
---@field primaryNades number Primary grenades count
---@field secondaryNades number Secondary grenades count
---@field isNotAffectedByGravity boolean Enable/disable biped gravity
---@field landing number Biped landing state, 0 when landing, stays on 0 when landing hard, null otherwise
---@field bumpedObjectId number Object ID that the biped is bumping, vehicles, bipeds, etc, keeps the previous value if not bumping a new object
---@field vehicleSeatIndex number Current vehicle seat index of this biped
---@field vehicleObjectId number Current vehicle objectId of this object
---@field walkingState number Biped walking state, 0 = not walking, 1 = walking, 2 = stoping walking, 3 = stationary
---@field motionState number Biped motion state, 0 = standing , 1 = walking , 2 = jumping/falling
---@field mostRecentDamagerPlayer number Id of the player that caused the most recent damage to this biped

local vehicleStructure = extendStructure(unitStructure, {
    isTireBlur = {type = "bit", offset = 0x4CC, bitLevel = 0},
    isHovering = {type = "bit", offset = 0x4CC, bitLevel = 1},
    isCrouched = {type = "bit", offset = 0x4CC, bitLevel = 2},
    isJumping = {type = "bit", offset = 0x4CC, bitLevel = 3},
    speed = {type = "float", offset = 0x4D4},
    slide = {type = "float", offset = 0x4D8},
    turn = {type = "float", offset = 0x4DC},
    tirePosition = {type = "float", offset = 0x4E0},
    threadPositionLeft = {type = "float", offset = 0x4E4},
    threadPositionRight = {type = "float", offset = 0x4E8},
    hover = {type = "float", offset = 0x4EC},
    thrust = {type = "float", offset = 0x4F0},
    hoverX = {type = "float", offset = 0x4FC},
    hoverY = {type = "float", offset = 0x500},
    hoverZ = {type = "float", offset = 0x504},
    respawnTimer = {type = "dword", offset = 0x5AC},
    respawnTime = {type = "word", offset = 0x5B0},
    respawnX = {type = "float", offset = 0x5B4},
    respawnY = {type = "float", offset = 0x5B8},
    respawnZ = {type = "float", offset = 0x5BC}
})

---@class vehicle : unit
---@field isTireBlur boolean Vehicle tire blur state
---@field isHovering boolean Vehicle hovering state
---@field isCrouched boolean Vehicle crouch state
---@field isJumping boolean Vehicle jumping state
---@field speed number Vehicle speed
---@field slide number Vehicle slide
---@field turn number Vehicle turn
---@field tirePosition number Vehicle tire position
---@field threadPositionLeft number Vehicle thread position left
---@field threadPositionRight number Vehicle thread position right
---@field hover number Vehicle hover
---@field thrust number Vehicle thrust
---@field hoverX number Vehicle hover X axis
---@field hoverY number Vehicle hover Y axis
---@field hoverZ number Vehicle hover Z axis
---@field respawnTimer number Vehicle respawn timer
---@field respawnTime number Vehicle respawn time
---@field respawnX number Vehicle respawn X axis
---@field respawnY number Vehicle respawn Y axis
---@field respawnZ number Vehicle respawn Z axis

-- Tag data header structure
local tagDataHeaderStructure = {
    array = {type = "dword", offset = 0x0},
    scenario = {type = "dword", offset = 0x4},
    count = {type = "word", offset = 0xC}
}

---@class tag
---@field class number Type of the tag
---@field index number Tag Index
---@field id number Tag ID
---@field path string Path of the tag
---@field data number Address of the tag data
---@field indexed boolean Is tag indexed on an external map file

-- Tag structure
local tagHeaderStructure = {
    class = {type = "dword", offset = 0x0},
    index = {type = "word", offset = 0xC},
    id = {type = "dword", offset = 0xC},
    path = {type = "dword", offset = 0x10},
    data = {type = "dword", offset = 0x14},
    indexed = {type = "dword", offset = 0x18}
}

---@class tagCollection
---@field count number Number of tags in the collection
---@field tagList table List of tags

-- tagCollection structure
local tagCollectionStructure = {
    count = {type = "byte", offset = 0x0},
    tagList = {type = "list", offset = 0x4, elementsType = "dword", jump = 0x10}
}

---@class unicodeStringList
---@field count number Number of unicode strings
---@field strings string[] List of unicode strings

-- UnicodeStringList structure
local unicodeStringListStructure = {
    count = {type = "byte", offset = 0x0},
    ---@deprecated
    stringList = {type = "list", offset = 0x4, elementsType = "pustring", jump = 0x14},
    -- Previous string list property works because of magic (well because of shit code haha)
    strings = {type = "list", offset = 0x4, elementsType = "sustring", jump = 0x14, noOffset = true}
}

---@class bitmapSequence
---@field name string
---@field firtBitmapIndex number
---@field bitmapCount number

---@class bitmap
---@field type number
---@field format number
---@field usage number
---@field usageFlags number
---@field detailFadeFactor number
---@field sharpenAmount number
---@field bumpHeight number
---@field spriteBudgetSize number
---@field spriteBudgetCount number
---@field colorPlateWidth number
---@field colorPlateHeight number 
---@field compressedColorPlate string
---@field processedPixelData string
---@field blurFilterSize number
---@field alphaBias number
---@field mipmapCount number
---@field spriteUsage number
---@field spriteSpacing number
---@field sequencesCount number
---@field sequences bitmapSequence[]
---@field bitmapsCount number
---@field bitmaps table

-- Bitmap structure
local bitmapStructure = {
    type = {type = "word", offset = 0x0},
    format = {type = "word", offset = 0x2},
    usage = {type = "word", offset = 0x4},
    usageFlags = {type = "word", offset = 0x6},
    detailFadeFactor = {type = "dword", offset = 0x8},
    sharpenAmount = {type = "dword", offset = 0xC},
    bumpHeight = {type = "dword", offset = 0x10},
    spriteBudgetSize = {type = "word", offset = 0x14},
    spriteBudgetCount = {type = "word", offset = 0x16},
    colorPlateWidth = {type = "word", offset = 0x18},
    colorPlateHeight = {type = "word", offset = 0x1A},
    -- compressedColorPlate = {offset = 0x1C},
    -- processedPixelData = {offset = 0x30},
    blurFilterSize = {type = "float", offset = 0x44},
    alphaBias = {type = "float", offset = 0x48},
    mipmapCount = {type = "word", offset = 0x4C},
    spriteUsage = {type = "word", offset = 0x4E},
    spriteSpacing = {type = "word", offset = 0x50},
    -- padding1 = {size = 0x2, offset = 0x52},
    sequencesCount = {type = "byte", offset = 0x54},
    sequences = {
        type = "table",
        offset = 0x58,
        jump = 0x40,
        rows = {
            name = {type = "string", offset = 0x0},
            firstBitmapIndex = {type = "word", offset = 0x20},
            bitmapCount = {type = "word", offset = 0x22}
            -- padding = {size = 0x10, offset = 0x24},
            --[[
            sprites = {
                type = "table",
                offset = 0x34,
                jump = 0x20,
                rows = {
                    bitmapIndex = {type = "word", offset = 0x0},
                    --padding1 = {size = 0x2, offset = 0x2},
                    --padding2 = {size = 0x4, offset = 0x4},
                    left = {type = "float", offset = 0x8},
                    right = {type = "float", offset = 0xC},
                    top = {type = "float", offset = 0x10},
                    bottom = {type = "float", offset = 0x14},
                    registrationX = {type = "float", offset = 0x18},
                    registrationY = {type = "float", offset = 0x1C}
                }
            }
            ]]
        }
    },
    bitmapsCount = {type = "byte", offset = 0x60},
    bitmaps = {
        type = "table",
        offset = 0x64,
        jump = 0x30,
        rows = {
            class = {type = "dword", offset = 0x0},
            width = {type = "word", offset = 0x4},
            height = {type = "word", offset = 0x6},
            depth = {type = "word", offset = 0x8},
            type = {type = "word", offset = 0xA},
            format = {type = "word", offset = 0xC},
            flags = {type = "word", offset = 0xE},
            x = {type = "word", offset = 0x10},
            y = {type = "word", offset = 0x12},
            mipmapCount = {type = "word", offset = 0x14},
            -- padding1 = {size = 0x2, offset = 0x16},
            pixelOffset = {type = "dword", offset = 0x18}
            -- padding2 = {size = 0x4, offset = 0x1C},
            -- padding3 = {size = 0x4, offset = 0x20},
            -- padding4 = {size = 0x4, offset= 0x24},
            -- padding5 = {size = 0x8, offset= 0x28}
        }
    }
}

---@class uiWidgetDefinitionChild
---@field widgetTag number Child uiWidgetDefinition reference
---@field name number Child widget name
---@field customControllerIndex number Custom controller index for this child widget
---@field verticalOffset number Offset in Y axis of this child, relative to the parent
---@field horizontalOffset number Offset in X axis of this child, relative to the parent

---@class uiWidgetDefinitionEventHandler
---@field eventType number Type of the event
---@field gameFunction number Game function of this event
---@field widgetTag number uiWidgetDefinition tag id of the event
---@field script string Name of the script function assigned to this event

---@class uiWidgetDefinition
---@field type number Type of widget
---@field controllerIndex number Index of the player controller
---@field name string Name of the widget
---@field top number Top bound of the widget
---@field left number Left bound of the widget
---@field bottom number Bottom bound of the widget
---@field right number Right bound of the widget
---@field boundsY number Top bound of the widget
---@field boundsX number Left bound of the widget
---@field height number Bottom bound of the widget
---@field width number Right bound of the widget
---@field passUnhandleEventsToFocusedChild boolean Pass unhandled events to focused child
---@field pauseGameTime boolean Pause game time
---@field flashBackgroundBitmap boolean Flash background bitmap
---@field dpadUpDownTabsThruChildren boolean Dpad up down tabs thru children
---@field dpadLeftRightTabsThruChildren boolean Dpad left right tabs thru children
---@field dpadUpDownTabsThruListItems boolean Dpad up down tabs thru list items
---@field dpadLeftRightTabsThruListItems boolean Dpad left right tabs thru list items
---@field dontFocusSpecificChildWidget boolean Don't focus specific child widget
---@field passUnhandledEventsToAllChildren boolean Pass unhandled events to all children
---@field renderRegardlessOfControllerIndex boolean Render regardless of controller index
---@field passHandledEventsToAllChildren boolean Pass handled events to all children
---@field returnToMainMenuIfNoHistory boolean Return to main menu if no history
---@field alwaysUseTagControllerIndex boolean Always use tag controller index
---@field alwaysUseNiftyRenderFx boolean Always use nifty render fx
---@field dontPushHistory boolean Don't push history
---@field forceHandleMouse boolean Force handle mouse
---@field backgroundBitmap number Tag ID of the background bitmap
---@field eventHandlers uiWidgetDefinitionEventHandler[] tag ID list of the child widgets
---@field unicodeStringListTag number Tag ID of the unicodeStringList from this widget
---@field fontTag number Tag ID of the font from this widget
---@field justification number Text justification of the text from this widget
---@field stringListIndex number Text index from the unicodeStringList tag from this widget
---@field textHorizontalOffset number Text offset in X axis from this widget
---@field textVerticalOffset number Text offset in Y axis from this widget
---@field childWidgetsCount number Number of child widgets
---@field childWidgets uiWidgetDefinitionChild[] List of the child widgets

local uiWidgetDefinitionStructure = {
    type = {type = "word", offset = 0x0},
    controllerIndex = {type = "word", offset = 0x2},
    name = {type = "string", offset = 0x4},
    top = {type = "short", offset = 0x24},
    left = {type = "short", offset = 0x26},
    bottom = {type = "short", offset = 0x28},
    right = {type = "short", offset = 0x2A},
    boundsY = {type = "short", offset = 0x24},
    boundsX = {type = "short", offset = 0x26},
    height = {type = "short", offset = 0x28},
    width = {type = "short", offset = 0x2A},
    passUnhandleEventsToFocusedChild = {type = "bit", offset = 0x2C, bitLevel = 0},
    pauseGameTime = {type = "bit", offset = 0x2C, bitLevel = 1},
    flashBackgroundBitmap = {type = "bit", offset = 0x2C, bitLevel = 2},
    dpadUpDownTabsThruChildren = {type = "bit", offset = 0x2C, bitLevel = 3},
    dpadLeftRightTabsThruChildren = {type = "bit", offset = 0x2C, bitLevel = 4},
    dpadUpDownTabsThruListItems = {type = "bit", offset = 0x2C, bitLevel = 5},
    dpadLeftRightTabsThruListItems = {type = "bit", offset = 0x2C, bitLevel = 6},
    dontFocusSpecificChildWidget = {type = "bit", offset = 0x2C, bitLevel = 7},
    passUnhandledEventsToAllChildren = {type = "bit", offset = 0x2C, bitLevel = 8},
    renderRegardlessOfControllerIndex = {type = "bit", offset = 0x2C, bitLevel = 9},
    passHandledEventsToAllChildren = {type = "bit", offset = 0x2C, bitLevel = 10},
    returnToMainMenuIfNoHistory = {type = "bit", offset = 0x2C, bitLevel = 11},
    alwaysUseTagControllerIndex = {type = "bit", offset = 0x2C, bitLevel = 12},
    alwaysUseNiftyRenderFx = {type = "bit", offset = 0x2C, bitLevel = 13},
    dontPushHistory = {type = "bit", offset = 0x2C, bitLevel = 14},
    forceHandleMouse = {type = "bit", offset = 0x2C, bitLevel = 15},
    backgroundBitmap = {type = "word", offset = 0x44},
    eventHandlers = {
        type = "table",
        offset = 0x54,
        jump = 0x48,
        rows = {
            -- TODO Add real flags support, or a subtyping of table instead
            -- flags = {type = "number", offset = 0x0},
            eventType = {type = "word", offset = 0x4},
            gameFunction = {type = "word", offset = 0x6},
            widgetTag = {type = "tagref", offset = 0x8},
            soundEffectTag = {type = "tagref", offset = 0x18},
            script = {type = "string", offset = 0x28}
        }
    },
    unicodeStringListTag = {type = "tagref", offset = 0xEC},
    fontTag = {type = "tagref", offset = 0xFC},
    -- TODO Add color support for hex and rgb values
    -- textColor = {type = "realargbcolor", offset = 0x10C},
    justification = {type = "word", offset = 0x11C},
    stringListIndex = {type = "short", offset = 0x12E},
    textHorizontalOffset = {type = "short", offset = 0x130},
    textVerticalOffset = {type = "short", offset = 0x132},
    ---@deprecated
    eventType = {type = "byte", offset = 0x03F0},
    ---@deprecated
    tagReference = {type = "word", offset = 0x400},
    childWidgetsCount = {type = "dword", offset = 0x03E0},
    ---@deprecated
    childWidgetsList = {type = "list", offset = 0x03E4, elementsType = "dword", jump = 0x50},
    childWidgets = {
        type = "table",
        offset = 0x03E4,
        jump = 0x50,
        rows = {
            widgetTag = {type = "tagref", offset = 0x0},
            name = {type = "string", offset = 0x10},
            -- flags = {type = "integer", offset = 0x30},
            customControllerIndex = {type = "short", offset = 0x34},
            verticalOffset = {type = "short", offset = 0x36},
            horizontalOffset = {type = "short", offset = 0x38}
        }
    }
}

---@class uiWidgetCollection
---@field count number Number of widgets in the collection
---@field tagList table Tag ID list of the widgets

-- uiWidgetCollection structure
local uiWidgetCollectionStructure = {
    count = {type = "byte", offset = 0x0},
    tagList = {type = "list", offset = 0x4, elementsType = "dword", jump = 0x10}
}

---@class crosshairOverlay
---@field x number
---@field y number
---@field widthScale number
---@field heightScale number
---@field defaultColorA number
---@field defaultColorR number
---@field defaultColorG number
---@field defaultColorB number
---@field sequenceIndex number

---@class crosshair
---@field type number
---@field mapType number
---@field bitmap number
---@field overlays crosshairOverlay[]

---@class weaponHudInterface
---@field childHud number
---@field totalAmmoCutOff number
---@field loadedAmmoCutOff number
---@field heatCutOff number
---@field ageCutOff number
---@field crosshairs crosshair[]

-- Weapon HUD Interface structure
local weaponHudInterfaceStructure = {
    childHud = {type = "dword", offset = 0xC},
    -- //TODO Check if this property should be moved to a nested property type
    usingParentHudFlashingParameters = {type = "bit", offset = "word", bitLevel = 1},
    -- padding1 = {type = "word", offset = 0x12},
    totalAmmoCutOff = {type = "word", offset = 0x14},
    loadedAmmoCutOff = {type = "word", offset = 0x16},
    heatCutOff = {type = "word", offset = 0x18},
    ageCutOff = {type = "word", offset = 0x1A},
    -- padding2 = {size = 0x20, offset = 0x1C},
    -- screenAlignment = {type = "word", },
    -- padding3 = {size = 0x2, offset = 0x3E},
    -- padding4 = {size = 0x20, offset = 0x40},
    crosshairs = {
        type = "table",
        offset = 0x88,
        jump = 0x68,
        rows = {
            type = {type = "word", offset = 0x0},
            mapType = {type = "word", offset = 0x2},
            -- padding1 = {size = 0x2, offset = 0x4},
            -- padding2 = {size = 0x1C, offset = 0x6},
            bitmap = {type = "dword", offset = 0x30},
            overlays = {
                type = "table",
                offset = 0x38,
                jump = 0x6C,
                rows = {
                    x = {type = "word", offset = 0x0},
                    y = {type = "word", offset = 0x2},
                    widthScale = {type = "float", offset = 0x4},
                    heightScale = {type = "float", offset = 0x8},
                    defaultColorB = {type = "byte", offset = 0x24},
                    defaultColorG = {type = "byte", offset = 0x25},
                    defaultColorR = {type = "byte", offset = 0x26},
                    defaultColorA = {type = "byte", offset = 0x27},
                    sequenceIndex = {type = "byte", offset = 0x46}
                }
            }
        }
    }
}

---@class spawnLocation
---@field x number
---@field y number
---@field z number
---@field rotation number
---@field type number
---@field teamIndex number

---@class vehicleLocation
---@field type number
---@field nameIndex string
---@field x number
---@field y number
---@field z number
---@field yaw number
---@field pitch number
---@field roll number

---@class cutsceneFlag
---@field name string
---@field x number
---@field y number
---@field z number
---@field vX number
---@field vY number

---@class scenarioScenery
---@field typeIndex number
---@field nameIndex string
---@field notPlaced boolean
---@field desiredPermutation number
---@field x number
---@field y number
---@field z number
---@field yaw number
---@field pitch number
---@field roll number

---@class scenarioBiped
---@field typeIndex number
---@field nameIndex string
---@field notPlaced boolean
---@field desiredPermutation number
---@field x number
---@field y number
---@field z number
---@field yaw number
---@field pitch number
---@field roll number

---@class scenario
---@field sceneryPaletteCount number Number of sceneries in the scenery palette
---@field sceneryPaletteList tagId[] Tag ID list of scenerys in the scenery palette
---@field spawnLocationCount number Number of spawns in the scenario
---@field spawnLocationList spawnLocation[] List of spawns in the scenario
---@field vehicleLocationCount number Number of vehicles locations in the scenario
---@field vehicleLocationList vehicleLocation[] List of vehicles locations in the scenario
---@field netgameEquipmentCount number Number of netgame equipments
---@field netgameEquipmentList table List of netgame equipments
---@field netgameFlagsCount number Number of netgame equipments
---@field netgameFlagsList table List of netgame equipments
---@field objectNamesCount number Count of the object names in the scenario
---@field objectNames string[] List of all the object names in the scenario
---@field sceneriesCount number Count of all the sceneries in the scenario
---@field sceneries scenarioScenery[] List of all the sceneries in the scenario
---@field bipedsCount number Count of all the bipeds in the scenario
---@field bipeds scenarioBiped[] List of all the bipeds in the scenario
---@field bipedPaletteCount number Count of all the bipeds in the biped palette
---@field bipedPaletteList tagId[] List of all the bipeds in the biped palette
---@field cutsceneFlagsCount number Count of all the cutscene flags in the scenario
---@field cutsceneFlags cutsceneFlag[] List of all the cutscene flags in the scenario
---@field actorPaletteCount number Count of all the actors in the actor palette
---@field encounterPaletteCount number Count of all the encounters in the encounter palette

-- Scenario structure
local scenarioStructure = {
    sceneryPaletteCount = {type = "byte", offset = 0x021C},
    sceneryPaletteList = {type = "list", offset = 0x0220, elementsType = "dword", jump = 0x30},
    spawnLocationCount = {type = "byte", offset = 0x354},
    spawnLocationList = {
        type = "table",
        offset = 0x358,
        jump = 0x34,
        rows = {
            x = {type = "float", offset = 0x0},
            y = {type = "float", offset = 0x4},
            z = {type = "float", offset = 0x8},
            rotation = {type = "float", offset = 0xC},
            teamIndex = {type = "byte", offset = 0x10},
            bspIndex = {type = "short", offset = 0x12},
            type = {type = "byte", offset = 0x14}
        }
    },
    vehicleLocationCount = {type = "byte", offset = 0x240},
    vehicleLocationList = {
        type = "table",
        offset = 0x244,
        jump = 0x78,
        rows = {
            type = {type = "word", offset = 0x0},
            nameIndex = {type = "word", offset = 0x2},
            x = {type = "float", offset = 0x8},
            y = {type = "float", offset = 0xC},
            z = {type = "float", offset = 0x10},
            yaw = {type = "float", offset = 0x14},
            pitch = {type = "float", offset = 0x18},
            roll = {type = "float", offset = 0x1C}
        }
    },
    netgameFlagsCount = {type = "byte", offset = 0x378},
    netgameFlagsList = {
        type = "table",
        offset = 0x37C,
        jump = 0x94,
        rows = {
            x = {type = "float", offset = 0x0},
            y = {type = "float", offset = 0x4},
            z = {type = "float", offset = 0x8},
            rotation = {type = "float", offset = 0xC},
            type = {type = "byte", offset = 0x10},
            teamIndex = {type = "word", offset = 0x12}
        }
    },
    netgameEquipmentCount = {type = "byte", offset = 0x384},
    netgameEquipmentList = {
        type = "table",
        offset = 0x388,
        jump = 0x90,
        rows = {
            levitate = {type = "bit", offset = 0x0, bitLevel = 0},
            type1 = {type = "word", offset = 0x4},
            type2 = {type = "word", offset = 0x6},
            type3 = {type = "word", offset = 0x8},
            type4 = {type = "word", offset = 0xA},
            teamIndex = {type = "byte", offset = 0xC},
            spawnTime = {type = "word", offset = 0xE},
            x = {type = "float", offset = 0x40},
            y = {type = "float", offset = 0x44},
            z = {type = "float", offset = 0x48},
            facing = {type = "float", offset = 0x4C},
            itemCollection = {type = "dword", offset = 0x5C}
        }
    },
    objectNamesCount = {type = "dword", offset = 0x204},
    objectNames = {
        type = "list",
        offset = 0x208,
        elementsType = "string",
        jump = 36,
        noOffset = true
    },
    sceneriesCount = {type = "dword", offset = 0x210},
    sceneries = {
        type = "table",
        offset = 0x214,
        jump = 0x48,
        rows = {
            typeIndex = {type = "word", offset = 0x0},
            nameIndex = {type = "word", offset = 0x2},
            notPlaced = {type = "bit", offset = 0x4, bitLevel = 0},
            desiredPermutation = {type = "byte", offset = 0x6},
            x = {type = "float", offset = 0x8},
            y = {type = "float", offset = 0xC},
            z = {type = "float", offset = 0x10},
            yaw = {type = "float", offset = 0x14},
            pitch = {type = "float", offset = 0x18},
            roll = {type = "float", offset = 0x1C}
        }
    },
    bipedsCount = {type = "dword", offset = 0x228},
    bipeds = {
        type = "table",
        offset = 0x0228 + 0x4,
        jump = 0x78,
        rows = {
            typeIndex = {type = "word", offset = 0x0},
            nameIndex = {type = "word", offset = 0x2},
            notPlaced = {type = "bit", offset = 0x4, bitLevel = 0},
            desiredPermutation = {type = "byte", offset = 0x6},
            x = {type = "float", offset = 0x8},
            y = {type = "float", offset = 0xC},
            z = {type = "float", offset = 0x10},
            yaw = {type = "float", offset = 0x14},
            pitch = {type = "float", offset = 0x18},
            roll = {type = "float", offset = 0x1C}
        }
    },
    bipedPaletteCount = {type = "byte", offset = 0x0234},
    bipedPaletteList = {type = "list", offset = 0x0238, elementsType = "dword", jump = 0x30},
    cutsceneFlagsCount = {type = "dword", offset = 0x4E4},
    cutsceneFlags = {
        type = "table",
        offset = 0x4E8,
        jump = 92,
        rows = {
            name = {type = "string", offset = 0x4},
            x = {type = "float", offset = 0x24},
            y = {type = "float", offset = 0x28},
            z = {type = "float", offset = 0x2C},
            vX = {type = "float", offset = 0x30},
            vY = {type = "float", offset = 0x34}
        }
    },
    actorPaletteCount = {type = "dword", offset = 0x0420},
    encounterPaletteCount = {type = "dword", offset = 0x042C}
}

---@class scenery
---@field model number
---@field modifierShader number

-- Scenery structure
local sceneryStructure = {
    model = {type = "word", offset = 0x28 + 0xC},
    modifierShader = {type = "word", offset = 0x90 + 0xC}
}

---@class collisionGeometry
---@field vertexCount number Number of vertex in the collision geometry
---@field vertexList table List of vertex in the collision geometry

-- Collision Model structure
local collisionGeometryStructure = {
    vertexCount = {type = "byte", offset = 0x408},
    vertexList = {
        type = "table",
        offset = 0x40C,
        jump = 0x10,
        rows = {
            x = {type = "float", offset = 0x0},
            y = {type = "float", offset = 0x4},
            z = {type = "float", offset = 0x8}
        }
    }
}

---@class animationClass
---@field name string Name of the animation
---@field type number Type of the animation
---@field frameCount number Frame count of the animation
---@field nextAnimation number Next animation id of the animation
---@field sound number Sound id of the animation

---@class modelAnimations
---@field fpAnimationCount number Number of first-person animations
---@field fpAnimationList number[] List of first-person animations
---@field animationCount number Number of animations of the model
---@field animationList animationClass[] List of animations of the model

-- Model Animation structure
local modelAnimationsStructure = {
    fpAnimationCount = {type = "byte", offset = 0x90},
    fpAnimationList = {
        type = "list",
        offset = 0x94,
        noOffset = true,
        elementsType = "byte",
        jump = 0x2
    },
    animationCount = {type = "byte", offset = 0x74},
    animationList = {
        type = "table",
        offset = 0x78,
        jump = 0xB4,
        rows = {
            name = {type = "string", offset = 0x0},
            type = {type = "word", offset = 0x20},
            frameCount = {type = "byte", offset = 0x22},
            nextAnimation = {type = "byte", offset = 0x38},
            sound = {type = "byte", offset = 0x3C}
        }
    }
}

---@class weapon : blamObject
---@field pressedReloadKey boolean Is weapon trying to reload
---@field isWeaponPunching boolean Is weapon playing melee or grenade animation
---@field ownerObjectId number Object ID of the weapon owner
---@field carrierObjectId number Object ID of the weapon owner
---@field isInInventory boolean Is weapon in inventory
---@field primaryTriggerState number Primary trigger state of the weapon
---@field totalAmmo number Total ammo of the weapon
---@field loadedAmmo number Loaded ammo of the weapon   

local weaponStructure = extendStructure(objectStructure, {
    pressedReloadKey = {type = "bit", offset = 0x230, bitLevel = 3},
    isWeaponPunching = {type = "bit", offset = 0x230, bitLevel = 4},
    ownerObjectId = {type = "dword", offset = 0x11C}, -- deprecated
    carrierObjectId = {type = "dword", offset = 0x11C},
    isInInventory = {type = "bit", offset = 0x1F4, bitLevel = 0},
    primaryTriggerState = {type = "byte", offset = 0x261},
    totalAmmo = {type = "word", offset = 0x2B6},
    loadedAmmo = {type = "word", offset = 0x2B8}
})

---@class weaponTag
---@field model number Tag ID of the weapon model

-- Weapon structure
local weaponTagStructure = {model = {type = "dword", offset = 0x34}}

-- @class modelMarkers
-- @field name string
-- @field nodeIndex number
-- TODO Add rotation fields, check Guerilla tag
-- @field x number
-- @field y number
-- @field z number

---@class modelPermutation
---@field name string

---@class modelRegion
---@field name string
---@field permutationCount number
---@field permutationsList modelPermutation[]

---@class modelNode
---@field x number
---@field y number
---@field z number

---@class gbxModel
---@field nodeCount number Number of nodes
---@field nodeList modelNode[] List of the model nodes
---@field regionCount number Number of regions
---@field regionList modelRegion[] List of regions

-- Model structure
local modelStructure = {
    nodeCount = {type = "dword", offset = 0xB8},
    nodeList = {
        type = "table",
        offset = 0xBC,
        jump = 0x9C,
        rows = {
            x = {type = "float", offset = 0x28},
            y = {type = "float", offset = 0x2C},
            z = {type = "float", offset = 0x30}
        }
    },
    regionCount = {type = "dword", offset = 0xC4},
    regionList = {
        type = "table",
        offset = 0xC8,
        -- jump = 0x50,
        jump = 0x4C,
        rows = {
            name = {type = "string", offset = 0x0},
            permutationCount = {type = "dword", offset = 0x40},
            permutationsList = {
                type = "table",
                offset = 0x44,
                jump = 0x58,
                rows = {
                    name = {type = "string", offset = 0x0}
                    -- markersList = {
                    --    type = "table",
                    --    offset = 0x4C,
                    --    jump = 0x0,
                    --    rows = {
                    --        name = {type = "string", offset = 0x0},
                    --        nodeIndex = {type = "word", offset = 0x20}
                    --    }
                    -- }
                }
            }
        }
    }
}

---@class projectile : blamObject
---@field action number Enumeration of denotation action
---@field attachedToObjectId number Id of the attached object
---@field armingTimer number PENDING
---@field xVel number Velocity in x direction
---@field yVel number Velocity in y direction
---@field zVel number Velocity in z direction
---@field yaw number Rotation in yaw direction
---@field pitch number Rotation in pitch direction
---@field roll number Rotation in roll direction

-- Projectile structure
local projectileStructure = extendStructure(objectStructure, {
    action = {type = "word", offset = 0x230},
    attachedToObjectId = {type = "dword", offset = 0x11C},
    armingTimer = {type = "float", offset = 0x248},
    --[[xVel = {type = "float", offset = 0x254},
    yVel = {type = "float", offset = 0x258},
    zVel = {type = "float", offset = 0x25C},]]
    pitch = {type = "float", offset = 0x264},
    yaw = {type = "float", offset = 0x268},
    roll = {type = "float", offset = 0x26C}
})

---@class player
---@field id number Get playerId of this player
---@field host number Check if player is host, 0 when host, null when not
---@field name string Name of this player
---@field team number Team color of this player, 0 when red, 1 when on blue team
---@field interactionObjectId number Object ID of the object this player is interacting with
---@field interactonObjectType number Type of the object this player is interacting with
---@field interactionObjectSeat number Seat of the object this player is interacting with
---@field respawnTime number Time in ticks until this player respawns
---@field respawnGrowthTime number Time in ticks until this player respawns
---@field objectId number Return the objectId associated to this player
---@field lastObjectId number Return the last objectId associated to this player
---@field lastFireTime number Return the last fire time associated to this player
---@field name2 string Name of this player
---@field color number Color of the player, only works on "Free for All" gametypes
---@field machineIndex number Machine index of this player
---@field controllerIndex number Controller index of this player
---@field team2 number Team color of this player, 0 when red, 1 when on blue team
---@field index number Local index of this player 0-15
---@field invisibilityTime number Time in ticks until this player is invisible
---@field speed number Current speed of this player
---@field teleporterFlagId number Unknown
---@field objectiveMode number Unknown
---@field objectivePlayerId number Unknown
---@field targetPlayerId number Player id the player is looking at
---@field targetTime number Some timer for fading in the name of the player being looked at
---@field lastDeathTime number Time in ticks since this player last died
---@field slayerTargetPlayerId number Unknown
---@field oddManOut number Is player odd man out
---@field killStreak number Current kill streak of this player
---@field multiKill number Current multi kill of this player
---@field lastKillTime number Time in ticks since this player last killed
---@field kills number Kills quantity done by this player
---@field ping number Ping amount from server of this player in milliseconds
---@field assists number Assists count of this player
---@field betraysAndSuicides number Betrays plus suicides count of this player
---@field deaths number Deaths count of this player
---@field suicides number Suicides count of this player

local playerStructure = {
    id = {type = "word", offset = 0x0},
    host = {type = "word", offset = 0x2},
    name = {type = "ustring", forced = true, offset = 0x4},
    unknown = {type = "byte", offset = 0x1C},
    team = {type = "byte", offset = 0x20},
    unknown2 = {type = "byte", offset = 0x21},
    unknown3 = {type = "byte", offset = 0x22},
    unknown4 = {type = "byte", offset = 0x23},
    interactionObjectId = {type = "dword", offset = 0x24},
    interactionObjectType = {type = "word", offset = 0x28},
    interactionObjectSeat = {type = "word", offset = 0x2A},
    respawnTime = {type = "dword", offset = 0x2C},
    respawnGrowthTime = {type = "dword", offset = 0x30},
    objectId = {type = "dword", offset = 0x34},
    lastObjectId = {type = "dword", offset = 0x38},
    unknown5 = {type = "dword", offset = 0x3C},
    unknown6 = {type = "dword", offset = 0x40},
    lastFireTime = {type = "dword", offset = 0x44},
    name2 = {type = "ustring", forced = true, offset = 0x48},
    color = {type = "word", offset = 0x60},
    unknown7 = {type = "word", offset = 0x62},
    machineIndex = {type = "byte", offset = 0x64},
    controllerIndex = {type = "byte", offset = 0x65},
    team2 = {type = "byte", offset = 0x66},
    index = {type = "byte", offset = 0x67},
    invisibilityTime = {type = "word", offset = 0x68},
    unknown8 = {type = "word", offset = 0x6A},
    speed = {type = "float", offset = 0x6C},
    teleporterFlagId = {type = "dword", offset = 0x70}, -- Unknown
    objectiveMode = {type = "dword", offset = 0x74}, -- Unknown
    objectivePlayerId = {type = "dword", offset = 0x78}, -- Unknown
    targetPlayerId = {type = "dword", offset = 0x7C}, -- Player id the player is looking at?
    targetTime = {type = "dword", offset = 0x80}, -- Some timer for fading in the name of the player being looked at?
    lastDeathTime = {type = "dword", offset = 0x84},
    slayerTargetPlayerId = {type = "dword", offset = 0x88},
    oddManOut = {type = "dword", offset = 0x8C}, -- Player is odd man out
    unknown9 = {type = "dword", offset = 0x90},
    unknown10 = {type = "word", offset = 0x94},
    killStreak = {type = "word", offset = 0x96},
    multiKill = {type = "word", offset = 0x98},
    lastKillTime = {type = "word", offset = 0x9A},
    kills = {type = "word", offset = 0x9C},
    ping = {type = "dword", offset = 0xDC},
    assists = {type = "word", offset = 0XA4},
    betraysAndSuicides = {type = "word", offset = 0xAC},
    deaths = {type = "word", offset = 0xAE},
    suicides = {type = "word", offset = 0XB0},
    --[[
        Appears to be some kind of tick or packet counter, when defined to specific value it will
        cause the player to desync and show the "connection problems icon"
        Counts up to 31 and then resets to 0
    ]]
    unknownTimer1 = {type = "dword", offset = 0xE8}
}

---@class firstPersonInterface
---@field firstPersonHands number

---@class multiplayerInformation
---@field flag number Tag ID of the flag object used for multiplayer games
---@field unit number Tag ID of the unit object used for multiplayer games

---@class globalsTag
---@field multiplayerInformation multiplayerInformation[]
---@field firstPersonInterface firstPersonInterface[]

local globalsTagStructure = {
    multiplayerInformation = {
        type = "table",
        jump = 0x0,
        offset = 0x168,
        rows = {flag = {type = "dword", offset = 0xC}, unit = {type = "dword", offset = 0x1C}}
    },
    firstPersonInterface = {
        type = "table",
        jump = 0x0,
        offset = 0x180,
        rows = {firstPersonHands = {type = "dword", offset = 0xC}}
    }
}

---@class firstPerson
---@field weaponObjectId number Weapon Id from the first person view

local firstPersonStructure = {weaponObjectId = {type = "dword", offset = 0x10}}

---@class bipedTag
---@field model number Gbxmodel tag Id of this biped tag
---@field disableCollision boolean Disable collision of this biped tag
---@field weaponCount number Number of weapons of this biped
---@field weaponList tagId[] List of weapons of this biped

local bipedTagStructure = {
    model = {type = "dword", offset = 0x34},
    disableCollision = {type = "bit", offset = 0x2F4, bitLevel = 5},
    weaponCount = {type = "byte", offset = 0x02D8},
    weaponList = {type = "list", offset = 0x02D8 + 0x4, jump = 0x24, elementsType = "dword"}

}

---@class deviceMachine : blamObject
---@field powerGroupIndex number Power index from the device groups table
---@field power number Position amount of this device machine
---@field powerChange number Power change of this device machine
---@field positionGroupIndex number Power index from the device groups table
---@field position number Position amount of this device machine
---@field positionChange number Position change of this device machine
local deviceMachineStructure = extendStructure(objectStructure, {
    powerGroupIndex = {type = "word", offset = 0x1F8},
    power = {type = "float", offset = 0x1FC},
    powerChange = {type = "float", offset = 0x200},
    positionGroupIndex = {type = "word", offset = 0x204},
    position = {type = "float", offset = 0x208},
    positionChange = {type = "float", offset = 0x20C}
})

---@class hudGlobals
---@field anchor number
---@field x number
---@field y number
---@field width number
---@field height number
---@field upTime number
---@field fadeTime number
---@field iconColorA number
---@field iconColorR number
---@field iconColorG number
---@field iconColorB number
---@field textSpacing number

local hudGlobalsStructure = {
    anchor = {type = "word", offset = 0x0},
    x = {type = "word", offset = 0x24},
    y = {type = "word", offset = 0x26},
    width = {type = "float", offset = 0x28},
    height = {type = "float", offset = 0x2C},
    upTime = {type = "float", offset = 0x68},
    fadeTime = {type = "float", offset = 0x6C},
    iconColorA = {type = "float", offset = 0x70},
    iconColorR = {type = "float", offset = 0x74},
    iconColorG = {type = "float", offset = 0x78},
    iconColorB = {type = "float", offset = 0x7C},
    textColorA = {type = "float", offset = 0x80},
    textColorR = {type = "float", offset = 0x84},
    textColorG = {type = "float", offset = 0x88},
    textColorB = {type = "float", offset = 0x8C},
    textSpacing = {type = "float", offset = 0x90}
}

---@class cinematicGlobals
---@field isInProgress boolean
---@field isShowingLetterbox boolean

local cinematicGlobalsStructure = {
    isInProgress = {type = "bit", offset = 0x9, bitLevel = 0},
    isShowingLetterbox = {type = "bit", offset = 0x8, bitLevel = 0}
}

------------------------------------------------------------------------------
-- LuaBlam globals
------------------------------------------------------------------------------

-- Provide with public blam! data tables
blam.addressList = addressList
blam.tagClasses = tagClasses
blam.objectClasses = objectClasses
blam.joystickInputs = joystickInputs
blam.dPadValues = dPadValues
blam.cameraTypes = cameraTypes
blam.consoleColors = consoleColors
blam.netgameFlagClasses = netgameFlagClasses
blam.gameTypeClasses = gameTypeClasses
blam.multiplayerTeamClasses = multiplayerTeamClasses
blam.unitTeamClasses = unitTeamClasses
blam.objectNetworkRoleClasses = objectNetworkRoleClasses

---@class tagDataHeader
---@field array any
---@field scenario string
---@field count number

---@type tagDataHeader
blam.tagDataHeader = createBindTable(addressList.tagDataHeader, tagDataHeaderStructure)

------------------------------------------------------------------------------
-- LuaBlam API
------------------------------------------------------------------------------

-- Add utilities to library
blam.dumpObject = dumpObject
blam.consoleOutput = consoleOutput
blam.null = null

---Get the current game camera type
---@return number?
function blam.getCameraType()
    local camera = read_word(addressList.cameraType)
    if camera then
        if camera == 22192 then
            return cameraTypes.scripted
        elseif camera == 30400 then
            return cameraTypes.firstPerson
        elseif camera == 30704 then
            return cameraTypes.devcam
            -- FIXME Validate this value, it seems to be wrong!
        elseif camera == 21952 then
            return cameraTypes.thirdPerson
        elseif camera == 23776 then
            return cameraTypes.deadCamera
        end
    end
    return nil
end

--- Get input from joystick assigned in the game
---@param joystickOffset number Offset input from the joystick data, use blam.joystickInputs
---@return boolean | number Value of the joystick input
function blam.getJoystickInput(joystickOffset)
    -- Based on aLTis controller method
    -- TODO Check if it is better to return an entire table with all input values 
    joystickOffset = joystickOffset or 0
    -- Nothing is pressed by default
    ---@type boolean | number
    local inputValue = 0
    -- Look for every input from every joystick available
    for controllerId = 0, 3 do
        local inputAddress = addressList.joystickInput + controllerId * 0xA0
        if (joystickOffset >= 30 and joystickOffset <= 38) then
            -- Sticks
            inputValue = inputValue + read_long(inputAddress + joystickOffset)
        elseif (joystickOffset > 96) then
            -- D-pad related
            local tempValue = read_word(inputAddress + 96)
            if (tempValue == joystickOffset - 100) then
                inputValue = true
            else
                inputValue = false
            end
        else
            inputValue = inputValue + read_byte(inputAddress + joystickOffset)
        end
    end
    return inputValue
end

--- Create a tag object from a given address, this object can't write data to game memory
---@param address integer
---@return tag?
function blam.tag(address)
    if (address and address ~= 0) then
        -- Generate a new tag object from class
        local tag = createBindTable(address, tagHeaderStructure)

        -- Get all the tag info
        local tagInfo = dumpObject(tag)

        -- Set up values
        tagInfo.address = address
        tagInfo.path = read_string(tagInfo.path)
        -- TODO Optimize this function
        -- Also review class prop type as it seems we are transforming it to a string but it is 
        -- a number in the binded structure
        tagInfo.class = tagClassFromInt(tagInfo.class --[[@as number]] )

        return tagInfo
    end
    return nil
end

--- Return a tag object given tagPath and tagClass or just tagId
---@param tagIdOrTagPath string | number
---@param tagClass? string
---@return tag?
function blam.getTag(tagIdOrTagPath, tagClass, ...)
    local tagId
    local tagPath

    -- Get arguments from table
    if (isNumber(tagIdOrTagPath)) then
        tagId = tagIdOrTagPath
    elseif (isString(tagIdOrTagPath)) then
        tagPath = tagIdOrTagPath
    elseif (not tagIdOrTagPath) then
        return nil
    end

    if (...) then
        consoleOutput(debug.traceback("Wrong number of arguments on get tag function", 2),
                      consoleColors.error)
    end

    local tagAddress

    -- Get tag address
    if (tagId) then
        if (tagId < 0xFFFF) then
            -- Calculate tag id
            tagId = read_dword(blam.tagDataHeader.array + (tagId * 0x20 + 0xC))
        end
        tagAddress = get_tag(tagId)
    elseif (tagClass and tagPath) then
        tagAddress = get_tag(tagClass, tagPath --[[@as string]] )
    end

    if tagAddress then
        return blam.tag(tagAddress)
    end
end

--- Create a player object given player entry table address
---@param address? number
---@return player?
function blam.player(address)
    if address and isValid(address) then
        return createBindTable(address, playerStructure)
    end
    return nil
end

--- Create a blamObject given address
---@param address? number
---@return blamObject?
function blam.object(address)
    if address and isValid(address) then
        return createBindTable(address, objectStructure)
    end
    return nil
end

--- Create a Projectile object given address
---@param address? number
---@return projectile?
function blam.projectile(address)
    if address and isValid(address) then
        return createBindTable(address, projectileStructure)
    end
    return nil
end

--- Create a Unit object from a given address
---@param address? number
---@return unit?
function blam.unit(address)
    if address and isValid(address) then
        return createBindTable(address, unitStructure)
    end
    return nil
end

--- Create a Biped object from a given address
---@param address? number
---@return biped?
function blam.biped(address)
    if address and isValid(address) then
        return createBindTable(address, bipedStructure)
    end
    return nil
end

--- Create a Vehicle object from a given address
---@param address? number
---@return vehicle?
function blam.vehicle(address)
    if address and isValid(address) then
        return createBindTable(address, vehicleStructure)
    end
    return nil
end

--- Create a biped tag from a tag path or id
---@param tag string | number
---@return bipedTag?
function blam.bipedTag(tag)
    if isValid(tag) then
        local bipedTag = blam.getTag(tag, tagClasses.biped)
        if (bipedTag) then
            return createBindTable(bipedTag.data, bipedTagStructure)
        end
    end
    return nil
end

--- Create a Unicode String List object from a tag path or id
---@param tag string | number
---@return unicodeStringList?
function blam.unicodeStringList(tag)
    if isValid(tag) then
        local unicodeStringListTag = blam.getTag(tag, tagClasses.unicodeStringList)
        if (unicodeStringListTag) then
            return createBindTable(unicodeStringListTag.data, unicodeStringListStructure)
        end
    end
    return nil
end

--- Create a bitmap object from a tag path or id
---@param tag string | number
---@return bitmap?
function blam.bitmap(tag)
    if isValid(tag) then
        local bitmapTag = blam.getTag(tag, tagClasses.bitmap)
        if (bitmapTag) then
            return createBindTable(bitmapTag.data, bitmapStructure)
        end
    end
end

--- Create a UI Widget Definition object from a tag path or id
---@param tag string | number
---@return uiWidgetDefinition?
function blam.uiWidgetDefinition(tag)
    if isValid(tag) then
        local uiWidgetDefinitionTag = blam.getTag(tag, tagClasses.uiWidgetDefinition)
        if (uiWidgetDefinitionTag) then
            return createBindTable(uiWidgetDefinitionTag.data, uiWidgetDefinitionStructure)
        end
    end
    return nil
end

--- Create a UI Widget Collection object from a tag path or id
---@param tag string | number
---@return uiWidgetCollection?
function blam.uiWidgetCollection(tag)
    if isValid(tag) then
        local uiWidgetCollectionTag = blam.getTag(tag, tagClasses.uiWidgetCollection)
        if (uiWidgetCollectionTag) then
            return createBindTable(uiWidgetCollectionTag.data, uiWidgetCollectionStructure)
        end
    end
    return nil
end

--- Create a Tag Collection object from a tag path or id
---@param tag string | number
---@return tagCollection?
function blam.tagCollection(tag)
    if isValid(tag) then
        local tagCollectionTag = blam.getTag(tag, tagClasses.tagCollection)
        if (tagCollectionTag) then
            return createBindTable(tagCollectionTag.data, tagCollectionStructure)
        end
    end
    return nil
end

--- Create a Weapon HUD Interface object from a tag path or id
---@param tag string | number
---@return weaponHudInterface?
function blam.weaponHudInterface(tag)
    if isValid(tag) then
        local weaponHudInterfaceTag = blam.getTag(tag, tagClasses.weaponHudInterface)
        if (weaponHudInterfaceTag) then
            return createBindTable(weaponHudInterfaceTag.data, weaponHudInterfaceStructure)
        end
    end
    return nil
end

--- Create a Scenario object from a tag path or id
---@param tag? string | number
---@return scenario?
function blam.scenario(tag)
    local scenarioTag = blam.getTag(tag or 0, tagClasses.scenario)
    if (scenarioTag) then
        return createBindTable(scenarioTag.data, scenarioStructure)
    end
end

--- Create a Scenery object from a tag path or id
---@param tag string | number
---@return scenery?
function blam.scenery(tag)
    if isValid(tag) then
        local sceneryTag = blam.getTag(tag, tagClasses.scenery)
        if (sceneryTag) then
            return createBindTable(sceneryTag.data, sceneryStructure)
        end
    end
    return nil
end

--- Create a Collision Geometry object from a tag path or id
---@param tag string | number
---@return collisionGeometry?
function blam.collisionGeometry(tag)
    if isValid(tag) then
        local collisionGeometryTag = blam.getTag(tag, tagClasses.collisionGeometry)
        if (collisionGeometryTag) then
            return createBindTable(collisionGeometryTag.data, collisionGeometryStructure)
        end
    end
    return nil
end

--- Create a Model Animations object from a tag path or id
---@param tag string | number
---@return modelAnimations?
function blam.modelAnimations(tag)
    if isValid(tag) then
        local modelAnimationsTag = blam.getTag(tag, tagClasses.modelAnimations)
        if (modelAnimationsTag) then
            return createBindTable(modelAnimationsTag.data, modelAnimationsStructure)
        end
    end
    return nil
end

--- Create a Weapon object from the given object address
---@param address? number
---@return weapon?
function blam.weapon(address)
    if address and isValid(address) then
        return createBindTable(address, weaponStructure)
    end
    return nil
end

--- Create a Weapon tag object from a tag path or id
---@param tag string | number
---@return weaponTag?
function blam.weaponTag(tag)
    if isValid(tag) then
        local weaponTag = blam.getTag(tag, tagClasses.weapon)
        if (weaponTag) then
            return createBindTable(weaponTag.data, weaponTagStructure)
        end
    end
    return nil
end

--- Create a model (gbxmodel) object from a tag path or id
---@param tag string | number
---@return gbxModel?
function blam.model(tag)
    if isValid(tag) then
        local modelTag = blam.getTag(tag, tagClasses.model)
        if (modelTag) then
            return createBindTable(modelTag.data, modelStructure)
        end
    end
    return nil
end
-- Alias
blam.gbxmodel = blam.model

--- Create a Globals tag object from a tag path or id, default globals path by default
---@param tag? string | number
---@return globalsTag?
function blam.globalsTag(tag)
    local tag = tag or "globals\\globals"
    if isValid(tag) then
        local globalsTag = blam.getTag(tag, tagClasses.globals)
        if (globalsTag) then
            return createBindTable(globalsTag.data, globalsTagStructure)
        end
    end
    return nil
end

--- Create a First person object from a given address, game known address by default
---@param address? number
---@return firstPerson
function blam.firstPerson(address)
    return createBindTable(address or addressList.firstPerson, firstPersonStructure)
end

--- Create a Device Machine object from a given address
---@param address? number
---@return deviceMachine?
function blam.deviceMachine(address)
    if address and isValid(address) then
        return createBindTable(address, deviceMachineStructure)
    end
    return nil
end

--- Create a HUD Globals tag object from a given address
---@param tag string | number
---@return hudGlobals?
function blam.hudGlobals(tag)
    if isValid(tag) then
        local hudGlobals = blam.getTag(tag, tagClasses.hudGlobals)
        if (hudGlobals) then
            return createBindTable(hudGlobals.data, hudGlobalsStructure)
        end
    end
    return nil
end

--- Return a blam object given object index or id.
--- Also returns objectId when given an object index.
---@param idOrIndex number
---@return blamObject?, number?
function blam.getObject(idOrIndex)
    local objectId
    local objectAddress

    -- Get object address
    if (idOrIndex) then
        -- Get object ID
        if (idOrIndex < 0xFFFF) then
            local index = idOrIndex

            -- Get objects table
            local table = createBindTable(addressList.objectTable, dataTableStructure)
            if (index > table.capacity) then
                return nil
            end

            -- Calculate object ID (this may be invalid, be careful)
            objectId =
                (read_word(table.firstElementAddress + index * table.elementSize) * 0x10000) + index
        else
            objectId = idOrIndex
        end

        objectAddress = get_object(objectId)

        if objectAddress then
            return blam.object(objectAddress), objectId
        end
    end
    return nil
end

--- Return an element from the device machines table
---@param index number
---@return number?
function blam.getDeviceGroup(index)
    -- Get object address
    if index then
        -- Get objects table
        local table = createBindTable(read_dword(addressList.deviceGroupsTable),
                                      deviceGroupsTableStructure)
        -- Calculate object ID (this may be invalid, be careful)
        local itemOffset = table.elementSize * index
        local item = read_float(table.firstElementAddress + itemOffset + 0x4)
        return item
    end
    return nil
end

local syncedObjectsTable = {
    maximumObjectsCount = {type = "dword", offset = 0x0},
    initialized = {type = "byte", offset = 0xC},
    objectsCount = {type = "dword", offset = 0x18},
    firstElementAddress = {type = "dword", offset = 0x28}
}

local function getSyncedObjectsTable()
    local tableAddress
    if blam.isGameSAPP() then
        tableAddress = addressList.syncedNetworkObjects
    else
        tableAddress = read_dword(addressList.syncedNetworkObjects)
        if tableAddress == 0 then
            console_out("Synced objects table is not accesible yet.")
            return nil
        end
    end

    return createBindTable(tableAddress, syncedObjectsTable)
end

--- Return the maximum allowed network objects count
---@return number
function blam.getMaximumNetworkObjects()
    local syncedObjectsTable = getSyncedObjectsTable()
    if not syncedObjectsTable then
        return engineConstants.defaultNetworkObjectsCount
    end

    -- For some reason fist element entry is always used, so we need to substract 1
    return syncedObjectsTable.maximumObjectsCount - 1
end

--- Return an element from the synced objects table
---@param index number
---@return number?
function blam.getObjectIdBySyncedIndex(index)
    if index then
        local syncedObjectsTable = getSyncedObjectsTable()
        if not syncedObjectsTable then
            return nil
        end

        if syncedObjectsTable.objectsCount == 0 then
            return nil
        end
        if not syncedObjectsTable.initialized == 1 then
            return nil
        end
        -- For some reason fist element entry is always used, so we need to substract 1
        if index >= syncedObjectsTable.maximumObjectsCount - 1 then
            return nil
        end

        local entryOffset = 4 * index
        -- Ignore first entry, it's always used so add 4 bytes offset
        local entryAddress = syncedObjectsTable.firstElementAddress + entryOffset + 0x4
        local objectId = read_dword(entryAddress)
        if blam.isNull(objectId) then
            return nil
        end
        return objectId
    end
    return nil
end

---@class blamRequest
---@field requestString string
---@field timeout number
---@field callback function<boolean, string>
---@field sentAt number

local rconEvents = {}
local maxRconDataLength = 60

blam.rcon = {}

---Define a request event callback
---@param eventName string
---@param callback fun(message?: string, playerIndex?: number): string?
function blam.rcon.event(eventName, callback)
    rconEvents[eventName:lower()] = callback
end

---Dispatch an rcon event to a client or server trough rcon.
---
--- As a client, you can only send messages to the server.
---
--- As a server, you can send messages to a specific client or all clients.
---@param eventName string Path or name of the resource we want to get
---@param message? string Message to send to the server
---@param playerIndex? number Player index to send the message to
---@overload fun(eventName: string, playerIndex: number)
---@return {callback: fun(callback: fun(response: string, playerIndex?: number))}
function blam.rcon.dispatch(eventName, message, playerIndex)
    -- if server_type ~= "dedicated" then
    --    console_out("Warning, requests only work while connected to a dedicated server.")
    -- end
    assert(eventName ~= nil, "Event must not be empty")
    assert(type(eventName) == "string", "Event must be a string")
    local message = message
    local playerIndex = playerIndex
    if message and type(message) == "number" then
        playerIndex = message
        message = nil
    end
    if eventName then
        if blam.isGameSAPP() then
            if playerIndex then
                rprint(playerIndex, ("?%s?%s"):format(eventName, message))
            else
                for i = 1, 16 do
                    rprint(i, ("?%s?%s"):format(eventName, message))
                end
            end
        else
            local request = ("?%s?%s"):format(eventName, message)
            assert(#request <= maxRconDataLength, "Rcon request is too long")
            if blam.isGameDedicated() then
                execute_script("rcon blam " .. request)
            else
                blam.rcon.handle(request)
            end
        end
        return {
            callback = function()
                blam.rcon.event(eventName .. "+", callback)
            end
        }
    end
    error("No event name provided")
end

---Evaluate rcon event and handle it as a request
---@param data string
---@param password? string
---@param playerIndex? number
---@return boolean | nil
function blam.rcon.handle(data, password, playerIndex)
    if data:sub(1, 1) == "?" then
        if blam.isGameSAPP() then
            if password ~= "blam" then
                return nil
            end
        end
        local data = split(data, "?")
        local eventName = data[2]
        local message = data[3]
        local event = rconEvents[eventName:lower()]
        if event then
            local response = event(message, playerIndex)
            if response then
                if blam.isGameSAPP() then
                    rprint(playerIndex, response)
                else
                    execute_script(("rcon blam ?%s?%s"):format(eventName .. "+", response))
                end
            end
            return false
        else
            error("No rcon event handler for " .. eventName)
        end
    end
    -- Pass request to the server
    return nil
end

local passwordAddress
local failMessageAddress

---Patch rcon server function to avoid failed rcon messages
function blam.rcon.patch()
    passwordAddress = read_dword(sig_scan("7740BA??????008D9B000000008A01") + 0x3)
    failMessageAddress = read_dword(sig_scan("B8????????E8??000000A1????????55") + 0x1)
    if passwordAddress and failMessageAddress then
        -- Remove "rcon command failure" message
        safe_write(true)
        write_byte(failMessageAddress, 0x0)
        safe_write(false)
        -- Read current rcon in the server
        local serverRcon = read_string(passwordAddress)
        if serverRcon then
            console_out("Server rcon password is: \"" .. serverRcon .. "\"")
        else
            console_out("Error, at getting server rcon, please set and enable rcon on the server.")
        end
    else
        console_out("Error, at obtaining rcon patches, please check SAPP version.")
    end
end

---Unpatch rcon server function to restore failed rcon messages
function blam.rcon.unpatch()
    if failMessageAddress then
        -- Restore "rcon command failure" message
        safe_write(true)
        write_byte(failMessageAddress, 0x72)
        safe_write(false)
    end
end

--- Find the path, index and id of a tag given partial tag path and tag type
---@param partialTagPath string
---@param searchTagType string
---@return tag? tag
function blam.findTag(partialTagPath, searchTagType)
    for tagIndex = 0, blam.tagDataHeader.count - 1 do
        local tag = blam.getTag(tagIndex)
        if (tag and tag.path:find(partialTagPath, 1, true) and tag.class == searchTagType) then
            return tag
        end
    end
    return nil
end

--- Find the path, index and id of a list of tags given partial tag path and tag type
---@param partialTagPath string
---@param searchTagType string
---@return tag[] tag
function blam.findTagsList(partialTagPath, searchTagType)
    local tagsList
    for tagIndex = 0, blam.tagDataHeader.count - 1 do
        local tag = blam.getTag(tagIndex)
        if (tag and tag.path:find(partialTagPath, 1, true) and tag.class == searchTagType) then
            if not tagsList then
                tagsList = {}
            end
            tagsList[#tagsList + 1] = tag
        end
    end
    return tagsList
end

--- Return the index of an id number
---@param id number
function blam.getIndexById(id)
    if id then
        return fmod(id, 0x10000)
    end
    return nil
end

---@class vector2D
---@field x number
---@field y number

---@class vector3D
---@field x number
---@field y number
---@field z number

---@class vector4D
---@field x number
---@field y number
---@field z number
---@field w number

---Returns game rotation vectors from euler angles, return optional rotation matrix, based on
---[source.](https://www.mecademic.com/en/how-is-orientation-in-space-represented-with-euler-angles)
--- @param yaw number
--- @param pitch number
--- @param roll number
--- @return vector3D, vector3D
local function eulerAnglesToVectors(yaw, pitch, roll)
    local yaw = rad(yaw)
    local pitch = rad(-pitch) -- Negative pitch due to Sapien handling anticlockwise pitch
    local roll = rad(roll)
    local matrix = {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}

    -- Roll, Pitch, Yaw = a, b, y
    local cosA = cos(roll)
    local sinA = sin(roll)
    local cosB = cos(pitch)
    local sinB = sin(pitch)
    local cosY = cos(yaw)
    local sinY = sin(yaw)

    matrix[1][1] = cosB * cosY
    matrix[1][2] = -cosB * sinY
    matrix[1][3] = sinB
    matrix[2][1] = cosA * sinY + sinA * sinB * cosY
    matrix[2][2] = cosA * cosY - sinA * sinB * sinY
    matrix[2][3] = -sinA * cosB
    matrix[3][1] = sinA * sinY - cosA * sinB * cosY
    matrix[3][2] = sinA * cosY + cosA * sinB * sinY
    matrix[3][3] = cosA * cosB

    local v1 = {x = matrix[1][1], y = matrix[2][1], z = matrix[3][1]}
    local v2 = {x = matrix[1][3], y = matrix[2][3], z = matrix[3][3]}

    return v1, v2
end

--- Get euler angles rotation from game rotation vectors
--- @param v1 vector3D Vector with first column values from rotation matrix
--- @param v2 vector3D Vector with third column values from rotation matrix
--- @return number yaw, number pitch, number roll
local function vectorsToEulerAngles(v1, v2)
    local v3 = {
        x = v1.y * v2.z - v1.z * v2.y,
        y = v1.z * v2.x - v1.x * v2.z,
        z = v1.x * v2.y - v1.y * v2.x
    }

    local matrix = {{v1.x, v3.x, v2.x}, {v1.y, v3.y, v2.y}, {v1.z, v3.z, v2.z}}

    -- Extract individual matrix elements
    local m11, m12, m13 = matrix[1][1], matrix[1][2], matrix[1][3]
    local m21, m22, m23 = matrix[2][1], matrix[2][2], matrix[2][3]
    local m31, m32, m33 = matrix[3][1], matrix[3][2], matrix[3][3]

    -- Calculate yaw (heading) angle
    local yaw = atan2(m12, m11)

    -- Calculate pitch (attitude) angle
    local pitch = atan2(-m13, sqrt(m23 ^ 2 + m33 ^ 2))

    -- Calculate roll (bank) angle
    local roll = -atan2(m23, m33)

    -- Convert angles from radians to degrees
    yaw = deg(yaw)
    pitch = deg(pitch)
    roll = deg(roll)

    -- Adjust angles to the range [0, 359]
    yaw = fmod(yaw + 360, 360)
    pitch = fmod(pitch + 360, 360)
    roll = fmod(roll + 360, 360)

    return yaw, pitch, roll
end

--- Get rotation angles from game object
---
--- Assuming clockwise rotation and absolute angles from 0 to 360
---@param object blamObject
---@return number yaw, number pitch, number roll
function blam.getObjectRotation(object)
    local v1 = {x = object.vX, y = object.vY, z = object.vZ}
    local v2 = {x = object.v2X, y = object.v2Y, z = object.v2Z}
    return vectorsToEulerAngles(v1, v2)
end

--- Get rotation angles from game vectors
---
--- Assuming clockwise rotation and absolute angles from 0 to 360
---@param v1 vector3D
---@param v2 vector3D
---@return number yaw, number pitch, number roll
function blam.getVectorRotation(v1, v2)
    return vectorsToEulerAngles(v1, v2)
end

--- Rotate object into desired angles
---
--- Assuming clockwise rotation and absolute angles from 0 to 360
---@param object blamObject
---@param yaw number
---@param pitch number
---@param roll number
function blam.rotateObject(object, yaw, pitch, roll)
    local v1, v2 = eulerAnglesToVectors(yaw, pitch, roll)
    object.vX = v1.x
    object.vY = v1.y
    object.vZ = v1.z
    object.v2X = v2.x
    object.v2Y = v2.y
    object.v2Z = v2.z
end

--- Get screen resolution
---@return {width: number, height: number, aspectRatio: number}
function blam.getScreenData()
    local height = read_word(addressList.screenResolution)
    local width = read_word(addressList.screenResolution + 0x2)
    return {width = width, height = height, aspectRatio = width / height}
end

--- Get the current game state
---@return {isLayerOpened: boolean, isGamePaused: boolean}
function blam.getGameState()
    return {
        isLayerOpened = read_byte(addressList.gameOnMenus) == 0,
        isGamePaused = read_byte(addressList.gamePaused) == 0
    }
end

--- Get object absolute coordinates
---Returns the absolute coordinates of an object, considering parent object coordinates if any.
---@param object blamObject
---@return vector3D
function blam.getAbsoluteObjectCoordinates(object)
    local coordinates = {x = object.x, y = object.y, z = object.z}
    if not isNull(object.parentObjectId) then
        local parentObject = blam.object(get_object(object.parentObjectId))
        if parentObject then
            coordinates.x = coordinates.x + parentObject.x
            coordinates.y = coordinates.y + parentObject.y
            coordinates.z = coordinates.z + parentObject.z
        end
    end
    return coordinates
end

--- Returns binded table to game cinematic globals
---@return cinematicGlobals
function blam.cinematicGlobals()
    return createBindTable(read_dword(addressList.cinematicGlobals), cinematicGlobalsStructure)
end

return blam

end,

["insurrection.core"] = function()
--------------------
-- Module: 'insurrection.core'
--------------------
local inspect = require "inspect"
local blam = require "blam"
local tagClasses = blam.tagClasses
local json = require "json"
local base64 = require "base64"
local color = require "color"
local balltze = Balltze
local engine = Engine

local mercury = require "insurrection.mercury"
local utils = require "insurrection.utils"

local currentWidgetIdAddress = 0x6B401C
local keyboardInputAddress = 0x64C550
local clientPortAddress = 0x6337F8
local friendlyClientPort = 2305
local profileNameAddress = 0x6ADE22
local profileColorAddress = 0x6ADF3A
local mouseInputAddress = 0x64C73C
local widgetCursorGlobals = 0x499E19

local core = {}

function core.loadMercuryPackages()
    local installedPackages = mercury.getInstalled()
    if (installedPackages) then
        console_out(inspect(installedPackages))
        local serverStringsTag = blam.findTag([[chimera_servers_menu\strings\options]],
                                              tagClasses.unicodeStringList)
        local serverStrings = blam.unicodeStringList(serverStringsTag.id)
        local newServers = serverStrings.stringList
        for stringIndex = 1, serverStrings.count do
            newServers[stringIndex] = " "
        end
        for packageIndex, packageLabel in pairs(table.keys(installedPackages)) do
            local package = installedPackages[packageLabel]
            newServers[packageIndex] = package.name .. " - " .. package.version
        end
        serverStrings.stringList = newServers
    end
end

--- Return the file name of a tag file path
---@param tagPath string
function core.getTagName(tagPath)
    local tagSplit = tagPath:split "\\"
    local tagName = tagSplit[#tagSplit]
    return tagName
end

function core.loadInsurrectionPatches()
    package.loaded["insurrection.version"] = nil
    local scriptVersion = "insurrection-" .. require "insurrection.version"

    -- Setting up version string
    local scriptVersionTag = blam.findTag("insurrection_version_footer",
                                          tagClasses.unicodeStringList)
    if scriptVersionTag then
        local scriptVersionString = blam.unicodeStringList(scriptVersionTag.id)
        if scriptVersionString then
            local strings = scriptVersionString.stringList
            -- Write string version to map tag
            strings[1] = scriptVersion
            scriptVersionString.stringList = strings
        end
        return true
    end
end

function core.saveCredentials(username, password)
    Balltze.filesystem.writeFile("credentials.json", json.encode(
                                     {username = username, password = base64.encode(password)}))
end

function core.loadCredentials()
    local credentialsFile = Balltze.filesystem.readFile("credentials.json")
    if credentialsFile then
        local success, credentials = pcall(json.decode, credentialsFile)
        if success and credentials then
            return credentials.username:trim(), base64.decode(credentials.password):trim()
        end
    end
end

function core.loadSettings()
    local settingsFile = Balltze.filesystem.readFile("settings.json")
    if settingsFile then
        local success, settings = pcall(json.decode, settingsFile)
        if success and settings then
            return settings
        end
    end
end

function core.updateSettings(newSettings)
    local previousSettings = core.loadSettings() or {}
    local settings = table.merge(previousSettings, newSettings)
    core.saveSettings(settings)
end

function core.saveSettings(settings)
    Balltze.filesystem.writeFile("settings.json", json.encode(settings))
end

function core.getRenderedUIWidgetTagId()
    -- TODO BALLTZE MIGRATE Ensure this works when the menu is not OPEN and does not crash
    local rootWidget = engine.userInterface.getRootWidget()
    local isPlayerOnMenu = rootWidget ~= nil
    if isPlayerOnMenu then
        return rootWidget.definitionTagHandle.value
    end
end

--- Get the tag widget of the current ui open in the game
function core.getCurrentUIWidgetTag()
    -- local widgetTagId = core.getRenderedUIWidgetTagId()
    local widget = engine.userInterface.getRootWidget()
    if widget then
        local tag = engine.tag.getTag(widget.definitionTagHandle.value, engine.tag.classes.uiWidgetDefinition)
        assert(tag, "No tag found for widget")
        -- TODO BALLTZE MIGRATE
        return {
            data = tag.data,
            id = widget.definitionTagHandle.value,
            tagPath = tag.path,
            tagClass = tag.primaryClass,
            index = tag.handle.index
        }
    end
    return nil
end

---Attempt to translate a input key code
---@param keyCode integer
---@return string | nil name of the given key code
function core.translateKeycode(keyCode)
    if keyCode == 29 then
        return "backspace"
    elseif keyCode == 72 then
        return "space"
    else
        return nil
    end
end

local capsLock
---Attempt to map keys to a text string
---@param pressedKey string | number | nil
---@param text string
---@return string | nil text Given text with mapped modifications applied
function core.mapKeyToText(pressedKey, text)
    if not pressedKey then
        return text
    end
    if pressedKey == "backspace" then
        return text:sub(1, #text - 1)
    elseif pressedKey == "space" then
        return text .. " "
    elseif pressedKey == "capslock" then
        capsLock = not capsLock
    elseif pressedKey > 31 and pressedKey < 127 and type(pressedKey) == "number" then
        if capsLock then
            return text .. string.char(pressedKey):upper()
        end
        return text .. string.char(pressedKey)
    end
end

function core.getStringFromWidget(widgetTagId)
    local widget = blam.uiWidgetDefinition(widgetTagId)
    assert(widget, "No widget found with tag id " .. widgetTagId)
    local virtualValue = VirtualInputValue[widgetTagId]
    if virtualValue then
        return virtualValue
    end
    local unicodeStrings = blam.unicodeStringList(widget.unicodeStringListTag)
    assert(unicodeStrings, "No unicodeStringList, can't get text from this widget")
    return unicodeStrings.strings[widget.stringListIndex + 1]
end

function core.setStringToWidget(text, widgetTagId, mask)
    local widgetDefinition = blam.uiWidgetDefinition(widgetTagId)
    if widgetDefinition then
        local unicodeStrings = blam.unicodeStringList(widgetDefinition.unicodeStringListTag)
        if unicodeStrings then
            if blam.isNull(unicodeStrings) then
                error("No unicodeStringList, can't assign text to this widget")
            end
            local stringListIndex = widgetDefinition.stringListIndex
            local newStrings = unicodeStrings.strings
            if mask then
                VirtualInputValue[widgetTagId] = text
                newStrings[stringListIndex + 1] = string.rep(mask, #text)
            else
                newStrings[stringListIndex + 1] = text
            end
            unicodeStrings.strings = newStrings
        end
    end
end

---Attempt to connect a game server
---@param host string
---@param port number
---@param password string
function core.connectServer(host, port, password)
    local command = "connect %s:%s \"%s\""
    engine.hsc.executeScript(command:format(host, port, password))
end

function core.getMyGamesHaloCEPath()
    local myGamesPath = read_string(0x00647830)
    return myGamesPath
end

function core.getWidgetValues(widgetTagId)
    if core.getCurrentUIWidgetTag() then
        return engine.userInterface.findWidget(widgetTagId)
    end
end

function core.setWidgetValues(widgetTagId, values)
    local function setValuesDOMSafe()
        -- Verify there is a widget loaded in the DOM
        local widget = engine.userInterface.findWidget(widgetTagId)
        if widget then
            for key, value in pairs(values) do
                if type(value) == "table" then
                    for subKey, subValue in pairs(value) do
                        widget[key][subKey] = subValue
                    end
                else
                    widget[key] = value
                end
            end
            return true
        end
    end
    if not setValuesDOMSafe() then
        -- If there is no widget loaded in the DOM, wait 33ms and try again
        -- (this is a workaround for the DOM not being loaded yet)
        -- TODO BALLTZE MIGRATE
        -- utils.delay(33, function()
        --    setValuesDOMSafe()
        -- end)
    end
end

-- TODO We do not need this, checkout replacements
function core.getWidgetHandle(widgetTagId)
    if core.getCurrentUIWidgetTag() then
        local sucess, widgetHandle = pcall(engine.userInterface.findWidgets, widgetTagId)
        if sucess and widgetHandle then
            return widgetHandle
        end
    end
end

function core.replaceWidgetInDom(widgetTagHandleValue, newWidgetTagHandleValue)
    local replaced, widget = pcall(engine.userInterface.findWidget, widgetTagHandleValue)
    if replaced and widget then
        engine.userInterface.replaceWidget(widget, newWidgetTagHandleValue)
    end
end

---Returns the current screen resolution
---@return number width, number height
function core.getScreenResolution()
    local width = read_word(0x637CF2)
    local height = read_word(0x637CF0)
    return width, height
end

local function isThreadRunning()
    if #Lanes == 0 then
        return false
    end
    return true
end

---Set game in loading state
---@param isLoading boolean
---@param text? string
---@param blockInput? boolean
function core.loading(isLoading, text, blockInput)
    if isLoading then
        -- There is already another thread running, do not modify loading status
        if isThreadRunning() then
            return
        end
        if blockInput then
            -- TODO BALLTZE MIGRATE
            -- harmony.menu.block_input(true)
        end
        LoadingText = text or "Loading..."
        log(LoadingText)
    else
        -- TODO BALLTZE MIGRATE
        -- harmony.menu.block_input(false)
        LoadingText = nil
    end
end

function core.setGameProfileName(name)
    local name = name
    if name then
        -- Limit name to 11 characters
        if #name > 11 then
            name = name:sub(1, 11)
        end
        blam.writeUnicodeString(profileNameAddress, name, true)
    end
    local profileName = blam.readUnicodeString(profileNameAddress, true)
    return profileName
end

---Get reference to any customization object available in the map
---@return number? objectId, table? regionPermutations
function core.getCustomizationObjectId()
    local scenario = blam.scenario(0)
    assert(scenario)
    for objectId, objectIndex in pairs(blam.getObjects()) do
        local object = blam.object(get_object(objectIndex))
        if object and scenario.objectNames[object.nameIndex + 1] == "customization_biped" then
            object.isNotCastingShadow = false
            return objectId
        end
    end
end

function core.getMouseState()
    return {
        right = read_int(mouseInputAddress),
        up = read_int(mouseInputAddress + 4),
        scroll = read_char(mouseInputAddress + 8),
        leftClick = read_byte(mouseInputAddress + 12),
        scrollClick = read_byte(mouseInputAddress + 13),
        rightClick = read_byte(mouseInputAddress + 14)
    }
end

function core.getWidgetCursorPosition()
    local cursorGlobals = read_dword(widgetCursorGlobals)
    if cursorGlobals then
        local cursorX = read_int(cursorGlobals + 0x4)
        local cursorY = read_int(cursorGlobals + 0x8)
        return cursorX, cursorY
    end
end

LastColorCustomization = {primary = 1, secondary = 1}
function core.getCustomizationObjectData()
    local customizationObjectId = core.getCustomizationObjectId()
    assert(customizationObjectId, "No customization biped found")
    local customizationBiped = blam.biped(get_object(customizationObjectId))
    assert(customizationBiped, "No customization biped found")
    local customizationBipedTag = blam.bipedTag(customizationBiped.tagId)
    assert(customizationBipedTag, "No customization biped tag found")
    local customizationModel = blam.model(customizationBipedTag.model)
    assert(customizationModel, "No customization biped model found")

    local primaryColor = color.decToHex(customizationBiped.colorCLowerRed,
                                        customizationBiped.colorCLowerGreen,
                                        customizationBiped.colorCLowerBlue)

    local secondaryColor = color.decToHex(customizationBiped.colorDLowerRed,
                                          customizationBiped.colorDLowerGreen,
                                          customizationBiped.colorDLowerBlue)

    local colors = table.flatten(constants.customColors)
    return {
        id = customizationObjectId,
        handle = customizationObjectId,
        biped = customizationBiped,
        bipedTag = customizationBipedTag,
        tag = blam.getTag(customizationBiped.tagId),
        model = customizationModel,
        regions = {
            customizationBiped.regionPermutation1,
            customizationBiped.regionPermutation2,
            customizationBiped.regionPermutation3,
            customizationBiped.regionPermutation4,
            customizationBiped.regionPermutation5,
            customizationBiped.regionPermutation6,
            customizationBiped.regionPermutation7,
            customizationBiped.regionPermutation8
        },
        color = {
            primary = primaryColor,
            secondary = secondaryColor,
            custom = {
                primary = colors[LastColorCustomization.primary],
                secondary = colors[LastColorCustomization.secondary]
            }
        },
        visor = customizationBiped.shaderPermutationIndex
    }
end

---Set the region permutation of a customization object
---
---NOTE: Assumes region index starts at 1
---@param object blamObject
---@param regionIndex number
---@param permutationIndex number
function core.setObjectPermutationSafely(object, regionIndex, permutationIndex)
    local objectBipedTag = blam.bipedTag(object.tagId)
    assert(objectBipedTag, "No biped tag found")
    local objectModel = blam.model(objectBipedTag.model)
    assert(objectModel, "No biped model found")

    -- This one does not need to be substracted by 1 because property name is Lua 1-based
    local maximumRegionIndex = objectModel.regionCount
    if regionIndex > maximumRegionIndex then
        console_debug("Region index " .. regionIndex .. " out of range, maximum is " ..
                          maximumRegionIndex)
        return
    end

    local maximumPermutationIndex = objectModel.regionList[regionIndex].permutationCount - 1
    if permutationIndex > maximumPermutationIndex then
        console_debug("Permutation index " .. permutationIndex .. " for region " .. regionIndex ..
                          " out of range, setting to 0")
        permutationIndex = 0
    end
    object["regionPermutation" .. regionIndex] = permutationIndex
end

function core.copyToClipboard(text)
    return balltze.misc.setClipboard(text)
end

function core.getClipboard()
    return balltze.misc.getClipboard()
end

---Get the current screen aspect ratio
---@return number aspectWidth, number aspectHeight
function core.getScreenAspectRatio()
    local screenWidth, screenHeight = core.getScreenResolution()
    log("Screen resolution: " .. screenWidth .. "x" .. screenHeight)
    -- Calculate the greatest common divisor (GCD) using Euclidean algorithm
    local function gcd(a, b)
        while b ~= 0 do
            a, b = b, a % b
        end
        return a
    end

    -- Calculate the aspect ratio
    local divisor = gcd(screenWidth, screenHeight)
    local aspectWidth = screenWidth / divisor
    local aspectHeight = screenHeight / divisor

    -- Format the aspect ratio as a string
    -- local aspectRatioString = string.format("%d:%d", aspectWidth, aspectHeight)

    return aspectWidth, aspectHeight
end

---Get current player profile
---@return {name: string, colorIndex: number}
function core.getPlayerProfile()
    local profile = {}
    profile.name = blam.readUnicodeString(profileNameAddress, true)
    profile.colorIndex = read_byte(profileColorAddress) + 1
    return profile
end

--- Set customization biped color if possible
--- @param primaryColorHex? string
--- @param secondaryColorHex? string
function core.setCustomizationBipedColor(primaryColorHex, secondaryColorHex)
    local customizationObjectData = core.getCustomizationObjectData()
    local customizationBiped = customizationObjectData.biped

    -- Set primary color
    if primaryColorHex then
        LastColorCustomization.primary = table.indexof(table.flatten(constants.customColors),
                                                      primaryColorHex)
        local r, g, b = color.hexToDec(primaryColorHex)
        customizationBiped.colorCLowerRed = r
        customizationBiped.colorCLowerGreen = g
        customizationBiped.colorCLowerBlue = b
    end

    -- Set secondary color
    if secondaryColorHex then
        LastColorCustomization.secondary = table.indexof(table.flatten(constants.customColors),
                                                        secondaryColorHex)
        r, g, b = color.hexToDec(secondaryColorHex)
        customizationBiped.colorDLowerRed = r
        customizationBiped.colorDLowerGreen = g
        customizationBiped.colorDLowerBlue = b
    end
end

function core.getCustomizationColorByValue(value)
    local colorIndex = table.indexof(table.flatten(constants.customColors), value)
    local colorName = table.keyof(constants.customColor, value)
    return colorIndex, colorName
end

return core

end,

["inspect"] = function()
--------------------
-- Module: 'inspect'
--------------------
local inspect ={
  _VERSION = 'inspect.lua 3.1.0',
  _URL     = 'http://github.com/kikito/inspect.lua',
  _DESCRIPTION = 'human-readable representations of tables',
  _LICENSE = [[
    MIT LICENSE

    Copyright (c) 2013 Enrique García Cota

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  ]]
}

local tostring = tostring

inspect.KEY       = setmetatable({}, {__tostring = function() return 'inspect.KEY' end})
inspect.METATABLE = setmetatable({}, {__tostring = function() return 'inspect.METATABLE' end})

local function rawpairs(t)
  return next, t, nil
end

-- Apostrophizes the string if it has quotes, but not aphostrophes
-- Otherwise, it returns a regular quoted string
local function smartQuote(str)
  if str:match('"') and not str:match("'") then
    return "'" .. str .. "'"
  end
  return '"' .. str:gsub('"', '\\"') .. '"'
end

-- \a => '\\a', \0 => '\\0', 31 => '\31'
local shortControlCharEscapes = {
  ["\a"] = "\\a",  ["\b"] = "\\b", ["\f"] = "\\f", ["\n"] = "\\n",
  ["\r"] = "\\r",  ["\t"] = "\\t", ["\v"] = "\\v"
}
local longControlCharEscapes = {} -- \a => nil, \0 => \000, 31 => \031
for i=0, 31 do
  local ch = string.char(i)
  if not shortControlCharEscapes[ch] then
    shortControlCharEscapes[ch] = "\\"..i
    longControlCharEscapes[ch]  = string.format("\\%03d", i)
  end
end

local function escape(str)
  return (str:gsub("\\", "\\\\")
             :gsub("(%c)%f[0-9]", longControlCharEscapes)
             :gsub("%c", shortControlCharEscapes))
end

local function isIdentifier(str)
  return type(str) == 'string' and str:match( "^[_%a][_%a%d]*$" )
end

local function isSequenceKey(k, sequenceLength)
  return type(k) == 'number'
     and 1 <= k
     and k <= sequenceLength
     and math.floor(k) == k
end

local defaultTypeOrders = {
  ['number']   = 1, ['boolean']  = 2, ['string'] = 3, ['table'] = 4,
  ['function'] = 5, ['userdata'] = 6, ['thread'] = 7
}

local function sortKeys(a, b)
  local ta, tb = type(a), type(b)

  -- strings and numbers are sorted numerically/alphabetically
  if ta == tb and (ta == 'string' or ta == 'number') then return a < b end

  local dta, dtb = defaultTypeOrders[ta], defaultTypeOrders[tb]
  -- Two default types are compared according to the defaultTypeOrders table
  if dta and dtb then return defaultTypeOrders[ta] < defaultTypeOrders[tb]
  elseif dta     then return true  -- default types before custom ones
  elseif dtb     then return false -- custom types after default ones
  end

  -- custom types are sorted out alphabetically
  return ta < tb
end

-- For implementation reasons, the behavior of rawlen & # is "undefined" when
-- tables aren't pure sequences. So we implement our own # operator.
local function getSequenceLength(t)
  local len = 1
  local v = rawget(t,len)
  while v ~= nil do
    len = len + 1
    v = rawget(t,len)
  end
  return len - 1
end

local function getNonSequentialKeys(t)
  local keys, keysLength = {}, 0
  local sequenceLength = getSequenceLength(t)
  for k,_ in rawpairs(t) do
    if not isSequenceKey(k, sequenceLength) then
      keysLength = keysLength + 1
      keys[keysLength] = k
    end
  end
  table.sort(keys, sortKeys)
  return keys, keysLength, sequenceLength
end

local function countTableAppearances(t, tableAppearances)
  tableAppearances = tableAppearances or {}

  if type(t) == 'table' then
    if not tableAppearances[t] then
      tableAppearances[t] = 1
      for k,v in rawpairs(t) do
        countTableAppearances(k, tableAppearances)
        countTableAppearances(v, tableAppearances)
      end
      countTableAppearances(getmetatable(t), tableAppearances)
    else
      tableAppearances[t] = tableAppearances[t] + 1
    end
  end

  return tableAppearances
end

local copySequence = function(s)
  local copy, len = {}, #s
  for i=1, len do copy[i] = s[i] end
  return copy, len
end

local function makePath(path, ...)
  local keys = {...}
  local newPath, len = copySequence(path)
  for i=1, #keys do
    newPath[len + i] = keys[i]
  end
  return newPath
end

local function processRecursive(process, item, path, visited)
  if item == nil then return nil end
  if visited[item] then return visited[item] end

  local processed = process(item, path)
  if type(processed) == 'table' then
    local processedCopy = {}
    visited[item] = processedCopy
    local processedKey

    for k,v in rawpairs(processed) do
      processedKey = processRecursive(process, k, makePath(path, k, inspect.KEY), visited)
      if processedKey ~= nil then
        processedCopy[processedKey] = processRecursive(process, v, makePath(path, processedKey), visited)
      end
    end

    local mt  = processRecursive(process, getmetatable(processed), makePath(path, inspect.METATABLE), visited)
    if type(mt) ~= 'table' then mt = nil end -- ignore not nil/table __metatable field
    setmetatable(processedCopy, mt)
    processed = processedCopy
  end
  return processed
end



-------------------------------------------------------------------

local Inspector = {}
local Inspector_mt = {__index = Inspector}

function Inspector:puts(...)
  local args   = {...}
  local buffer = self.buffer
  local len    = #buffer
  for i=1, #args do
    len = len + 1
    buffer[len] = args[i]
  end
end

function Inspector:down(f)
  self.level = self.level + 1
  f()
  self.level = self.level - 1
end

function Inspector:tabify()
  self:puts(self.newline, string.rep(self.indent, self.level))
end

function Inspector:alreadyVisited(v)
  return self.ids[v] ~= nil
end

function Inspector:getId(v)
  local id = self.ids[v]
  if not id then
    local tv = type(v)
    id              = (self.maxIds[tv] or 0) + 1
    self.maxIds[tv] = id
    self.ids[v]     = id
  end
  return tostring(id)
end

function Inspector:putKey(k)
  if isIdentifier(k) then return self:puts(k) end
  self:puts("[")
  self:putValue(k)
  self:puts("]")
end

function Inspector:putTable(t)
  if t == inspect.KEY or t == inspect.METATABLE then
    self:puts(tostring(t))
  elseif self:alreadyVisited(t) then
    self:puts('<table ', self:getId(t), '>')
  elseif self.level >= self.depth then
    self:puts('{...}')
  else
    if self.tableAppearances[t] > 1 then self:puts('<', self:getId(t), '>') end

    local nonSequentialKeys, nonSequentialKeysLength, sequenceLength = getNonSequentialKeys(t)
    local mt                = getmetatable(t)

    self:puts('{')
    self:down(function()
      local count = 0
      for i=1, sequenceLength do
        if count > 0 then self:puts(',') end
        self:puts(' ')
        self:putValue(t[i])
        count = count + 1
      end

      for i=1, nonSequentialKeysLength do
        local k = nonSequentialKeys[i]
        if count > 0 then self:puts(',') end
        self:tabify()
        self:putKey(k)
        self:puts(' = ')
        self:putValue(t[k])
        count = count + 1
      end

      if type(mt) == 'table' then
        if count > 0 then self:puts(',') end
        self:tabify()
        self:puts('<metatable> = ')
        self:putValue(mt)
      end
    end)

    if nonSequentialKeysLength > 0 or type(mt) == 'table' then -- result is multi-lined. Justify closing }
      self:tabify()
    elseif sequenceLength > 0 then -- array tables have one extra space before closing }
      self:puts(' ')
    end

    self:puts('}')
  end
end

function Inspector:putValue(v)
  local tv = type(v)

  if tv == 'string' then
    self:puts(smartQuote(escape(v)))
  elseif tv == 'number' or tv == 'boolean' or tv == 'nil' or
         tv == 'cdata' or tv == 'ctype' then
    self:puts(tostring(v))
  elseif tv == 'table' then
    self:putTable(v)
  else
    self:puts('<', tv, ' ', self:getId(v), '>')
  end
end

-------------------------------------------------------------------

function inspect.inspect(root, options)
  options       = options or {}

  local depth   = options.depth   or math.huge
  local newline = options.newline or '\n'
  local indent  = options.indent  or '  '
  local process = options.process

  if process then
    root = processRecursive(process, root, {}, {})
  end

  local inspector = setmetatable({
    depth            = depth,
    level            = 0,
    buffer           = {},
    ids              = {},
    maxIds           = {},
    newline          = newline,
    indent           = indent,
    tableAppearances = countTableAppearances(root)
  }, Inspector_mt)

  inspector:putValue(root)

  return table.concat(inspector.buffer)
end

setmetatable(inspect, { __call = function(_, ...) return inspect.inspect(...) end })

return inspect


end,

["json"] = function()
--------------------
-- Module: 'json'
--------------------
--
-- json.lua
--
-- Copyright (c) 2020 rxi
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of
-- this software and associated documentation files (the "Software"), to deal in
-- the Software without restriction, including without limitation the rights to
-- use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
-- of the Software, and to permit persons to whom the Software is furnished to do
-- so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in all
-- copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local json = { _version = "0.1.2" }

-------------------------------------------------------------------------------
-- Encode
-------------------------------------------------------------------------------

local encode

local escape_char_map = {
  [ "\\" ] = "\\",
  [ "\"" ] = "\"",
  [ "\b" ] = "b",
  [ "\f" ] = "f",
  [ "\n" ] = "n",
  [ "\r" ] = "r",
  [ "\t" ] = "t",
}

local escape_char_map_inv = { [ "/" ] = "/" }
for k, v in pairs(escape_char_map) do
  escape_char_map_inv[v] = k
end


local function escape_char(c)
  return "\\" .. (escape_char_map[c] or string.format("u%04x", c:byte()))
end


local function encode_nil(val)
  return "null"
end


local function encode_table(val, stack)
  local res = {}
  stack = stack or {}

  -- Circular reference?
  if stack[val] then error("circular reference") end

  stack[val] = true

  if rawget(val, 1) ~= nil or next(val) == nil then
    -- Treat as array -- check keys are valid and it is not sparse
    local n = 0
    for k in pairs(val) do
      if type(k) ~= "number" then
        error("invalid table: mixed or invalid key types")
      end
      n = n + 1
    end
    if n ~= #val then
      error("invalid table: sparse array")
    end
    -- Encode
    for i, v in ipairs(val) do
      table.insert(res, encode(v, stack))
    end
    stack[val] = nil
    return "[" .. table.concat(res, ",") .. "]"

  else
    -- Treat as an object
    for k, v in pairs(val) do
      if type(k) ~= "string" then
        error("invalid table: mixed or invalid key types")
      end
      table.insert(res, encode(k, stack) .. ":" .. encode(v, stack))
    end
    stack[val] = nil
    return "{" .. table.concat(res, ",") .. "}"
  end
end


local function encode_string(val)
  return '"' .. val:gsub('[%z\1-\31\\"]', escape_char) .. '"'
end


local function encode_number(val)
  -- Check for NaN, -inf and inf
  if val ~= val or val <= -math.huge or val >= math.huge then
    error("unexpected number value '" .. tostring(val) .. "'")
  end
  return string.format("%.14g", val)
end


local type_func_map = {
  [ "nil"     ] = encode_nil,
  [ "table"   ] = encode_table,
  [ "string"  ] = encode_string,
  [ "number"  ] = encode_number,
  [ "boolean" ] = tostring,
}


encode = function(val, stack)
  local t = type(val)
  local f = type_func_map[t]
  if f then
    return f(val, stack)
  end
  error("unexpected type '" .. t .. "'")
end


function json.encode(val)
  return ( encode(val) )
end


-------------------------------------------------------------------------------
-- Decode
-------------------------------------------------------------------------------

local parse

local function create_set(...)
  local res = {}
  for i = 1, select("#", ...) do
    res[ select(i, ...) ] = true
  end
  return res
end

local space_chars   = create_set(" ", "\t", "\r", "\n")
local delim_chars   = create_set(" ", "\t", "\r", "\n", "]", "}", ",")
local escape_chars  = create_set("\\", "/", '"', "b", "f", "n", "r", "t", "u")
local literals      = create_set("true", "false", "null")

local literal_map = {
  [ "true"  ] = true,
  [ "false" ] = false,
  [ "null"  ] = nil,
}


local function next_char(str, idx, set, negate)
  for i = idx, #str do
    if set[str:sub(i, i)] ~= negate then
      return i
    end
  end
  return #str + 1
end


local function decode_error(str, idx, msg)
  local line_count = 1
  local col_count = 1
  for i = 1, idx - 1 do
    col_count = col_count + 1
    if str:sub(i, i) == "\n" then
      line_count = line_count + 1
      col_count = 1
    end
  end
  error( string.format("%s at line %d col %d", msg, line_count, col_count) )
end


local function codepoint_to_utf8(n)
  -- http://scripts.sil.org/cms/scripts/page.php?site_id=nrsi&id=iws-appendixa
  local f = math.floor
  if n <= 0x7f then
    return string.char(n)
  elseif n <= 0x7ff then
    return string.char(f(n / 64) + 192, n % 64 + 128)
  elseif n <= 0xffff then
    return string.char(f(n / 4096) + 224, f(n % 4096 / 64) + 128, n % 64 + 128)
  elseif n <= 0x10ffff then
    return string.char(f(n / 262144) + 240, f(n % 262144 / 4096) + 128,
                       f(n % 4096 / 64) + 128, n % 64 + 128)
  end
  error( string.format("invalid unicode codepoint '%x'", n) )
end


local function parse_unicode_escape(s)
  local n1 = tonumber( s:sub(1, 4),  16 )
  local n2 = tonumber( s:sub(7, 10), 16 )
   -- Surrogate pair?
  if n2 then
    return codepoint_to_utf8((n1 - 0xd800) * 0x400 + (n2 - 0xdc00) + 0x10000)
  else
    return codepoint_to_utf8(n1)
  end
end


local function parse_string(str, i)
  local res = ""
  local j = i + 1
  local k = j

  while j <= #str do
    local x = str:byte(j)

    if x < 32 then
      decode_error(str, j, "control character in string")

    elseif x == 92 then -- `\`: Escape
      res = res .. str:sub(k, j - 1)
      j = j + 1
      local c = str:sub(j, j)
      if c == "u" then
        local hex = str:match("^[dD][89aAbB]%x%x\\u%x%x%x%x", j + 1)
                 or str:match("^%x%x%x%x", j + 1)
                 or decode_error(str, j - 1, "invalid unicode escape in string")
        res = res .. parse_unicode_escape(hex)
        j = j + #hex
      else
        if not escape_chars[c] then
          decode_error(str, j - 1, "invalid escape char '" .. c .. "' in string")
        end
        res = res .. escape_char_map_inv[c]
      end
      k = j + 1

    elseif x == 34 then -- `"`: End of string
      res = res .. str:sub(k, j - 1)
      return res, j + 1
    end

    j = j + 1
  end

  decode_error(str, i, "expected closing quote for string")
end


local function parse_number(str, i)
  local x = next_char(str, i, delim_chars)
  local s = str:sub(i, x - 1)
  local n = tonumber(s)
  if not n then
    decode_error(str, i, "invalid number '" .. s .. "'")
  end
  return n, x
end


local function parse_literal(str, i)
  local x = next_char(str, i, delim_chars)
  local word = str:sub(i, x - 1)
  if not literals[word] then
    decode_error(str, i, "invalid literal '" .. word .. "'")
  end
  return literal_map[word], x
end


local function parse_array(str, i)
  local res = {}
  local n = 1
  i = i + 1
  while 1 do
    local x
    i = next_char(str, i, space_chars, true)
    -- Empty / end of array?
    if str:sub(i, i) == "]" then
      i = i + 1
      break
    end
    -- Read token
    x, i = parse(str, i)
    res[n] = x
    n = n + 1
    -- Next token
    i = next_char(str, i, space_chars, true)
    local chr = str:sub(i, i)
    i = i + 1
    if chr == "]" then break end
    if chr ~= "," then decode_error(str, i, "expected ']' or ','") end
  end
  return res, i
end


local function parse_object(str, i)
  local res = {}
  i = i + 1
  while 1 do
    local key, val
    i = next_char(str, i, space_chars, true)
    -- Empty / end of object?
    if str:sub(i, i) == "}" then
      i = i + 1
      break
    end
    -- Read key
    if str:sub(i, i) ~= '"' then
      decode_error(str, i, "expected string for key")
    end
    key, i = parse(str, i)
    -- Read ':' delimiter
    i = next_char(str, i, space_chars, true)
    if str:sub(i, i) ~= ":" then
      decode_error(str, i, "expected ':' after key")
    end
    i = next_char(str, i + 1, space_chars, true)
    -- Read value
    val, i = parse(str, i)
    -- Set
    res[key] = val
    -- Next token
    i = next_char(str, i, space_chars, true)
    local chr = str:sub(i, i)
    i = i + 1
    if chr == "}" then break end
    if chr ~= "," then decode_error(str, i, "expected '}' or ','") end
  end
  return res, i
end


local char_func_map = {
  [ '"' ] = parse_string,
  [ "0" ] = parse_number,
  [ "1" ] = parse_number,
  [ "2" ] = parse_number,
  [ "3" ] = parse_number,
  [ "4" ] = parse_number,
  [ "5" ] = parse_number,
  [ "6" ] = parse_number,
  [ "7" ] = parse_number,
  [ "8" ] = parse_number,
  [ "9" ] = parse_number,
  [ "-" ] = parse_number,
  [ "t" ] = parse_literal,
  [ "f" ] = parse_literal,
  [ "n" ] = parse_literal,
  [ "[" ] = parse_array,
  [ "{" ] = parse_object,
}


parse = function(str, idx)
  local chr = str:sub(idx, idx)
  local f = char_func_map[chr]
  if f then
    return f(str, idx)
  end
  decode_error(str, idx, "unexpected character '" .. chr .. "'")
end


function json.decode(str)
  if type(str) ~= "string" then
    error("expected argument of type string, got " .. type(str))
  end
  local res, idx = parse(str, next_char(str, 1, space_chars, true))
  idx = next_char(str, idx, space_chars, true)
  if idx <= #str then
    decode_error(str, idx, "trailing garbage")
  end
  return res
end


return json

end,

["base64"] = function()
--------------------
-- Module: 'base64'
--------------------
--[[

 base64 -- v1.5.3 public domain Lua base64 encoder/decoder
 no warranty implied; use at your own risk

 Needs bit32.extract function. If not present it's implemented using BitOp
 or Lua 5.3 native bit operators. For Lua 5.1 fallbacks to pure Lua
 implementation inspired by Rici Lake's post:
   http://ricilake.blogspot.co.uk/2007/10/iterating-bits-in-lua.html

 author: Ilya Kolbin (iskolbin@gmail.com)
 url: github.com/iskolbin/lbase64

 COMPATIBILITY

 Lua 5.1+, LuaJIT

 LICENSE

 See end of file for license information.

--]]


local base64 = {}

local extract = _G.bit32 and _G.bit32.extract -- Lua 5.2/Lua 5.3 in compatibility mode
if not extract then
	if _G.bit then -- LuaJIT
		local shl, shr, band = _G.bit.lshift, _G.bit.rshift, _G.bit.band
		extract = function( v, from, width )
			return band( shr( v, from ), shl( 1, width ) - 1 )
		end
	elseif _G._VERSION == "Lua 5.1" then
		extract = function( v, from, width )
			local w = 0
			local flag = 2^from
			for i = 0, width-1 do
				local flag2 = flag + flag
				if v % flag2 >= flag then
					w = w + 2^i
				end
				flag = flag2
			end
			return w
		end
	else -- Lua 5.3+
		extract = load[[return function( v, from, width )
			return ( v >> from ) & ((1 << width) - 1)
		end]]()
	end
end


function base64.makeencoder( s62, s63, spad )
	local encoder = {}
	for b64code, char in pairs{[0]='A','B','C','D','E','F','G','H','I','J',
		'K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y',
		'Z','a','b','c','d','e','f','g','h','i','j','k','l','m','n',
		'o','p','q','r','s','t','u','v','w','x','y','z','0','1','2',
		'3','4','5','6','7','8','9',s62 or '+',s63 or'/',spad or'='} do
		encoder[b64code] = char:byte()
	end
	return encoder
end

function base64.makedecoder( s62, s63, spad )
	local decoder = {}
	for b64code, charcode in pairs( base64.makeencoder( s62, s63, spad )) do
		decoder[charcode] = b64code
	end
	return decoder
end

local DEFAULT_ENCODER = base64.makeencoder()
local DEFAULT_DECODER = base64.makedecoder()

local char, concat = string.char, table.concat

function base64.encode( str, encoder, usecaching )
	encoder = encoder or DEFAULT_ENCODER
	local t, k, n = {}, 1, #str
	local lastn = n % 3
	local cache = {}
	for i = 1, n-lastn, 3 do
		local a, b, c = str:byte( i, i+2 )
		local v = a*0x10000 + b*0x100 + c
		local s
		if usecaching then
			s = cache[v]
			if not s then
				s = char(encoder[extract(v,18,6)], encoder[extract(v,12,6)], encoder[extract(v,6,6)], encoder[extract(v,0,6)])
				cache[v] = s
			end
		else
			s = char(encoder[extract(v,18,6)], encoder[extract(v,12,6)], encoder[extract(v,6,6)], encoder[extract(v,0,6)])
		end
		t[k] = s
		k = k + 1
	end
	if lastn == 2 then
		local a, b = str:byte( n-1, n )
		local v = a*0x10000 + b*0x100
		t[k] = char(encoder[extract(v,18,6)], encoder[extract(v,12,6)], encoder[extract(v,6,6)], encoder[64])
	elseif lastn == 1 then
		local v = str:byte( n )*0x10000
		t[k] = char(encoder[extract(v,18,6)], encoder[extract(v,12,6)], encoder[64], encoder[64])
	end
	return concat( t )
end

function base64.decode( b64, decoder, usecaching )
	decoder = decoder or DEFAULT_DECODER
	local pattern = '[^%w%+%/%=]'
	if decoder then
		local s62, s63
		for charcode, b64code in pairs( decoder ) do
			if b64code == 62 then s62 = charcode
			elseif b64code == 63 then s63 = charcode
			end
		end
		pattern = ('[^%%w%%%s%%%s%%=]'):format( char(s62), char(s63) )
	end
	b64 = b64:gsub( pattern, '' )
	local cache = usecaching and {}
	local t, k = {}, 1
	local n = #b64
	local padding = b64:sub(-2) == '==' and 2 or b64:sub(-1) == '=' and 1 or 0
	for i = 1, padding > 0 and n-4 or n, 4 do
		local a, b, c, d = b64:byte( i, i+3 )
		local s
		if usecaching then
			local v0 = a*0x1000000 + b*0x10000 + c*0x100 + d
			s = cache[v0]
			if not s then
				local v = decoder[a]*0x40000 + decoder[b]*0x1000 + decoder[c]*0x40 + decoder[d]
				s = char( extract(v,16,8), extract(v,8,8), extract(v,0,8))
				cache[v0] = s
			end
		else
			local v = decoder[a]*0x40000 + decoder[b]*0x1000 + decoder[c]*0x40 + decoder[d]
			s = char( extract(v,16,8), extract(v,8,8), extract(v,0,8))
		end
		t[k] = s
		k = k + 1
	end
	if padding == 1 then
		local a, b, c = b64:byte( n-3, n-1 )
		local v = decoder[a]*0x40000 + decoder[b]*0x1000 + decoder[c]*0x40
		t[k] = char( extract(v,16,8), extract(v,8,8))
	elseif padding == 2 then
		local a, b = b64:byte( n-3, n-2 )
		local v = decoder[a]*0x40000 + decoder[b]*0x1000
		t[k] = char( extract(v,16,8))
	end
	return concat( t )
end

return base64

--[[
------------------------------------------------------------------------------
This software is available under 2 licenses -- choose whichever you prefer.
------------------------------------------------------------------------------
ALTERNATIVE A - MIT License
Copyright (c) 2018 Ilya Kolbin
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
------------------------------------------------------------------------------
ALTERNATIVE B - Public Domain (www.unlicense.org)
This is free and unencumbered software released into the public domain.
Anyone is free to copy, modify, publish, use, compile, sell, or distribute this
software, either in source code form or as a compiled binary, for any purpose,
commercial or non-commercial, and by any means.
In jurisdictions that recognize copyright laws, the author or authors of this
software dedicate any and all copyright interest in the software to the public
domain. We make this dedication for the benefit of the public at large and to
the detriment of our heirs and successors. We intend this dedication to be an
overt act of relinquishment in perpetuity of all present and future rights to
this software under copyright law.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
------------------------------------------------------------------------------
--]]

end,

["color"] = function()
--------------------
-- Module: 'color'
--------------------
local color = {}

--- Convert to decimal rgb color from hex string color
---@param hex string
---@param alpha? number
function color.hexToDec(hex, alpha)
    local redColor, greenColor, blueColor = hex:gsub("#", ""):match("(..)(..)(..)")
    redColor, greenColor, blueColor = tonumber(redColor, 16) / 255, tonumber(greenColor, 16) / 255,
                                      tonumber(blueColor, 16) / 255
    redColor, greenColor, blueColor = math.floor(redColor * 100) / 100,
                                      math.floor(greenColor * 100) / 100,
                                      math.floor(blueColor * 100) / 100
    if alpha == nil then
        return redColor, greenColor, blueColor
    elseif alpha > 1 then
        alpha = alpha / 100
    end
    return redColor, greenColor, blueColor, alpha
end

--- Convert to decimal rgb color from byte rgb color
---@param r number
---@param g number
---@param b number
---@param alpha number
function color.rgbToDec(r, g, b, alpha)
    local redColor, greenColor, blueColor = r / 255, g / 255, b / 255
    redColor, greenColor, blueColor = math.floor(redColor * 100) / 100,
                                      math.floor(greenColor * 100) / 100,
                                      math.floor(blueColor * 100) / 100
    if alpha == nil then
        return redColor, greenColor, blueColor
    elseif alpha > 1 then
        alpha = alpha / 100
    end
    return redColor, greenColor, blueColor, alpha
end

--- Convert to rgb bytes to hex string
---@param r number
---@param g number
---@param b number
function color.rgbToHex(r, g, b)
    local rgb = (r * 0x10000) + (g * 0x100) + b
    return string.format("%06x", rgb)
end

--- Convert to rgb bytes to hex string
---@param r number
---@param g number
---@param b number
function color.decToHex(r, g, b)
    r = math.max(0, math.min(1, r))
    g = math.max(0, math.min(1, g))
    b = math.max(0, math.min(1, b))

    r = math.ceil(r * 255)
    g = math.ceil(g * 255)
    b = math.ceil(b * 255)

    return string.format("#%02X%02X%02X", r, g, b)
end

return color

end,

["insurrection.mercury"] = function()
--------------------
-- Module: 'insurrection.mercury'
--------------------
------------------------------------------------------------------------------
-- Mercury Communication
-- Gelatinoso, Sledmine
-- Module to communicate with Mercury available on the host system
------------------------------------------------------------------------------
local json = require "json"
local inspect = require "inspect"

local mercury = {}

---@class packageMetadata
---@field name string
---@field label string
---@field author string
---@field version string
---@field internalVersion string
---@field category string
---@field conflicts string[]
---@field mirrors string[]
---@field nextVersion string

---@class mercDependencies
---@field label string
---@field version string

---@class mercFiles
---@field path string
---@field type string
---@field outputPath string

---@class mercUpdates
---@field path string
---@field diffPath string
---@field type string
---@field outputPath string

---@class packageMercury
---@field name string
---@field label string
---@field description string
---@field author string
---@field version string
---@field targetVersion string
---@field internalVersion string
---@field manifestVersion string
---@field files mercFiles[]
---@field updates mercUpdates[]
---@field dependencies mercDependencies[]

--- Get an array of installed Mercury packages
---@return packageMercury[] | nil
function mercury.getInstalled()
    local installedPackages
    local pipe = io.popen("mercury list -j", "r")
    if (pipe) then
        local output = pipe:read("*all")
        console_out(output)
        pipe:close()
        if (output and output ~= "" and not output:find("Warning,")) then
            installedPackages = json.decode(output)
            return installedPackages
        end
    end
end

-- Fetch the latest package index available on the repository
---@return packageMetadata[]
function mercury.fetchPackages()
    -- Get the package index output on json format
    local pipe = io.popen("mercury fetch -j")
    if (pipe) then
        local output = pipe:read("*all")
        pipe:close()
        print(inspect(output))
        if (output and output ~= "" and not output:find("Error,")) then
            local fetchedPackages = json.decode(output)
            return fetchedPackages
        end
    end
    return {}
end

-- Installs selected package
function mercury.installPackage(label)
    local pipe = io.popen("mercury install " .. label)
    if (pipe) then
        local response = pipe:read("*all")
        pipe:close()
        return response
    end
end

function mercury.removePackage(label)
    local result = os.execute("mercury remove " .. label)
    return result
end

return mercury

end,

["insurrection.utils"] = function()
--------------------
-- Module: 'insurrection.utils'
--------------------
local utils = {}

---Returns the path, filename or name, and extension of a path
---@param path string
---@return {path: string, name: string, extension?: string}
function utils.path(path)
    local split = path:split("\\")
    local pathElements = table.copy(split)
    table.remove(pathElements, #pathElements)
    local path = table.concat(pathElements, "\\")
    local name = split[#split]
    local extension
    if name:includes "." then
        split = name:split "."
        name = split[1]
        extension = split[#split]
    end
    return {path = path, name = name, extension = extension}
end

---Executes a function after a delay
---@param milliseconds number
---@param callback function
function utils.delay(milliseconds, callback)
    local timer
    timer = Balltze.misc.setTimer(milliseconds, function()
        -- Prevent the entire game from crashing as Balltze does not handle errors in timers lol
        local success, message = pcall(callback)
        if not success then
            logger:error("Error in delay callback: %s", message)
        end
        timer.stop()
    end)
end

---Returns an upper sentence case string from a snake case string
---@param s string
function utils.snakeCaseToUpperTitleCase(s)
    local str = s:upper():replace("_", " ")
    return str
end

---Returns a tile case string from a snake case string
---@param s string
function utils.snakeCaseToTitleCase(s)
    local str = s:replace("_", " "):gsub("(%l)(%w*)", function(a, b)
        return a:upper() .. b
    end)
    return str
end

---Returns a table sorted alphabetically by key
---@generic K,V
---@param t table<K,V>
---@return table<K,V>
function utils.sortTableAlphabetically(t)
    local sorted = table.copy(t)
    table.sort(sorted, function(a, b)
        return a.label < b.label
    end)
    return sorted
end

---Returns a snake case string from a camel case string
---@param s string
---@return string
function utils.camelCaseToSnakeCase(s)
    -- Convert camelCase to snake_case but keep numbers
    s = s:sub(1, 1):lower() .. s:sub(2)
    local new = s:gsub("([A-Z])", function(c)
        return "_" .. c:lower()
    end)
    if new:sub(1, 1) == "_" then
        new = new:sub(2)
    end
    return new
end

return utils

end,

["insurrection.version"] = function()
--------------------
-- Module: 'insurrection.version'
--------------------
local releaseVersion = "2.8.1"
local metadata = "5f1165a." .. os.date("%Y%m%d")
local version = releaseVersion
if DebugMode then
    return releaseVersion .. "-dev+" .. metadata
end
return version

end,

["insurrection.specialEvents"] = function()
--------------------
-- Module: 'insurrection.specialEvents'
--------------------
local constants = require "insurrection.constants"
local core = require "insurrection.core"
local engine = Engine
local balltze = Balltze
local blam = require "blam"

local specialEvents = {}

local christmasHats = {}

local isDecember = function()
    local date = os.date("*t")
    -- return date.month == 12 and date.day >= 24 and date.day <= 26
    return date.month == 12
end

function specialEvents.onPostMapLoad()
    if engine.map.getCurrentMapHeader().name == "ui" then
        return
    end
    if not isDecember() then
        return
    end
    local treeScenerys = table.filter(table.map(engine.tag.findTags("tree",
                                                                    engine.tag.classes.scenery),
                                                function(tag)
        return {path = tag.path, tagHandle = tag.handle}
    end), function(tag)
        return not tag.path:includes("insurrection")
    end)

    -- TODO Using this will result in a memory leak due to how accessing elements trough Balltze
    -- works, basically this creates 999999+ elements in the table, which will never be garbage
    -- collected, so we need to use the for loop below instead to use actual elements count

    -- local xmasTreeCollection = table.map(constants.scenery.xmasObjects.data.tags.elements, function(tag)
    --    return {path = tag.reference.path, handle = tag/.reference.tagHandle}
    -- end)

    --- TODO Add support to read "dynamic arrays" in Luna, so we can use tags property instead of
    --- elements
    ---@type MetaEngineTagDataTagCollectionTag[]
    local xmasTreeCollection = {}
    for i = 1, constants.scenery.xmasObjects.data.tags.count do
        local tag = constants.scenery.xmasObjects.data.tags.elements[i]
        table.insert(xmasTreeCollection,
                     {path = tag.reference.path, handle = tag.reference.tagHandle})
    end

    -- Erase all insurrection trees if any (helps with reloading plugin)
    for objectIndex = 0, 2047 do
        local object = engine.gameState.getObject(objectIndex)
        if object then
            if object.type == engine.tag.objectType.scenery then
                local scenery = engine.gameState.getObject(objectIndex,
                                                           engine.tag.objectType.scenery)
                assert(scenery, "Failed to get scenery object")
                local sceneryTag = engine.tag.getTag(scenery.tagHandle.value,
                                                     engine.tag.classes.scenery)
                assert(sceneryTag, "Failed to get scenery tag")
                if sceneryTag.path:includes("insurrection") and sceneryTag.path:includes("tree") then
                    local _, objectHandle = blam.getObject(objectIndex)
                    assert(objectHandle, "Failed to get object handle")
                    engine.gameState.deleteObject(objectHandle)
                end
            end
        end
    end

    -- Create xmas trees
    for objectIndex = 0, 2047 do
        local object = engine.gameState.getObject(objectIndex)
        if object then
            if object.type == engine.tag.objectType.scenery then
                local scenery = engine.gameState.getObject(objectIndex,
                                                           engine.tag.objectType.scenery)
                assert(scenery, "Failed to get scenery object")
                local sceneryTag = engine.tag.getTag(scenery.tagHandle.value,
                                                     engine.tag.classes.scenery)
                assert(sceneryTag, "Failed to get scenery tag")
                if not sceneryTag.path:includes("insurrection") then
                    local sceneryName = core.getTagName(sceneryTag.path)
                    local sceneryToReplace = table.find(xmasTreeCollection, function(xmasTree)
                        local xmasTreeName = core.getTagName(xmasTree.path)
                        return xmasTreeName == sceneryName
                    end)
                    if sceneryToReplace then
                        logger:debug("Replacing {} with {}", sceneryTag.path, sceneryToReplace.path)
                        -- sceneryTag.data.model.tagHandle.value = sceneryToReplace.handle
                        local newSceneryHandle = engine.gameState.createObject(
                                                     sceneryToReplace.handle, nil, {
                                x = scenery.position.x,
                                y = scenery.position.y,
                                z = scenery.position.z
                            })
                        local _, objectHandle = blam.getObject(objectIndex)
                        assert(objectHandle, "Failed to get object handle")
                        engine.gameState.objectAttachToMarker(objectHandle, "",
                                                              newSceneryHandle.value, "")
                    end

                end
            end
        end
    end
end

function specialEvents.onTick()
    if engine.map.getCurrentMapHeader().name == "ui" then
        return
    end
    if not isDecember() then
        return
    end
    for playerIndex = 0, constants.limits.maximumPlayers do
        local player = engine.gameState.getPlayer(playerIndex)
        if player then
            local biped = engine.gameState.getObject(player.objectHandle.value,
                                                     engine.tag.objectType.biped)
            if biped then
                if not christmasHats[playerIndex] then
                    local hatHandle = engine.gameState.createObject(
                                          constants.scenery.christmasHat.handle.value, nil, {
                            x = biped.position.x,
                            y = biped.position.y,
                            z = biped.position.z + 1
                        })
                    christmasHats[playerIndex] = hatHandle
                    engine.gameState.objectAttachToMarker(player.objectHandle.value, "head",
                                                          hatHandle, "")
                end
                if christmasHats[playerIndex] then
                    local hat = engine.gameState.getObject(christmasHats[playerIndex],
                                                           engine.tag.objectType.scenery)
                    if not hat then
                        christmasHats[playerIndex] = nil
                    end
                    if hat and hat.parentObject.value ~= player.objectHandle.value then
                        christmasHats[playerIndex] = nil
                    end
                end
            end
        end
    end
end

return specialEvents

end,

["insurrection.constants"] = function()
--------------------
-- Module: 'insurrection.constants'
--------------------
local constants = {}

local engine = Engine

constants.path = {
    pauseMenu = [[insurrection\ui\menus\pause\pause_menu]],
    nameplateCollection = [[insurrection\ui\shared\nameplates]],
    dialog = [[insurrection\ui\menus\dialog\dialog_menu]],
    customSounds = [[insurrection\sound\custom_sounds]],
    tester = [[insurrection\ui\menus\tester\tester_menu]],
    christmasHat = [[insurrection\hats\christmas\christmas_hat]],
    xmasObjects = [[insurrection\scenery\season\xmas\xmas_objects]]
}

constants.customColor = {
    black = "#141414",
    charcoal = "#323232",
    gray = "#787878",
    silver = "#c8c8c8",
    white = "#ffffff",
    --
    brown = "#413823",
    paleBrown = "#8c7a53",
    ash = "#7a7966",
    sage = "#a2a089",
    misr = "#d1d0c1",
    --
    blue = "#001473",
    cobalt = "#303c8f",
    deluge = "#5966ac",
    ice = "#8493c8",
    cerulean = "#b0c0e4",
    --
    indigo = "#0051a7",
    mariner = "#2f79bc",
    moonstone = "#57a3d1",
    malibu = "#80cde6",
    lightCyan = "#abf8fc",
    --
    green = "#117300",
    grass = "#18a20b",
    alien = "#40c937",
    algae = "#82ec7a",
    pastelGreen = "#b9f9b4",
    --
    muddyGreen = "#687434",
    drab = "#7a8932",
    moss = "#8f9e4d",
    olive = "#acbc6a",
    caper = "#ddf1a4",
    --
    mustard = "#918d09",
    bile = "#bfbf00",
    banana = "#eaea4d",
    buff = "#f7f79a",
    ecru = "#ffffc7",
    --
    rum = "#89642c",
    gold = "#b27e10",
    bee = "#eaae14",
    mango = "#f7bd44",
    chardonnay = "#FCCB77",
    --
    rust = "#8b3900",
    burnt = "#b14e00",
    bamboo = "#e86600",
    tiger = "#f78335",
    peach = "#f9a37a",
    --
    berry = "#8b0000",
    scarlet = "#b50000",
    corsa = "#ea0000",
    sunset = "#f35d5d",
    geraldine = "#ff8c8c",
    --
    eggplant = "#a70069",
    redViolet = "#ce0081",
    cerise = "#ff00a0",
    pink = "#ff7eb9",
    candy = "#ffbad5",
    --
    purple = "#321861",
    meteorite = "#462f6f",
    lilac = "#8b70a5",
    violet = "#d0a9f6",
    lavender = "#e1cdf6"
}

constants.customColors = {
    {
        constants.customColor.black,
        constants.customColor.charcoal,
        constants.customColor.gray,
        constants.customColor.silver,
        constants.customColor.white
    },
    {
        constants.customColor.brown,
        constants.customColor.paleBrown,
        constants.customColor.ash,
        constants.customColor.sage,
        constants.customColor.misr
    },
    {
        constants.customColor.blue,
        constants.customColor.cobalt,
        constants.customColor.deluge,
        constants.customColor.ice,
        constants.customColor.cerulean
    },
    {
        constants.customColor.indigo,
        constants.customColor.mariner,
        constants.customColor.moonstone,
        constants.customColor.malibu,
        constants.customColor.lightCyan
    },
    {
        constants.customColor.green,
        constants.customColor.grass,
        constants.customColor.alien,
        constants.customColor.algae,
        constants.customColor.pastelGreen
    },
    {
        constants.customColor.muddyGreen,
        constants.customColor.drab,
        constants.customColor.moss,
        constants.customColor.olive,
        constants.customColor.caper
    },
    {
        constants.customColor.mustard,
        constants.customColor.bile,
        constants.customColor.banana,
        constants.customColor.buff,
        constants.customColor.ecru
    },
    {
        constants.customColor.rum,
        constants.customColor.gold,
        constants.customColor.bee,
        constants.customColor.mango,
        constants.customColor.chardonnay
    },
    {
        constants.customColor.rust,
        constants.customColor.burnt,
        constants.customColor.bamboo,
        constants.customColor.tiger,
        constants.customColor.peach
    },
    {
        constants.customColor.berry,
        constants.customColor.scarlet,
        constants.customColor.corsa,
        constants.customColor.sunset,
        constants.customColor.geraldine
    },
    {
        constants.customColor.eggplant,
        constants.customColor.redViolet,
        constants.customColor.cerise,
        constants.customColor.pink,
        constants.customColor.candy
    },
    {
        constants.customColor.purple,
        constants.customColor.meteorite,
        constants.customColor.lilac,
        constants.customColor.violet,
        constants.customColor.lavender
    }
}

constants.color = {
    white = "#FFFFFF",
    black = "#000000",
    red = "#FE0000",
    blue = "#0201E3",
    gray = "#707E71",
    yellow = "#FFFF01",
    green = "#00FF01",
    pink = "#FF56B9",
    purple = "#AB10F4",
    cyan = "#01FFFF",
    cobalt = "#6493ED",
    orange = "#FF7F00",
    teal = "#1ECC91",
    sage = "#006401",
    brown = "#603814",
    tan = "#C69C6C",
    maroon = "#9D0B0E",
    salmon = "#F5999E"
}

constants.colors = {
    constants.color.white,
    constants.color.black,
    constants.color.red,
    constants.color.blue,
    constants.color.gray,
    constants.color.yellow,
    constants.color.green,
    constants.color.pink,
    constants.color.purple,
    constants.color.cyan,
    constants.color.cobalt,
    constants.color.orange,
    constants.color.teal,
    constants.color.sage,
    constants.color.brown,
    constants.color.tan,
    constants.color.maroon,
    constants.color.salmon
}

constants.customization = {
    rotation = {default = 133.144, left_shoulder = 80, right_shoulder = 190, arms = 80, gear = 0, color = 118},
}

constants.limits = {
    maximumPlayers = 15,
}

constants.widgets = {}

function constants.get()
    local blam = require "blam"
    local tagClasses = blam.tagClasses
    local findTag = blam.findTag
    local function findWidgetTag(partialName)
        return findTag(partialName, blam.tagClasses.uiWidgetDefinition)
    end
    local core = require "insurrection.core"

    constants.widgets = {
        --[[
        The following widgets are not used by the game, but are used by Insurrection.
        They are gathered here for later use.
        Ideally, we only want root widgets, so we can search for nested widgets later.
    ]]
        intro = findWidgetTag("intro\\intro_menu"),
        main = findWidgetTag("main\\main_menu"),
        dialog = findWidgetTag("dialog\\dialog_menu"),
        login = findWidgetTag("login_menu"),
        lobby = findWidgetTag("lobby_menu"),
        lobbyClient = findWidgetTag("lobby_client_menu"),
        dashboard = findWidgetTag("dashboard\\dashboard_menu"),
        customization = findWidgetTag("customization_menu"),
        pause = findWidgetTag("pause\\pause_menu"),
        nameplate = findWidgetTag("nameplate_current_profile"),
        tester = findWidgetTag("tester_menu"),
        settings = findWidgetTag("settings\\settings_menu"),
        chimera = findWidgetTag("chimera\\chimera_mod_menu"),
        color = findWidgetTag("customization_color_menu"),
        team = findWidgetTag("pause_choose_team_menu"),
        optic = findWidgetTag("optic\\optic_mod_menu"),
        biped = findWidgetTag("customization_biped\\customization_biped_menu"),
        browser = findWidgetTag("lobby_browser_menu"),
        legacyModalError = findWidgetTag("error_modal_fullscreen"),
        balltze = findWidgetTag("balltze\\balltze_mod_menu"),
        -- TODO Rename colors to color
        bipedColor = findWidgetTag("customization_biped_colors_menu"),
    }

    constants.sounds = {
        error = findTag("flag_failure", tagClasses.sound),
        back = findTag("back", tagClasses.sound),
        success = findTag("forward", tagClasses.sound),
        join = findTag("player_join", tagClasses.sound),
        leave = findTag("player_leave", tagClasses.sound)
    }

    constants.tagCollections = {
        nameplates = findTag("nameplates", tagClasses.tagCollection),
        maps = findTag("insurrection_maps", tagClasses.tagCollection)
    }

    constants.widgetCollections = {
        multiplayer = findTag("ui\\shell\\multiplayer", tagClasses.uiWidgetCollection)
    }

    if constants.tagCollections.nameplates then
        local nameplatesTagCollection = blam.tagCollection(constants.tagCollections.nameplates.id)
        if nameplatesTagCollection then
            ---@type table<string, tag>
            local nameplateBitmapTags = {}
            for _, tagId in ipairs(nameplatesTagCollection.tagList) do
                local tag = blam.getTag(tagId) --[[@as tag]]
                local nameplateId = core.getTagName(tag.path)
                if nameplateId and not nameplateBitmapTags[nameplateId] then
                    nameplateBitmapTags[nameplateId] = tag
                end
            end
            constants.nameplates = nameplateBitmapTags
        end
    end

    constants.bitmaps = {
        unknownMapPreview = findTag("unknown_map_preview", tagClasses.bitmap),
        customization = {
            left_shoulder = findTag("customization_left_shoulder_icons", tagClasses.bitmap),
            right_shoulder = findTag("customization_right_shoulder_icons", tagClasses.bitmap),
            regions = findTag("customization_icons", tagClasses.bitmap),
            helmet = findTag("customization_helmet_icons", tagClasses.bitmap),
            chest = findTag("customization_chest_icons", tagClasses.bitmap),
            gear = findTag("customization_gear_icons", tagClasses.bitmap),
            legs = findTag("customization_legs_icons", tagClasses.bitmap)
        }
    }
    local fontName = "geogrotesque-regular-"
    constants.fonts = {
        text = findTag(fontName .. "text", tagClasses.font),
        title = findTag(fontName .. "title", tagClasses.font),
        subtitle = findTag(fontName .. "subtitle", tagClasses.font),
        button = findTag(fontName .. "button", tagClasses.font),
        shadow = findTag(fontName .. "shadow", tagClasses.font)
    }

    constants.scenery = {
        christmasHat = engine.tag.findTags(constants.path.christmasHat, engine.tag.classes.scenery)[1],
        xmasObjects = engine.tag.findTags(constants.path.xmasObjects, engine.tag.classes.tagCollection)[1]
    }
    log("Loaded constants")
end

return constants

end,

["luna"] = function()
--------------------
-- Module: 'luna'
--------------------
local luna = {_VERSION = "2.5.0"}

luna.string = {}

local unpack = table.unpack or unpack
local find = string.find
local insert = table.insert
local format = string.format
local char = string.char

--- Split a string into a table of substrings by `sep`, or by each character if `sep` is not provided.
---@param s string
---@param sep? string
---@return string[]
---@nodiscard
function string.split(s, sep)
    assert(s ~= nil, "string.split: s must not be nil")
    local elements = {}
    -- Support splitting by any character or string.
    if sep == nil or sep == "" then
        for i = 1, #s do
            elements[i] = s:sub(i, i)
        end
    else
        -- Avoid using a pattern
        local position = 0
        for st, sp in function()
            return s:find(sep, position, true)
        end do
            insert(elements, s:sub(position, st - 1))
            position = sp + 1
        end
        insert(elements, s:sub(position))
    end
    return elements
end

--- Return a string with all leading whitespace removed.
---@param s string
---@return string
---@nodiscard
function string.ltrim(s)
    assert(s ~= nil, "string.ltrim: s must not be nil")
    return s:match "^%s*(.-)$"
end

--- Return a string with all trailing whitespace removed.
---@param s string
---@return string
---@nodiscard
function string.rtrim(s)
    assert(s ~= nil, "string.rtrim: s must not be nil")
    return s:match "^(.-)%s*$"
end

--- Return a string with all leading and trailing whitespace removed.
---@param s string
---@return string
---@nodiscard
function string.trim(s)
    assert(s ~= nil, "string.trim: s must not be nil")
    -- return s:match "^%s*(.-)%s*$"
    return string.ltrim(string.rtrim(s))
end

--- Return a string with all ocurrences of `pattern` replaced with `replacement`.
---
--- **NOTE**: Pattern is a plain string, not a Lua pattern. Use `string.gsub` for Lua patterns.
---@param s string
---@param pattern string
---@param replacement string
---@return string
---@nodiscard
function string.replace(s, pattern, replacement)
    assert(s ~= nil, "string.replace: s must not be nil")
    assert(pattern ~= nil, "string.replace: pattern must not be nil")
    assert(replacement ~= nil, "string.replace: replacement must not be nil")
    local pattern = pattern:escapep()
    local replaced, _ = s:gsub(pattern, replacement:escapep())
    return replaced
end

--- Return a hex encoded string.
---@param s string | number
---@return string
---@nodiscard
function string.tohex(s)
    assert(s ~= nil, "string.hex: s must not be nil")
    if type(s) == "number" then
        return format("%08.8x", s)
    end
    return (s:gsub(".", function(c)
        return format("%02x", string.byte(c))
    end))
end

--- Return a hex decoded string.
---@param s string
---@return string
---@nodiscard
function string.fromhex(s)
    assert(s ~= nil, "string.fromhex: s must not be nil")
    return (s:gsub("..", function(cc)
        return char(tonumber(cc, 16))
    end))
end

--- Resturn if a string starts with a given substring.
---@param s string
---@param start string
---@return boolean
---@nodiscard
function string.startswith(s, start)
    assert(s ~= nil, "string.startswith: s must not be nil")
    assert(start ~= nil, "string.startswith: start must not be nil")
    return string.sub(s, 1, string.len(start)) == start
end

--- Resturn if a string ends with a given substring.
---@param s string
---@param ending string
---@return boolean
---@nodiscard
function string.endswith(s, ending)
    assert(s ~= nil, "string.endswith: s must not be nil")
    assert(ending ~= nil, "string.endswith: ending must not be nil")
    return ending == "" or string.sub(s, -string.len(ending)) == ending
end

--- Return a string with template variables replaced with values from a table.
---@param s string
---@param t table<string, string | number | boolean>
---@return string
---@nodiscard
function string.template(s, t)
    assert(s ~= nil, "string.template: s must not be nil")
    assert(t ~= nil, "string.template: t must not be nil")
    return (s:gsub("{(.-)}", function(k)
        return t[k] or ""
    end))
end

--- Return if a string includes a given substring.
---@param s string
---@param substring string
---@return boolean
---@nodiscard
function string.includes(s, substring)
    assert(s ~= nil, "string.includes: s must not be nil")
    assert(substring ~= nil, "string.includes: substring must not be nil")
    return find(s, substring, 1, true) ~= nil
end

--- Return a string with all lua pattern characters escaped.
---@param s string
---@return string
---@nodiscard
function string.escapep(s)
    assert(s ~= nil, "string.escape: s must not be nil")
    return (s:gsub("%%", "%%%%"):gsub("%z", "%%z"):gsub("([%^%$%(%)%.%[%]%*%+%-%?])", "%%%1"))
end

--- Return how many times a substring appears in a string.
---@param s string
---@param substring string
---@return number
---@nodiscard
function string.count(s, substring)
    assert(s ~= nil, "string.count: s must not be nil")
    assert(substring ~= nil, "string.count: substring must not be nil")
    local count = 0
    for _ in s:gmatch(string.escapep(substring)) do
        count = count + 1
    end
    return count
end

luna.string.split = string.split
luna.string.ltrim = string.ltrim
luna.string.rtrim = string.rtrim
luna.string.trim = string.trim
luna.string.replace = string.replace
luna.string.tohex = string.tohex
luna.string.fromhex = string.fromhex
luna.string.startswith = string.startswith
luna.string.endswith = string.endswith
luna.string.template = string.template
luna.string.includes = string.includes
luna.string.escapep = string.escapep
luna.string.count = string.count

luna.table = {}

--- Return a deep copy of a table.
---
--- If the table contains other tables, they are copied as well, even metatables.
---@generic T
---@param t T
---@return T
---@nodiscard
function table.copy(t)
    assert(t ~= nil, "table.copy: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.copy: t must be a table")
    local u = {}
    ---@diagnostic disable-next-line: param-type-mismatch
    for k, v in pairs(t) do
        u[k] = type(v) == "table" and table.copy(v) or v
    end
    return setmetatable(u, getmetatable(t))
end

--- Find and return first index of `value` in `t`.
---@generic V
---@param t table<number, V>: { [number]: V }
---@param value V
---@return number?
---@nodiscard
function table.indexof(t, value)
    assert(t ~= nil, "table.find: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.find: t must be a table")
    ---@diagnostic disable-next-line: param-type-mismatch
    for i, v in ipairs(t) do
        if v == value then
            return i
        end
    end
end

--- Return a table with all keys and values swapped.
---@generic K, V
---@param t table<K, V>
---@return {[V]: K}
---@nodiscard
function table.flip(t)
    assert(t ~= nil, "table.flip: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.flip: t must be a table")
    local u = {}
    ---@diagnostic disable-next-line: param-type-mismatch
    for k, v in pairs(t) do
        u[v] = k
    end
    return u
end

--- Returns the first element of `t` that satisfies the predicate `f`.
--- If `f` is a value, it will return the first element that is equal to `f`.
---@generic K, V
---@param t table<K, V>
---@param f fun(v: V, k: K): boolean || `V`
---@return V?
---@nodiscard
function table.find(t, f)
    assert(t ~= nil, "table.find: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.find: t must be a table")
    assert(f ~= nil, "table.find: f must not be nil")
    if type(f) ~= "function" then
        local value = f
        f = function(v)
            return v == value
        end
    end
    ---@diagnostic disable-next-line: param-type-mismatch
    for k, v in pairs(t) do
        if f(v, k) then
            return v
        end
    end
end

--- Returns an array of all keys in `t`.
---@generic K, V
---@param t table<K, V>
---@return K[]
---@nodiscard
function table.keys(t)
    assert(t ~= nil, "table.keys: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.keys: t must be a table")
    local keys = {}
    ---@diagnostic disable-next-line: param-type-mismatch
    for k in pairs(t) do
        keys[#keys + 1] = k
    end
    return keys
end

--- Returns an array of all values in `t`.
---@generic K, V
---@param t table<K, V>
---@return V[]
---@nodiscard
function table.values(t)
    assert(t ~= nil, "table.values: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.values: t must be a table")
    local values = {}
    ---@diagnostic disable-next-line: param-type-mismatch
    for _, v in pairs(t) do
        values[#values + 1] = v
    end
    return values
end

--- Returns a table with all elements of `t` that satisfy the predicate `f`.
---@generic K, V
---@param t table<K, V>
---@param f fun(v: V, k: K): boolean
---@return {[K]: V}
---@nodiscard
function table.filter(t, f)
    assert(t ~= nil, "table.filter: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.filter: t must be a table")
    assert(f ~= nil, "table.filter: f must not be nil")
    assert(type(f) == "function", "table.filter: f must be a function")
    local filtered = {}
    ---@diagnostic disable-next-line: param-type-mismatch
    for k, v in pairs(t) do
        if f(v, k) then
            filtered[#filtered + 1] = v
        end
    end
    return filtered
end

--- Returns a table with all elements of `t` that satisfy the predicate `f`.
---
--- **NOTE**: It keeps original keys in the new table.
---@generic K, V
---@param t table<K, V>
---@param f fun(v: V, k: K): boolean
---@return {[K]: V}
---@nodiscard
function table.kfilter(t, f)
    assert(t ~= nil, "table.kfilter: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.kfilter: t must be a table")
    assert(f ~= nil, "table.kfilter: f must not be nil")
    assert(type(f) == "function", "table.kfilter: f must be a function")
    local filtered = {}
    ---@diagnostic disable-next-line: param-type-mismatch
    for k, v in pairs(t) do
        if f(v, k) then
            filtered[k] = v
        end
    end
    return filtered
end

--- Returns a table with all elements of `t` mapped by function `f`.
---
--- **NOTE**: It keeps original keys in the new table.
---@generic K, V
---@generic R
---@param t table<K, V>
---@param f fun(v: V, k: K): R
---@return {[K]: R}
-- @return R[]
---@nodiscard
function table.map(t, f)
    assert(t ~= nil, "table.map: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.map: t must be a table")
    assert(f ~= nil, "table.map: f must not be nil")
    assert(type(f) == "function", "table.map: f must be a function")
    local mapped = {}
    ---@diagnostic disable-next-line: param-type-mismatch
    for k, v in pairs(t) do
        mapped[k] = f(v, k)
    end
    return mapped
end

--- Returns a table merged from all tables passed as arguments.
---@generic K, V
---@vararg table<K, V>
---@return {[K]: V}
---@nodiscard
function table.merge(...)
    local merged = {}
    for _, t in ipairs {...} do
        for k, v in pairs(t) do
            merged[k] = v
        end
    end
    return merged
end

--- Returns a table with all values extended from all tables passed as arguments.
---@generic K, V
---@param t table<K, V>
---@vararg table<K, V>
---@return V[]
---@nodiscard
function table.extend(t, ...)
    assert(t ~= nil, "table.extend: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.extend: t must be a table")
    local extended = table.copy(t)
    for _, t in ipairs {...} do
        for _, v in pairs(t) do
            extended[#extended + 1] = v
        end
    end
    ---@diagnostic disable-next-line: return-type-mismatch
    return extended
end

--- Append values to a table.
--- It will append all values to the end of the table.
---@generic V
---@param t V[]
---@vararg V
---@return V[]
function table.append(t, ...)
    assert(t ~= nil, "table.append: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.append: t must be a table")
    local appended = table.copy(t)
    for _, v in ipairs {...} do
        appended[#appended + 1] = v
    end
    ---@diagnostic disable-next-line: return-type-mismatch
    return appended
end

--- Returns a table with all elements in reversed order.
---@generic T
---@param t T
---@return T
---@nodiscard
function table.reverse(t)
    assert(t ~= nil, "table.reverse: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.reverse: t must be a table")
    local reversed = {}
    for i = #t, 1, -1 do
        reversed[#reversed + 1] = t[i]
    end
    return reversed
end

--- Return a slice of a table, from `start` to `stop`
---
--- If `stop` is not provided, it will slice to the end of the table.
---@generic K, V
---@param t table<K, V>
---@param start number Index to start slice from.
---@param stop? number Index to stop slice at.
---@return {[K]: V}
---@nodiscard
function table.slice(t, start, stop)
    assert(t ~= nil, "table.slice: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.slice: t must be a table")
    assert(start ~= nil, "table.slice: start must not be nil")
    assert(type(start) == "number", "table.slice: start must be a number")
    if stop then
        assert(type(stop) == "number", "table.slice: stop must be a number")
    end
    local sliced = {}
    for i = start, stop or #t do
        sliced[#sliced + 1] = t[i]
    end
    return sliced
end

--- Return an array of chunks from a table, each chunk containing `size` elements.
---
--- If `t` is not evenly divisible by `size`, last chunk will contain the remaining elements.
---@generic K, V
---@param t table<K, V>
---@param size number
---@return {[K]: V[]}
---@nodiscard
function table.chunks(t, size)
    assert(t ~= nil, "table.chunks: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.chunks: t must be a table")
    assert(size ~= nil, "table.chunks: size must not be nil")
    assert(type(size) == "number", "table.chunks: size must be a number")
    local chunks = {}
    for i = 1, #t, size do
        ---@diagnostic disable-next-line: param-type-mismatch
        chunks[#chunks + 1] = table.slice(t, i, i + size - 1)
    end
    return chunks
end

--- Return the number of elements in a table.
--- Optionally `v` can be provided to count the number of occurrences of `v` in `t`.
---@generic T, V
---@param t T
---@param v? V
---@return number
function table.count(t, v)
    assert(t ~= nil, "table.count: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.count: t must be a table")
    if v == nil then
        return #t
    end
    local count = 0
    ---@diagnostic disable-next-line: param-type-mismatch
    for _, value in pairs(t) do
        if value == v then
            count = count + 1
        end
    end
    return count
end

--- Return the key of a value in a table.
--- If the value is not found, it will return `nil`.
---@generic K, V
---@param t table<K, V>
---@param v V
---@return K?
---@nodiscard
function table.keyof(t, v)
    assert(t ~= nil, "table.keyof: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.keyof: t must be a table")
    ---@diagnostic disable-next-line: param-type-mismatch
    for k, value in pairs(t) do
        if value == v then
            return k
        end
    end
end

--- Return a flattened table from a nested table.
--- All nested tables will be flattened into a single table.
--- If `t` is not a table, it will return `t`.
--- If `t` is a table with no nested tables, it will return `t`.
---@generic K, V
---@param t table<K, V>
---@return V[]
---@nodiscard
function table.flatten(t)
    assert(t ~= nil, "table.flatten: t must not be nil")
    assert(type(t) == "table" or type(t) == "userdata", "table.flatten: t must be a table")
    local flattened = {}
    ---@diagnostic disable-next-line: param-type-mismatch
    for _, v in pairs(t) do
        if type(v) == "table" then
            for _, value in pairs(table.flatten(v)) do
                flattened[#flattened + 1] = value
            end
        else
            flattened[#flattened + 1] = v
        end
    end
    return flattened
end

luna.table.copy = table.copy
luna.table.indexof = table.indexof
luna.table.flip = table.flip
luna.table.find = table.find
luna.table.keys = table.keys
luna.table.values = table.values
luna.table.filter = table.filter
luna.table.map = table.map
luna.table.merge = table.merge
luna.table.reverse = table.reverse
luna.table.slice = table.slice
luna.table.chunks = table.chunks
luna.table.count = table.count
luna.table.keyof = table.keyof
luna.table.flatten = table.flatten
luna.table.extend = table.extend
luna.table.append = table.append

luna.file = {}

--- Read a file into a string.
---@param path string
---@return string?
---@nodiscard
function luna.file.read(path)
    assert(path ~= nil, "file.read: path must not be nil")
    local file = io.open(path, "r")
    if file then
        local content = file:read "*a"
        file:close()
        return content
    end
end

--- Write a file from a string.
---@param path string
---@param content string
---@return boolean
function luna.file.write(path, content)
    assert(path ~= nil, "file.write: path must not be nil")
    assert(content ~= nil, "file.write: content must not be nil")
    local file = io.open(path, "w")
    if file then
        file:write(content)
        file:close()
        return true
    end
    return false
end

--- Append a string to a file.
---@param path string
---@param content string
---@return boolean
function luna.file.append(path, content)
    assert(path ~= nil, "file.append: path must not be nil")
    assert(content ~= nil, "file.append: content must not be nil")
    local file = io.open(path, "a")
    if file then
        file:write(content)
        file:close()
        return true
    end
    return false
end

--- Return if a file exists.
---@param path string
---@return boolean
---@nodiscard
function luna.file.exists(path)
    assert(path ~= nil, "file.exists: path must not be nil")
    local file = io.open(path, "r")
    if file then
        file:close()
        return true
    end
    return false
end

--- Attempt to remove a file.
---@param path string
---@return boolean
function luna.file.remove(path)
    assert(path ~= nil, "file.remove: path must not be nil")
    return os.remove(path)
end

--- Write file from a byte array.
---@param path string
---@param bytes number[]
---@return boolean
function luna.file.frombytes(path, bytes)
    assert(path ~= nil, "file.frombytes: path must not be nil")
    assert(bytes ~= nil, "file.frombytes: bytes must not be nil")
    local file = io.open(path, "wb")
    if file then
        file:write(char(unpack(bytes)))
        file:close()
        return true
    end
    return false
end

--- Return a byte array from a file.
---@param path string
---@return number[]?
---@nodiscard
function luna.file.tobytes(path)
    assert(path ~= nil, "file.tobytes: path must not be nil")
    local file = io.open(path, "rb")
    if file then
        local bytes = {string.byte(file:read "*a", 1, -1)}
        file:close()
        return bytes
    end
end

luna.binary = {}

--- Read a binary file into a string.
---@param path string
---@return string?
---@nodiscard
function luna.binary.read(path)
    assert(path ~= nil, "binary.read: path must not be nil")
    local file = io.open(path, "rb")
    if file then
        local content = file:read "*a"
        file:close()
        return content
    end
end

--- Write a binary file from a string.
---@param path string
---@param content string
---@return boolean
function luna.binary.write(path, content)
    assert(path ~= nil, "binary.write: path must not be nil")
    assert(content ~= nil, "binary.write: content must not be nil")
    local file = io.open(path, "wb")
    if file then
        file:write(content)
        file:close()
        return true
    end
    return false
end

--- Return a boolean from `v` if it is a boolean like value.
---@param v string | boolean | number
---@return boolean
function luna.bool(v)
    assert(v ~= nil, "bool: v must not be nil")
    return v == true or v == "true" or v == 1 or v == "1"
end

--- Return an integer from `v` if possible.
---
--- If `v` is not a number, it will return `fail`.
---@param v string
---@return integer
function tointeger(v)
    assert(v ~= nil, "int: v must not be nil")
    return tonumber(v, 10)
end
luna.int = tointeger

return luna

end,

["insurrection.mods.chimera"] = function()
--------------------
-- Module: 'insurrection.mods.chimera'
--------------------
local blam = require "blam"
local tagClasses = blam.tagClasses
local ini = require "ini"
local constants = require "insurrection.constants"
local luna = require "luna"
local split = luna.string.split
local engine = Engine

local core = require "insurrection.core"

local chimera = {}

---@class chimeraServer
---@field ip string
---@field port number
---@field password string
---@field dns string?

---@type chimeraServer[]
local servers = {}

---@class chimeraConfiguration
---@field server_list Serverlist
---@field custom_console Customconsole
---@field memory Memory
---@field controller any[]
---@field font_override Fontoverride
---@field custom_chat Customchat
---@field scoreboard Scoreboard
---@field hotkey Hotkey
---@field video_mode Videomode
---@field name Name
---@field error_handling any[]
---@field halo Halo

---@class Halo
---@field hash string
---@field background_playback number
---@field optimal_defaults number
---@field console number

---@class Name
---@field font string

---@class Videomode
---@field height string
---@field borderless number
---@field refresh_rate number
---@field windowed number
---@field width string
---@field enabled number
---@field vsync number

---@class Hotkey
---@field ctrl_alt_shift_3 string
---@field alt_6 string
---@field ctrl_1 string
---@field alt_5 string
---@field ctrl_6 string
---@field ctrl_0 string
---@field alt_shift_3 string
---@field ctrl_alt_shift_2 string
---@field ctrl_4 string
---@field alt_1 string
---@field enabled number
---@field ctrl_7 string
---@field alt_0 string
---@field alt_8 string
---@field ctrl_9 string
---@field alt_shift_4 string
---@field ctrl_2 string
---@field alt_4 string
---@field ctrl_alt_shift_1 string
---@field ctrl_3 string
---@field f12 string
---@field ctrl_8 string
---@field alt_3 string
---@field alt_shift_2 string
---@field alt_9 string
---@field alt_shift_1 string
---@field alt_7 string
---@field ctrl_5 string
---@field alt_2 string

---@class Scoreboard
---@field font string
---@field fade_time number

---@class Customchat
---@field chat_message_hide_on_console number
---@field server_message_color_r number
---@field chat_input_anchor string
---@field chat_input_x number
---@field chat_input_color_b number
---@field chat_message_color_red_g number
---@field server_message_anchor string
---@field chat_message_color_red_a number
---@field chat_message_anchor string
---@field server_fade_out_time number
---@field chat_fade_out_time number
---@field server_slide_time_length number
---@field chat_input_color_a number
---@field server_message_font string
---@field server_time_up number
---@field chat_input_font string
---@field chat_message_y number
---@field chat_message_w number
---@field chat_slide_time_length number
---@field chat_input_w number
---@field chat_message_color_red_b number
---@field chat_message_h number
---@field chat_message_x number
---@field chat_message_color_blue_g number
---@field server_message_hide_on_console number
---@field server_message_color_a number
---@field server_message_h_chat_open number
---@field chat_message_color_ffa_g number
---@field chat_input_color_r number
---@field server_message_color_g number
---@field chat_message_color_ffa_r number
---@field chat_message_color_ffa_a number
---@field chat_message_font string
---@field chat_input_color_g number
---@field chat_line_height number
---@field chat_message_color_ffa_b number
---@field chat_message_color_blue_a number
---@field chat_message_color_blue_b number
---@field chat_input_y number
---@field server_message_w number
---@field chat_message_h_chat_open number
---@field chat_message_color_red_r number
---@field server_message_color_b number
---@field server_message_y number
---@field chat_message_color_blue_r number
---@field server_message_x number
---@field chat_time_up number
---@field server_line_height number
---@field server_message_h number

---@class Fontoverride
---@field ticker_font_weight number
---@field large_font_override number
---@field large_font_shadow_offset_y number
---@field small_font_offset_x number
---@field smaller_font_family string
---@field smaller_font_offset_y number
---@field ticker_font_offset_x number
---@field smaller_font_shadow_offset_x number
---@field small_font_shadow_offset_y number
---@field system_font_shadow_offset_x number
---@field system_font_shadow_offset_y number
---@field smaller_font_offset_x number
---@field small_font_offset_y number
---@field ticker_font_offset_y number
---@field ticker_font_family string
---@field smaller_font_override number
---@field small_font_family string
---@field console_font_shadow_offset_y number
---@field system_font_y_offset number
---@field large_font_family string
---@field console_font_offset_y number
---@field system_font_override number
---@field system_font_weight number
---@field large_font_size number
---@field ticker_font_override number
---@field small_font_override number
---@field console_font_size number
---@field ticker_font_shadow_offset_x number
---@field large_font_weight number
---@field console_font_family string
---@field enabled number
---@field console_font_weight number
---@field ticker_font_shadow_offset_y number
---@field smaller_font_size number
---@field system_font_size number
---@field small_font_weight number
---@field console_font_override number
---@field ticker_font_size number
---@field smaller_font_shadow_offset_y number
---@field small_font_size number
---@field large_font_offset_y number
---@field system_font_x_offset number
---@field large_font_shadow_offset_x number
---@field smaller_font_weight number
---@field large_font_offset_x number
---@field console_font_shadow_offset_x number
---@field console_font_offset_x number
---@field system_font_family string
---@field small_font_shadow_offset_x number

---@class Memory
---@field download_font string
---@field enable_map_memory_buffer number
---@field map_size number

---@class Customconsole
---@field fade_time number
---@field fade_start number
---@field buffer_size_soft number
---@field enable_scrollback number
---@field line_height number
---@field enabled number
---@field buffer_size number
---@field x_margin number

---@class Serverlist
---@field auto_query number

local insurrectionFontOverride = {
    console_font_family = "Hack Bold",
    console_font_offset_x = 0,
    console_font_offset_y = 0,
    console_font_override = 1,
    console_font_shadow_offset_x = 2,
    console_font_shadow_offset_y = 2,
    console_font_size = 10,
    console_font_weight = 400,
    enabled = 1,
    large_font_family = "Geogrotesque-Regular",
    large_font_offset_x = 0,
    large_font_offset_y = 3,
    large_font_override = 1,
    large_font_shadow_offset_x = 2,
    large_font_shadow_offset_y = 2,
    large_font_size = 13,
    large_font_weight = 400,
    small_font_family = "Geogrotesque-Regular",
    small_font_offset_x = 0,
    small_font_offset_y = 3,
    small_font_override = 1,
    small_font_shadow_offset_x = 2,
    small_font_shadow_offset_y = 2,
    small_font_size = 10,
    small_font_weight = 400,
    smaller_font_family = "Geogrotesque-Regular",
    smaller_font_offset_x = 0,
    smaller_font_offset_y = 4,
    smaller_font_override = 1,
    smaller_font_shadow_offset_x = 0,
    smaller_font_shadow_offset_y = 0,
    smaller_font_size = 10,
    smaller_font_weight = 400,
    system_font_family = "Geogrotesque-Regular",
    system_font_override = 1,
    system_font_shadow_offset_x = 2,
    system_font_shadow_offset_y = 2,
    system_font_size = 11,
    system_font_weight = 400,
    system_font_x_offset = 0,
    system_font_y_offset = 1,
    ticker_font_family = "Geogrotesque-Regular",
    ticker_font_offset_x = 0,
    ticker_font_offset_y = 0,
    ticker_font_override = 1,
    ticker_font_shadow_offset_x = 2,
    ticker_font_shadow_offset_y = 2,
    ticker_font_size = 18,
    ticker_font_weight = 400
}

local chimeraFontOverride = {
    console_font_family = "Hack Bold",
    console_font_offset_x = 0,
    console_font_offset_y = 0,
    console_font_override = 1,
    console_font_shadow_offset_x = 2,
    console_font_shadow_offset_y = 2,
    console_font_size = 14,
    console_font_weight = 400,
    enabled = 1,
    large_font_family = "Interstate-Bold",
    large_font_offset_x = 0,
    large_font_offset_y = 1,
    large_font_override = 1,
    large_font_shadow_offset_x = 2,
    large_font_shadow_offset_y = 2,
    large_font_size = 20,
    large_font_weight = 400,
    small_font_family = "Interstate-Bold",
    small_font_offset_x = 0,
    small_font_offset_y = 3,
    small_font_override = 1,
    small_font_shadow_offset_x = 2,
    small_font_shadow_offset_y = 2,
    small_font_size = 15,
    small_font_weight = 400,
    smaller_font_family = "Interstate-Bold",
    smaller_font_offset_x = 0,
    smaller_font_offset_y = 4,
    smaller_font_override = 1,
    smaller_font_shadow_offset_x = 2,
    smaller_font_shadow_offset_y = 2,
    smaller_font_size = 11,
    smaller_font_weight = 400,
    system_font_family = "Interstate-Bold",
    system_font_override = 1,
    system_font_shadow_offset_x = 2,
    system_font_shadow_offset_y = 2,
    system_font_size = 20,
    system_font_weight = 400,
    system_font_x_offset = 0,
    system_font_y_offset = 1,
    ticker_font_family = "Lucida Console",
    ticker_font_offset_x = 0,
    ticker_font_offset_y = 0,
    ticker_font_override = 1,
    ticker_font_shadow_offset_x = 2,
    ticker_font_shadow_offset_y = 2,
    ticker_font_size = 11,
    ticker_font_weight = 400
}

---Get chimera configuration
---@return chimeraConfiguration?
function chimera.getConfiguration()
    local configIni = luna.file.read("chimera.ini")
    if configIni then
        ---@type chimeraConfiguration
        local configuration = ini.decode(configIni)
        return configuration
    end
end

---Save chimera configuration
---@param configuration chimeraConfiguration
function chimera.saveConfiguration(configuration)
    local configIni = ini.encode(configuration)
    return luna.file.write("chimera.ini", configIni)
end

function chimera.setupFonts(revert)
    local chimeraIni = luna.file.read("chimera.ini")
    if chimeraIni then
        local configuration = chimera.getConfiguration()
        if configuration then
            configuration.font_override = insurrectionFontOverride
            if revert then
                configuration.font_override = chimeraFontOverride
            end
            return chimera.saveConfiguration(configuration)
        end
    end
    return false
end

function chimera.enableBlockServerIp()
    local preferences = chimera.getPreferences()
    if preferences then
        preferences.chimera_block_server_ip = 1
        return chimera.savePreferences(preferences)
    end
    return false
end

---@class chimeraPreferences
---@field chimera_devmode 1|0
---@field chimera_budget 1|0
---@field chimera_widescreen_fix 1|0
---@field chimera_block_gametype_rules 1
---@field chimera_block_gametype_indicator true | false
---@field chimera_uncap_cinematic 1|0
---@field chimera_fov number | "auto"
---@field chimera_af 1|0
---@field chimera_invert_shader_flags 1|0
---@field chimera_show_coordinates 1|0
---@field chimera_mouse_sensitivity number[]
---@field chimera_throttle_fps number
---@field chimera_block_zoom_blur 1|0
---@field chimera_block_server_ip 1|0
---@field chimera_block_buffering 1|0
---@field chimera_show_fps 1|0
---@field chimera_block_loading_screen 1|0
---@field chimera_block_hold_f1 1|0
---@field chimera_block_mouse_acceleration 1|0

---Get chimera preferences
---@return chimeraPreferences?
function chimera.getPreferences()
    local preferencesTxt = luna.file.read(
                               core.getMyGamesHaloCEPath() .. "\\chimera\\preferences.txt")
    if preferencesTxt then
        local preferences = {}
        -- Split the file into lines and iterate over them
        for line in preferencesTxt:gmatch("[^\n]+") do
            -- Search for line that starts with "chimera"
            if line:sub(1, 7) == "chimera" then
                -- Get key and value from line separeted by a space, support numbers as well
                local key, value = line:match("([^ ]+) (.+)")
                if key and value then
                    -- Convert value to number if possible
                    local number = tonumber(value)
                    if number then
                        value = number
                    end
                    -- Convert boolean to flag number if possible
                    if value == "true" then
                        value = 1
                    elseif value == "false" then
                        value = 0
                    end
                    preferences[key] = value
                end
            end
        end
        return preferences
    end
end

function chimera.savePreferences(preferences)
    local preferencesTxt = ""
    for key, value in pairs(preferences) do
        preferencesTxt = preferencesTxt .. key .. " " .. value .. "\n"
    end
    return luna.file.write(core.getMyGamesHaloCEPath() .. "\\chimera\\preferences.txt",
                          preferencesTxt)
end

function chimera.executeCommand(command)
    if not execute_chimera_command then
        engine.core.consolePrint("execute_chimera_command is not available.")
        return false
    end
    local result, error = pcall(execute_chimera_command, command, true)
    if result then
        execute_script("cls")
        return true
    end
    engine.core.consolePrint(error)
    return false
end

local function safeFontOverride(...)
    pcall(create_font_override, ...)
end

function chimera.fontOverride()
    logger:debug("Overriding Chimera font...")
    if create_font_override then
        if constants.fonts.text then
            safeFontOverride(constants.fonts.text.id, "Geogrotesque-Regular", 14, 400, 2, 2, 1, 1)
        end
        if constants.fonts.title then
            safeFontOverride(constants.fonts.title.id, "Geogrotesque-Regular", 18, 400, 2, 2, 0, 0)
        end
        if constants.fonts.subtitle then
            safeFontOverride(constants.fonts.subtitle.id, "Geogrotesque-Regular", 10, 400, 2, 2, 0, 0)
        end
        if constants.fonts.button then
            safeFontOverride(constants.fonts.button.id, "Geogrotesque-Regular", 13, 400, 2, 2, 1, 1)
        end
        if constants.fonts.shadow then
            safeFontOverride(constants.fonts.shadow.id, "Geogrotesque-Regular", 10, 400, 0, 0, 0, 0)
        end
        return true
    end
    logger:error("create_font_override is not available.")
    return false
end

return chimera

end,

["ini"] = function()
--------------------
-- Module: 'ini'
--------------------
-------------------------------------------------------------------------------
--- INI Module
--- Dynodzzo, Sledmine
--- It has never been that simple to use ini files with Lua
----------------------------------------------------------------------------------
local ini = {
    _VERSION = 1.0,
    _LICENSE = [[
	Copyright (c) 2012 Carreras Nicolas
	
	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:
	
	The above copyright notice and this permission notice shall be included in all
	copies or substantial portions of the Software.
	
	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER G
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
	SOFTWARE.
]]
}

--- Returns a table containing all the data from an ini string
---@param fileString string Ini encoded string
---@return table Table containing all data from the ini string
function ini.decode(fileString)
    local position, lines = 0, {}
    for st, sp in function()
        return string.find(fileString, "\n", position, true)
    end do
        table.insert(lines, string.sub(fileString, position, st - 1))
        position = sp + 1
    end
    table.insert(lines, string.sub(fileString, position))
    local data = {}
    local section
    for lineNumber, line in pairs(lines) do
        if string.sub(line, 1, 1) == "[" then
            section = string.sub(line, 2, string.find(line, "]") - 1)
            data[section] = {}
        elseif section then
            --local key, value = string.match(line, "([^=]*)=(.*)")
            local key, value = string.match(line, "^([%w|_]+)%s-=%s-(.+)$")
            if key and value then
                value = value:gsub("\r", "")
                if (tonumber(value)) then
                    value = tonumber(value)
                elseif (value == "true") then
                    value = true
                elseif (value == "false") then
                    value = false
                end
                if (tonumber(key)) then
                    key = tonumber(key)
                end
                data[section][key] = value
            end
        end
    end
    return data
end

--- Returns a ini encoded string given data
---@param data table Table containing all data from the ini file
---@return string String encoded as an ini file
function ini.encode(data)
    local content = ""
    for section, param in pairs(data) do
        content = content .. ("[%s]\n"):format(section)
        for key, value in pairs(param) do
            content = content .. ("%s=%s\n"):format(key, tostring(value))
        end
        content = content .. "\n"
    end
    return content
end

--- Returns a table containing all the data from an ini file
---@param fileName string Path to the file to load
---@return table Table containing all data from the ini file
function ini.load(fileName)
    assert(type(fileName) == "string", "Parameter \"fileName\" must be a string.")
    local file = assert(io.open(fileName, "r"), "Error loading file : " .. fileName)
    local data = ini.decode(file:read("*"))
    file:close()
    return data
end

--- Saves all the data from a table to an ini file
---@param fileName string The name of the ini file to fill
---@param data table The table containing all the data to store
function ini.save(fileName, data)
    assert(type(fileName) == "string", "Parameter \"fileName\" must be a string.")
    assert(type(data) == "table", "Parameter \"data\" must be a table.")
    local file = assert(io.open(fileName, "w+b"), "Error loading file :" .. fileName)
    file:write(ini.encode(data))
    file:close()
end

return ini

end,

["insurrection.interface"] = function()
--------------------
-- Module: 'insurrection.interface'
--------------------
local balltze = Balltze
local engine = Engine
local openWidget = engine.userInterface.openWidget
local playSound = engine.userInterface.playSound
local components = require "insurrection.components"
local menus = require "insurrection.menus"
local button = require "insurrection.components.button"
local translations = require "insurrection.translations"
local blam = require "blam"
local core = require "insurrection.core"
local uiWidgetTag = blam.uiWidgetDefinition
local uiWidgetCollection = blam.uiWidgetCollection
local constants = require "insurrection.constants"
local chimera = require "insurrection.mods.chimera"

local interface = {}
interface.shared = {}

shared = interface.shared

function interface.load()
    components.free()
    -- constants.get()
    IsUICompatible = true
    -- local scenery = engine.tag.findTags("halo", engine.tag.classes.scenery)[1]
    -- assert(scenery, "Error, no scenery tag found")
    -- engine.core.consolePrint("{}", tostring(inspect(table.keys(scenery.data))))
    if IsUICompatible then

        -- Load Insurrection features
        log("Loading Insurrection patches...")
        core.loadInsurrectionPatches()

        -- Components initialization
        log("Initializing components...")
        interface.loadProfileNameplate()
        components.cleanAllEditableWidgets()

        -- interface.animate()
        if constants.widgets.login then
            require "insurrection.components.dynamic.dialog"()
            require "insurrection.components.dynamic.customizationColorMenu"()
            require "insurrection.components.dynamic.settingsMenu"()
            require "insurrection.components.dynamic.loginMenu"()
            require "insurrection.components.dynamic.dashboardMenu"()
            require "insurrection.components.dynamic.customizationMenu"()
            require "insurrection.components.dynamic.lobbyMenu"()
            require "insurrection.components.dynamic.lobbyMenuClient"()
            -- TODO Find a better way to toggle biped preview generation
            require "insurrection.components.dynamic.customizationBipedMenu" {
                isBipedPreviewGenEnabled = false
            }
            require "insurrection.components.dynamic.lobbyBrowserMenu"()
            require "insurrection.components.dynamic.customizationBipedColorMenu"()

            local errorModalLegacy = components.new(constants.widgets.legacyModalError.id)
            errorModalLegacy:onOpen(function()
                log("Checking if lobby is active...")
                if api.session.lobbyKey and engine.map.getCurrentMapHeader().name == "ui" then
                    api.lobby(api.session.lobbyKey)
                end
            end)

            local pause = components.new(constants.widgets.pause.id)
            pause:onClose(function()
                interface.blur(false)
            end)

            -- TODO BALLTZE MIGRATE
            local tester = components.new(constants.widgets.tester.id)
            local testerAnimTest = components.new(tester:findChildWidgetTag("anim_test").id)
            testerAnimTest:animate()
            -- testerAnimTest:setAnimation(0.6, "horizontal", 100, 300, "ease in")
        end

        if constants.widgets.chimera then
            require "insurrection.components.dynamic.chimeraMenu"()
        end
        if constants.widgets.optic then
            require "insurrection.components.dynamic.opticMenu"()
        end
        if constants.widgets.balltze then
            require "insurrection.components.dynamic.balltzeMenu"()
        end

        -- Insurrection is running outside the UI
        if constants.widgetCollections.multiplayer then
            local multiplayerWidgetsCollection = uiWidgetCollection(
                                                     constants.widgetCollections.multiplayer.id)
            if multiplayerWidgetsCollection then
                local pause = components.new(multiplayerWidgetsCollection.tagList[1])
                if pause then
                    log(multiplayerWidgetsCollection.tagList[1])
                    if constants.widgets.pause then
                        log("Insurrection may load in external map...")
                        require "insurrection.components.dynamic.dialog"()
                        local insurrectionPause = components.new(constants.widgets.pause.id)
                        local resumeButton = button.new(
                                                 insurrectionPause:findChildWidgetTag(
                                                     "resume_game_button").id)
                        local stockResumeButton = button.new(pause:findChildWidgetTag("resume").id)
                        local exitButton = button.new(
                                               insurrectionPause:findChildWidgetTag("exit_button").id)
                        resumeButton:onClick(function()
                            log("Resume button clicked")
                            interface.blur(false)
                            interface.sound("back")
                        end)
                        stockResumeButton:onClick(function()
                            log("Stock resume button clicked")
                            interface.sound("back")
                        end)
                        exitButton:onClick(function()
                            api.deleteLobby()
                        end)
                        local insurrectionChooseTeam = components.new(constants.widgets.team.id)
                        local blueTeamButton = button.new(
                                                   insurrectionChooseTeam:findChildWidgetTag(
                                                       "blue_team_button").id)
                        local redTeamButton = button.new(
                                                  insurrectionChooseTeam:findChildWidgetTag(
                                                      "red_team_button").id)
                        blueTeamButton:onClick(function()
                            interface.blur(false)
                        end)
                        redTeamButton:onClick(function()
                            interface.blur(false)
                        end)
                        pause:onOpen(function()
                            log("Opening stock pause menu...")
                            if not InvalidatePauseOverride then
                                -- if engine.map.getCurrentMapHeader().name ~= "ui" and (engine.netgame.getServerType() == "dedicated" or DebugMode) then
                                if engine.map.getCurrentMapHeader().name ~= "ui" then
                                    log("Opening Insurrection pause menu...")
                                    interface.blur(true)
                                    balltze.features.setUIAspectRatio(16, 9)
                                    menus.pause()
                                end
                            else
                                balltze.features.setUIAspectRatio(4, 3)
                            end
                            InvalidatePauseOverride = false
                        end)
                        insurrectionPause:onClose(function()
                            interface.blur(false)
                            balltze.features.setUIAspectRatio(4, 3)
                        end)
                        local openMapPauseButton = button.new(
                                                       insurrectionPause:findChildWidgetTag(
                                                           "open_map_pause").id)
                        openMapPauseButton:onClick(function()
                            interface.blur(false)
                            InvalidatePauseOverride = true
                            menus.open(pause.tagId)
                        end)
                    end
                end

            end
        end

        -- Set up some chimera configs
        if engine.map.getCurrentMapHeader().name == "ui" and false then
            local preferences = chimera.getPreferences() or {}
            -- TODO Check forced server name preference
            local notServerIpBlocking = not preferences.chimera_block_server_ip or
                                            preferences.chimera_block_server_ip == 0
            if notServerIpBlocking then
                interface.shared.dialog:onClose(function()
                    preferences.chimera_block_server_ip = 1
                    chimera.savePreferences(preferences)
                    if not chimera.executeCommand("chimera_block_server_ip 1") then
                        execute_script("quit")
                    end
                end)
                interface.dialog("WARNING", translations.eng.block_server_ips_subtitle,
                                 translations.eng.block_server_ips_message)
            end
        end
    end
end

function interface.loadProfileNameplate(nameplateId)
    if not constants.tagCollections.nameplates then
        log("Error, no nameplates collection found")
        return
    end
    local nameplate = components.new(constants.widgets.nameplate.id)
    local nameplatesTagCollection = blam.tagCollection(constants.tagCollections.nameplates.id)
    if nameplatesTagCollection then
        local nameplateBitmapTags = {}
        for _, tagId in ipairs(nameplatesTagCollection.tagList) do
            local tag = blam.getTag(tagId) --[[@as tag]]
            local nameplateId = core.getTagName(tag.path)
            if nameplateId and not nameplateBitmapTags[nameplateId] then
                nameplateBitmapTags[nameplateId] = tag
            end
        end
        nameplate:animate()
        if nameplateId then
            if not nameplateBitmapTags[nameplateId] then
                log("Invalid nameplate id: " .. nameplateId)
                return
            end
            nameplate.widgetDefinition.backgroundBitmap = nameplateBitmapTags[nameplateId].id
            return
        end
        log("Loading nameplate from settings...")
        local settings = core.loadSettings()
        if settings and settings.nameplate and nameplateBitmapTags[settings.nameplate] then
            nameplate.widgetDefinition.backgroundBitmap = nameplateBitmapTags[settings.nameplate].id
        end
    end
end

---Animates UI elements by animating background bitmap
---@param widgetTagHandleValue number
---@param willRepeat? boolean
function interface.animateUIWidgetBackground(widgetTagHandleValue, willRepeat)
    -- local willRepeat = willRepeat or true
    local widget = engine.userInterface.findWidget(widgetTagHandleValue)
    if widget then
        local widgetTag = engine.tag.getTag(widgetTagHandleValue,
                                            engine.tag.classes.uiWidgetDefinition)
        -- TODO Change balltze to catch exceptions in timer as they propagate and crash the game
        assert(widgetTag, "Error, widget tag not found")
        -- print("Animating widget {}", widgetTag.path)
        local bitmapTag = engine.tag.getTag(widgetTag.data.backgroundBitmap.tagHandle.value,
                                            engine.tag.classes.bitmap)
        if bitmapTag then
            local bitmapTagData = bitmapTag.data
            if bitmapTagData.bitmapData.count > 1 then
                if widget.bitmapIndex < bitmapTagData.bitmapData.count then
                    widget.bitmapIndex = widget.bitmapIndex + 1
                else
                    widget.bitmapIndex = 0
                end
            end
        end
    end
end

---Show a dialog message on the screen
---@param titleText "WARNING" | "INFORMATION" | "ERROR" | string
---@param subtitleText string
---@param bodyText string
function interface.dialog(titleText, subtitleText, bodyText)
    if constants.sounds then
        if titleText == "WARNING" or titleText == "ERROR" then
            log(constants.sounds.error.path)
            playSound(constants.sounds.error.id)
        else
            playSound(constants.sounds.success.id)
        end
    end
    local dialog = shared.dialog

    local title = components.new(dialog:findChildWidgetTag("title").id)
    title:setText(titleText)

    local subtitle = components.new(dialog:findChildWidgetTag("subtitle").id)
    subtitle:setText(subtitleText)

    local body = components.new(dialog:findChildWidgetTag("text").id)
    body:setText(bodyText)

    if titleText == "ERROR" then
        openWidget(constants.widgets.dialog.id, false)
    else
        openWidget(constants.widgets.dialog.id, true)
    end
end

---Play a special interface sound
---@param sound "error" | "success" | "back" | "join" | "leave"
function interface.sound(sound)
    if not (constants.sounds.error and constants.sounds.success and constants.sounds.back and
        constants.sounds.join and constants.sounds.leave) then
        log("Error, no custom sounds found", "error")
        return
    end
    if sound == "error" then
        playSound(constants.sounds.error.id)
    elseif sound == "success" then
        playSound(constants.sounds.success.id)
    elseif sound == "back" then
        playSound(constants.sounds.back.id)
    elseif sound == "join" then
        playSound(constants.sounds.join.id)
    elseif sound == "leave" then
        playSound(constants.sounds.leave.id)
    else
        log("Invalid sound: " .. sound, "error")
    end
end

---Blur UI background
---@param enable boolean
function interface.blur(enable)
    if enable then
        execute_script([[(begin
        (show_hud false)
        (cinematic_screen_effect_start true)
        (cinematic_screen_effect_set_convolution 3 1 1 2 0)
        (cinematic_screen_effect_start false)
    )]])
    else
        execute_script([[(begin
            (show_hud true)
            (cinematic_stop)
        )]])
    end
end

---Close current interface widget
---@param closeAllWidgets boolean
function interface.close(closeAllWidgets)
    if closeAllWidgets then
        while core.getCurrentUIWidgetTag() do
            engine.userInterface.closeWidget()
        end
        return
    end
    engine.userInterface.closeWidget()
    return
end

-- TODO Move this variable to a better global namespace
BipedRotation = 0
function interface.rotateCustomizationBiped()
    local mouse = core.getMouseState()

    if mouse.rightClick > 0 then
        local objectId = core.getCustomizationObjectId()
        if objectId then
            local object = blam.object(get_object(objectId))
            assert(object)
            BipedRotation = BipedRotation + mouse.right * 3
            if BipedRotation > 360 then
                BipedRotation = 0
            end
            blam.rotateObject(object, BipedRotation, 0, 0)
        end
    end
end

---Handle interface on tick events
function interface.onTick()
    local currentWidgetTag = core.getCurrentUIWidgetTag()
    if not currentWidgetTag then
        return
    end
    if constants.widgets.biped then
        if currentWidgetTag.id == constants.widgets.biped.id or currentWidgetTag.id ==
            constants.widgets.bipedColor.id then
            interface.rotateCustomizationBiped()
        end
    end
    -- Animate UI widgets
    -- TODO Try to animate only widgets that are visible
    for tagId, component in pairs(components.widgets) do
        if component.isBackgroundAnimated then
            interface.animateUIWidgetBackground(tagId)
        end
    end
    -- Inventory protptype code
    -- if core.getMouseState().scrollClick > 0 then
    --    if focusedWidgetTagId then
    --        local widget = blam.uiWidgetDefinition(focusedWidgetTagId)
    --        console_out(widget.width .. " " .. widget.height)
    --        local x, y = core.getWidgetCursorPosition()
    --        -- console_out("X: " .. x .. " Y: " .. y)
    --        local props = core.getWidgetValues(focusedWidgetTagId)
    --        -- console_out("Focused widget: " .. focusedWidgetTagId .. " X: " .. props.left_bound .. " Y: " .. props.top_bound)
    --        core.setWidgetValues(focusedWidgetTagId, {left_bound = x - (widget.width / 2), top_bound = y - (widget.height / 2)})
    --    end
    -- end
end

function interface.changeAspectRatio()
    local widgetTag = core.getCurrentUIWidgetTag()
    if widgetTag then
        if widgetTag.data.bounds.right > 640 then
            -- Change UI aspect ratio
            log("Setting UI aspect ratio to 16:9")
            balltze.features.setUIAspectRatio(16, 9)
        end
        -- Enable menu blur
        execute_script("menu_blur_on")

        -- Set network timeout to 5 seconds (keeps connection alive at loading huge maps)
        -- NOTE! This is meant to help server side loading time, not client side
        execute_script("network_connect_timeout 15000")
    else
        balltze.features.setUIAspectRatio(4, 3)
    end
end

---Fade screen in or out
---@param type "in" | "out"
---@param duration number
function interface.fade(type, duration)
    local type = type == "in" and "in" or "out"
    execute_script("fade_" .. type .. " 0 0 0 " .. duration)
end

return interface

end,

["insurrection.menus"] = function()
--------------------
-- Module: 'insurrection.menus'
--------------------
local engine = Engine
local constants = require "insurrection.constants"
local playSound = engine.userInterface.playSound
local core = require "insurrection.core"

local menus = {}

local function openWidget(...)
    playSound(constants.sounds.success.id)
    engine.userInterface.openWidget(...)
end

function menus.dashboard()
    openWidget(constants.widgets.dashboard.id, true)
end

function menus.customization()
    openWidget(constants.widgets.customization.id, true)
end

--- Open the lobby widget
---@param client? boolean
function menus.lobby(client)
    local currentWidgetTag = core.getCurrentUIWidgetTag()
    local isRejoiningLobby = false
    if currentWidgetTag then
        -- This is a workaround for the game's behavior when rejoining the lobby
        --
        -- Game throws a legacy message saying "The game has closed down" when rejoining the lobby
        -- Because the normal game ended, we need to rejoin the lobby and get rid of this message
        isRejoiningLobby = currentWidgetTag.id == constants.widgets.legacyModalError.id
        log("Rejoining lobby: {}", isRejoiningLobby)
    end
    if client then
        logger:debug("Opening lobby client")
        openWidget(constants.widgets.lobbyClient.id, not isRejoiningLobby)
        return
    end
    openWidget(constants.widgets.lobby.id, not isRejoiningLobby)
end

function menus.pause()
    openWidget(constants.widgets.pause.id, false)
end

function menus.open(widget, replace)
    openWidget(widget, replace or false)
end

function menus.biped()
    openWidget(constants.widgets.biped.id, true)
end

function menus.lobbies()
    openWidget(constants.widgets.browser.id, true)
end

function menus.bipedColor()
    openWidget(constants.widgets.bipedColor.id, true)
end

return menus


end,

["insurrection.components.button"] = function()
--------------------
-- Module: 'insurrection.components.button'
--------------------
---@diagnostic disable: duplicate-set-field, duplicate-doc-field
local core = require "insurrection.core"
local components = require "insurrection.components"

---@class uiComponentButtonClass : uiComponent
local button = setmetatable({}, {__index = components})

---@class uiComponentButtonEvents : uiComponentEvents
---@field onClick fun(value?: string | boolean | number):boolean | nil

---@class uiComponentButton : uiComponentButtonClass
---@field events uiComponentButtonEvents

---@param tagId number
---@return uiComponentButton
function button.new(tagId)
    local instance = setmetatable(components.new(tagId), {__index = button}) --[[@as uiComponentButton]]
    return instance
end

---@param self uiComponentButton
function button.onClick(self, callback)
    self.events.onClick = callback
end

return button

end,

["insurrection.translations"] = function()
--------------------
-- Module: 'insurrection.translations'
--------------------
return {
    eng = {
        block_server_ips_subtitle = "Chimera is not blocking server IP addresses.",
        block_server_ips_message = "Insurrection might close the game to enable this feature.\nThis is a security measure to prevent server IP addresses being leaked."
    }
}

end,

["insurrection.components.dynamic.dialog"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.dialog'
--------------------
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local button = require "insurrection.components.button"

return function()
    local dialog = components.new(constants.widgets.dialog.id)
    local dialogBackButton = button.new(dialog:findChildWidgetTag("ok").id)
    dialogBackButton:onClick(function()
        if dialog.events.onClose then
            dialog.events.onClose()
        end
    end)
    shared.dialog = dialog
end

end,

["insurrection.components.dynamic.customizationColorMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.customizationColorMenu'
--------------------
local color = require "color"
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local list = require "insurrection.components.list"
local button = require "insurrection.components.button"
local blam = require "blam"
local core = require "insurrection.core"
local getWidgetValues = core.getWidgetValues
local setWidgetValues = core.setWidgetValues
local menus = require "insurrection.menus"
local delay = require"insurrection.utils".delay

return function()
    local customizationColor = components.new(constants.widgets.color.id)

    local optionsId = customizationColor:findChildWidgetTag("options").id
    local customizationColorListOptions = list.new(optionsId)

    local actionsId = customizationColor:findChildWidgetTag("actions").id
    local customizationColorListActions = list.new(actionsId)

    local saveId = customizationColorListActions:findChildWidgetTag("save").id
    local customizationColorSaveButton = button.new(saveId)
    local colorButtons = customizationColorListOptions:getChildWidgetTags()

    local page = 0
    local multiplier = #colorButtons - 2

    local scrollDown = button.new(colorButtons[1].id)
    local scrollUp = button.new(colorButtons[#colorButtons - 1].id)
    scrollDown:onClick(function()
    end)
    scrollUp:onClick(function()
    end)

    updateColorMenu = function()
        Engine.core.consolePrint("Updating color menu")

        local profile = core.getPlayerProfile()
        local colorFromGame = constants.colors[profile.colorIndex]
        local colorName = table.flip(constants.color)[colorFromGame]:lower()
        logger:debug("Color: {}", colorName)

        local customizationBiped = core.getCustomizationObjectData().biped
        local r, g, b = color.hexToDec(colorFromGame)
        customizationBiped.colorCLowerRed = r
        customizationBiped.colorCLowerGreen = g
        customizationBiped.colorCLowerBlue = b

        customizationBiped.colorDLowerRed = r
        customizationBiped.colorDLowerGreen = g
        customizationBiped.colorDLowerBlue = b

        for buttonIndex, tag in pairs(colorButtons) do
            if buttonIndex > 1 and buttonIndex < #colorButtons - 1 then
                local colorButton = button.new(tag.id)
                local colorButtonText = button.new(colorButton:findChildWidgetTag("_text").id)
                local colorIcon = button.new(colorButton:findChildWidgetTag("_icon").id)
                --local colorValue = constants.colors[]
                --if colorValue then
                --   local colorIndex = table.flip(constants.colors)[colorValue] - 1
                --   setWidgetValues(colorIcon.tag.id, {bitmapIndex = colorIndex})
                    colorButton:onClick(function()
                        Engine.core.consolePrint(profile.colorIndex)
                        Engine.core.consolePrint(core.getPlayerProfile().colorIndex)
                        --local r, g, b = color.hexToDec(colorValue)
                        --customizationBiped.colorCLowerRed = r
                        --customizationBiped.colorCLowerGreen = g
                        --customizationBiped.colorCLowerBlue = b
                    end)
                --end
            end
        end
        return false
    end
    customizationColor:onOpen(function()
        execute_script "object_create customization_biped"

        -- delay(45, updateColorMenu)
        updateColorMenu()
    end)
end

end,

["insurrection.components.list"] = function()
--------------------
-- Module: 'insurrection.components.list'
--------------------
---@diagnostic disable: inject-field
local isNull = require"blam".isNull
local components = require "insurrection.components"
local button = require "insurrection.components.button"
local core = require "insurrection.core"
local interface = require "insurrection.interface"

---@class uiComponentListClass : uiComponent
local list = setmetatable({
    ---@type string
    type = "list",
    ---@type number
    firstWidgetIndex = nil,
    ---@type number
    lastWidgetIndex = nil,
    ---@type number
    currentItemIndex = 1,
    ---@type number
    lastSelectedItemIndex = nil,
    ---@type uiComponentListItem[]
    items = {},
    ---@type uiWidgetDefinitionChild[]
    backupChildWidgets = {},
    ---@type boolean
    isScrollable = true,
    ---@type uiComponentBar
    scrollBar = nil
}, {__index = components})

---@class uiComponentListItem
---@field label? string
---@field value string | boolean | number | any
---@field bitmap? number | fun(uiComponent: uiComponent)

---@class uiComponentListEvents : uiComponentEvents
---@field onSelect fun(item: uiComponentListItem)
---@field onScroll fun(item: uiComponentListItem)

---@class uiComponentList : uiComponentListClass
---@field events uiComponentListEvents

---@param tagId number
---@param firstWidgetIndex? number
---@param lastWidgetIndex? number
---@return uiComponentList
function list.new(tagId, firstWidgetIndex, lastWidgetIndex)
    local instance = setmetatable(components.new(tagId), {__index = list}) --[[@as uiComponentList]]
    instance.firstWidgetIndex = firstWidgetIndex or 1
    instance.lastWidgetIndex = lastWidgetIndex or instance.widgetDefinition.childWidgetsCount
    return instance
end

---@param self uiComponentList
---@param callback fun(item: uiComponentListItem)
function list.onSelect(self, callback)
    self.events.onSelect = callback
end

---@param self uiComponentList
---@param callback fun(item: uiComponentListItem)
function list.onScroll(self, callback)
    self.events.onScroll = callback
end

---@param self uiComponentList
function list.scroll(self, direction, isFromMouse)
    local itemIndex = self.currentItemIndex + direction
    if itemIndex < 1 then
        if not isFromMouse then
            interface.sound("error")
        end
        itemIndex = 1
    elseif itemIndex > #self.items then
        itemIndex = #self.items
    end

    local lastWidgetIndex = self.lastWidgetIndex
    if self.isScrollable then
        lastWidgetIndex = lastWidgetIndex - 2
    end
    local maximumDisplayableIndex = #self.items - lastWidgetIndex + 1
    if maximumDisplayableIndex < 1 then
        maximumDisplayableIndex = 1
    end
    if itemIndex > maximumDisplayableIndex then
        if not isFromMouse then
            interface.sound("error")
        end
        itemIndex = maximumDisplayableIndex
    end

    self.currentItemIndex = itemIndex
    if self.events.onScroll then
        self.events.onScroll(self.items[itemIndex])
    end
    self:refresh()
end

local floor = math.floor
local function round(x, p)
    p = p or 1
    return floor(x / p + .5) * p
end

---@param self uiComponentList
function list.refresh(self)
    local items = self.items
    local itemIndex = self.currentItemIndex
    local widgetDefinition = self.widgetDefinition
    local firstWidgetIndex = self.firstWidgetIndex
    local lastWidgetIndex = self.lastWidgetIndex
    if self.isScrollable then
        firstWidgetIndex = firstWidgetIndex + 1
        lastWidgetIndex = lastWidgetIndex - 1

        if self.scrollBar then
            local scroll = self.scrollBar
            local scrollBackground = scroll.widgetDefinition
            local scrollBar = scroll:findChildWidgetDefinition("bar_value")
            local elementsCount = #items
            local visibleElementsCount = lastWidgetIndex - firstWidgetIndex + 1
            local isHorizontal = self.widgetDefinition.dpadLeftRightTabsThruChildren
            local size = scrollBackground.height
            if isHorizontal then
                size = scrollBackground.width
            end
            barSizePerElement = size / elementsCount
            local isScrollBarVisible = elementsCount > visibleElementsCount
            if isScrollBarVisible then
                local scrollPosition = round((itemIndex - 1) * barSizePerElement)
                if elementsCount > 0 then
                    if isHorizontal then
                        scrollBar.width = round(barSizePerElement * visibleElementsCount)
                        scroll:setBarValues{position = {x = scrollPosition}}
                    else
                        scrollBar.height = round(barSizePerElement * visibleElementsCount)
                        scroll:setBarValues{position = {y = scrollPosition}}
                    end
                end
            else
                if isHorizontal then
                    scrollBar.width = scrollBackground.width
                    scroll:setBarValues{position = {x = 0}}
                else
                    scrollBar.height = scrollBackground.height
                    scroll:setBarValues{position = {y = 0}}
                end
            end
        end
    end
    for widgetIndex = firstWidgetIndex, lastWidgetIndex do
        local item = items[itemIndex]
        local childWidget = widgetDefinition.childWidgets[widgetIndex]
        if item then
            core.setWidgetValues(childWidget.widgetTag, {neverReceiveEvents = false, visible = true})
            if childWidget and not isNull(childWidget.widgetTag) then
                local listButton = button.new(childWidget.widgetTag)
                if item.label then
                    listButton:setText(item.label)
                end
                local onSelect = self.events.onSelect
                -- TODO Check if we need to apply a select event even if no onSelect callback is provided
                if onSelect then
                    local lastSelectedItemIndex = itemIndex
                    listButton:onClick(function()
                        self.lastSelectedItemIndex = lastSelectedItemIndex
                        onSelect(item)
                    end)
                end
                if item.bitmap then
                    if type(item.bitmap) == "number" then
                        -- TODO We might need to animate bitmaps when selected by a function
                        listButton:animate()
                        listButton.widgetDefinition.backgroundBitmap = item.bitmap --[[@as number]]
                    elseif type(item.bitmap) == "function" then
                        item.bitmap(listButton)
                    end
                end
                itemIndex = itemIndex + 1
            end
        else
            core.setWidgetValues(childWidget.widgetTag, {neverReceiveEvents = true, visible = false})
        end
    end
end

---@param self uiComponentList
---@param items uiComponentListItem[]
function list.setItems(self, items)
    local widgetDefinition = self.widgetDefinition
    if not widgetDefinition.type == 3 then
        log("Widget: " .. self.tag.path .. " is being used as a list but is not a column_list")
    end
    -- if not (#items > 0) then
    --    error("setItems requires at least one item")
    -- end
    if not self.backupChildWidgets then
        self.backupChildWidgets = table.map(widgetDefinition.childWidgets, function(childWidget)
            return {
                widgetTag = childWidget.widgetTag,
                name = childWidget.name,
                customControllerIndex = childWidget.customControllerIndex,
                verticalOffset = childWidget.verticalOffset,
                horizontalOffset = childWidget.horizontalOffset
            }
        end)
    end
    self.items = items
    -- if self.currentItemIndex > #items then
    --    self.currentItemIndex = 1
    -- end
    for widgetIndex = self.firstWidgetIndex, self.lastWidgetIndex do
        local widgetTagId = widgetDefinition.childWidgets[widgetIndex].widgetTag
        button.new(widgetTagId)
    end
    self.currentItemIndex = 1
    self.lastSelectedItemIndex = nil
    if self.isScrollable then
        local firstWidgetTagId = widgetDefinition.childWidgets[self.firstWidgetIndex].widgetTag
        local lastWidgetTagId = widgetDefinition.childWidgets[self.lastWidgetIndex].widgetTag
        local firstWidget = button.new(firstWidgetTagId)
        local lastWidget = button.new(lastWidgetTagId)
        firstWidget:onClick(function()
            self:scroll(-1)
        end)
        lastWidget:onClick(function()
            self:scroll(1)
        end)
    end
    self:refresh()
end

---@param self uiComponentList
function list.getSelectedItem(self)
    if self:getWidgetValues() then
        return self.items[self.lastSelectedItemIndex]
    end
end

---@param self uiComponentList
---@param isScrollable boolean
function list.scrollable(self, isScrollable)
    self.isScrollable = isScrollable
end

---@param self uiComponentList
function list.getCurrentItem(self)
    return self.items[self.currentItemIndex]
end

---@param self uiComponentList
---@param itemIndex number
function list.setCurrentItemIndex(self, itemIndex)
    self.currentItemIndex = itemIndex
    self:refresh()
end

---@param self uiComponentList
---@param scrollBar uiComponentBar
function list.setScrollBar(self, scrollBar)
    self.scrollBar = scrollBar
end

---@param self uiComponentList
function list.isHorizontal(self)
    return self.widgetDefinition.dpadLeftRightTabsThruChildren
end

return list

end,

["insurrection.components.dynamic.settingsMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.settingsMenu'
--------------------
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local list = require "insurrection.components.list"
local button = require "insurrection.components.button"
local blam = require "blam"

return function()
    -- Hard code settings description text change, because the game doesn't support it
    local settings = components.new(constants.widgets.settings.id)
    local settingsOptions = list.new(settings:findChildWidgetTag("settings_menu_options").id)
    -- TODO Add extended description widget support to lua-blam
    local settingsDescription = components.new(blam.findTag("settings_elements_description",
                                                            blam.tagClasses.uiWidgetDefinition).id)
    local settingsDescriptionText = components.new(
                                        settingsDescription:findChildWidgetTag(
                                            "settings_elements_description_data").id)
    for i = 1, settingsOptions.widgetDefinition.childWidgetsCount - 1 do
        local childWidget = settingsOptions.widgetDefinition.childWidgets[i]
        local button = button.new(childWidget.widgetTag)
        button:onFocus(function()
            settingsDescriptionText.widgetDefinition.stringListIndex = i - 1
        end)
    end
end

end,

["insurrection.components.dynamic.loginMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.loginMenu'
--------------------
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local button = require "insurrection.components.button"
local core = require "insurrection.core"
local interface = require "insurrection.interface"
local engine = Engine

return function()
    local login = components.new(constants.widgets.login.id)
    local usernameInput = button.new(login:findChildWidgetTag("username_input").id)
    local passwordInput = button.new(login:findChildWidgetTag("password_input").id)
    login:onOpen(function(previousWidget)
        -- Start discord presence only if script is loaded in the UI map, prevent crashes in other maps
        if engine.map.getCurrentMapHeader().name == "ui" and (previousWidget and previousWidget.handle.value == constants.widgets.main.id) then
            if not discord.ready then
                discord.startPresence()
            end
        end
    end)
    -- Load login data
    local savedUserName, savedPassword = core.loadCredentials()
    if savedUserName and savedPassword then
        usernameInput:setText(savedUserName)
        passwordInput:setText(savedPassword, "*")
    end
    local loginButton = button.new(login:findChildWidgetTag("login_button").id)
    loginButton:onClick(function()
        local username, password = usernameInput:getText(), passwordInput:getText()
        if username and password and username ~= "" and password ~= "" then
            core.saveCredentials(username, password)
            api.login(username, password)
        else
            interface.dialog("WARNING", "", "Please enter a username and password.")
        end
    end)
    local registerButton = button.new(login:findChildWidgetTag("register_button").id)
    -- dialogBackButton:onClick(function()
    --    os.execute("start https://discord.shadowmods.net")
    --    dialogBackButton:onClick(nil)
    -- end)
    registerButton:onClick(function()
        interface.dialog("INFORMATION", "Join us on our Discord server!",
                         "We have a Discord Bot for the registering process:\n\n\nhttps://discord.shadowmods.net")
    end)
end

end,

["insurrection.components.dynamic.dashboardMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.dashboardMenu'
--------------------
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local button = require "insurrection.components.button"
local input = require "insurrection.components.input"
local interface = require "insurrection.interface"
local menus = require "insurrection.menus"

return function()
    local dashboard = components.new(constants.widgets.dashboard.id)
    local createLobbyButton = button.new(dashboard:findChildWidgetTag("create_lobby_button").id)
    createLobbyButton:onClick(function()
        api.lobby()
    end)
    local joinLobbyButton = button.new(dashboard:findChildWidgetTag("join_lobby_button").id)
    local joinLobbyInput = input.new(dashboard:findChildWidgetTag("lobby_key_input").id)
    joinLobbyButton:onClick(function()
        local lobbyKey = joinLobbyInput:getText()
        if lobbyKey ~= "" then
            api.lobby(lobbyKey)
        else
            interface.dialog("WARNING", "", "Please specify a lobby key to join.")
        end
    end)
    local browseLobby = button.new(dashboard:findChildWidgetTag("browse_lobby").id)
    browseLobby:onClick(function()
        menus.lobbies()
    end)

    dashboard:onOpen(function()
        execute_script("set_ui_background")
        api.stopRefreshLobby()
        discord.clearParty()
        discord.setState("Playing Insurrection", "In the dashboard")
    end)
    dashboard:onClose(function()
        api.stopRefreshLobby()
        discord.clearPresence()
    end)

    local customizationButton = button.new(dashboard:findChildWidgetTag("customization_button").id)
    customizationButton:onClick(function()
        interface.fade("in", 60)
    end)
end

end,

["insurrection.components.input"] = function()
--------------------
-- Module: 'insurrection.components.input'
--------------------
local button = require "insurrection.components.button"

---@class uiComponentInputClass : uiComponentButton
local input = setmetatable({type = "input"}, {__index = button})

---@class uiComponentInputEvents : uiComponentEvents
---@field onInputText fun(text: string) | nil

---@class uiComponentInput : uiComponentInputClass
---@field events uiComponentInputEvents

---@param tagId number
---@return uiComponentInput
function input.new(tagId)
    local instance = setmetatable(button.new(tagId), {__index = input}) --[[@as uiComponentInput]]
    return instance
end

---@param self uiComponentInput
function input.onInputText(self, callback)
    if self.widgetDefinition.type == 1 then
        self.events.onInputText = callback
    else
        error("onInputText can only be used on uiWidgetDefinition of type 1")
    end
end

return input

end,

["insurrection.components.dynamic.customizationMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.customizationMenu'
--------------------
local balltze = Balltze
local engine = Engine
local blam = require "blam"
local bar = require "insurrection.components.bar"
local interface = require "insurrection.interface"
local tagClasses = blam.tagClasses
local findTag = blam.findTag
local color = require "color"
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local button = require "insurrection.components.button"
local list = require "insurrection.components.list"
local utils = require "insurrection.utils"
local menus = require "insurrection.menus"
local core = require "insurrection.core"
local getState = require "insurrection.redux.getState"
local t = utils.snakeCaseToUpperTitleCase

return function()
    local selectedProject
    local selectedBiped
    local profile = core.getPlayerProfile()
    local state = getState()

    -- Get customization widget menu
    local customization = components.new(constants.widgets.customization.id)

    local scrollBar = bar.new(customization:findChildWidgetTag("customization_scroll").id, "scroll")
    -- Get nameplate list widget
    local nameplatesList = list.new(customization:findChildWidgetTag("nameplates_options").id, 1, 9)
    nameplatesList:setScrollBar(scrollBar)

    -- Get select bipeds widget
    local selectBipedsWrapper = components.new(findTag("select_bipeds",
                                                       tagClasses.uiWidgetDefinition).id)
    local selectProjectsList = list.new(selectBipedsWrapper:findChildWidgetTag("select_project").id)
    selectProjectsList:setScrollBar(scrollBar)
    local bipedsList = list.new(selectBipedsWrapper:findChildWidgetTag("select_project_biped").id)

    local customizationOptions = components.new(customization:findChildWidgetTag(
                                                    "customization_options").id)
    local backButton = button.new(customizationOptions:findChildWidgetTag("back").id)
    local customizationTypesList = components.new(customization:findChildWidgetTag("types").id)
    local selectNameplateButton = button.new(
                                      customizationTypesList:findChildWidgetTag("nameplates").id)

    local selectBipedButton = button.new(customizationTypesList:findChildWidgetTag("bipeds").id)

    local currentBipedLabel = components.new(
                                  customization:findChildWidgetTag("current_biped_label").id)
    -- Get save customization button
    local saveCustomizationButton = button.new(
                                        customization:findChildWidgetTag("save_customization").id)

    nameplatesList:onSelect(function(item)
        log(item.value)
        interface.loadProfileNameplate(item.value)
        -- nameplatePreview.widgetDefinition.backgroundBitmap = item.bitmap --[[@as number]]
    end)
    local sortedNameplates = table.keys(constants.nameplates)
    table.sort(sortedNameplates)
    sortedNameplates = table.map(sortedNameplates, function(nameplateId)
        return {value = nameplateId, bitmap = constants.nameplates[nameplateId].id}
    end)
    nameplatesList:setItems(sortedNameplates)

    ---Select a customization biped
    ---@param bipedPath string
    ---@param regions? number[]
    ---@param visor integer
    local handleSelectBiped = function(bipedPath, regions, visor)
        log("Selected biped path: {}", bipedPath)
        log("Regions: {}", inspect(regions))
        selectedBiped = bipedPath

        execute_script("object_create customization_biped")

        local tagEntry = engine.tag.findTags(bipedPath, engine.tag.classes.biped)[1]
        if not tagEntry then
            logger:error("Custom external biped tag: {} not found", bipedPath)
            return
        end

        local bipedData = tagEntry.data
        if not bipedData then
            logger:error("Biped tag: {} has no data", bipedPath)
            return
        end

        local bipedTag = engine.tag.getTag(tagEntry.handle.value, engine.tag.classes.biped)
        assert(bipedTag, "Biped tag data not found")
        -- TODO Check if this is the right way to remove creation effect
        bipedTag.data.creationEffect.tagHandle.value = 0xFFFFFFFF
        -- TODO Remove this when biped animations are fixed in coop evolved
        if not (tagEntry.path:includes "marine" or tagEntry.path:includes "grunt") then
            -- FIXME This does not work with Balltze as weapons count is read only
            -- bipedTag.data.weapons.count = 0

            -- TODO Try to use unit remove weapon function or something similar
            local bipedTag = blam.bipedTag(tagEntry.handle.value)
            bipedTag.weaponCount = 0
        end

        local scenario = blam.scenario(0)
        assert(scenario)
        -- Respawn biped object from scenario as it is safer than doing it from lua
        for _, biped in pairs(scenario.bipeds) do
            local sceneryName = scenario.objectNames[biped.nameIndex + 1]
            if sceneryName == "customization_biped" then
                local newPaletteList = scenario.bipedPaletteList
                -- Replace scenario biped tag with custom biped tag
                if newPaletteList[biped.typeIndex + 1] ~= tagEntry.handle.value then
                    newPaletteList[biped.typeIndex + 1] = tagEntry.handle.value
                    scenario.bipedPaletteList = newPaletteList
                    execute_script "object_destroy customization_biped"
                    execute_script "object_create customization_biped"
                    execute_script "fade_screen_in"
                    logger:debug("Biped tag replaced")
                end
                break
            end
        end

        local customizationObjectData = core.getCustomizationObjectData()
        local customizationBiped = customizationObjectData.biped
        assert(customizationBiped, "No customization biped found")

        local bipedName = t(utils.path(bipedPath:replace("_mp", "")).name)
        currentBipedLabel:setText(bipedName)

        -- local colorFromGame = constants.colors[profile.colorIndex]
        local colorFromInsurrection = core.getCustomizationObjectData().color.custom
        log("Color from Insurrection: {}", colorFromInsurrection)

        local r, g, b = color.hexToDec(colorFromInsurrection.primary)
        customizationBiped.colorCLowerRed = r
        customizationBiped.colorCLowerGreen = g
        customizationBiped.colorCLowerBlue = b

        local r, g, b = color.hexToDec(colorFromInsurrection.secondary)
        customizationBiped.colorDLowerRed = r
        customizationBiped.colorDLowerGreen = g
        customizationBiped.colorDLowerBlue = b
        if regions then
            logger:debug("Setting regions")
            for regionIndex, permutationIndex in pairs(regions) do
                core.setObjectPermutationSafely(customizationBiped, regionIndex, permutationIndex)
            end
        end

        if visor then
            customizationBiped.shaderPermutationIndex = visor
        end
    end

    local function handleLoadNameplates()
        execute_script("set_customization_background")
        nameplatesList:refresh()
        currentBipedLabel:setText("")
        selectBipedsWrapper:replace(nameplatesList.tagId)
    end

    ---@param projectName? string
    local function handleLoadProject(projectName)
        local savedBipeds = api.getSavedBipeds()
        local lastSavedProject = (core.loadSettings() or {}).lastSavedProject
        local defaultProject = state.available.customization[table.keys(state.available
                                                                            .customization)[1]]
        -- TODO Load last selected project
        local projectName = projectName or lastSavedProject or
                                table.keys(state.available.customization)[1]
        selectedProject = projectName
        local project = state.available.customization[projectName] or defaultProject
        logger:debug("Selected project: {}", projectName)
        local bipeds = table.map(project.tags, function(bipedPath)
            return {label = "CUSTOMIZE", value = bipedPath:replace(".biped", "")}
        end)
        bipeds = utils.sortTableAlphabetically(bipeds)
        bipedsList:setItems(bipeds)
        local regions
        local bipedPath = bipeds[1].value
        local savedBiped = savedBipeds[projectName]
        local visor = 0
        if savedBiped then
            local bipedIsStillAvailable = table.find(bipeds, function(biped)
                return biped.value == savedBiped.path
            end)
            if bipedIsStillAvailable then
                bipedsList:setCurrentItemIndex(table.indexof(bipeds,
                                                             table.find(bipeds, function(biped)
                    return biped.value == savedBiped.path
                end)) or 1)
                bipedPath = savedBiped.path
                regions = savedBiped.regions
                visor = savedBiped.visor
            end
        end
        -- log(savedBiped)
        handleSelectBiped(bipedPath, regions, visor)
    end

    ---@param openList boolean?
    local function handleLoadBipeds(openList)
        if openList then
            nameplatesList:replace(selectBipedsWrapper.tagId)
        end
        local state = getState()

        local projects = table.keys(state.available.customization)
        table.sort(projects)

        selectProjectsList:onSelect(function(item)
            handleLoadProject(item.value)
        end)

        selectProjectsList:setItems(table.map(projects, function(project)
            return {label = utils.snakeCaseToTitleCase(project), value = project}
        end))

        bipedsList:onScroll(function(item)
            handleSelectBiped(item.value)
        end)

        bipedsList:onSelect(function(item)
            -- FIXME State from components should have been reset at this point
            -- Due to current implementation, setItems is not executed when coming back from another
            -- widget, so we need to reset the state manually or prevent scenarios where we get
            -- and invalid state
            log(item.value)
            if core.getCustomizationObjectId() then
                menus.biped()
            end
        end)

    end

    selectBipedButton:onClick(function()
        handleLoadProject()
        handleLoadBipeds(true)
    end)

    selectNameplateButton:onClick(function()
        handleLoadNameplates()
    end)

    saveCustomizationButton:onClick(function()
        local currentNameplateId
        if settings and settings.nameplate then
            currentNameplateId = settings.nameplate
        end
        local selectedNameplateItem = nameplatesList:getSelectedItem() or
                                          {value = currentNameplateId}
        local nameplate = selectedNameplateItem.value
        local bipeds
        if selectedNameplateItem then
            nameplate = selectedNameplateItem.value
        end
        if selectedProject and selectedBiped then
            local objectId = core.getCustomizationObjectId()
            local objectData = core.getCustomizationObjectData()
            local visor = objectData.visor
            local regions = objectData.regions
            if objectId and regions then
                local project = selectedProject
                bipeds = {
                    [project] = selectedBiped .. "+" .. table.concat(regions, "+") .. "+" .. visor
                }
            end
        end
        local color = core.getCustomizationObjectData().color.custom
        log("Nameplate: {}", nameplate)
        log("Bipeds: {}", inspect(bipeds))
        log("Color: {}", inspect(color))
        if IsDebugLocalCustomizationEnabled then
            logger:warning("Local customization can not be saved")
            bipeds = {}
        end
        api.playerProfileEdit({
            nameplate = nameplate,
            bipeds = bipeds,
            color = {primary = color.primary, secondary = color.secondary}
        })
    end)

    customization:onClose(function()
        interface.fade("in", 30)
        interface.loadProfileNameplate()
        if api.session.player.color then
            LastColorCustomization.primary = core.getCustomizationColorByValue(api.session.player
                                                                                   .color.primary)
            LastColorCustomization.secondary = core.getCustomizationColorByValue(api.session.player
                                                                                     .color
                                                                                     .secondary)
        end
    end)

    customization:onOpen(function(previousWidgetTag)
        log("Customization menu opened")
        discord.setState("Playing Insurrection", "In the customization menu")
        profile = core.getPlayerProfile()
        if previousWidgetTag then
            if previousWidgetTag.handle.value == constants.widgets.biped.id then
                handleLoadBipeds(true)
                return
            end
        end
        local savedBipeds = api.getSavedBipeds()
        if savedBipeds then
            handleLoadProject()
            handleLoadBipeds()
        end
        handleLoadNameplates()
    end)

    backButton:onClick(function()
        customization.events.onClose()
    end)

    return function()
    end
end

end,

["insurrection.components.bar"] = function()
--------------------
-- Module: 'insurrection.components.bar'
--------------------
---@diagnostic disable: duplicate-set-field, duplicate-doc-field, inject-field
local core       = require "insurrection.core"
local components = require "insurrection.components"

---@class uiComponentBar : uiComponent
local bar = setmetatable({
    ---@type string
    type = "bar",
    ---@type "progress" | "scroll"
    barType = "progress",
    ---@type "horizontal" | "vertical"
    orientation = "horizontal",
}, {__index = components})

---@param tagId number
---@param barType? "progress" | "scroll"
---@return uiComponentBar
function bar.new(tagId, barType)
    local instance = setmetatable(components.new(tagId), {__index = bar}) --[[@as uiComponentBar]]
    instance.barType = barType or "scroll"
    instance.orientation = instance.widgetDefinition.width > instance.widgetDefinition.height and "horizontal" or "vertical"
    return instance
end

---@param self uiComponentBar
---@param values uiWidgetValues
function bar.setBarValues(self, values)
    core.setWidgetValues(self:findChildWidgetTag("bar_value").id, values)
end

return bar

end,

["insurrection.redux.getState"] = function()
--------------------
-- Module: 'insurrection.redux.getState'
--------------------
local store = require "insurrection.redux.store"

---Returns the current state of the redux store
---@return interfaceState
return function ()
    ---@type interfaceState
    local state = store:getState()
    return state
end
end,

["insurrection.redux.store"] = function()
--------------------
-- Module: 'insurrection.redux.store'
--------------------
local createStore = require "redux".createStore
local interfaceReducer = require "insurrection.redux.reducers.interfaceReducer"
local updateInterface = require "insurrection.interface".lobbyInit

local store = createStore(interfaceReducer)
--store:subscribe(updateInterface)

return store
end,

["redux"] = function()
--------------------
-- Module: 'redux'
--------------------
local actionTypes = {INIT = "@@lua-redux/INIT"}

local function inverse(table)
    local newTable = {}
    for k, v in pairs(table) do
        newTable[v] = k
    end
    return newTable
end

---@class reduxAction
---@field type string
---@field payload any

---Create a redux store for the application state
---@param reducer fun(action:string, payload:any)
---@param preloadedState any
local function createStore(reducer, preloadedState)
    local store = {reducer = reducer, state = preloadedState, subscribers = {}}

    ---Subscribe a function callback to this store
    ---@param callback function
    ---@return function unsubscribe Unsubscribe function for this store
    function store:subscribe(callback)
        local i = table.insert(self.subscribers, callback)
        return function()
            table.remove(self.subscribers, inverse(self.subscribers)[callback])
        end
    end

    ---Dispatch an action event to the store
    ---@param action reduxAction
    function store:dispatch(action)
        self.state = self.reducer(self.state, action)
        for k, v in pairs(self.subscribers) do
            v()
        end
    end

    ---Return the state of the store
    ---@return any
    function store:getState()
        return self.state
    end

    ---Replace the reducer of this store with a given one
    ---@return any
    function store:replaceReducer(reducer)
        self.reducer = reducer
        self:dispatch({type = actionTypes.INIT})
    end

    store:dispatch({type = actionTypes.INIT})

    return store
end

return {actionTypes = actionTypes, createStore = createStore}

end,

["insurrection.redux.reducers.interfaceReducer"] = function()
--------------------
-- Module: 'insurrection.redux.reducers.interfaceReducer'
--------------------
local redux = require "redux"
local actions = require "insurrection.redux.actions"
local glue = require "glue"
local chunks = glue.chunks

---@class interfaceState
---@field isLoading boolean
---@field player insurrectionPlayer?
---@field available availableParameters?
---@field lobby insurrectionLobby?
---@field lobbies insurrectionLobby[]?

---@type interfaceState
local defaultState = {
    isLoading = false,
    lobbyKey = nil,
    available = nil,
    lobby = nil,
    player = nil,
    lobbies = nil
}

---Game interface reducer
---@param state interfaceState
---@param action reduxAction
---@return interfaceState
local function interfaceReducer(state, action)
    if logger then
        log(action.type)
    end
    if action.type == redux.actionTypes.INIT then
        return table.copy(defaultState)
    elseif action.type == actions.types.SET_IS_LOADING then
        state.isLoading = action.payload
        return state
    elseif action.type == actions.types.SET_AVAILABLE_RESOURCES then
        state.available = action.payload
        return state
    elseif action.type == actions.types.SET_LOBBY then
        state.lobby = action.payload.lobby
        return state
    elseif action.type == actions.types.SET_LOBBIES then
        state.lobbies = action.payload
        return state
    else
        error("Undefined redux action type!")
    end
end

return interfaceReducer

end,

["insurrection.redux.actions"] = function()
--------------------
-- Module: 'insurrection.redux.actions'
--------------------
local redux = require "redux"
local actions = {}

actions.types = {
    SET_IS_LOADING = "SET_IS_LOADING",
    SET_LOBBY = "SET_LOBBY",
    SET_AVAILABLE_RESOURCES = "SET_AVAILABLE_RESOURCES",
    RESET = "RESET",
    SET_PLAYER = "SET_PLAYER",
    SET_LOBBIES = "SET_LOBBIES"
}

function actions.setIsLoading(loading)
    return {type = actions.types.SET_IS_LOADING, payload = loading}
end

function actions.setLobby(key, lobby)
    return {type = actions.types.SET_LOBBY, payload = {key = key, lobby = lobby}}
end

function actions.reset()
    return {type = redux.actionTypes.INIT}
end

function actions.setAvailableResources(available)
    return {type = actions.types.SET_AVAILABLE_RESOURCES, payload = available}
end

function actions.setPlayer(player)
    return {type = actions.types.SET_PLAYER, payload = player}
end

function actions.setLobbies(lobbies)
    return {type = actions.types.SET_LOBBIES, payload = lobbies}
end

return actions

end,

["glue"] = function()
--------------------
-- Module: 'glue'
--------------------
-- Lua extended vocabulary of basic tools.
-- Written by Cosmin Apreutesei. Public domain.
-- Modifications by Sled
local glue = {}

local min, max, floor, ceil, log = math.min, math.max, math.floor, math.ceil, math.log
---@diagnostic disable-next-line: deprecated
local select, unpack, pairs, rawget = select, unpack, pairs, rawget

-- math -----------------------------------------------------------------------

function glue.round(x, p)
    p = p or 1
    return floor(x / p + .5) * p
end

function glue.floor(x, p)
    p = p or 1
    return floor(x / p) * p
end

function glue.ceil(x, p)
    p = p or 1
    return ceil(x / p) * p
end

glue.snap = glue.round

function glue.clamp(x, x0, x1)
    return min(max(x, x0), x1)
end

function glue.lerp(x, x0, x1, y0, y1)
    return y0 + (x - x0) * ((y1 - y0) / (x1 - x0))
end

function glue.nextpow2(x)
    return max(0, 2 ^ (ceil(log(x) / log(2))))
end

-- varargs --------------------------------------------------------------------

if table.pack then
    glue.pack = table.pack
else
    function glue.pack(...)
        return {n = select("#", ...), ...}
    end
end

-- always use this because table.unpack's default j is #t not t.n.
function glue.unpack(t, i, j)
    return unpack(t, i or 1, j or t.n or #t)
end

-- tables ---------------------------------------------------------------------

---Count the keys in a table with an optional upper limit
---@param t table
---@param maxn integer
---@return integer
function glue.count(t, maxn)
    local maxn = maxn or (1 / 0)
    local n = 0
    for _ in pairs(t) do
        n = n + 1
        if n >= maxn then
            break
        end
    end
    return n
end

---Reverse keys with values
---@param t table<any, any>
---@return table
function glue.index(t)
    local dt = {}
    for k, v in pairs(t) do
        dt[v] = k
    end
    return dt
end

local function desc_cmp(a, b)
    return a > b
end

---Put keys in a list, optionally sorted
---@param t table
---@param cmp? boolean | '"asc"' | '"desc"'
---@return string[] | integer[] | any[]
function glue.keys(t, cmp)
    local dt = {}
    for k in pairs(t) do
        dt[#dt + 1] = k
    end
    if cmp == true or cmp == "asc" then
        table.sort(dt)
    elseif cmp == "desc" then
        table.sort(dt, desc_cmp)
    elseif cmp then
        table.sort(dt, cmp)
    end
    return dt
end

---Stateless pairs() that iterate elements in key order
---@param t table
---@param cmp? boolean | '"asc"' | '"desc"'
---@return function
function glue.sortedpairs(t, cmp)
    local kt = glue.keys(t, cmp or true)
    local i = 0
    return function()
        i = i + 1
        return kt[i], t[kt[i]]
    end
end

---Update a table with the contents of other table(s)
---@param dt table
---@param ... any
---@return table
function glue.update(dt, ...)
    for i = 1, select("#", ...) do
        local t = select(i, ...)
        if t then
            for k, v in pairs(t) do
                dt[k] = v
            end
        end
    end
    return dt
end

---Add the contents of other table(s) without overwriting
---@param dt table
---@param ... any
---@return table
function glue.merge(dt, ...)
    for i = 1, select("#", ...) do
        local t = select(i, ...)
        if t then
            for k, v in pairs(t) do
                if rawget(dt, k) == nil then
                    dt[k] = v
                end
            end
        end
    end
    return dt
end

---Copy the content of a table and create a new one without references
---@param orig table
---@return table
function glue.deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == "table" then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[glue.deepcopy(orig_key)] = glue.deepcopy(orig_value)
        end
        setmetatable(copy, glue.deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

---Get the value of a table field, and if the field is not present in the\
---table, create it as an empty table, and return it
---@param t table
---@param k any
---@param v0 any
---@return any
function glue.attr(t, k, v0)
    local v = t[k]
    if v == nil then
        if v0 == nil then
            v0 = {}
        end
        v = v0
        t[k] = v
    end
    return v
end

-- lists ----------------------------------------------------------------------

-- extend a list with the elements of other lists.
function glue.extend(dt, ...)
    for j = 1, select("#", ...) do
        local t = select(j, ...)
        if t then
            local j = #dt
            for i = 1, #t do
                dt[j + i] = t[i]
            end
        end
    end
    return dt
end

-- append non-nil arguments to a list.
function glue.append(dt, ...)
    local j = #dt
    for i = 1, select("#", ...) do
        dt[j + i] = select(i, ...)
    end
    return dt
end

-- insert n elements at i, shifting elemens on the right of i (i inclusive)
-- to the right.
local function insert(t, i, n)
    if n == 1 then -- shift 1
        table.insert(t, i, false)
        return
    end
    for p = #t, i, -1 do -- shift n
        t[p + n] = t[p]
    end
end

-- remove n elements at i, shifting elements on the right of i (i inclusive)
-- to the left.
local function remove(t, i, n)
    n = min(n, #t - i + 1)
    if n == 1 then -- shift 1
        table.remove(t, i)
        return
    end
    for p = i + n, #t do -- shift n
        t[p - n] = t[p]
    end
    for p = #t, #t - n + 1, -1 do -- clean tail
        t[p] = nil
    end
end

-- shift all the elements on the right of i (i inclusive) to the left
-- or further to the right.
function glue.shift(t, i, n)
    if n > 0 then
        insert(t, i, n)
    elseif n < 0 then
        remove(t, i, -n)
    end
    return t
end

-- map f over t or extract a column from a list of records.
function glue.map(t, f, ...)
    local dt = {}
    if #t == 0 then -- treat as hashmap
        if type(f) == "function" then
            for k, v in pairs(t) do
                dt[k] = f(k, v, ...)
            end
        else
            for k, v in pairs(t) do
                local sel = v[f]
                if type(sel) == "function" then -- method to apply
                    dt[k] = sel(v, ...)
                else -- field to pluck
                    dt[k] = sel
                end
            end
        end
    else -- treat as array
        if type(f) == "function" then
            for i, v in ipairs(t) do
                dt[i] = f(v, ...)
            end
        else
            for i, v in ipairs(t) do
                local sel = v[f]
                if type(sel) == "function" then -- method to apply
                    dt[i] = sel(v, ...)
                else -- field to pluck
                    dt[i] = sel
                end
            end
        end
    end
    return dt
end

-- arrays ---------------------------------------------------------------------

---Scan list for value, works with ffi arrays too given i and j
---@param v any
---@param t any[]
---@param eq fun(key:any, value:any)
---@param i integer
---@param j integer
---@return integer
function glue.indexof(v, t, eq, i, j)
    i = i or 1
    j = j or #t
    if eq then
        for i = i, j do
            if eq(t[i], v) then
                return i
            end
        end
    else
        for i = i, j do
            if t[i] == v then
                return i
            end
        end
    end
end

---Return the index of a table/array if value exists
---@param array any[]
---@param value any
function glue.arrayhas(array, value)
    for k, v in pairs(array) do
        if (v == value) then
            return k
        end
    end
    return nil
end

---Get new values of an array compared to another
---@param oldarray any[]
---@param newarray any[]
function glue.arraynv(oldarray, newarray)
    local newvalues = {}
    for k, v in pairs(newarray) do
        if (not glue.arrayhas(oldarray, v)) then
            glue.append(newvalues, v)
        end
    end
    return newvalues
end

---Reverse elements of a list in place, works with ffi arrays too given i and j
---@param t any[]
---@param i integer
---@param j integer
---@return any[]
function glue.reverse(t, i, j)
    i = i or 1
    j = (j or #t) + 1
    for k = 1, (j - i) / 2 do
        t[i + k - 1], t[j - k] = t[j - k], t[i + k - 1]
    end
    return t
end

--- Get all the values of a key recursively
---@param t table
---@param dp any
function glue.childsbyparent(t, dp)
    for p, ch in pairs(t) do
        if (p == dp) then
            return ch
        end
        if (ch) then
            local found = glue.childsbyparent(ch, dp)
            if (found) then
                return found
            end
        end
    end
    return nil
end

-- Get the key of a value recursively
---@param t table
---@param dp any
function glue.parentbychild(t, dp)
    for p, ch in pairs(t) do
        if (ch[dp]) then
            return p
        end
        if (ch) then
            local found = glue.parentbychild(ch, dp)
            if (found) then
                return found
            end
        end
    end
    return nil
end

--- Split a list/array into small parts of given size
---@param list any[]
---@param chunks number
function glue.chunks(list, chunks)
    local chunkcounter = 0
    local chunk = {}
    local chunklist = {}
    -- Append chunks to the list in the specified amount of elements
    for k, v in pairs(list) do
        if (chunkcounter == chunks) then
            glue.append(chunklist, chunk)
            chunk = {}
            chunkcounter = 0
        end
        glue.append(chunk, v)
        chunkcounter = chunkcounter + 1
    end
    -- If there was a chunk that was not completed append it
    if (chunkcounter ~= 0) then
        glue.append(chunklist, chunk)
    end
    return chunklist
end

-- binary search for an insert position that keeps the table sorted.
-- works with ffi arrays too if lo and hi are provided.
local cmps = {}
cmps["<"] = function(t, i, v)
    return t[i] < v
end
cmps[">"] = function(t, i, v)
    return t[i] > v
end
cmps["<="] = function(t, i, v)
    return t[i] <= v
end
cmps[">="] = function(t, i, v)
    return t[i] >= v
end
local less = cmps["<"]
function glue.binsearch(v, t, cmp, lo, hi)
    lo, hi = lo or 1, hi or #t
    cmp = cmp and cmps[cmp] or cmp or less
    local len = hi - lo + 1
    if len == 0 then
        return nil
    end
    if len == 1 then
        return not cmp(t, lo, v) and lo or nil
    end
    while lo < hi do
        local mid = floor(lo + (hi - lo) / 2)
        if cmp(t, mid, v) then
            lo = mid + 1
            if lo == hi and cmp(t, lo, v) then
                return nil
            end
        else
            hi = mid
        end
    end
    return lo
end

-- strings --------------------------------------------------------------------

-- string submodule. has its own namespace which can be merged with _G.string.
glue.string = {}

--- Split a string list/array given a separator string
function glue.string.split(s, sep)
    if (sep == nil or sep == "") then
        return 1
    end
    local position, array = 0, {}
    for st, sp in function()
        return string.find(s, sep, position, true)
    end do
        table.insert(array, string.sub(s, position, st - 1))
        position = sp + 1
    end
    table.insert(array, string.sub(s, position))
    return array
end

-- split a string by a separator that can be a pattern or a plain string.
-- return a stateless iterator for the pieces.
local function iterate_once(s, s1)
    return s1 == nil and s or nil
end
function glue.string.gsplit(s, sep, start, plain)
    start = start or 1
    plain = plain or false
    if not s:find(sep, start, plain) then
        return iterate_once, s:sub(start)
    end
    local done = false
    local function pass(i, j, ...)
        if i then
            local seg = s:sub(start, i - 1)
            start = j + 1
            return seg, ...
        else
            done = true
            return s:sub(start)
        end
    end
    return function()
        if done then
            return
        end
        if sep == "" then
            done = true;
            return s:sub(start)
        end
        return pass(s:find(sep, start, plain))
    end
end

---Split a string into lines, optionally including the line terminator
---@param s string
---@param opt any | '"*L"'
---@return function
function glue.lines(s, opt)
    local term = opt == "*L"
    local patt = term and "([^\r\n]*()\r?\n?())" or "([^\r\n]*)()\r?\n?()"
    local next_match = s:gmatch(patt)
    local empty = s == ""
    local ended -- string ended with no line ending
    return function()
        local s, i1, i2 = next_match()
        if s == nil then
            return
        end
        if s == "" and not empty and ended then
            s = nil
        end
        ended = i1 == i2
        return s
    end
end

---String trim12, source: http://lua-users.org/wiki/StringTrim
---@param s any
---@return string
function glue.string.trim(s)
    local from = s:match("^%s*()")
    return from > #s and "" or s:match(".*%S", from)
end

-- escape a string so that it can be matched literally inside a pattern.
local function format_ci_pat(c)
    return ("[%s%s]"):format(c:lower(), c:upper())
end
---Escape a string to match it inside a lua pattern
---@param s string
---@param mode? any
---@return any
function glue.string.esc(s, mode) -- escape is a reserved word in Terra
    s = s:gsub("%%", "%%%%"):gsub("%z", "%%z"):gsub("([%^%$%(%)%.%[%]%*%+%-%?])", "%%%1")
    if mode == "*i" then
        s = s:gsub("[%a]", format_ci_pat)
    end
    return s
end

---Convert string or number to hex
---@param s string
---@param upper? boolean
---@return string
function glue.string.tohex(s, upper)
    if type(s) == "number" then
        return (upper and "%08.8X" or "%08.8x"):format(s)
    end
    if upper then
        return (s:gsub(".", function(c)
            return ("%02X"):format(c:byte())
        end))
    else
        return (s:gsub(".", function(c)
            return ("%02x"):format(c:byte())
        end))
    end
end

-- hex to binary string.
function glue.string.fromhex(s)
    if #s % 2 == 1 then
        return glue.string.fromhex("0" .. s)
    end
    return (s:gsub("..", function(cc)
        return string.char(tonumber(cc, 16))
    end))
end

function glue.string.starts(s, p) -- 5x faster than s:find'^...' in LuaJIT 2.1
    return s:sub(1, #p) == p
end

function glue.string.ends(s, p)
    return p == "" or s:sub(-#p) == p
end

function glue.string.subst(s, t) -- subst('{foo} {bar}', {foo=1, bar=2}) -> '1 2'
    return s:gsub("{([_%w]+)}", t)
end

-- publish the string submodule in the glue namespace.
glue.update(glue, glue.string)

-- iterators ------------------------------------------------------------------

-- run an iterator and collect the n-th return value into a list.
local function select_at(i, ...)
    return ..., select(i, ...)
end
local function collect_at(i, f, s, v)
    local t = {}
    repeat
        v, t[#t + 1] = select_at(i, f(s, v))
    until v == nil
    return t
end
local function collect_first(f, s, v)
    local t = {}
    repeat
        v = f(s, v);
        t[#t + 1] = v
    until v == nil
    return t
end
function glue.collect(n, ...)
    if type(n) == "number" then
        return collect_at(n, ...)
    else
        return collect_first(n, ...)
    end
end

-- closures -------------------------------------------------------------------

-- no-op filters.
function glue.pass(...)
    return ...
end
function glue.noop()
end

-- memoize for 0, 1, 2-arg and vararg and 1 retval functions.
local function memoize0(fn) -- for strict no-arg functions
    local v, stored
    return function()
        if not stored then
            v = fn();
            stored = true
        end
        return v
    end
end
local nilkey = {}
local nankey = {}
local function memoize1(fn) -- for strict single-arg functions
    local cache = {}
    return function(arg)
        local k = arg == nil and nilkey or arg ~= arg and nankey or arg
        local v = cache[k]
        if v == nil then
            v = fn(arg);
            cache[k] = v == nil and nilkey or v
        else
            if v == nilkey then
                v = nil
            end
        end
        return v
    end
end
local function memoize2(fn) -- for strict two-arg functions
    local cache = {}
    return function(a1, a2)
        local k1 = a1 ~= a1 and nankey or a1 == nil and nilkey or a1
        local cache2 = cache[k1]
        if cache2 == nil then
            cache2 = {}
            cache[k1] = cache2
        end
        local k2 = a2 ~= a2 and nankey or a2 == nil and nilkey or a2
        local v = cache2[k2]
        if v == nil then
            v = fn(a1, a2)
            cache2[k2] = v == nil and nilkey or v
        else
            if v == nilkey then
                v = nil
            end
        end
        return v
    end
end
local function memoize_vararg(fn, minarg, maxarg)
    local cache = {}
    local values = {}
    return function(...)
        local key = cache
        local narg = min(max(select("#", ...), minarg), maxarg)
        for i = 1, narg do
            local a = select(i, ...)
            local k = a ~= a and nankey or a == nil and nilkey or a
            local t = key[k]
            if not t then
                t = {};
                key[k] = t
            end
            key = t
        end
        local v = values[key]
        if v == nil then
            v = fn(...);
            values[key] = v == nil and nilkey or v
        end
        if v == nilkey then
            v = nil
        end
        return v
    end
end
local memoize_narg = {[0] = memoize0, memoize1, memoize2}
local function choose_memoize_func(func, narg)
    if narg then
        local memoize_narg = memoize_narg[narg]
        if memoize_narg then
            return memoize_narg
        else
            return memoize_vararg, narg, narg
        end
    else
        local info = debug.getinfo(func, "u")
        if info.isvararg then
            return memoize_vararg, info.nparams, 1 / 0
        else
            return choose_memoize_func(func, info.nparams)
        end
    end
end
function glue.memoize(func, narg)
    local memoize, minarg, maxarg = choose_memoize_func(func, narg)
    return memoize(func, minarg, maxarg)
end

-- memoize a function with multiple return values.
function glue.memoize_multiret(func, narg)
    local memoize, minarg, maxarg = choose_memoize_func(func, narg)
    local function wrapper(...)
        return glue.pack(func(...))
    end
    local func = memoize(wrapper, minarg, maxarg)
    return function(...)
        return glue.unpack(func(...))
    end
end

local tuple_mt = {__call = glue.unpack}
function tuple_mt:__tostring()
    local t = {}
    for i = 1, self.n do
        t[i] = tostring(self[i])
    end
    return string.format("(%s)", table.concat(t, ", "))
end
function glue.tuples(narg)
    return glue.memoize(function(...)
        return setmetatable(glue.pack(...), tuple_mt)
    end)
end

-- objects --------------------------------------------------------------------

-- set up dynamic inheritance by creating or updating a table's metatable.
function glue.inherit(t, parent)
    local meta = getmetatable(t)
    if meta then
        meta.__index = parent
    elseif parent ~= nil then
        setmetatable(t, {__index = parent})
    end
    return t
end

-- prototype-based dynamic inheritance with __call constructor.
function glue.object(super, o, ...)
    o = o or {}
    o.__index = super
    o.__call = super and super.__call
    glue.update(o, ...) -- add mixins, defaults, etc.
    return setmetatable(o, o)
end

local function install(self, combine, method_name, hook)
    rawset(self, method_name, combine(self[method_name], hook))
end
local function before(method, hook)
    if method then
        return function(self, ...)
            hook(self, ...)
            return method(self, ...)
        end
    else
        return hook
    end
end
function glue.before(self, method_name, hook)
    install(self, before, method_name, hook)
end
local function after(method, hook)
    if method then
        return function(self, ...)
            method(self, ...)
            return hook(self, ...)
        end
    else
        return hook
    end
end
function glue.after(self, method_name, hook)
    install(self, after, method_name, hook)
end
local function override(method, hook)
    local method = method or glue.noop
    return function(...)
        return hook(method, ...)
    end
end
function glue.override(self, method_name, hook)
    install(self, override, method_name, hook)
end

-- return a metatable that supports virtual properties.
-- can be used with setmetatable() and ffi.metatype().
function glue.gettersandsetters(getters, setters, super)
    local get = getters and function(t, k)
        local get = getters[k]
        if get then
            return get(t)
        end
        return super and super[k]
    end
    local set = setters and function(t, k, v)
        local set = setters[k]
        if set then
            set(t, v);
            return
        end
        rawset(t, k, v)
    end
    return {__index = get, __newindex = set}
end

-- i/o ------------------------------------------------------------------------

-- check if a file exists and can be opened for reading or writing.
function glue.canopen(name, mode)
    local f = io.open(name, mode or "rb")
    if f then
        f:close()
    end
    return f ~= nil and name or nil
end

-- read a file into a string (in binary mode by default).
function glue.readfile(name, mode, open)
    open = open or io.open
    local f, err = open(name, mode == "t" and "r" or "rb")
    if not f then
        return nil, err
    end
    local s, err = f:read "*a"
    if s == nil then
        return nil, err
    end
    f:close()
    return s
end

-- read the output of a command into a string.
function glue.readpipe(cmd, mode, open)
    return glue.readfile(cmd, mode, open or io.popen)
end

-- like os.rename() but behaves like POSIX on Windows too.
if jit then

    local ffi = require "ffi"

    if ffi.os == "Windows" then

        ffi.cdef [[
			int MoveFileExA(
				const char *lpExistingFileName,
				const char *lpNewFileName,
				unsigned long dwFlags
			);
			int GetLastError(void);
		]]

        local MOVEFILE_REPLACE_EXISTING = 1
        local MOVEFILE_WRITE_THROUGH = 8
        local ERROR_FILE_EXISTS = 80
        local ERROR_ALREADY_EXISTS = 183

        function glue.replacefile(oldfile, newfile)
            if ffi.C.MoveFileExA(oldfile, newfile, 0) ~= 0 then
                return true
            end
            local err = ffi.C.GetLastError()
            if err == ERROR_FILE_EXISTS or err == ERROR_ALREADY_EXISTS then
                if ffi.C.MoveFileExA(oldfile, newfile,
                                     bit.bor(MOVEFILE_WRITE_THROUGH, MOVEFILE_REPLACE_EXISTING)) ~=
                    0 then
                    return true
                end
                err = ffi.C.GetLastError()
            end
            return nil, "WinAPI error " .. err
        end

    else

        function glue.replacefile(oldfile, newfile)
            return os.rename(oldfile, newfile)
        end

    end

end

-- write a string, number, table or the results of a read function to a file.
-- uses binary mode by default.
function glue.writefile(filename, s, mode, tmpfile)
    if tmpfile then
        local ok, err = glue.writefile(tmpfile, s, mode)
        if not ok then
            return nil, err
        end
        local ok, err = glue.replacefile(tmpfile, filename)
        if not ok then
            os.remove(tmpfile)
            return nil, err
        else
            return true
        end
    end
    local f, err = io.open(filename, mode == "t" and "w" or "wb")
    if not f then
        return nil, err
    end
    local ok, err
    if type(s) == "table" then
        for i = 1, #s do
            ok, err = f:write(s[i])
            if not ok then
                break
            end
        end
    elseif type(s) == "function" then
        local read = s
        while true do
            ok, err = xpcall(read, debug.traceback)
            if not ok or err == nil then
                break
            end
            ok, err = f:write(err)
            if not ok then
                break
            end
        end
    else -- string or number
        ok, err = f:write(s)
    end
    f:close()
    if not ok then
        os.remove(filename)
        return nil, err
    else
        return true
    end
end

-- virtualize the print function.
function glue.printer(out, format)
    format = format or tostring
    return function(...)
        local n = select("#", ...)
        for i = 1, n do
            out(format((select(i, ...))))
            if i < n then
                out "\t"
            end
        end
        out "\n"
    end
end

-- dates & timestamps ---------------------------------------------------------

-- compute timestamp diff. to UTC because os.time() has no option for UTC.
function glue.utc_diff(t)
    local d1 = os.date("*t", 3600 * 24 * 10)
    local d2 = os.date("!*t", 3600 * 24 * 10)
    d1.isdst = false
    return os.difftime(os.time(d1), os.time(d2))
end

-- overloading os.time to support UTC and get the date components as separate args.
function glue.time(utc, y, m, d, h, M, s, isdst)
    if type(utc) ~= "boolean" then -- shift arg#1
        utc, y, m, d, h, M, s, isdst = nil, utc, y, m, d, h, M, s
    end
    if type(y) == "table" then
        local t = y
        if utc == nil then
            utc = t.utc
        end
        y, m, d, h, M, s, isdst = t.year, t.month, t.day, t.hour, t.min, t.sec, t.isdst
    end
    local utc_diff = utc and glue.utc_diff() or 0
    if not y then
        return os.time() + utc_diff
    else
        s = s or 0
        local t = os.time {
            year = y,
            month = m or 1,
            day = d or 1,
            hour = h or 0,
            min = M or 0,
            sec = s,
            isdst = isdst
        }
        return t and t + s - floor(s) + utc_diff
    end
end

-- get the time at the start of the week of a given time, plus/minus a number of weeks.
function glue.sunday(utc, t, offset)
    if type(utc) ~= "boolean" then -- shift arg#1
        utc, t, offset = false, utc, t
    end
    local d = os.date(utc and "!*t" or "*t", t)
    return glue.time(false, d.year, d.month, d.day - (d.wday - 1) + (offset or 0) * 7)
end

-- get the time at the start of the day of a given time, plus/minus a number of days.
function glue.day(utc, t, offset)
    if type(utc) ~= "boolean" then -- shift arg#1
        utc, t, offset = false, utc, t
    end
    local d = os.date(utc and "!*t" or "*t", t)
    return glue.time(false, d.year, d.month, d.day + (offset or 0))
end

-- get the time at the start of the month of a given time, plus/minus a number of months.
function glue.month(utc, t, offset)
    if type(utc) ~= "boolean" then -- shift arg#1
        utc, t, offset = false, utc, t
    end
    local d = os.date(utc and "!*t" or "*t", t)
    return glue.time(false, d.year, d.month + (offset or 0))
end

-- get the time at the start of the year of a given time, plus/minus a number of years.
function glue.year(utc, t, offset)
    if type(utc) ~= "boolean" then -- shift arg#1
        utc, t, offset = false, utc, t
    end
    local d = os.date(utc and "!*t" or "*t", t)
    return glue.time(false, d.year + (offset or 0))
end

-- error handling -------------------------------------------------------------

-- allocation-free assert() with string formatting.
-- NOTE: unlike standard assert(), this only returns the first argument
-- to avoid returning the error message and it's args along with it so don't
-- use it with functions returning multiple values when you want those values.
function glue.assert(v, err, ...)
    if v then
        return v
    end
    err = err or "assertion failed!"
    if select("#", ...) > 0 then
        err = string.format(err, ...)
    end
    error(err, 2)
end

-- pcall with traceback. LuaJIT and Lua 5.2 only.
local function pcall_error(e)
    return debug.traceback("\n" .. tostring(e))
end
function glue.pcall(f, ...)
    return xpcall(f, pcall_error, ...)
end

local function unprotect(ok, result, ...)
    if not ok then
        return nil, result, ...
    end
    if result == nil then
        result = true
    end -- to distinguish from error.
    return result, ...
end

---Wrap a function that raises errors on failure into a function that follows\
---the Lua convention of returning nil, err on failure
---@param func function
---@return function
function glue.protect(func)
    return function(...)
        return unprotect(pcall(func, ...))
    end
end

-- pcall with finally and except "clauses":
--		local ret,err = fpcall(function(finally, except)
--			local foo = getfoo()
--			finally(function() foo:free() end)
--			except(function(err) io.stderr:write(err, '\n') end)
--		emd)
-- NOTE: a bit bloated at 2 tables and 4 closures. Can we reduce the overhead?
local function fpcall(f, ...)
    local fint, errt = {}, {}
    local function finally(f)
        fint[#fint + 1] = f
    end
    local function onerror(f)
        errt[#errt + 1] = f
    end
    local function err(e)
        for i = #errt, 1, -1 do
            errt[i](e)
        end
        for i = #fint, 1, -1 do
            fint[i]()
        end
        return tostring(e) .. "\n" .. debug.traceback()
    end
    local function pass(ok, ...)
        if ok then
            for i = #fint, 1, -1 do
                fint[i]()
            end
        end
        return ok, ...
    end
    return pass(xpcall(f, err, finally, onerror, ...))
end

function glue.fpcall(...)
    return unprotect(fpcall(...))
end

-- fcall is like fpcall() but without the protection (i.e. raises errors).
local function assert_fpcall(ok, ...)
    if not ok then
        error(..., 2)
    end
    return ...
end
function glue.fcall(...)
    return assert_fpcall(fpcall(...))
end

-- modules --------------------------------------------------------------------

-- create a module table that dynamically inherits another module.
-- naming the module returns the same module table for the same name.
function glue.module(name, parent)
    if type(name) ~= "string" then
        name, parent = parent, name
    end
    if type(parent) == "string" then
        parent = require(parent)
    end
    parent = parent or _M
    local parent_P = parent and assert(parent._P, "parent module has no _P") or _G
    local M = package.loaded[name]
    if M then
        return M, M._P
    end
    local P = {__index = parent_P}
    M = {__index = parent, _P = P}
    P._M = M
    M._M = M
    P._P = P
    setmetatable(P, P)
    setmetatable(M, M)
    if name then
        package.loaded[name] = M
        P[name] = M
    end
    ---@diagnostic disable-next-line: deprecated
    setfenv(2, P)
    return M, P
end

-- setup a module to load sub-modules when accessing specific keys.
function glue.autoload(t, k, v)
    local mt = getmetatable(t) or {}
    if not mt.__autoload then
        local old_index = mt.__index
        local submodules = {}
        mt.__autoload = submodules
        mt.__index = function(t, k)
            -- overriding __index...
            if type(old_index) == "function" then
                local v = old_index(t, k)
                if v ~= nil then
                    return v
                end
            elseif type(old_index) == "table" then
                local v = old_index[k]
                if v ~= nil then
                    return v
                end
            end
            if submodules[k] then
                local mod
                if type(submodules[k]) == "string" then
                    mod = require(submodules[k]) -- module
                else
                    mod = submodules[k](k) -- custom loader
                end
                submodules[k] = nil -- prevent loading twice
                if type(mod) == "table" then -- submodule returned its module table
                    assert(mod[k] ~= nil) -- submodule has our symbol
                    t[k] = mod[k]
                end
                return rawget(t, k)
            end
        end
        setmetatable(t, mt)
    end
    if type(k) == "table" then
        glue.update(mt.__autoload, k) -- multiple key -> module associations.
    else
        mt.__autoload[k] = v -- single key -> module association.
    end
    return t
end

-- portable way to get script's directory, based on arg[0].
-- NOTE: the path is not absolute, but relative to the current directory!
-- NOTE: for bundled executables, this returns the executable's directory.
local dir = rawget(_G, "arg") and arg[0] and arg[0]:gsub("[/\\]?[^/\\]+$", "") or "" -- remove file name
glue.bin = dir == "" and "." or dir

-- portable way to add more paths to package.path, at any place in the list.
-- negative indices count from the end of the list like string.sub().
-- index 'after' means 0.
function glue.luapath(path, index, ext)
    ext = ext or "lua"
    index = index or 1
    local psep = package.config:sub(1, 1) -- '/'
    local tsep = package.config:sub(3, 3) -- ';'
    local wild = package.config:sub(5, 5) -- '?'
    local paths = glue.collect(glue.gsplit(package.path, tsep, nil, true))
    path = path:gsub("[/\\]", psep) -- normalize slashes
    if index == "after" then
        index = 0
    end
    if index < 1 then
        index = #paths + 1 + index
    end
    table.insert(paths, index, path .. psep .. wild .. psep .. "init." .. ext)
    table.insert(paths, index, path .. psep .. wild .. "." .. ext)
    package.path = table.concat(paths, tsep)
end

-- portable way to add more paths to package.cpath, at any place in the list.
-- negative indices count from the end of the list like string.sub().
-- index 'after' means 0.
function glue.cpath(path, index)
    index = index or 1
    local psep = package.config:sub(1, 1) -- '/'
    local tsep = package.config:sub(3, 3) -- ';'
    local wild = package.config:sub(5, 5) -- '?'
    local ext = package.cpath:match("%.([%a]+)%" .. tsep .. "?") -- dll | so | dylib
    local paths = glue.collect(glue.gsplit(package.cpath, tsep, nil, true))
    path = path:gsub("[/\\]", psep) -- normalize slashes
    if index == "after" then
        index = 0
    end
    if index < 1 then
        index = #paths + 1 + index
    end
    table.insert(paths, index, path .. psep .. wild .. "." .. ext)
    package.cpath = table.concat(paths, tsep)
end

-- allocation -----------------------------------------------------------------

-- freelist for Lua tables.
local function create_table()
    return {}
end
function glue.freelist(create, destroy)
    create = create or create_table
    destroy = destroy or glue.noop
    local t = {}
    local n = 0
    local function alloc()
        local e = t[n]
        if e then
            t[n] = false
            n = n - 1
        end
        return e or create()
    end
    local function free(e)
    ---@diagnostic disable-next-line: redundant-parameter
        destroy(e)
        n = n + 1
        t[n] = e
    end
    return alloc, free
end

-- ffi ------------------------------------------------------------------------

if jit then

    local ffi = require "ffi"

    -- static, auto-growing buffer allocation pattern (ctype must be vla).
    function glue.buffer(ctype)
        local vla = ffi.typeof(ctype)
        local buf, len = nil, -1
        return function(minlen)
            if minlen == false then
                buf, len = nil, -1
            elseif minlen > len then
                len = glue.nextpow2(minlen)
                buf = vla(len)
            end
            return buf, len
        end
    end

    -- like glue.buffer() but preserves data on reallocations
    -- also returns minlen instead of capacity.
    function glue.dynarray(ctype)
        local buffer = glue.buffer(ctype)
        local elem_size = ffi.sizeof(ctype, 1)
        local buf0, minlen0
        return function(minlen)
            local buf, len = buffer(minlen)
            if buf ~= buf0 and buf ~= nil and buf0 ~= nil then
                ffi.copy(buf, buf0, minlen0 * elem_size)
            end
            buf0, minlen0 = buf, minlen
            return buf, minlen
        end
    end

    local intptr_ct = ffi.typeof "intptr_t"
    local intptrptr_ct = ffi.typeof "const intptr_t*"
    local intptr1_ct = ffi.typeof "intptr_t[1]"
    local voidptr_ct = ffi.typeof "void*"

    -- x86: convert a pointer's address to a Lua number.
    local function addr32(p)
        return tonumber(ffi.cast(intptr_ct, ffi.cast(voidptr_ct, p)))
    end

    -- x86: convert a number to a pointer, optionally specifying a ctype.
    local function ptr32(ctype, addr)
        if not addr then
            ctype, addr = voidptr_ct, ctype
        end
        return ffi.cast(ctype, addr)
    end

    -- x64: convert a pointer's address to a Lua number or possibly string.
    local function addr64(p)
        local np = ffi.cast(intptr_ct, ffi.cast(voidptr_ct, p))
        local n = tonumber(np)
        if ffi.cast(intptr_ct, n) ~= np then
            -- address too big (ASLR? tagged pointers?): convert to string.
            return ffi.string(intptr1_ct(np), 8)
        end
        return n
    end

    -- x64: convert a number or string to a pointer, optionally specifying a ctype.
    local function ptr64(ctype, addr)
        if not addr then
            ctype, addr = voidptr_ct, ctype
        end
        if type(addr) == "string" then
            return ffi.cast(ctype, ffi.cast(voidptr_ct, ffi.cast(intptrptr_ct, addr)[0]))
        else
            return ffi.cast(ctype, addr)
        end
    end

    glue.addr = ffi.abi "64bit" and addr64 or addr32
    glue.ptr = ffi.abi "64bit" and ptr64 or ptr32

end -- if jit

if bit then

    local band, bor, bnot = bit.band, bit.bor, bit.bnot

    -- extract the bool value of a bitmask from a value.
    function glue.getbit(from, mask)
        return band(from, mask) == mask
    end

    -- set a single bit of a value without affecting other bits.
    function glue.setbit(over, mask, yes)
        return bor(yes and mask or 0, band(over, bnot(mask)))
    end

    local function bor_bit(bits, k, mask, strict)
        local b = bits[k]
        if b then
            return bit.bor(mask, b)
        elseif strict then
            error(string.format("invalid bit %s", k))
        else
            return mask
        end
    end
    function glue.bor(flags, bits, strict)
        local mask = 0
        if type(flags) == "number" then
            return flags -- passthrough
        elseif type(flags) == "string" then
            for k in flags:gmatch "[^%s]+" do
                mask = bor_bit(bits, k, mask, strict)
            end
        elseif type(flags) == "table" then
            for k, v in pairs(flags) do
                k = type(k) == "number" and v or k
                mask = bor_bit(bits, k, mask, strict)
            end
        else
            error "flags expected"
        end
        return mask
    end

end

return glue

end,

["insurrection.components.dynamic.lobbyMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.lobbyMenu'
--------------------
local component = require "insurrection.components"
local constants = require "insurrection.constants"
local button = require "insurrection.components.button"
local list = require "insurrection.components.list"
local input = require "insurrection.components.input"
local core = require "insurrection.core"
local blam = require "blam"
local getState = require "insurrection.redux.getState"
local checkbox = require "insurrection.components.checkbox"

local gametypeIcons = {
    "unknown",
    "assault",
    "ctf",
    "forge",
    "infection",
    "juggernaut",
    "king",
    "oddball",
    "race",
    "slayer",
    "team_slayer"
}

local templateIcons = {
    "stock",
    "fiesta"
}

local bitmaps = {
    gametypeIcons = blam.findTag("lobby_gametype_icon", blam.tagClasses.bitmap),
    templateIcons = blam.findTag("lobby_template_icon", blam.tagClasses.bitmap)
}

return function()
    local state = getState()
    local lobby = state.lobby
    local definition = "map"

    local isPlayerLobbyOwner = api.session.player and state.lobby and api.session.player.publicId ==
                                   state.lobby.owner

    local lobbyMenu = component.new(constants.widgets.lobby.id)
    local summary = component.new(
                        component.new(lobbyMenu:findChildWidgetTag("summary").id):findChildWidgetTag(
                            "text").id)

    local options = component.new(lobbyMenu:findChildWidgetTag("options").id)

    local definitionList = component.new(options:findChildWidgetTag("definitions").id)
    local template = button.new(definitionList:findChildWidgetTag("template").id)

    local map = button.new(definitionList:findChildWidgetTag("map").id)
    local gametype = button.new(definitionList:findChildWidgetTag("gametype").id)

    -- local lobbySettings = button.new(lobbyDefs:findChildWidgetTag("settings").id)

    local elementsList = list.new(options:findChildWidgetTag("elements").id)
    local mapsList = list.new(blam.findTag("lobby_maps_options", blam.tagClasses.uiWidgetDefinition)
                                  .id)
    local fullMapList = component.new(blam.findTag("lobby_maps_wrapper",
                                                   blam.tagClasses.uiWidgetDefinition).id)
    local mapDescription = component.new(blam.findTag("map_small",
                                                      blam.tagClasses.uiWidgetDefinition).id)
    local mapOverlay =
        component.new(mapDescription:findChildWidgetTag("overlay_scanner").id):animate()
    local search = input.new(options:findChildWidgetTag("search").id)
    local play = button.new(options:findChildWidgetTag("play").id)
    local back = button.new(options:findChildWidgetTag("back").id)
    local makePublic = checkbox.new(options:findChildWidgetTag("make_public").id)
    local key = input.new(options:findChildWidgetTag("lobby_key").id)
    key:onFocus(function()
        key:setText(api.session.lobbyKey)
    end)
    key:onClick(function()
        core.copyToClipboard(api.session.lobbyKey)
        key:setText(string.rep("*", #api.session.lobbyKey))
    end)

    local mapPreview = component.new(blam.findTag("map_small_preview",
                                                  blam.tagClasses.uiWidgetDefinition).id)
    local playersList = list.new(lobbyMenu:findChildWidgetTag("players").id)
    playersList:scrollable(false)

    summary:setText("Play with your friends, define your rules and enjoy.")

    local function getMapBackgroundBitmap(mapName)
        local mapCollection = blam.tagCollection(constants.tagCollections.maps.id)
        assert(mapCollection, "No map preview collection found")
        for k, v in pairs(mapCollection.tagList) do
            local bitmapTag = blam.getTag(v) --[[@as tag]]
            local mapBitmaName = core.getTagName(bitmapTag.path):lower()
            if mapBitmaName == mapName:lower() then
                return bitmapTag.id
            end
        end
        return constants.bitmaps.unknownMapPreview.id
    end

    local function setMapBackgroundBitmap(mapName)
        mapPreview.widgetDefinition.backgroundBitmap = getMapBackgroundBitmap(mapName)
    end

    if lobby and isPlayerLobbyOwner then
        elementsList:onSelect(function(item)
            item.value:setText(item.label)
            api.editLobby(api.session.lobbyKey, {
                template = template:getText(),
                map = map:getText(),
                gametype = gametype:getText()
            })
        end)
        mapsList:onSelect(function(item)
            item.value:setText(item.label)
            api.editLobby(api.session.lobbyKey, {
                template = template:getText(),
                map = map:getText(),
                gametype = gametype:getText()
            })
            setMapBackgroundBitmap(item.label)
        end)

        local handleDefinition = function(lobbyDef, newDefinition)
            log("Changing definition to " .. newDefinition)
            search:setText("")
            local component = elementsList
            if newDefinition == "map" then
                component = mapsList
            end
            component:setItems(table.map(state.available[newDefinition .. "s"], function(element)
                ---@type uiComponentListItem
                local item = {label = element, value = lobbyDef}
                if newDefinition == "template" or newDefinition == "gametype" then
                    item.bitmap = function(uiComponent)
                        local icon = component.new(uiComponent:findChildWidgetTag("button_icon").id)
                        local iconBitmaps = newDefinition == "template" and bitmaps.templateIcons or
                                                   bitmaps.gametypeIcons
                        assert(iconBitmaps, "No icon bitmaps found")
                        icon.widgetDefinition.backgroundBitmap = iconBitmaps.id
                        local iconsToUse = newDefinition == "template" and templateIcons or gametypeIcons
                        local iconToUse = table.find(iconsToUse, function(icon)
                            return element:includes(icon)
                        end)
                        local backgroundBitmapIndex =
                            (table.indexof(iconsToUse, iconToUse) or 1) - 1
                        if backgroundBitmapIndex then
                            icon:setWidgetValues({bitmapIndex = backgroundBitmapIndex})
                        end
                    end
                end
                return item
            end))
            definition = newDefinition
        end

        template:onClick(function()
            fullMapList:replace(elementsList.tagId)
            handleDefinition(template, "template")
        end)
        template:onFocus(function()
            summary:setText(
                "Template defines a set of changes to the base server that will be applied when the lobby is created.")
        end)

        -- Force selection of map definition when opening the lobby menu
        map:onClick(function()
            elementsList:replace(fullMapList.tagId)
            handleDefinition(map, "map")
        end)
        map:onFocus(function()
            summary:setText(
                "Choose a map from the available list to play on, you need to have the map installed.")
        end)
        map.events.onClick()

        gametype:onClick(function()
            fullMapList:replace(elementsList.tagId)
            handleDefinition(gametype, "gametype")
        end)
        gametype:onFocus(function()
            summary:setText(
                "Game type defines the rules of the game, defines team play, scoring, etc.")
        end)

        play:onClick(function()
            if isPlayerLobbyOwner then
                local template = template:getText()
                local map = map:getText()
                local gametype = gametype:getText()
                api.borrow(template:lower(), map, gametype:lower())
            else
                interface.dialog("WARNING", "", "You are not the owner of the lobby.")
            end
        end)

    end

    local definitionsToComponent = {template = template, map = map, gametype = gametype}
    search:onInputText(function(text)
        local elements = state.available[definition .. "s"] --[=[@as string[]]=]
        elements = table.filter(elements, function(element)
            return element:lower():includes(text:lower())
        end)
        local component = elementsList
        if definition == "map" then
            component = mapsList
        end

        component:setItems(table.map(elements, function(element)
            ---@type uiComponentListItem
            local item = {label = element, value = definitionsToComponent[definition]}
            if definition ~= "map" then
                item.bitmap = function(uiComponent)
                    local icon = component.new(uiComponent:findChildWidgetTag("button_icon").id)
                    local iconToUse = table.find(gametypeIcons, function(icon)
                        return element:includes(icon)
                    end)
                    local backgroundBitmapIndex = (table.indexof(gametypeIcons, iconToUse) or 1) - 1
                    if backgroundBitmapIndex then
                        icon:setWidgetValues({bitmapIndex = backgroundBitmapIndex})
                    end
                end
            end
            return item
        end))
    end)

    lobbyMenu:onOpen(function()
        if map.events.onClick then
            map.events.onClick()
        end
    end)
    lobbyMenu:onClose(function()
        api.deleteLobby()
    end)
    back:onClick(function()
        lobbyMenu.events.onClose()
    end)
    makePublic:onToggle(function(value)
        api.editLobby(api.session.lobbyKey, {isPublic = value})
    end)

    return function()
        template:setText(state.lobby.template)
        map:setText(state.lobby.map)
        gametype:setText(state.lobby.gametype)
        makePublic:setValue(state.lobby.isPublic)
        if api.session.lobbyKey then
            key:setText(string.rep("*", #api.session.lobbyKey))
        end
        setMapBackgroundBitmap(state.lobby.map)

        playersList:setItems(table.map(state.lobby.players, function(player)
            local nameplateTag = constants.nameplates[player.nameplate] or {}
            return {label = player.name, value = player, bitmap = nameplateTag.id}
        end))
    end
end

end,

["insurrection.components.checkbox"] = function()
--------------------
-- Module: 'insurrection.components.checkbox'
--------------------
local core = require "insurrection.core"
local components = require "insurrection.components"

---@class uiComponentCheckboxClass : uiComponent
local checkbox = setmetatable({
    ---@type string
    type = "checkbox",
    ---@type number
    checkboxTagId = nil,
    ---@type boolean
    value = false
}, {__index = components})

---@class uiComponentCheckboxEvents : uiComponentEvents
---@field onToggle fun(value: boolean):boolean | nil

---@class uiComponentCheckbox : uiComponentCheckboxClass
---@field events uiComponentCheckboxEvents

---@param tagId number
---@return uiComponentCheckbox
function checkbox.new(tagId)
    local instance = setmetatable(components.new(tagId), {__index = checkbox}) --[[@as uiComponentCheckbox]]
    assert(instance.tag.path:find("checkbox", 1, true),
           "Tag " .. instance.tag.path .. " is not a checkbox")
    instance.checkboxTagId = instance:findChildWidgetTag("checkbox").id
    return instance
end

---@param self uiComponentCheckbox
---@return boolean
function checkbox.getValue(self)
    local widgetValues = core.getWidgetValues(self.checkboxTagId)
    assert(widgetValues, "Checkbox " .. self.tag.path .. " does not exist in DOM tree")
    return widgetValues.bitmapIndex == 1
end

---@param self uiComponentCheckbox
function checkbox.setValue(self, value)
    local value = value and 1 or 0
    core.setWidgetValues(self.checkboxTagId, {bitmapIndex = value})
end

---@param self uiComponentCheckbox
function checkbox.toggle(self)
    local value = self:getValue()
    self:setValue(not value)
end

---@param self uiComponentCheckbox
function checkbox.onToggle(self, callback)
    self.events.onClick = function()
        local value = self:getValue()
        self:toggle()
        return callback(not value)
    end
end

return checkbox

end,

["insurrection.components.dynamic.lobbyMenuClient"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.lobbyMenuClient'
--------------------
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local button = require "insurrection.components.button"
local list = require "insurrection.components.list"
local input = require "insurrection.components.input"
local actions = require "insurrection.redux.actions"
local getState = require "insurrection.redux.getState"
local core = require "insurrection.core"
local blam = require "blam"
local findTag = blam.findTag
local tagClasses = blam.tagClasses
local s = require"insurrection.utils".snakeCaseToUpperTitleCase

local gametypeIcons = {
    "unknown",
    "assault",
    "ctf",
    "forge",
    "infection",
    "juggernaut",
    "king",
    "oddball",
    "race",
    "slayer",
    "team_slayer"
}

local function lobbyMenuClient()
    local state = getState()

    local lobby = components.new(constants.widgets.lobbyClient.id)
    local template = components.new(lobby:findChildWidgetTag("template").id)
    local mapPreview = components.new(findTag("lobby_client_map", tagClasses.uiWidgetDefinition).id)
    local gametypeIcon = components.new(lobby:findChildWidgetTag("gametype_icon").id)
    local description = components.new(lobby:findChildWidgetTag("description").id)
    local playersList = list.new(lobby:findChildWidgetTag("players").id)
    playersList:scrollable(false)

    -- Render
    return function()
        mapPreview.widgetDefinition.backgroundBitmap = constants.bitmaps.unknownMapPreview.id
        local mapCollection = blam.tagCollection(constants.tagCollections.maps.id)
        assert(mapCollection, "No map preview collection found")
        for k, v in pairs(mapCollection.tagList) do
            local bitmapTag = blam.getTag(v) --[[@as tag]]
            local mapName = core.getTagName(bitmapTag.path):lower()
            if mapName == state.lobby.map then
                mapPreview.widgetDefinition.backgroundBitmap = bitmapTag.id
            end
        end

        local iconToUse = table.find(gametypeIcons, function(icon)
            if state.lobby.gametype:find(icon, 1, true) then
                return true
            end
            return false
        end)
        local backgroundBitmapIndex = (table.indexof(gametypeIcons, iconToUse) or 1) - 1
        if backgroundBitmapIndex then
            gametypeIcon:setWidgetValues({bitmapIndex = backgroundBitmapIndex})
        end

        description:setText(s(state.lobby.gametype:upper()) .. " on " .. s(state.lobby.map))

        playersList:setItems(table.map(state.lobby.players, function(player)
            local nameplateTag = constants.nameplates[player.nameplate] or {}
            return {label = player.name, value = player, bitmap = nameplateTag.id}
        end))

        template:setText(s(state.lobby.template))
    end
end

return lobbyMenuClient

end,

["insurrection.components.dynamic.customizationBipedMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.customizationBipedMenu'
--------------------
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local button = require "insurrection.components.button"
local list = require "insurrection.components.list"
local utils = require "insurrection.utils"
local blam = require "blam"
local core = require "insurrection.core"
local color = require "color"
local bar = require "insurrection.components.bar"
local menus = require "insurrection.menus"
local t = utils.snakeCaseToUpperTitleCase

local staticRegions = {
    "body",
    "chest",
    "left_shoulder",
    "right_shoulder",
    "arms",
    "knees",
    "helmet",
    "visor",
    "color",
    "gear",
    "legs",
    "bundle"
}

local staticVisors = {
    "Sulfur",
    "Silver",
    "Blue",
    "Green",
    "Purple",
    "Golden",
    "Red",
    "Orange",
    "Black",
    "White",
    "Cyanotic",
    "Nightfall",
    "Matrix",
    "Sunshine",
    "Chimera",
    "Iridescent",
    "Coalescence",
    "Hologram",
    "Blizzard",
    "Legend",
    "Frostbite",
    "Scattered",
    "Pink Pop",
    "Moonlight",
    "Gilver"
}

local dynamicRegions = staticRegions

local specificRegionCameras = {
    left_shoulder = "shoulders",
    right_shoulder = "shoulders",
    left_leg = "legs",
    right_leg = "legs"
}

local interpolationTicks = 30
---@type "regions"|"permutations"|"visor"
local editing = "regions"
local currentRegionIndex = 1

local getCustomizationObjectData = core.getCustomizationObjectData

---@param biped biped
---@param colorInHex string
local function setBipedColor(biped, colorInHex)
    local r, g, b = color.hexToDec(colorInHex)
    biped.colorCLowerRed = r
    biped.colorCLowerGreen = g
    biped.colorCLowerBlue = b

    biped.colorDLowerRed = r
    biped.colorDLowerGreen = g
    biped.colorDLowerBlue = b
end

local function setCamera(region, ticks)
    local command = "camera_set customization_{region}_generic {ticks}"

    local customizationObjectData = getCustomizationObjectData()
    if customizationObjectData.tag.path:find("keymind") then
        command = "camera_set customization_{region} {ticks}"
    end
    -- FIXME This is a hack to use the lobby cammera, find a better way to do this
    if region == "lobby" then
        command = "camera_set customization_{region} {ticks}"
    elseif region == "visor" then
        command = "camera_set customization_helmet {ticks}"
    end

    execute_script(command:template{
        region = specificRegionCameras[region] or region,
        ticks = ticks or interpolationTicks
    })
end

local function setEditingGeometry(region, ticks)
    local customizationObjectId = core.getCustomizationObjectId()
    assert(customizationObjectId, "No customization biped found")
    local object = blam.getObject(customizationObjectId)
    if object then
        blam.rotateObject(object, constants.customization.rotation.default, 0, 0)
    end

    -- Set camera
    setCamera(region, ticks)

    BipedRotation = constants.customization.rotation[region] or
                        constants.customization.rotation.default
    if BipedRotation and object then
        blam.rotateObject(object, BipedRotation, 0, 0)
    end
end

local function getBipedName(tagPath)
    local name = utils.path(tagPath).name
    return name:replace("_mp", "")
end

local function getBitmapIndexForRegion(region)
    return (table.indexof(staticRegions, region) or 12) - 1
end

return function(props)
    props = props or {}
    local isBipedPreviewGenEnabled = props.isBipedPreviewGenEnabled
    -- Get customization widget menu
    local customization = components.new(constants.widgets.biped.id)
    local geometryName = components.new(customization:findChildWidgetTag("geometry_name").id)
    local options = list.new(customization:findChildWidgetTag("geometry_list").id, 1, 8)
    local scrollOptionsBar = bar.new(customization:findChildWidgetTag("geometry_scroll").id,
                                     "scroll")
    options:setScrollBar(scrollOptionsBar)
    local back = button.new(options:findChildWidgetTag("back").id)

    local function loadRegions()
        local customizationObjectData = getCustomizationObjectData()
        local customizationModel = customizationObjectData.model

        local bipedName = getBipedName(customizationObjectData.tag.path)
        geometryName:setText(t(bipedName))

        dynamicRegions = table.map(customizationModel.regionList, function(region)
            local regionName = region.name:trim()
            if regionName:includes("+") then
                regionName = regionName:split("+")[2]
            end
            return regionName
        end)

        local regions = table.map(dynamicRegions, function(region)
            local regionName = t(region)
            return {
                value = region,
                label = regionName,
                bitmap = function(uiComponent)
                    local icon = components.new(uiComponent:findChildWidgetTag("button_icon").id)
                    -- Default bitmap
                    icon.widgetDefinition.backgroundBitmap =
                        constants.bitmaps.customization.regions.id
                    -- Set bitmap index
                    icon:setWidgetValues({bitmapIndex = getBitmapIndexForRegion(region)})
                end
            }
        end)

        local isTheFloodProject = customizationObjectData.tag.path:includes("keymind")
        if isTheFloodProject then
            table.insert(regions, 1, {
                value = "visor",
                label = t("visor"),
                bitmap = function(uiComponent)
                    local icon = components.new(uiComponent:findChildWidgetTag("button_icon").id)
                    icon.widgetDefinition.backgroundBitmap =
                        constants.bitmaps.customization.regions.id
                    icon:setWidgetValues({bitmapIndex = getBitmapIndexForRegion("visor")})
                end
            })
        end
        table.insert(regions, 1, {
            value = "color",
            label = t("color"),
            bitmap = function(uiComponent)
                local icon = components.new(uiComponent:findChildWidgetTag("button_icon").id)
                icon.widgetDefinition.backgroundBitmap = constants.bitmaps.customization.regions.id
                icon:setWidgetValues({bitmapIndex = getBitmapIndexForRegion("color")})
            end
        })

        editing = "regions"
        options:setItems(regions)
    end

    local function loadPermutations(region)
        geometryName:setText(t(region))
        if region == "visor" then
            local visors = {}
            -- TODO Add visor permutations dynamically
            for _, visorName in ipairs(staticVisors) do
                table.insert(visors, {
                    value = table.indexof(staticVisors, visorName) - 1,
                    label = visorName:upper(),
                    bitmap = function(uiComponent)
                        local icon =
                            components.new(uiComponent:findChildWidgetTag("button_icon").id)
                        icon.widgetDefinition.backgroundBitmap =
                            constants.bitmaps.customization.regions.id
                        icon:setWidgetValues({bitmapIndex = getBitmapIndexForRegion("visor")})
                    end
                })
            end
            editing = "visor"
            options:setItems(visors)
            return
        end
        if region == "color" then
            menus.bipedColor()
            return
        end

        local customizationObjectData = getCustomizationObjectData()
        local customizationModel = customizationObjectData.model

        currentRegionIndex = table.indexof(dynamicRegions, region) --[[@as number]]

        -- TODO Check if region exists
        local lastPermutation = customizationModel.regionList[currentRegionIndex].permutationCount -
                                    1

        local permutations = {}
        for permutationIndex = 0, lastPermutation do
            local permutation =
                customizationModel.regionList[currentRegionIndex].permutationsList[permutationIndex +
                    1]
            local permutationName = permutation.name
            if permutationName:includes("+") then
                permutationName = permutationName:split("+")[4]
            end
            permutationName = t(permutationName):upper()
            local isTheFloodProject = customizationObjectData.tag.path:includes("keymind")
            table.insert(permutations, {
                value = permutationIndex,
                label = permutationName,
                bitmap = function(uiComponent)
                    local icon = components.new(uiComponent:findChildWidgetTag("button_icon").id)
                    local index = getBitmapIndexForRegion(region)
                    local permutationsBitmapTag = constants.bitmaps.customization[region]
                    if isTheFloodProject and permutationsBitmapTag then
                        icon.widgetDefinition.backgroundBitmap = permutationsBitmapTag.id
                        index = permutationIndex
                    end
                    icon:setWidgetValues({bitmapIndex = index})
                end
            })
        end
        editing = "permutations"
        options:setItems(permutations)
    end

    local function setPermutation(permutationIndex)
        local customizationObjectId = core.getCustomizationObjectId()
        assert(customizationObjectId, "No customization biped found")
        local customizationBiped = blam.biped(get_object(customizationObjectId))
        assert(customizationBiped, "No customization biped found")

        core.setObjectPermutationSafely(customizationBiped, currentRegionIndex, permutationIndex)
    end

    local function setVisor(visorIndex)
        local customizationBiped = getCustomizationObjectData().biped
        assert(customizationBiped, "No customization biped found")

        customizationBiped.shaderPermutationIndex = visorIndex
    end

    local function generateBipedPreviews()
        IsUIPhotoSessionRunning = true
        log("object_create green_screen")
        log("Generating biped previews")
        execute_script("cls")
        -- core.setWidgetValues(core.getCurrentUIWidgetTag().id, {opacity = 0.04})
        core.setWidgetValues(core.getCurrentUIWidgetTag().id, {opacity = 0})

        local customizationObjectData = getCustomizationObjectData()
        local customizationModel = customizationObjectData.model
        local customizationBiped = customizationObjectData.biped
        assert(customizationBiped, "No customization biped found")
        customizationBiped.animationFrame = 0
        setBipedColor(customizationBiped, "#A6B0B5")

        local delay = 600
        for regionIndex, region in ipairs(customizationModel.regionList) do
            local currentRegionName = region.name:split("+")[2] or region.name
            utils.delay(delay, function()
                for i = 1, 8 do
                    customizationBiped["regionPermutation" .. i] = 0
                end
                setEditingGeometry(currentRegionName, 1)
                if currentRegionName == "helmet" then
                    customizationBiped["regionPermutation" .. regionIndex] = blam.null
                end

                local screenShotName = region.name .. "+00.png"
                local screenShotPath = "/home/sledmine/Pictures/CaptureCustomization/" ..
                                           screenShotName
                os.execute(("cmd /c start /unix /usr/bin/gnome-screenshot -w -f %s"):format(
                               screenShotPath))
                -- os.execute("convert *.png -fuzz 5% -transparent 'rgb(0, 255, 0)' -set filename:base "%[basename]" "%[filename:base]_transparent.png"")
            end)
            delay = delay + 600
        end

        for regionIndex, region in ipairs(customizationModel.regionList) do
            local currentRegionName = region.name:split("+")[2] or region.name
            for permutationIndex, permutation in ipairs(region.permutationsList) do
                utils.delay(delay, function()
                    -- Hide all non editing regions
                    for otherRegionIndex, region in ipairs(customizationModel.regionList) do
                        local otherRegionName = region.name:split("+")[2] or region.name
                        if currentRegionName ~= otherRegionName then
                            customizationBiped["regionPermutation" .. otherRegionIndex] = blam.null
                        end
                    end

                    setEditingGeometry(currentRegionName, 1)
                    core.setObjectPermutationSafely(customizationBiped, regionIndex,
                                                    permutationIndex - 1)
                    -- log("Generating biped preview for " .. regionName .. " " .. permutation.name)
                    customizationBiped.animationFrame = 0
                    local screenShotName = region.name .. "+" ..
                                               string.format("%02d", permutationIndex) .. ".png"
                    local screenShotPath = "/home/sledmine/Pictures/CaptureCustomization/" ..
                                               screenShotName
                    os.execute(("cmd /c start /unix /usr/bin/gnome-screenshot -w -f %s"):format(
                                   screenShotPath))
                end)
                delay = delay + 600
            end
        end

        utils.delay(delay, function()
            log("Finished generating biped previews")
            execute_script("object_destroy green_screen")
            IsUIPhotoSessionRunning = false
            core.setWidgetValues(core.getCurrentUIWidgetTag().id, {opacity = 1})
            for i = 1, 8 do
                customizationBiped["regionPermutation" .. i] = 0
            end
        end)
    end

    options:onSelect(function(item)
        if editing == "regions" then
            local geometry = item.value
            setEditingGeometry(geometry)
            loadPermutations(geometry)
        elseif editing == "permutations" then
            local permutation = item.value
            setPermutation(permutation)
        elseif editing == "visor" then
            local visor = item.value
            setVisor(visor)
        else
            menus.bipedColor()
        end
    end)

    local function onClose()
        local customizationBiped = core.getCustomizationObjectData().biped
        if customizationBiped then
            BipedRotation = constants.customization.rotation.default
            blam.rotateObject(customizationBiped, BipedRotation, 0, 0)
        end
        --- TODO Mix permutations and visor in one concept
        if editing == "permutations" or editing == "visor" then
            customization.events.onOpen()
            return false
        end
        setCamera("lobby")
    end

    back:onClick(function()
        return onClose()
    end)

    customization:onOpen(function()
        setEditingGeometry("body")
        loadRegions()
        if isBipedPreviewGenEnabled then
            generateBipedPreviews()
        end
    end)
    customization:onClose(function()
        return onClose()
    end)
end

end,

["insurrection.components.dynamic.lobbyBrowserMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.lobbyBrowserMenu'
--------------------
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local button = require "insurrection.components.button"
local input = require "insurrection.components.input"
local list = require "insurrection.components.list"
local interface = require "insurrection.interface"
local getState = require "insurrection.redux.getState"
local blam = require "blam"
local utils = require "insurrection.utils"
local t = require "insurrection.utils".snakeCaseToUpperTitleCase

return function()
    local state = getState()
    local browser = components.new(constants.widgets.browser.id)

    local lobbies = list.new(browser:findChildWidgetTag("lobby_browser_list").id, 1, 6)
    lobbies:onSelect(function(item)
        local lobby = state.lobbies[item.value]
        if lobby then
            api.lobby(lobby.key)
        end
    end)

    browser:onOpen(function(previousWidgetTag)
        if previousWidgetTag and previousWidgetTag.handle.value == constants.widgets.dashboard.id then
            api.getLobbies()
        end
        api.stopRefreshLobby()
    end)

    return function()
        lobbies:setItems(table.map((state.lobbies or {}), function(lob, lobbyIndex)
            return {
                -- label = v.name,
                value = lobbyIndex,
                bitmap = function(element)
                    local mapPreview = components.new(element:findChildWidgetTag("preview").id)
                    local template = components.new(element:findChildWidgetTag("template").id)
                    --local gametypeIcon = components.new(element:findChildWidgetTag("gametype_icon").id)
                    local description = components.new(element:findChildWidgetTag("description").id)
                    local owner = components.new(element:findChildWidgetTag("owner").id)
                    local players = components.new(element:findChildWidgetTag("players").id)

                    mapPreview.widgetDefinition.backgroundBitmap = constants.bitmaps
                                                                       .unknownMapPreview.id
                    local mapCollection = blam.tagCollection(constants.tagCollections.maps.id)
                    assert(mapCollection, "No map preview collection found")
                    for k, v in pairs(mapCollection.tagList) do
                        local bitmapTag = blam.getTag(v)
                        assert(bitmapTag, "No bitmap tag found")
                        local mapName = utils.path(bitmapTag.path).name:lower()
                        if mapName == lob.map then
                            mapPreview.widgetDefinition.backgroundBitmap = bitmapTag.id
                        end
                    end

                    template:setText(t(lob.template))
                    description:setText(t(lob.gametype) .. " on " .. t(lob.map))
                    local lobbyOwner = table.find(lob.players, function(player)
                        return player.publicId == lob.owner
                    end)
                    owner:setText(lobbyOwner.name)
                    players:setText(#lob.players .. "/" .. 16)
                end
            }
        end))
    end
end

end,

["insurrection.components.dynamic.customizationBipedColorMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.customizationBipedColorMenu'
--------------------
local color = require "color"
local components = require "insurrection.components"
local constants = require "insurrection.constants"
local list = require "insurrection.components.list"
local button = require "insurrection.components.button"
local core = require "insurrection.core"
local utils = require "insurrection.utils"
local blam  = require "blam"
local l = function(str)
    local snakeCase = utils.camelCaseToSnakeCase(str)
    local titleCase = utils.snakeCaseToTitleCase(snakeCase)
    return titleCase:upper()
end

local function getColorName(color)
    return table.keyof(constants.customColor, color)
end

return function()
    local customizationColor = components.new(constants.widgets.bipedColor.id)

    local primaryColorLabel = components.new(customizationColor:findChildWidgetTag(
                                                 "primary_color_subtitle").id)
    
    local secondaryColorLabel = components.new(customizationColor:findChildWidgetTag ("secondary_color_subtitle").id)

    local customizationBipedColorOptionsHandle = customizationColor:findChildWidgetTag(
                                                     "customization_biped_colors_options").id
    local customizationBipedColorOptions = components.new(customizationBipedColorOptionsHandle)

    for columnIndex = 1, #constants.customColors do
        local colorPrimaryColumnHandle = customizationBipedColorOptions:findChildWidgetTag(
                                             "primary_colors_column_" .. columnIndex).id
        local colorPrimaryColumnList = list.new(colorPrimaryColumnHandle)
        colorPrimaryColumnList:scrollable(false)
        colorPrimaryColumnList:onSelect(function(item)
            primaryColorLabel:setText(l(getColorName(item.value)))
            core.setCustomizationBipedColor(item.value)
        end)
        colorPrimaryColumnList:setItems(table.map(
                                            table.reverse(constants.customColors[columnIndex]),
                                            function(color)
                return {value = color, text = color}
            end))

        local colorSecondaryColumnHandle = customizationBipedColorOptions:findChildWidgetTag(
                                               "secondary_colors_column_" .. columnIndex).id
        local colorSecondaryColumnList = list.new(colorSecondaryColumnHandle)
        colorSecondaryColumnList:scrollable(false)
        colorSecondaryColumnList:onSelect(function(item)
            secondaryColorLabel:setText(l(getColorName(item.value)))
            core.setCustomizationBipedColor(nil, item.value)
        end)
        colorSecondaryColumnList:setItems(table.map(table.reverse(
                                                        constants.customColors[columnIndex]),
                                                    function(color)
            return {value = color, text = color}
        end))

    end

    customizationColor:onOpen(function()
        BipedRotation = constants.customization.rotation.color
        local customizationBiped = core.getCustomizationObjectData().biped
        assert(customizationBiped, "No customization biped found")
        blam.rotateObject(customizationBiped, BipedRotation, 0 , 0)
        execute_script "camera_set customization_color 30"
    end)
end

end,

["insurrection.components.dynamic.chimeraMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.chimeraMenu'
--------------------
local components = require "insurrection.components"
local blam = require "blam"
local constants = require "insurrection.constants"
local chimera = require "insurrection.mods.chimera"
local interface = require "insurrection.interface"
local checkbox = require "insurrection.components.checkbox"
local spinner = require "insurrection.components.spinner"
local core = require "insurrection.core"

return function()
    local chimeraMod = components.new(constants.widgets.chimera.id)
    local chimeraOptions = components.new(chimeraMod:findChildWidgetTag("chimera_mod_options").id)

    local config = chimera.getConfiguration() or {}
    local preferences = chimera.getPreferences() or {}

    local elements = {}

    local aspectWidth = core.getScreenAspectRatio()
    local isUltraWide = aspectWidth > 16

    local elementsData = {
        ["USE VSYNC"] = {
            value = config.video_mode.vsync,
            change = function(value)
                config.video_mode.vsync = value and 1 or 0
                chimera.saveConfiguration(config)
            end
        },
        ["SHOW FPS"] = {
            value = preferences.chimera_show_fps,
            change = function(value)
                preferences.chimera_show_fps = value and 1 or 0
                chimera.executeCommand("chimera_show_fps " .. (value and 1 or 0))
            end
        },
        ["WINDOWED MODE"] = {
            value = config.video_mode.windowed,
            change = function(value)
                config.video_mode.windowed = value and 1 or 0
                chimera.saveConfiguration(config)
            end
        },
        ["BORDERLESS"] = {
            value = config.video_mode.borderless,
            change = function(value)
                config.video_mode.borderless = value and 1 or 0
                chimera.saveConfiguration(config)
            end
        },
        ["LOAD MAPS ON RAM"] = {
            value = config.memory.enable_map_memory_buffer,
            change = function(value)
                config.memory.enable_map_memory_buffer = value and 1 or 0
                chimera.saveConfiguration(config)
                interface.dialog("INFORMATION", "You have changed a critical Chimera setting.",
                                 "You need to restart the game for the changes to take effect.")
            end
        },
        ["ANISOTROPIC FILTER"] = {
            value = preferences.chimera_af,
            change = function(value)
                preferences.chimera_af = value and 1 or 0
                chimera.executeCommand("chimera_af " .. (value and 1 or 0))
            end
        },
        ["BLOCK BUFFERING"] = {
            value = preferences.chimera_block_buffering,
            change = function(value)
                preferences.chimera_block_buffering = value and 1 or 0
                chimera.executeCommand("chimera_block_buffering " .. (value and 1 or 0))
            end
        },
        ["BLOCK HOLD F1 AT START"] = {
            value = preferences.chimera_block_hold_f1,
            change = function(value)
                preferences.chimera_block_hold_f1 = value and 1 or 0
                chimera.executeCommand("chimera_block_hold_f1 " .. (value and 1 or 0))
            end
        },
        ["BLOCK LOADING SCREEN"] = {
            value = preferences.chimera_block_loading_screen,
            change = function(value)
                preferences.chimera_block_loading_screen = value and 1 or 0
                chimera.executeCommand("chimera_block_loading_screen " .. (value and 1 or 0))
            end
        },
        ["BLOCK ZOOM BLUR"] = {
            value = preferences.chimera_block_zoom_blur,
            change = function(value)
                preferences.chimera_block_zoom_blur = value and 1 or 0
                chimera.executeCommand("chimera_block_zoom_blur " .. (value and 1 or 0))
            end
        },
        -- ["BLOCK MOUSE ACCELERATION"] = function(value)
        --    preferences.chimera_block_mouse_acceleration = value and 1 or 0
        --    chimera.executeCommand("chimera_block_mouse_acceleration " .. (value and 1 or 0))
        -- end,
        ["DEVMODE"] = {
            value = preferences.chimera_devmode,
            change = function(value)
                preferences.chimera_devmode = value and 1 or 0
                chimera.executeCommand("chimera_devmode " .. (value and 1 or 0))
            end
        },
        ["SHOW BUDGET"] = {
            value = preferences.chimera_budget,
            change = function(value)
                preferences.chimera_budget = value and 1 or 0
                chimera.executeCommand("chimera_budget " .. (value and 1 or 0))
            end
        },
        ["FOV"] = {
            value = preferences.chimera_fov,
            change = function(value)
                preferences.chimera_fov = value
                chimera.executeCommand("chimera_fov " .. value)
                log("Setting horizontal FOV")
            end
        }
    }

    for i = 1, chimeraOptions.widgetDefinition.childWidgetsCount - 1 do
        local childWidget = chimeraOptions.widgetDefinition.childWidgets[i]
        local tag = blam.getTag(childWidget.widgetTag)
        assert(tag)
        if tag.path:includes "checkbox" then
            local check = checkbox.new(childWidget.widgetTag)
            elements[check:getText()] = check
            check:onToggle(function(value)
                local optionName = check:getText()
                if elementsData[optionName] then
                    elementsData[optionName].change(value)
                end
            end)
        elseif tag.path:includes "spinner" then
            local spin = spinner.new(childWidget.widgetTag)
            elements[spin:getText()] = spin
            spin:onScroll(function(value, index)
                local optionName = spin:getText()
                if elementsData[optionName] then
                    elementsData[optionName].change(value)
                end
            end)
        end
    end
    chimeraMod:onOpen(function()
        log("Aspect width: " .. aspectWidth)
        config = chimera.getConfiguration() or {}
        preferences = chimera.getPreferences() or {}

        -- Create fovs list with values from 60 to 120 in steps of 5
        --log(preferences.chimera_fov)
        local fovs = {}
        for fov = 60, 120, 1 do
            table.insert(fovs, tostring(fov) .. (isUltraWide and "v" or ""))
        end

        elements["FOV"]:setValues(fovs)

        elementsData["USE VSYNC"].value = config.video_mode.vsync
        elementsData["SHOW FPS"].value = preferences.chimera_show_fps
        elementsData["WINDOWED MODE"].value = config.video_mode.windowed
        elementsData["BORDERLESS"].value = config.video_mode.borderless
        elementsData["LOAD MAPS ON RAM"].value = config.memory.enable_map_memory_buffer
        elementsData["ANISOTROPIC FILTER"].value = preferences.chimera_af
        elementsData["BLOCK BUFFERING"].value = preferences.chimera_block_buffering
        elementsData["BLOCK HOLD F1 AT START"].value = preferences.chimera_block_hold_f1
        elementsData["BLOCK LOADING SCREEN"].value = preferences.chimera_block_loading_screen
        elementsData["BLOCK ZOOM BLUR"].value = preferences.chimera_block_zoom_blur
        -- elementsData["BLOCK MOUSE ACCELERATION"].value = preferences.chimera_block_mouse_acceleration
        elementsData["DEVMODE"].value = preferences.chimera_devmode
        elementsData["SHOW BUDGET"].value = preferences.chimera_budget
        elementsData["FOV"].value = preferences.chimera_fov
        for k, component in pairs(elements) do
            local data = elementsData[k]
            if data then
                if component.type == "checkbox" then
                    component:setValue(data.value == 1)
                elseif component.type == "spinner" then
                    component:setValue(data.value)
                end
            else
                -- error("Element from list not found: " .. k)
            end
        end
        log("chimeraMod:onOpen")
    end)
end

end,

["insurrection.components.spinner"] = function()
--------------------
-- Module: 'insurrection.components.spinner'
--------------------
---@diagnostic disable: inject-field
local blam = require "blam"
local uiWidgetDefinition = blam.uiWidgetDefinition
local unicodeStringList = blam.unicodeStringList
local isNull = blam.isNull
local components = require "insurrection.components"
local button = require "insurrection.components.button"

---@class uiComponentSpinnerClass : uiComponent
local spinner = setmetatable({
    ---@type string
    type = "spinner",
    ---@type string
    value = nil,
    ---@type string[]
    values = {},
    ---@type number
    leftArrowTagId = nil,
    ---@type number
    rightArrowTagId = nil,
    ---@type number
    currentValueIndex = 1
}, {__index = components})

---@class uiComponentSpinnerEvents : uiComponentEvents
---@field onScroll fun(value: string, index: number):boolean | nil

---@class uiComponentSpinner : uiComponentSpinnerClass
---@field events uiComponentSpinnerEvents

---@param tagId number
---@return uiComponentSpinner
function spinner.new(tagId)
    local instance = setmetatable(components.new(tagId), {__index = spinner}) --[[@as uiComponentSpinner]]
    assert(instance.tag.path:find("spinner", 1, true),
           "Tag " .. instance.tag.path .. " is not a spinner")
    instance.leftArrowTagId = instance:findChildWidgetTag("arrow_left").id
    instance.rightArrowTagId = instance:findChildWidgetTag("arrow_right").id
    local leftArrow = button.new(instance.leftArrowTagId)
    leftArrow:onClick(function()
        instance:scroll(-1)
    end)
    local rightArrow = button.new(instance.rightArrowTagId)
    rightArrow:onClick(function()
        instance:scroll(1)
    end)
    return instance
end

---@param self uiComponentSpinner
---@param text string
local function setValueText(self, text)
    local childUnicodeStrings
    local childWidgetDefinition
    local labelTagId = self:findChildWidgetTag("label").id
    local widgetDefinition = uiWidgetDefinition(labelTagId)
    assert(widgetDefinition, "Invalid widgetDefinition")
    local unicodeStrings = unicodeStringList(self.widgetDefinition.unicodeStringListTag)
    if not (unicodeStrings and not isNull(unicodeStrings)) then
        error("No unicodeStringList found for widgetDefinition " .. self.tag.path)
    end
    local stringListIndex = widgetDefinition.stringListIndex
    local newStrings = unicodeStrings.strings
    newStrings[stringListIndex + 1] = text
    unicodeStrings.strings = newStrings
end

---@param self uiComponentSpinner
---@return string
function spinner.getValue(self)
    return self.values[self.currentValueIndex]
end

---@param self uiComponentSpinner
---@return number
function spinner.getValueIndex(self)
    return self.currentValueIndex
end

---@param self uiComponentSpinner
---@param value string
function spinner.setValue(self, value)
    local value = tostring(value)
    local index = table.indexof(self.values, value)
    if index then
        self.currentValueIndex = index
        setValueText(self, value)
    end
end

---@param self uiComponentSpinner
---@param values string[]
function spinner.setValues(self, values)
    self.values = values
    self:setValue(values[1])
    setValueText(self, values[1])
end

---@param self uiComponentSpinner
---@param direction number
function spinner.scroll(self, direction)
    local itemIndex = self.currentValueIndex + direction
    -- TODO Add flag to define if the spinner should loop or not
    if itemIndex < 1 then
        itemIndex = #self.values
    elseif itemIndex > #self.values then
        itemIndex = 1
    end
    self.currentValueIndex = itemIndex
    setValueText(self, self.values[itemIndex])
    if self.events.onScroll then
        self.events.onScroll(self.values[itemIndex], itemIndex)
    end
end

---@param self uiComponentSpinner
---@param callback fun(value: string, index: number):boolean | nil
function spinner.onScroll(self, callback)
    self.events.onScroll = callback
end

return spinner

end,

["insurrection.components.dynamic.opticMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.opticMenu'
--------------------
local components = require "insurrection.components"
local checkbox = require "insurrection.components.checkbox"
local constants = require "insurrection.constants"
local chimera = require "insurrection.mods.chimera"
local interface = require "insurrection.interface"
local blam = require "blam"
local optic = require "insurrection.mods.optic"
local menus = require "insurrection.menus"

return function()
    local opticMod = components.new(constants.widgets.optic.id)
    local options = components.new(opticMod:findChildWidgetTag("options").id)
    local checkboxes = {}
    local config = optic.getConfiguration() or {}
    for i = 1, options.widgetDefinition.childWidgetsCount - 3 do
        local childWidget = options.widgetDefinition.childWidgets[i]
        local check = checkbox.new(childWidget.widgetTag)
        checkboxes[check:getText()] = check
        check:onToggle(function(value)
            local optionName = check:getText()
            local optionsToggle = {
                ["ENABLE SOUND"] = function(value)
                    execute_script("optic_test")
                    config.enableSound = value
                    optic.saveConfiguration(config)
                end,
                ["HITMARKER"] = function(value)
                    config.hitmarker = value
                    optic.saveConfiguration(config)
                end,
                ["HUD MESSAGES"] = function(value)
                    config.hudMessages = value
                    optic.saveConfiguration(config)
                end
            }
            if optionsToggle[optionName] then
                optionsToggle[optionName](value)
            end
        end)
    end
    opticMod:onOpen(function()
        log("opticMod:onOpen")
        if not optic.isInstalled() then
            interface.dialog("ERROR", "Optic mod is not installed.",
                             "Please install it with Mercury and try again.")
            return
        end
        config = optic.getConfiguration() or {}
        log(config)
        local optionsMapping = {
            ["ENABLE SOUND"] = config.enableSound,
            ["HITMARKER"] = config.hitmarker,
            ["HUD MESSAGES"] = config.hudMessages
            -- ["VOLUME"] = config.volume,
        }
        for k, check in pairs(checkboxes) do
            check:setValue(optionsMapping[k])
        end
    end)
end

end,

["insurrection.mods.optic"] = function()
--------------------
-- Module: 'insurrection.mods.optic'
--------------------
local luna = require "luna"
local getMyGamesHaloCEPath = require"insurrection.core".getMyGamesHaloCEPath
local json = require "json"

local optic = {}

local configPath = "\\chimera\\lua\\data\\global\\optic\\optic.json"
local installPath = "\\chimera\\lua\\scripts\\global\\optic.lua"

---@class opticConfiguration
---@field enableSound boolean
---@field hitmarker boolean
---@field hudMessages boolean
---@field volume number

---Get optic configuration
---@return opticConfiguration?
function optic.getConfiguration()
    local configJson = luna.file.read(getMyGamesHaloCEPath() .. configPath)
    if configJson then
        ---@type opticConfiguration
        local configuration = json.decode(configJson)
        return configuration
    end
end

---Save optic configuration
---@param configuration opticConfiguration
function optic.saveConfiguration(configuration)
    local configJson = json.encode(configuration)
    return luna.file.write(getMyGamesHaloCEPath() .. configPath, configJson)
end

---Check if optic is installed
---@return boolean
function optic.isInstalled()
    return luna.file.read(getMyGamesHaloCEPath() .. installPath) ~= nil
end

return optic

end,

["insurrection.components.dynamic.balltzeMenu"] = function()
--------------------
-- Module: 'insurrection.components.dynamic.balltzeMenu'
--------------------
local components = require "insurrection.components"
local checkbox = require "insurrection.components.checkbox"
local constants = require "insurrection.constants"
local chimera = require "insurrection.mods.chimera"
local interface = require "insurrection.interface"
local blam = require "blam"
local balltze = require "insurrection.mods.balltze"
local menus = require "insurrection.menus"
local spinner = require "insurrection.components.spinner"
local utils = require "insurrection.utils"

-- I don't like this, but I'll keep it for now
local loadingScreenDescription =
    "Select this option to enable the classic Halo: Combat Evolved loading screen before a map it's opened.\r\nWARNING: There is currently a bug where pressing a key in the loading screen might crash the game.\r\nAvoid pressing any key in the meantime."
local preloadMapTexturesDescription =
    "Select this option to preload a map textures before it's opened.\r\nWARNING: It could take a long time on low specs computers."
local gamepadProfileDescription =
    "Select a profile of gamepad buttons to use in the game.\r\nFor example, Balltze will use this to know which buttons to show in the HUD messages."
local reminderDescription = "\r\n\r\nNOTE: You must restart the game to apply changes."

return function()
    local balltzeMod = components.new(constants.widgets.balltze.id)
    local options = components.new(balltzeMod:findChildWidgetTag("options").id)
    local footer = components.new(balltzeMod:findChildWidgetTag("footer").id)
    local description = components.new(footer:findChildWidgetTag("text").id)

    local config = balltze.getConfiguration() or {}

    local elements = {}
    local elementsData = {
        ["LOADING SCREEN"] = {
            value = false,
            change = function(value)
                config.loading_screen.enable = value
                balltze.saveConfiguration(config)
            end,
            focus = function()
                description:setText(loadingScreenDescription .. reminderDescription)
            end
        },
        ["PRELOAD MAP TEXTURES"] = {
            value = false,
            change = function(value)
                config.preload_map_textures.enable = value
                balltze.saveConfiguration(config)
            end,
            focus = function()
                description:setText(preloadMapTexturesDescription .. reminderDescription)
            end
        },
        ["GAMEPAD PROFILE"] = {
            value = "xbox_360_windows",
            change = function(value)
                config.gamepad = value
                balltze.saveConfiguration(config)
            end,
            focus = function()
                description:setText(gamepadProfileDescription .. reminderDescription)
            end
        }
    }

    for i = 1, options.widgetDefinition.childWidgetsCount - 1 do
        local childWidget = options.widgetDefinition.childWidgets[i]
        local tag = blam.getTag(childWidget.widgetTag)
        assert(tag)
        if tag.path:includes "checkbox" then
            local check = checkbox.new(childWidget.widgetTag)
            elements[check:getText()] = check
            check:onToggle(function(value)
                local optionName = check:getText()
                if elementsData[optionName] then
                    elementsData[optionName].change(value)
                end
            end)
            check:onFocus(function()
                local optionName = check:getText()
                if elementsData[optionName] then
                    elementsData[optionName].focus()
                end
            end)
        elseif tag.path:includes "spinner" then
            local spin = spinner.new(childWidget.widgetTag)
            console_debug(spin:getText())
            elements[spin:getText()] = spin
            spin:onScroll(function(value, index)
                local optionName = spin:getText()
                if elementsData[optionName] then
                    elementsData[optionName].change(balltze.getGamepadProfiles()[index])
                end
            end)
            spin:onFocus(function()
                local optionName = spin:getText()
                if elementsData[optionName] then
                    elementsData[optionName].focus()
                end
            end)
        end
    end
    balltzeMod:onOpen(function()
        log("balltzeMod:onOpen")
        if not balltze.isInstalled() then
            interface.dialog("ERROR", "Balltze mod is not installed.",
                             "Please install it with Mercury and try again.")
            return
        end
        config = balltze.getConfiguration() or {}

        local gamepadProfiles = table.map(balltze.getGamepadProfiles(), function(profile)
            local title = utils.snakeCaseToTitleCase(profile)
            return title
        end)

        elements["GAMEPAD PROFILE"]:setValues(gamepadProfiles)

        elementsData["LOADING SCREEN"].value = config.loading_screen.enable
        elementsData["PRELOAD MAP TEXTURES"].value = config.preload_map_textures.enable
        elementsData["GAMEPAD PROFILE"].value = utils.snakeCaseToTitleCase(config.gamepad)

        for k, component in pairs(elements) do
            local data = elementsData[k]
            if data then
                component:setValue(data.value)
            else
                error("Element from list not found: " .. k)
            end
        end
    end)
end

end,

["insurrection.mods.balltze"] = function()
--------------------
-- Module: 'insurrection.mods.balltze'
--------------------
local luna = require "luna"
local getMyGamesHaloCEPath = require"insurrection.core".getMyGamesHaloCEPath
local json = require "json"

local balltze = {}

local configPath = "\\balltze\\config\\settings.json"
local installPath = "mods\\balltze.dll"

---@class balltzeConfiguration
---@field gamepad string
---@field loading_screen {enable: boolean}
---@field preload_map_textures {enable: boolean, min_map_size: number}

---Get balltze configuration
---@return balltzeConfiguration?
function balltze.getConfiguration()
    local configJson = luna.file.read(getMyGamesHaloCEPath() .. configPath)
    if configJson then
        ---@type balltzeConfiguration
        local configuration = json.decode(configJson)
        return configuration
    end
end

---Get balltze gamepad profiles
---@return string[]
function balltze.getGamepadProfiles()
    ---@type string[]
    --return list_directory(getMyGamesHaloCEPath() .. "\\balltze\\gamepad")
    --return {"Xbox 360 Windows","Dual Shock 1 Windows"}
    return {"xbox_360_windows","dual_shock_1_windows"}
end

---Save balltze configuration
---@param configuration balltzeConfiguration
function balltze.saveConfiguration(configuration)
    local configJson = json.encode(configuration)
    return luna.file.write(getMyGamesHaloCEPath() .. configPath, configJson)
end

---Check if balltze is installed
---@return boolean
function balltze.isInstalled()
    return luna.file.read(installPath) ~= nil
end

return balltze

end,

["async"] = function()
--------------------
-- Module: 'async'
--------------------
local lanes = require "lanes"

local async = {}

async._VERSION = "1.0.2"

local asyncLibs = "*"
THREAD_LANES = {}

---Perform an async function
---@generic T, U
---@param inputFunction fun(await: fun(callback: fun(...:U): (T?), ...:U): T)
function async.async(inputFunction)
    local co
    local await = function(asyncCallback, ...)
        if not asyncCallback or type(asyncCallback) ~= "function" then
            error("Await callback is not a function or is nil", 2)
            return
        end
        table.insert(THREAD_LANES, {
            thread = lanes.gen(asyncLibs, asyncCallback)(...),
            callback = function(ret)
                local ok = coroutine.resume(co, ret)
                if not ok then
                    error(debug.traceback(co), 0)
                end
            end,
            co = co
        })
        return coroutine.yield()
    end
    ---@return boolean success Async function has finished successfully
    return function()
        co = coroutine.create(inputFunction)
        local ok, message = coroutine.resume(co, await)
        if not ok then
            error(debug.traceback(co, message), 0)
        end
        return ok
    end
end

---Configure the async module to use the specified libraries
---@param libs any
function async.configure(libs)
    asyncLibs = libs
end

---Dispatches the async functions and returns the number of async functions still running
function async.dispatch()
    for index, lane in ipairs(THREAD_LANES) do
        --print(lane.thread.status)
        if lane.thread.status == "done" then
            table.remove(THREAD_LANES, index)
            lane.callback(lane.thread[1])
        elseif lane.thread.status == "error" then
            local threadError = lane.thread[1]
            table.remove(THREAD_LANES, index)
            error(threadError)
        end
    end
    return #THREAD_LANES
end

return async
end,

["lanes"] = function()
--------------------
-- Module: 'lanes'
--------------------
--
-- LANES.LUA
--
-- Multithreading and -core support for Lua
--
-- Authors: Asko Kauppi <akauppi@gmail.com>
--          Benoit Germain <bnt.germain@gmail.com>
--
-- History: see CHANGES
--
--[[
===============================================================================

Copyright (C) 2007-10 Asko Kauppi <akauppi@gmail.com>
Copyright (C) 2010-13 Benoit Germain <bnt.germain@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

===============================================================================
]]--

local core = require "lanes.core"
-- Lua 5.1: module() creates a global variable
-- Lua 5.2: module() is gone
-- almost everything module() does is done by require() anyway
-- -> simply create a table, populate it, return it, and be done
local lanesMeta = {}
local lanes = setmetatable( {}, lanesMeta)

-- this function is available in the public interface until it is called, after which it disappears
lanes.configure = function( settings_)

    -- This check is for sublanes requiring Lanes
    --
    -- TBD: We could also have the C level expose 'string.gmatch' for us. But this is simpler.
    --
    if not string then
        error( "To use 'lanes', you will also need to have 'string' available.", 2)
    end
    -- Configure called so remove metatable from lanes
    setmetatable( lanes, nil)
    --
    -- Cache globals for code that might run under sandboxing
    --
    local assert = assert( assert)
    local string_gmatch = assert( string.gmatch)
    local string_format = assert( string.format)
    local select = assert( select)
    local type = assert( type)
    local pairs = assert( pairs)
    local tostring = assert( tostring)
    local error = assert( error)

    local default_params =
    {
        nb_keepers = 1,
        on_state_create = nil,
        shutdown_timeout = 0.25,
        with_timers = true,
        track_lanes = false,
        demote_full_userdata = nil,
        verbose_errors = false,
        -- LuaJIT provides a thread-unsafe allocator by default, so we need to protect it when used in parallel lanes
        allocator = (package.loaded.jit and jit.version) and "protected" or nil
    }
    local boolean_param_checker = function( val_)
        -- non-'boolean-false' should be 'boolean-true' or nil
        return val_ and (val_ == true) or true
    end
    local param_checkers =
    {
        nb_keepers = function( val_)
            -- nb_keepers should be a number > 0
            return type( val_) == "number" and val_ > 0
        end,
        with_timers = boolean_param_checker,
        allocator = function( val_)
            -- can be nil, "protected", or a function
            return val_ and (type( val_) == "function" or val_ == "protected") or true
        end,
        on_state_create = function( val_)
            -- on_state_create may be nil or a function
            return val_ and type( val_) == "function" or true
        end,
        shutdown_timeout = function( val_)
            -- shutdown_timeout should be a number >= 0
            return type( val_) == "number" and val_ >= 0
        end,
        track_lanes = boolean_param_checker,
        demote_full_userdata = boolean_param_checker,
        verbose_errors = boolean_param_checker
    }

    local params_checker = function( settings_)
        if not settings_ then
            return default_params
        end
        -- make a copy of the table to leave the provided one unchanged, *and* to help ensure it won't change behind our back
        local settings = {}
        if type( settings_) ~= "table" then
            error "Bad parameter #1 to lanes.configure(), should be a table"
        end
        -- any setting unknown to Lanes raises an error
        for setting, _ in pairs( settings_) do
            if not param_checkers[setting] then
            error( "Unknown parameter '" .. setting .. "' in configure options")
            end
        end
        -- any setting not present in the provided parameters takes the default value
        for key, checker in pairs( param_checkers) do
            local my_param = settings_[key]
            local param
            if my_param ~= nil then
                param = my_param
            else
                param = default_params[key]
            end
            if not checker( param) then
                error( "Bad " .. key .. ": " .. tostring( param), 2)
            end
            settings[key] = param
        end
        return settings
    end
    local settings = core.configure and core.configure( params_checker( settings_)) or core.settings
    local core_lane_new = assert( core.lane_new)
    local max_prio = assert( core.max_prio)

    lanes.ABOUT =
    {
        author= "Asko Kauppi <akauppi@gmail.com>, Benoit Germain <bnt.germain@gmail.com>",
        description= "Running multiple Lua states in parallel",
        license= "MIT/X11",
        copyright= "Copyright (c) 2007-10, Asko Kauppi; (c) 2011-19, Benoit Germain",
        version = assert( core.version)
    }


    -- Making copies of necessary system libs will pass them on as upvalues;
    -- only the first state doing "require 'lanes'" will need to have 'string'
    -- and 'table' visible.
    --
    local function WR(str)
        io.stderr:write( str.."\n" )
    end

    local function DUMP( tbl )
        if not tbl then return end
        local str=""
        for k,v in pairs(tbl) do
            str= str..k.."="..tostring(v).."\n"
        end
        WR(str)
    end


    ---=== Laning ===---

    -- lane_h[1..n]: lane results, same as via 'lane_h:join()'
    -- lane_h[0]:    can be read to make sure a thread has finished (always gives 'true')
    -- lane_h[-1]:   error message, without propagating the error
    --
    --      Reading a Lane result (or [0]) propagates a possible error in the lane
    --      (and execution does not return). Cancelled lanes give 'nil' values.
    --
    -- lane_h.state: "pending"/"running"/"waiting"/"done"/"error"/"cancelled"
    --
    -- Note: Would be great to be able to have '__ipairs' metamethod, that gets
    --      called by 'ipairs()' function to custom iterate objects. We'd use it
    --      for making sure a lane has ended (results are available); not requiring
    --      the user to precede a loop by explicit 'h[0]' or 'h:join()'.
    --
    --      Or, even better, 'ipairs()' should start valuing '__index' instead
    --      of using raw reads that bypass it.
    --
    -----
    -- lanes.gen( [libs_str|opt_tbl [, ...],] lane_func ) ( [...] ) -> h
    --
    -- 'libs': nil:     no libraries available (default)
    --         "":      only base library ('assert', 'print', 'unpack' etc.)
    --         "math,os": math + os + base libraries (named ones + base)
    --         "*":     all standard libraries available
    --
    -- 'opt': .priority:  int (-3..+3) smaller is lower priority (0 = default)
    --
    --        .globals:  table of globals to set for a new thread (passed by value)
    --
    --        .required: table of packages to require
    --
    --        .gc_cb:    function called when the lane handle is collected
    --
    --        ... (more options may be introduced later) ...
    --
    -- Calling with a function parameter ('lane_func') ends the string/table
    -- modifiers, and prepares a lane generator.

    local valid_libs =
    {
        ["package"] = true,
        ["table"] = true,
        ["io"] = true,
        ["os"] = true,
        ["string"] = true,
        ["math"] = true,
        ["debug"] = true,
        ["bit32"] = true, -- Lua 5.2 only, ignored silently under 5.1
        ["utf8"] = true, -- Lua 5.3 only, ignored silently under 5.1 and 5.2
        ["bit"] = true, -- LuaJIT only, ignored silently under PUC-Lua
        ["jit"] = true, -- LuaJIT only, ignored silently under PUC-Lua
        ["ffi"] = true, -- LuaJIT only, ignored silently under PUC-Lua
        --
        ["base"] = true,
        ["coroutine"] = true, -- part of "base" in Lua 5.1
        ["lanes.core"] = true
    }

    local raise_option_error = function( name_, tv_, v_)
        error( "Bad '" .. name_ .. "' option: " .. tv_ .. " " .. string_format( "%q", tostring( v_)), 4)
    end

    local opt_validators =
    {
        priority = function( v_)
            local tv = type( v_)
            return (tv == "number") and v_ or raise_option_error( "priority", tv, v_)
        end,
        globals = function( v_)
            local tv = type( v_)
            return (tv == "table") and v_ or raise_option_error( "globals", tv, v_)
        end,
        package = function( v_)
            local tv = type( v_)
            return (tv == "table") and v_ or raise_option_error( "package", tv, v_)
        end,
        required = function( v_)
            local tv = type( v_)
            return (tv == "table") and v_ or raise_option_error( "required", tv, v_)
        end,
        gc_cb = function( v_)
            local tv = type( v_)
            return (tv == "function") and v_ or raise_option_error( "gc_cb", tv, v_)
        end
    }

    -- PUBLIC LANES API
    -- receives a sequence of strings and tables, plus a function
    local gen = function( ...)
        -- aggregrate all strings together, separated by "," as well as tables
        -- the strings are a list of libraries to open
        -- the tables contain the lane options
        local opt = {}
        local libs = nil

        local n = select( '#', ...)

        -- we need at least a function
        if n == 0 then
            error( "No parameters!", 2)
        end

        -- all arguments but the last must be nil, strings, or tables
        for i = 1, n - 1 do
            local v = select( i, ...)
            local tv = type( v)
            if tv == "string" then
                libs = libs and libs .. "," .. v or v
            elseif tv == "table" then
                for k, vv in pairs( v) do
                    opt[k]= vv
                end
            elseif v == nil then
                -- skip
            else
                error( "Bad parameter " .. i .. ": " .. tv .. " " .. string_format( "%q", tostring( v)), 2)
            end
        end

        -- the last argument should be a function or a string
        local func = select( n, ...)
        local functype = type( func)
        if functype ~= "function" and functype ~= "string" then
            error( "Last parameter not function or string: " .. functype .. " " .. string_format( "%q", tostring( func)), 2)
        end

        -- check that the caller only provides reserved library names, and those only once
        -- "*" is a special case that doesn't require individual checking
        if libs and libs ~= "*" then
            local found = {}
            for s in string_gmatch(libs, "[%a%d.]+") do
                if not valid_libs[s] then
                    error( "Bad library name: " .. s, 2)
                else
                    found[s] = (found[s] or 0) + 1
                    if found[s] > 1 then
                        error( "libs specification contains '" .. s .. "' more than once", 2)
                    end
                end
            end
        end

        -- validate that each option is known and properly valued
        for k, v in pairs( opt) do
            local validator = opt_validators[k]
            if not validator then
                error( (type( k) == "number" and "Unkeyed option: " .. type( v) .. " " .. string_format( "%q", tostring( v)) or "Bad '" .. tostring( k) .. "' option"), 2)
            else
                opt[k] = validator( v)
            end
        end

        local priority, globals, package, required, gc_cb = opt.priority, opt.globals, opt.package or package, opt.required, opt.gc_cb
        return function( ...)
            -- must pass functions args last else they will be truncated to the first one
            return core_lane_new( func, libs, priority, globals, package, required, gc_cb, ...)
        end
    end -- gen()

    ---=== Timers ===---

    -- PUBLIC LANES API
    local timer = function() error "timers are not active" end
    local timers = timer
    local timer_lane = nil

    -- timer_gateway should always exist, even when the settings disable the timers
    local timer_gateway = assert( core.timer_gateway)

    -----
    -- <void> = sleep( [seconds_])
    --
    -- PUBLIC LANES API
    local sleep = function( seconds_)
        seconds_ = seconds_ or 0.0 -- this causes false and nil to be a valid input, equivalent to 0.0, but that's ok
        if type( seconds_) ~= "number" then
            error( "invalid duration " .. string_format( "%q", tostring(seconds_)))
        end
        -- receive data on a channel no-one ever sends anything, thus blocking for the specified duration
        return timer_gateway:receive( seconds_, "ac100de1-a696-4619-b2f0-a26de9d58ab8")
    end


    if settings.with_timers ~= false then

    --
    -- On first 'require "lanes"', a timer lane is spawned that will maintain
    -- timer tables and sleep in between the timer events. All interaction with
    -- the timer lane happens via a 'timer_gateway' Linda, which is common to
    -- all that 'require "lanes"'.
    --
    -- Linda protocol to timer lane:
    --
    --  TGW_KEY: linda_h, key, [wakeup_at_secs], [repeat_secs]
    --
    local TGW_KEY= "(timer control)"    -- the key does not matter, a 'weird' key may help debugging
    local TGW_QUERY, TGW_REPLY = "(timer query)", "(timer reply)"
    local first_time_key= "first time"

    local first_time = timer_gateway:get( first_time_key) == nil
    timer_gateway:set( first_time_key, true)

    --
    -- Timer lane; initialize only on the first 'require "lanes"' instance (which naturally
    -- has 'table' always declared)
    --
    if first_time then

        local now_secs = core.now_secs
        assert( type( now_secs) == "function")
        -----
        -- Snore loop (run as a lane on the background)
        --
        -- High priority, to get trustworthy timings.
        --
        -- We let the timer lane be a "free running" thread; no handle to it
        -- remains.
        --
        local timer_body = function()
            set_debug_threadname( "LanesTimer")
            --
            -- { [deep_linda_lightuserdata]= { [deep_linda_lightuserdata]=linda_h,
            --                                 [key]= { wakeup_secs [,period_secs] } [, ...] },
            -- }
            --
            -- Collection of all running timers, indexed with linda's & key.
            --
            -- Note that we need to use the deep lightuserdata identifiers, instead
            -- of 'linda_h' themselves as table indices. Otherwise, we'd get multiple
            -- entries for the same timer.
            --
            -- The 'hidden' reference to Linda proxy is used in 'check_timers()' but
            -- also important to keep the Linda alive, even if all outside world threw
            -- away pointers to it (which would ruin uniqueness of the deep pointer).
            -- Now we're safe.
            --
            local collection = {}
            local table_insert = assert( table.insert)

            local get_timers = function()
                local r = {}
                for deep, t in pairs( collection) do
                    -- WR( tostring( deep))
                    local l = t[deep]
                    for key, timer_data in pairs( t) do
                        if key ~= deep then
                            table_insert( r, {l, key, timer_data})
                        end
                    end
                end
                return r
            end -- get_timers()

            --
            -- set_timer( linda_h, key [,wakeup_at_secs [,period_secs]] )
            --
            local set_timer = function( linda, key, wakeup_at, period)
                assert( wakeup_at == nil or wakeup_at > 0.0)
                assert( period == nil or period > 0.0)

                local linda_deep = linda:deep()
                assert( linda_deep)

                -- Find or make a lookup for this timer
                --
                local t1 = collection[linda_deep]
                if not t1 then
                    t1 = { [linda_deep] = linda}     -- proxy to use the Linda
                    collection[linda_deep] = t1
                end

                if wakeup_at == nil then
                    -- Clear the timer
                    --
                    t1[key]= nil

                    -- Remove empty tables from collection; speeds timer checks and
                    -- lets our 'safety reference' proxy be gc:ed as well.
                    --
                    local empty = true
                    for k, _ in pairs( t1) do
                        if k ~= linda_deep then
                            empty = false
                            break
                        end
                    end
                    if empty then
                        collection[linda_deep] = nil
                    end

                    -- Note: any unread timer value is left at 'linda[key]' intensionally;
                    --       clearing a timer just stops it.
                else
                    -- New timer or changing the timings
                    --
                    local t2 = t1[key]
                    if not t2 then
                        t2= {}
                        t1[key]= t2
                    end

                    t2[1] = wakeup_at
                    t2[2] = period   -- can be 'nil'
                end
            end -- set_timer()

            -----
            -- [next_wakeup_at]= check_timers()
            -- Check timers, and wake up the ones expired (if any)
            -- Returns the closest upcoming (remaining) wakeup time (or 'nil' if none).
            local check_timers = function()
                local now = now_secs()
                local next_wakeup

                for linda_deep,t1 in pairs(collection) do
                    for key,t2 in pairs(t1) do
                        --
                        if key==linda_deep then
                            -- no 'continue' in Lua :/
                        else
                            -- 't2': { wakeup_at_secs [,period_secs] }
                            --
                            local wakeup_at= t2[1]
                            local period= t2[2]     -- may be 'nil'

                            if wakeup_at <= now then
                                local linda= t1[linda_deep]
                                assert(linda)

                                linda:set( key, now )

                                -- 'pairs()' allows the values to be modified (and even
                                -- removed) as far as keys are not touched

                                if not period then
                                    -- one-time timer; gone
                                    --
                                    t1[key]= nil
                                    wakeup_at= nil   -- no 'continue' in Lua :/
                                else
                                    -- repeating timer; find next wakeup (may jump multiple repeats)
                                    --
                                    repeat
                                            wakeup_at= wakeup_at+period
                                    until wakeup_at > now

                                    t2[1]= wakeup_at
                                end
                            end

                            if wakeup_at and ((not next_wakeup) or (wakeup_at < next_wakeup)) then
                                next_wakeup= wakeup_at
                            end
                        end
                    end -- t2 loop
                end -- t1 loop

                return next_wakeup  -- may be 'nil'
            end -- check_timers()

            local timer_gateway_batched = timer_gateway.batched
            set_finalizer( function( err, stk)
                if err and type( err) ~= "userdata" then
                    WR( "LanesTimer error: "..tostring(err))
                --elseif type( err) == "userdata" then
                --	WR( "LanesTimer after cancel" )
                --else
                --	WR("LanesTimer finalized")
                end
            end)
            while true do
                local next_wakeup = check_timers()

                -- Sleep until next timer to wake up, or a set/clear command
                --
                local secs
                if next_wakeup then
                    secs =  next_wakeup - now_secs()
                    if secs < 0 then secs = 0 end
                end
                local key, what = timer_gateway:receive( secs, TGW_KEY, TGW_QUERY)

                if key == TGW_KEY then
                    assert( getmetatable( what) == "Linda") -- 'what' should be a linda on which the client sets a timer
                    local _, key, wakeup_at, period = timer_gateway:receive( 0, timer_gateway_batched, TGW_KEY, 3)
                    assert( key)
                    set_timer( what, key, wakeup_at, period and period > 0 and period or nil)
                elseif key == TGW_QUERY then
                    if what == "get_timers" then
                        timer_gateway:send( TGW_REPLY, get_timers())
                    else
                        timer_gateway:send( TGW_REPLY, "unknown query " .. what)
                    end
                --elseif secs == nil then -- got no value while block-waiting?
                --	WR( "timer lane: no linda, aborted?")
                end
            end
        end -- timer_body()
        timer_lane = gen( "*", { package= {}, priority = max_prio}, timer_body)() -- "*" instead of "io,package" for LuaJIT compatibility...
    end -- first_time

    -----
    -- = timer( linda_h, key_val, date_tbl|first_secs [,period_secs] )
    --
    -- PUBLIC LANES API
    timer = function( linda, key, a, period )
        if getmetatable( linda) ~= "Linda" then
            error "expecting a Linda"
        end
        if a == 0.0 then
            -- Caller expects to get current time stamp in Linda, on return
            -- (like the timer had expired instantly); it would be good to set this
            -- as late as possible (to give most current time) but also we want it
            -- to precede any possible timers that might start striking.
            --
            linda:set( key, core.now_secs())

            if not period or period==0.0 then
                timer_gateway:send( TGW_KEY, linda, key, nil, nil )   -- clear the timer
                return  -- nothing more to do
            end
            a= period
        end

        local wakeup_at= type(a)=="table" and core.wakeup_conv(a)    -- given point of time
                                           or (a and core.now_secs()+a or nil)
        -- queue to timer
        --
        timer_gateway:send( TGW_KEY, linda, key, wakeup_at, period )
    end

    -----
    -- {[{linda, slot, when, period}[,...]]} = timers()
    --
    -- PUBLIC LANES API
    timers = function()
        timer_gateway:send( TGW_QUERY, "get_timers")
        local _, r = timer_gateway:receive( TGW_REPLY)
        return r
    end

    end -- settings.with_timers

    -- avoid pulling the whole core module as upvalue when cancel_error is enough
    local cancel_error = assert( core.cancel_error)

    ---=== Lock & atomic generators ===---

    -- These functions are just surface sugar, but make solutions easier to read.
    -- Not many applications should even need explicit locks or atomic counters.

    --
    -- [true [, ...]= trues(uint)
    --
    local function trues( n)
        if n > 0 then
            return true, trues( n - 1)
        end
    end

    --
    -- lock_f = lanes.genlock( linda_h, key [,N_uint=1] )
    --
    -- = lock_f( +M )   -- acquire M
    --      ...locked...
    -- = lock_f( -M )   -- release M
    --
    -- Returns an access function that allows 'N' simultaneous entries between
    -- acquire (+M) and release (-M). For binary locks, use M==1.
    --
    -- PUBLIC LANES API
    local genlock = function( linda, key, N)
        -- clear existing data and set the limit
        N = N or 1
        if linda:set( key) == cancel_error or linda:limit( key, N) == cancel_error then
            return cancel_error
        end

        -- use an optimized version for case N == 1
        return (N == 1) and
        function( M, mode_)
            local timeout = (mode_ == "try") and 0 or nil
            if M > 0 then
                -- 'nil' timeout allows 'key' to be numeric
                return linda:send( timeout, key, true)    -- suspends until been able to push them
            else
                local k = linda:receive( nil, key)
                -- propagate cancel_error if we got it, else return true or false
                return k and ((k ~= cancel_error) and true or k) or false
            end
        end
        or
        function( M, mode_)
            local timeout = (mode_ == "try") and 0 or nil
            if M > 0 then
                -- 'nil' timeout allows 'key' to be numeric
                return linda:send( timeout, key, trues(M))    -- suspends until been able to push them
            else
                local k = linda:receive( nil, linda.batched, key, -M)
                -- propagate cancel_error if we got it, else return true or false
                return k and ((k ~= cancel_error) and true or k) or false
            end
        end
    end


    --
    -- atomic_f = lanes.genatomic( linda_h, key [,initial_num=0.0])
    --
    -- int|cancel_error = atomic_f( [diff_num = 1.0])
    --
    -- Returns an access function that allows atomic increment/decrement of the
    -- number in 'key'.
    --
    -- PUBLIC LANES API
    local genatomic = function( linda, key, initial_val)
        -- clears existing data (also queue). the slot may contain the stored value, and an additional boolean value
        if linda:limit( key, 2) == cancel_error or linda:set( key, initial_val or 0.0) == cancel_error then
            return cancel_error
        end

        return function( diff)
            -- 'nil' allows 'key' to be numeric
            -- suspends until our 'true' is in
            if linda:send( nil, key, true) == cancel_error then
                return cancel_error
            end
            local val = linda:get( key)
            if val ~= cancel_error then
                val = val + (diff or 1.0)
                -- set() releases the lock by emptying queue
                if linda:set( key, val) == cancel_error then
                    val = cancel_error
                end
            end
            return val
        end
    end

    -- activate full interface
    lanes.require = core.require
    lanes.register = core.register
    lanes.gen = gen
    lanes.linda = core.linda
    lanes.cancel_error = core.cancel_error
    lanes.nameof = core.nameof
    lanes.set_singlethreaded = core.set_singlethreaded
    lanes.threads = core.threads or function() error "lane tracking is not available" end -- core.threads isn't registered if settings.track_lanes is false
    lanes.set_thread_priority = core.set_thread_priority
    lanes.set_thread_affinity = core.set_thread_affinity
    lanes.timer = timer
    lanes.timer_lane = timer_lane
    lanes.timers = timers
    lanes.sleep = sleep
    lanes.genlock = genlock
    lanes.now_secs = core.now_secs
    lanes.genatomic = genatomic
    lanes.configure = nil -- no need to call configure() ever again
    return lanes
end -- lanes.configure

lanesMeta.__index = function( t, k)
    -- This is called when some functionality is accessed without calling configure()
    lanes.configure() -- initialize with default settings
    -- Access the required key
    return lanes[k]
end

-- no need to force calling configure() manually excepted the first time (other times will reuse the internally stored settings of the first call)
if core.settings then
    return lanes.configure()
else
    return lanes
end

--the end

end,

["insurrection.api"] = function()
--------------------
-- Module: 'insurrection.api'
--------------------
local engine = Engine
local lanes = require "lanes"
local json = require "json"
local getState = require "insurrection.redux.getState"
local react = require "insurrection.react"
local blam = require "blam"
local requests = require "requestscurl"
local interface = require "insurrection.interface"
local glue = require "glue"
local exists = glue.canopen
local actions = require "insurrection.redux.actions"
local core = require "insurrection.core"
local menus = require "insurrection.menus"
local shared = interface.shared
local constants = require "insurrection.constants"
local loading = core.loading
local luna = require "luna"
local utils = require "insurrection.utils"

local api = {}

api.variables = {refreshRate = 3000, refreshTimer = nil}

---@class insurrectionSession
---@field token? string
---@field lobbyKey? string
---@field username? string
---@field player? insurrectionPlayer

---@type insurrectionSession
api.session = {token = nil, lobbyKey = nil, username = nil, player = nil}

-- Models

---@class requestResult
---@field message string

---@class insurrectionPlayer
---@field name string
---@field nameplate number
---@field publicId string
---@field rank number
---@field bipeds table<string, string>
---@field color {primary: string, secondary: string}

---@class loginResponse
---@field message string
---@field token? string
---@field player insurrectionPlayer
---@field secondsToExpire number

---@class availableParameters
---@field maps string[]
---@field gametypes string[]
---@field templates string[]
---@field customization table<string, {maps: string[], tags: string[]}>

---@class serverInstance
---@field password string
---@field template string
---@field host string
---@field port integer
---@field map string
---@field gametype string
---@field cpu integer
---@field owner string
---@field lobbyKey string

---@class insurrectionLobby
---@field owner string
---@field players insurrectionPlayer[]
---@field map string
---@field gametype string
---@field template string
---@field server serverInstance
---@field isPublic boolean

-- function async(func, callback, ...)
--    if #Lanes == 0 then
--        Lanes[#Lanes + 1] = {thread = lanes.gen(asyncLibs, func)(...), callback = callback}
--    else
--        log("Warning! An async function is trying to add another thread!")
--    end
-- end

-- @param inputFunction fun(await: fun(callback: fun(...): (T), ...): T)

local function connect(desiredMap, host, port, password)
    api.stopRefreshLobby()
    api.session.lobbyKey = nil
    if not engine.map.getCurrentMapHeader().name == "ui" then
        engine.core.consolePrint("Can't connect to a server while in-game.")
        return
    end
    logger:debug("Connecting to " .. tostring(host) .. ":" .. tostring(port) .. " with password " ..
                     tostring(password))
    local mapList = engine.map.getMapList()
    if table.indexof(mapList, desiredMap) then
        -- Force game profile name to be the same as the player's name
        core.setGameProfileName(api.session.player.name)
        core.connectServer(host, port, password)
    else
        interface.dialog("ERROR", "LOCAL MAP NOT FOUND", "Map \"" .. desiredMap ..
                             "\" was not found on your game files.\nPlease download it from Mercury or map repositories.")
    end
end

local function showErrorDialog(logs)
    if type(logs) ~= "string" then
        if type(logs) == "table" and logs.json then
            logs = tostring(inspect(logs)) .. "\n" .. tostring(inspect(logs.json()))
        else
            logs = tostring(inspect(logs))
            logger:error("Unknown error: " .. logs, "error")
        end
    end
    interface.dialog("ERROR", "UNKNOWN ERROR",
                     "An unknown error has ocurred, please check logs and try again later.")
    if logs then
        local log = Balltze.filesystem.readFile("insurrection.log") or ""
        -- Check if log is over 100,000 kilobytes and reset it
        if #log > 100000 then
            log = ""
        end
        log = log .. "\n" .. debug.traceback()
        log = log .. "\n" .. logs
        Balltze.filesystem.writeFile("insurrection.log", log)
    end
end

---Prevent players from getting stuck in the joining screen if the server connection fails
local function preventStuckLobby()
    -- Check after 30 seconds if we are still in the UI map
    -- TODO BALLTZE MIGRATE
    -- utils.delay(30000, function()
    --    if map == "ui" then
    --        -- If we were trying to join a lobby and the widget is closed, delete the lobby
    --        if api.session.lobbyKey then
    --            console_out(
    --                "It seems like there is a problem joining the server, deleting lobby...",
    --                table.unpack(blam.consoleColors.warning))
    --            api.deleteLobby()
    --        end
    --    end
    -- end)
end

---Load Insurrection URL to be used by the API
---@param host? string
function api.loadUrl(host)
    api.version = "v1"
    api.host = Balltze.filesystem.readFile("insurrection_host") or host or "http://localhost:4343/"
    if APILocalMode then
        api.host = "http://localhost:4343/"
    end
    api.url = api.host .. api.version
end

function api.login(username, password)
    local login = async(function(await)
        loading(true, "Logging in...")
        ---@type httpResponse<loginResponse>?
        local response = await(requests.postform, api.url .. "/login",
                               {username = username, password = password})
        log("onLoginResponse")
        loading(false)
        if not response then
            logger:error("No response")
            showErrorDialog("No response")
            return
        end
        if response.code == 200 then
            local jsonResponse = response.json()
            api.session.token = jsonResponse.token
            api.session.player = jsonResponse.player

            if api.session.player.color then
                local primaryColorIndex = core.getCustomizationColorByValue(api.session.player.color
                                                                                .primary) or
                                              LastColorCustomization.primary
                local secondaryColorIndex = core.getCustomizationColorByValue(api.session.player
                                                                                  .color.secondary) or
                                                LastColorCustomization.secondary
                LastColorCustomization.primary = primaryColorIndex
                LastColorCustomization.secondary = secondaryColorIndex
            end

            requests.headers = {"Authorization: Bearer " .. api.session.token}
            -- Save last defined nameplate
            core.updateSettings({nameplate = jsonResponse.player.nameplate})
            interface.loadProfileNameplate()
            api.available()
            menus.dashboard()
        elseif response.code == 401 then
            local jsonResponse = response.json()
            interface.dialog("ATTENTION", "ERROR " .. response.code, jsonResponse.message)
        end
    end)
    login()
end

function api.available()
    loading(true, "Loading available parameters...")
    logger:debug("Loading available parameters...")
    if IsDebugLocalCustomizationEnabled then
        local customization = {}
        for mapName, data in pairs(CustomBipedPaths) do
            customization[mapName] = {
                maps = {mapName},
                tags = table.map(data, function(tagPath)
                    return tagPath .. ".biped"
                end)
            }
        end
        store:dispatch(actions.setAvailableResources({
            maps = {"bloodgulch", "beavercreek", "dangercanyon", "deathisland", "icefields"},
            gametypes = {"slayer", "ctf", "oddball"},
            templates = {"default"},
            customization = customization
        }))
        return
    end
    local available = async(function(await)
        ---@type httpResponse<availableParameters>?
        local response = await(requests.get, api.url .. "/available")
        if not response then
            showErrorDialog("No response")
            return
        end
        loading(false)
        if response then
            if response.code == 200 then
                local jsonResponse = response.json()
                store:dispatch(actions.setAvailableResources(jsonResponse))
                return
            end
        end
        showErrorDialog(response)
    end)
    available()
end

---Create or join a lobby
---@param lobbyKey? string
function api.lobby(lobbyKey)
    if not lobbyKey then
        log("Creating lobby")
        local lobby = async(function(await)
            loading(true, "Creating lobby...")
            ---@type httpResponse<lobbyResponse>?
            local response = await(requests.get, api.url .. "/lobby")
            if not response then
                showErrorDialog("No response")
                return
            end
            if response.code == requests.codes.ok then
                ---@class lobbyResponse
                ---@field key string
                ---@field lobby insurrectionLobby

                local data = response.json()
                if data then
                    -- We asked for a new lobby room
                    if data.key then
                        api.session.lobbyKey = data.key
                        store:dispatch(actions.setLobby(data.key, data.lobby))
                    else
                        -- We have to joined an existing lobby
                        local lobby = data --[[@as insurrectionLobby]]
                        store:dispatch(actions.setLobby(nil, lobby))
                        -- There is a server already running for this lobby, connect to it
                        if lobby.server then
                            local isPlayerLobbyOwner = api.session.player and
                                                           api.session.player.publicId ==
                                                           lobby.owner
                            if isPlayerLobbyOwner then
                                discord.setParty(lobby.server.lobbyKey, #lobby.players, 16,
                                                 lobby.map, isPlayerLobbyOwner)
                            end
                            connect(lobby.server.map, lobby.server.host, lobby.server.port,
                                    lobby.server.password)
                            return
                        end
                    end
                    api.startLobbyRefresh()
                    local state = getState()
                    local isPlayerLobbyOwner = api.session.player and api.session.player.publicId ==
                                                   state.lobby.owner
                    if isPlayerLobbyOwner then
                        menus.lobby()
                        react.render("lobbyMenu")
                    else
                        menus.lobby(true)
                        react.render("lobbyMenuClient")
                    end
                    discord.setParty(api.session.lobbyKey, #state.lobby.players, 16,
                                     state.lobby.map, isPlayerLobbyOwner)
                end
                return
            elseif response.code == requests.codes.forbidden then
                local jsonResponse = response.json()
                if jsonResponse and jsonResponse.key then
                    api.lobby(jsonResponse.key)
                end
                return
            else
                api.session.lobbyKey = nil
                local jsonResponse = response.json()
                interface.dialog("ATTENTION", "ERROR " .. response.code, jsonResponse.message)
                return
            end
        end)
        lobby()
        return
    end

    log("Joining lobby with key: " .. lobbyKey)
    api.session.lobbyKey = lobbyKey
    async(function(await)
        ---@type httpResponse<insurrectionLobby>?
        local response = await(requests.get, api.url .. "/lobby/" .. lobbyKey)
        if not response then
            showErrorDialog("No response")
            return
        end
        if response.code == requests.codes.ok then
            local lobby = response.json()
            if lobby then
                store:dispatch(actions.setLobby(lobbyKey, lobby))
                local state = getState()
                local isPlayerLobbyOwner = api.session.player and api.session.player.publicId ==
                                               state.lobby.owner
                if isPlayerLobbyOwner then
                    menus.lobby()
                    react.render("lobbyMenu")
                else
                    menus.lobby(true)
                    react.render("lobbyMenuClient")
                end
                discord.setParty(api.session.lobbyKey, #lobby.players, 16, lobby.map,
                                 isPlayerLobbyOwner)
                api.startLobbyRefresh()
                -- Lobby already has a server running, connect to it
                if lobby.server and engine.netgame.getServerType() ~= "dedicated" then
                    log("Connecting to lobby server...")
                    api.stopRefreshLobby()
                    connect(lobby.server.map, lobby.server.host, lobby.server.port,
                            lobby.server.password)
                    preventStuckLobby()
                end
            end
        end
    end)()
end

function api.startLobbyRefresh()
    -- Start a timer to pull lobby data every certain time
    if api.variables.refreshTimer then
        api.stopRefreshLobby()
    end
    -- Create global function to be called by the timer
    RefreshLobby = function()
        if api.session.lobbyKey then
            api.refreshLobby()
        end
    end
    api.variables.refreshTimer = Balltze.misc.setTimer(api.variables.refreshRate, RefreshLobby)
end

function api.refreshLobby()
    if not api.session.lobbyKey then
        return
    end
    logger:debug("Refreshing lobby data...")
    loading(true, "Refreshing lobby...", false)
    local refresh = async(function(await)
        ---@type httpResponse<insurrectionLobby | requestResult>?
        local response = await(requests.get, api.url .. "/lobby/" .. api.session.lobbyKey)
        loading(false)
        if not response then
            api.stopRefreshLobby()
            showErrorDialog("No response")
            return
        end
        if response.code == 200 then
            local lobby = response.json()
            if lobby then
                -- Update previously joined lobby data
                store:dispatch(actions.setLobby(api.session.lobbyKey, lobby))
                local state = getState()
                local isPlayerLobbyOwner = api.session.player and api.session.player.publicId ==
                                               state.lobby.owner
                if isPlayerLobbyOwner then
                    react.render("lobbyMenu")
                else
                    react.render("lobbyMenuClient")
                end
                discord.setParty(api.session.lobbyKey, #lobby.players, 16, lobby.map,
                                 isPlayerLobbyOwner)
                -- Lobby already has a server running, connect to it
                if lobby.server and engine.netgame.getServerType() ~= "dedicated" then
                    api.stopRefreshLobby()
                    connect(lobby.server.map, lobby.server.host, lobby.server.port,
                            lobby.server.password)
                    preventStuckLobby()
                end
            end
            return
        else
            api.stopRefreshLobby()
            -- TODO Add a generic error handling function for this
            local jsonResponse = response.json()
            interface.dialog("ERROR", "ERROR " .. response.code, jsonResponse.message)
            return
        end
    end)
    refresh()
end
function api.stopRefreshLobby()
    if api.session.lobbyKey and api.variables.refreshTimer then
        api.variables.refreshTimer.stop()
        api.variables.refreshTimer = nil
    end
end
function api.deleteLobby()
    if api.session.lobbyKey then
        log("DELETING lobby")
        if api.variables.refreshTimer then
            api.variables.refreshTimer.stop()
        end
        api.variables.refreshTimer = nil
        api.session.lobbyKey = nil
        store:dispatch(actions.setLobby(nil, nil))
    end
end

function api.borrow(template, map, gametype)
    local borrow = async(function(await)
        loading(true, "Borrowing game server...", false)
        ---@type httpResponse<serverBorrowResponse>?
        local response = await(requests.get,
                               api.url .. "/borrow/" .. template .. "/" .. map .. "/" .. gametype ..
                                   "/" .. api.session.lobbyKey)
        if not response then
            showErrorDialog("No response")
            return
        end
        if response.code == 200 then
            -- Prevent lobby from refreshing while we are waiting for the server to start
            -- This is critical to avoid crashing the game due to multitasking stuff
            api.stopRefreshLobby()
            ---@class serverBorrowResponse
            ---@field password string
            ---@field message string
            ---@field port number
            ---@field host string
            ---@field map string

            local jsonResponse = response.json()
            if jsonResponse then
                connect(jsonResponse.map, jsonResponse.host, jsonResponse.port,
                        jsonResponse.password)
                preventStuckLobby()
            end
            return
        elseif response.code == 404 then
            local jsonResponse = response.json()
            interface.dialog("ATTENTION", "ERROR " .. response.code, jsonResponse.message)
            return
        else
            api.stopRefreshLobby()
            if response.code == 500 then
                interface.dialog("ATTENTION", "ERROR " .. response.code, "Internal Server Error")
                return
            else
                local jsonResponse = response.json()
                interface.dialog("ATTENTION", "ERROR " .. response.code, jsonResponse.message)
                return
            end
        end
    end)
    borrow()
end

---Edit player nameplate
---@param data {nameplate: string, bipeds: table<string, string>}
function api.playerProfileEdit(data)
    loading(true, "Editing profile...", false)
    local edit = async(function(await)
        ---@type httpResponse<any>?
        local response = await(requests.patch, api.url .. "/players", data)
        if not response then
            showErrorDialog("No response")
            return
        end
        loading(false)

        if response.code == 200 then
            local bipedProjectName = table.keys(data.bipeds)[1]
            interface.dialog("INFORMATION", "CONGRATULATIONS", "Profile customized successfully.")
            interface.loadProfileNameplate(data.nameplate)
            core.updateSettings({nameplate = data.nameplate, lastSavedProject = bipedProjectName})
            -- Refresh player data
            api.session.player.bipeds = table.merge(api.session.player.bipeds, data.bipeds)
            return
        else
            local jsonResponse = response.json()
            if jsonResponse then
                interface.dialog("WARNING", "ERROR " .. response.code, jsonResponse.message)
            end
            return
        end
    end)
    edit()
end

function api.editLobby(lobbyKey, data)
    loading(true, "Editing lobby...", false)
    local edit = async(function(await)
        ---@type httpResponse<any>?
        local response = await(requests.patch, api.url .. "/lobby/" .. lobbyKey, data)
        if not response then
            showErrorDialog("No response")
            return
        end
        if response.code == 200 then
        else
            local jsonResponse = response.json()
            if jsonResponse then
                interface.dialog("ATTENTION", "ERROR " .. response.code, jsonResponse.message)
            end
        end
    end)
    edit()
end

function api.getLobbies()
    loading(true, "Loading lobbies...", false)
    local get = async(function(await)
        ---@type httpResponse<insurrectionLobby[] | requestResult>?
        local response = await(requests.get, api.url .. "/lobbies")
        if not response then
            showErrorDialog("No lobbies server response")
            return
        end
        if response.code == 200 then
            local jsonResponse = response.json()
            if jsonResponse then
                if #jsonResponse == 0 then
                    interface.dialog("ERROR", "NO LOBBIES FOUND",
                                     "There are no lobbies available at the moment.")
                    return
                else
                    store:dispatch(actions.setLobbies(jsonResponse))
                    react.render("lobbyBrowserMenu")
                end
                return
            end
        else
            local jsonResponse = response.json()
            if jsonResponse then
                interface.dialog("ATTENTION", "ERROR " .. response.code, jsonResponse.message)
            end
        end
    end)
    get()
end

function api.getSavedBipeds()
    if api.session and api.session.player and api.session.player.bipeds then
        return table.map(api.session.player.bipeds, function(data)
            local elements = data:split("+")
            return {
                path = elements[1],
                regions = table.map(table.slice(elements, 2, 9), tointeger),
                visor = tointeger(elements[10] or "0")
            }
        end)
    end
end

return api

end,

["insurrection.react"] = function()
--------------------
-- Module: 'insurrection.react'
--------------------
local react = {}

local mounted = {}

function react.mount(component)
    log("Mounting component " .. component)
    if not mounted[component] then
        mounted[component] = require("insurrection.components.dynamic." .. component)()
    end
    if not mounted[component] then
        error("Failed to mount component " .. component)
    end
    return mounted[component]
end

function react.render(component)
    log("Rendering component " .. component)
    local render = react.mount(component)
    return render()
end

function react.unmountAll()
    for k, v in pairs(mounted) do
        log("Unmounting component " .. k)
        mounted[k] = nil
    end
end

return react
end,

["requestscurl"] = function()
--------------------
-- Module: 'requestscurl'
--------------------
local json = require "json"
local requests = {}

requests.codes = {
    ok = 200,
    created = 201,
    accepted = 202,
    no_content = 204,
    moved_permanently = 301,
    found = 302,
    see_other = 303,
    not_modified = 304,
    temporary_redirect = 307,
    bad_request = 400,
    unauthorized = 401,
    forbidden = 403,
    not_found = 404,
    method_not_allowed = 405,
    not_acceptable = 406,
    request_timeout = 408,
    conflict = 409,
    gone = 410,
    length_required = 411,
    precondition_failed = 412,
    payload_too_large = 413,
    uri_too_long = 414,
    unsupported_media_type = 415,
    range_not_satisfiable = 416,
    expectation_failed = 417,
    im_a_teapot = 418,
    unprocessable_entity = 422,
    too_many_requests = 429,
    internal_server_error = 500,
    not_implemented = 501,
    bad_gateway = 502,
    service_unavailable = 503,
    gateway_timeout = 504
}

requests.headers = {}
requests.timeout = 300

---@class httpResponse<T> : { url: string, code: number, text: string, headers: table<string, string>, json : fun(): T }

local function paramsToBodyString(params)
    local body = ""
    for k, v in pairs(params) do
        body = body .. k .. "=" .. v .. "&"
    end
    -- Remove last "&"
    body = body:sub(1, #body - 1)
    return body
end

-- Probably an alternative due to performance
-- https://gist.github.com/liukun/f9ce7d6d14fa45fe9b924a3eed5c3d99
local function urlencode(str)
    -- str = string.gsub(str, "([^0-9a-zA-Z !'()*._~-])", -- locale independent
    str = string.gsub(str, "([ '])", -- locale independent
    function(c)
        return string.format("%%%02X", string.byte(c))
    end)
    str = string.gsub(str, " ", "+")
    return str
end

---Perform a GET request
---@param url string
---@return httpResponse?
function requests.get(url)
    local curl = require "lcurl.safe"

    local requestHeaders = {}
    for k, v in pairs(requests.headers) do
        requestHeaders[#requestHeaders + 1] = v
    end
    requestHeaders[#requestHeaders + 1] = "Content-Type: application/x-www-form-urlencoded"

    local responseBody = {}
    local responseHeaders = {}

    local request = curl.easy {
        url = urlencode(url),
        timeout = requests.timeout,
        httpheader = requests.headers,
        writefunction = function(input)
            table.insert(responseBody, input)
        end,
        headerfunction = function(header)
            local key = header:match("^([^:]+):")
            if key then
                responseHeaders[key] = header:sub(#key + 2, #header - 2)
            end
        end
    }:perform()

    if request then
        local url = request:getinfo_effective_url()
        local code = request:getinfo_response_code()
        local responseBody = table.concat(responseBody)
        request:close()
        return {
            url = url,
            code = code,
            text = responseBody,
            headers = responseHeaders,
            json = function()
                return json.decode(responseBody)
            end
        }
    end
end

---Perform a POST request
---@param url string
---@param params? table<string, string | number>
---@return httpResponse?
function requests.postform(url, params)
    -- TODO Implement different types of POST
    local curl = require "lcurl.safe"

    local requestHeaders = {}
    for k, v in pairs(requests.headers) do
        requestHeaders[#requestHeaders + 1] = v
    end
    requestHeaders[#requestHeaders + 1] = "Content-Type: application/x-www-form-urlencoded"

    local responseBody = {}
    local responseHeaders = {}

    local request = curl.easy {
        url = urlencode(url),
        timeout = requests.timeout,
        httpheader = headers,
        writefunction = function(input)
            table.insert(responseBody, input)
        end,
        headerfunction = function(header)
            local key = header:match("^([^:]+):")
            if key then
                responseHeaders[key] = header:sub(#key + 2, #header - 2)
            end
        end,
        postfields = paramsToBodyString(params)
    }:perform()
    if request then
        local url = request:getinfo_effective_url()
        local code = request:getinfo_response_code()
        local responseBody = table.concat(responseBody)
        request:close()
        return {
            url = url,
            code = code,
            text = responseBody,
            headers = responseHeaders,
            json = function()
                return json.decode(responseBody)
            end
        }
    end
end

---Perform a PATCH request
---@param url string
---@param params? table<string, string | number>
---@return httpResponse?
function requests.patch(url, params)
    local curl = require "lcurl.safe"
    
    local requestHeaders = {}
    for k, v in pairs(requests.headers) do
        requestHeaders[#requestHeaders + 1] = v
    end
    requestHeaders[#requestHeaders + 1] = "Content-Type: application/json"

    local responseBody = {}
    local responseHeaders = {}

    local request = curl.easy {
        url = urlencode(url),
        timeout = requests.timeout,
        httpheader = requestHeaders,
        writefunction = function(input)
            table.insert(responseBody, input)
        end,
        headerfunction = function(header)
            local key = header:match("^([^:]+):")
            if key then
                responseHeaders[key] = header:sub(#key + 2, #header - 2)
            end
        end,
        postfields = json.encode(params),
        customrequest = "PATCH"
    }:perform()

    if request then
        local url = request:getinfo_effective_url()
        local code = request:getinfo_response_code()
        local responseBody = table.concat(responseBody)
        request:close()
        return {
            url = url,
            code = code,
            text = responseBody,
            headers = responseHeaders,
            json = function()
                return json.decode(responseBody)
            end
        }
    end
end

return requests

end,

["insurrection.discord"] = function()
--------------------
-- Module: 'insurrection.discord'
--------------------
local base64 = require "base64"
local discordRPC = require "discordRPC"
local utils = require "insurrection.utils"

local core = require "insurrection.core"
local interface = require "insurrection.interface"
local getState = require "insurrection.redux.getState"

local discord = {}

---@type BalltzeTimer?
DiscordPresenceTimer = nil
---@type BalltzeTimer?
DiscordCheckTimer = nil

discord.presence = {
    state = "",
    details = "",
    largeImageKey = "insurrection",
    largeImageText = "Insurrection",
    startTimestamp = os.time(os.date("*t") --[[@as osdateparam]] ),
    -- startTimestamp = os.date("*t", os.time())
    smallImageKey = "insurrection",
    smallImageText = "Insurrection"
    -- instance = 0
}

discord.attempted = false
discord.ready = false

function discordRPC.ready(userId, username, discriminator, avatar)
    log(string.format("Discord: ready (%s, %s, %s, %s)", userId, username, discriminator, avatar))
    core.loading(false)
    discord.ready = true
end

function discordRPC.disconnected(errorCode, message)
    log(string.format("Discord: disconnected (%d: %s)", errorCode, message))
    discord.ready = false
end

function discordRPC.joinGame(joinSecret)
    log("Discord: join game (" .. joinSecret .. ")")
    api.lobby(joinSecret)
end

function discordRPC.joinRequest(userId, username, discriminator, avatar)
    log(string.format("Discord: join request (%s, %s, %s, %s)", userId, username, discriminator,
                         avatar))
    discordRPC.respond(userId, "yes")
end

function discord.initialize()
    discordRPC.initialize(base64.decode(Balltze.filesystem.readFile("micro")), true)
end

function discord.startPresence()
    if not discord.ready then
        -- discordRPC.shutdown()
        core.loading(true, "Starting Discord Presence...")
        discord.initialize()
    end

    if DiscordPresenceTimer then
        DiscordPresenceTimer.stop()
        DiscordPresenceTimer = nil
    end

    -- Routines to handle Discord presence
    function DiscordUpdate()
        --log("DiscordUpdate")
        discordRPC.runCallbacks()
        if discord.ready then
            if DiscordCheckTimer then
                DiscordCheckTimer.stop()
                DiscordCheckTimer = nil
            end
            discordRPC.updatePresence(discord.presence)
        end
    end
    function DiscordCheck()
        if not discord.ready then
            core.loading(false)
            if not DebugMode and not discord.attempted then
                --interface.dialog("WARNING", "Discord Rich Presence failed to start.",
                --                 "Please make sure you have Discord running and allow Discord to display your game status.")
                discord.attempted = true
            end
        end
        return false
    end
    DiscordPresenceTimer = Balltze.misc.setTimer(120, DiscordUpdate)
    DiscordCheckTimer = Balltze.misc.setTimer(5000, DiscordCheck)
end

---Set presence state and details
---@param state string
---@param details? string
---@param image? string
function discord.setState(state, details, image)
    log("discord.setState: " .. state .. ", " .. details)
    if not discord.ready then
        return
    end
    discord.presence.state = state .. "\n"
    discord.presence.details = details
    if not image then
        discord.presence.largeImageKey = "insurrection"
        discord.presence.largeImageText = "Insurrection"
    end
end

---Set presence party info
---@param partyId string?
---@param partySize number?
---@param partyMax number?
---@param map? string
---@param isLobbyOpen? boolean
function discord.setParty(partyId, partySize, partyMax, map, isLobbyOpen)
    -- Party ID sometines is nil, so we need to check for it
    log(
        "discord.setParty: " .. partyId .. ", " .. partySize .. ", " .. partyMax .. ", " .. map ..
            ", " .. tostring(isLobbyOpen))
    if partyId then
        -- TODO Replace with a proper party unique ID
        discord.presence.partyId = partyId .. partyId:reverse()
        discord.presence.joinSecret = nil
        if isLobbyOpen then
            discord.presence.joinSecret = partyId
        end
    end
    discord.presence.partySize = partySize
    discord.presence.partyMax = partyMax
    if map then
        local mapName = utils.snakeCaseToTitleCase(map)
        discord.presence.details = mapName
        discord.presence.largeImageKey = map
        discord.presence.largeImageText = map
    end
end

function discord.clearParty()
    log("discord.clearParty")

    discord.presence.partyId = nil
    discord.presence.joinSecret = nil
    discord.presence.partySize = nil
    discord.presence.partyMax = nil
end

--- Clear the presence info
function discord.clearPresence()
    log("discord.clearPresence")

    discord.presence.state = nil
    discord.presence.details = nil
    discord.presence.partyId = nil
    discord.presence.joinSecret = nil
    discord.presence.partySize = nil
    discord.presence.partyMax = nil
    discord.presence.largeImageKey = nil
    discord.presence.largeImageText = nil
end

function discord.stopPresence()
    if DiscordPresenceTimer then
        pcall(stop_timer, DiscordPresenceTimer)
    end
    discordRPC.clearPresence()
    discordRPC.shutdown()
end

return discord

end,

["discordRPC"] = function()
--------------------
-- Module: 'discordRPC'
--------------------
local ffi = require "cffi"
local discordRPClib = ffi.load "discord-rpc"

function newproxy(new_meta)
    local proxy = {}

    if (new_meta == true) then
        local mt = {}
        setmetatable(proxy, mt)
    elseif (new_meta == false) then
    else
        -- new_meta must have a metatable.
        local mt = getmetatable(new_meta)
        setmetatable(proxy, mt)
    end

    return proxy
end

ffi.cdef [[
typedef struct DiscordRichPresence {
    const char* state;   /* max 128 bytes */
    const char* details; /* max 128 bytes */
    int64_t startTimestamp;
    int64_t endTimestamp;
    const char* largeImageKey;  /* max 32 bytes */
    const char* largeImageText; /* max 128 bytes */
    const char* smallImageKey;  /* max 32 bytes */
    const char* smallImageText; /* max 128 bytes */
    const char* partyId;        /* max 128 bytes */
    int partySize;
    int partyMax;
    const char* matchSecret;    /* max 128 bytes */
    const char* joinSecret;     /* max 128 bytes */
    const char* spectateSecret; /* max 128 bytes */
    int8_t instance;
} DiscordRichPresence;

typedef struct DiscordUser {
    const char* userId;
    const char* username;
    const char* discriminator;
    const char* avatar;
} DiscordUser;

typedef void (*readyPtr)(const DiscordUser* request);
typedef void (*disconnectedPtr)(int errorCode, const char* message);
typedef void (*erroredPtr)(int errorCode, const char* message);
typedef void (*joinGamePtr)(const char* joinSecret);
typedef void (*spectateGamePtr)(const char* spectateSecret);
typedef void (*joinRequestPtr)(const DiscordUser* request);

typedef struct DiscordEventHandlers {
    readyPtr ready;
    disconnectedPtr disconnected;
    erroredPtr errored;
    joinGamePtr joinGame;
    spectateGamePtr spectateGame;
    joinRequestPtr joinRequest;
} DiscordEventHandlers;

void Discord_Initialize(const char* applicationId,
                        DiscordEventHandlers* handlers,
                        int autoRegister,
                        const char* optionalSteamId);

void Discord_Shutdown(void);

void Discord_RunCallbacks(void);

void Discord_UpdatePresence(const DiscordRichPresence* presence);

void Discord_ClearPresence(void);

void Discord_Respond(const char* userid, int reply);

void Discord_UpdateHandlers(DiscordEventHandlers* handlers);
]]

local discordRPC = {} -- module table

-- proxy to detect garbage collection of the module
discordRPC.gcDummy = newproxy(true)

local function unpackDiscordUser(request)
    return ffi.string(request.userId), ffi.string(request.username),
           ffi.string(request.discriminator), ffi.string(request.avatar)
end

-- callback proxies
-- note: callbacks are not JIT compiled (= SLOW), try to avoid doing performance critical tasks in them
-- luajit.org/ext_ffi_semantics.html
local ready_proxy = ffi.cast("readyPtr", function(request)
    if discordRPC.ready then
        discordRPC.ready(unpackDiscordUser(request))
    end
end)

local disconnected_proxy = ffi.cast("disconnectedPtr", function(errorCode, message)
    if discordRPC.disconnected then
        discordRPC.disconnected(errorCode, ffi.string(message))
    end
end)

local errored_proxy = ffi.cast("erroredPtr", function(errorCode, message)
    if discordRPC.errored then
        discordRPC.errored(errorCode, ffi.string(message))
    end
end)

local joinGame_proxy = ffi.cast("joinGamePtr", function(joinSecret)
    if discordRPC.joinGame then
        discordRPC.joinGame(ffi.string(joinSecret))
    end
end)

local spectateGame_proxy = ffi.cast("spectateGamePtr", function(spectateSecret)
    if discordRPC.spectateGame then
        discordRPC.spectateGame(ffi.string(spectateSecret))
    end
end)

local joinRequest_proxy = ffi.cast("joinRequestPtr", function(request)
    if discordRPC.joinRequest then
        discordRPC.joinRequest(unpackDiscordUser(request))
    end
end)

-- helpers
local function checkArg(arg, argType, argName, func, maybeNil)
    assert(type(arg) == argType or (maybeNil and arg == nil), string.format(
               "Argument \"%s\" to function \"%s\" has to be of type \"%s\"", argName, func, argType))
end

local function checkStrArg(arg, maxLen, argName, func, maybeNil)
    if maxLen then
        assert(type(arg) == "string" and arg:len() <= maxLen or (maybeNil and arg == nil),
               string.format(
                   "Argument \"%s\" of function \"%s\" has to be of type string with maximum length %d",
                   argName, func, maxLen))
    else
        checkArg(arg, "string", argName, func, true)
    end
end

local function checkIntArg(arg, maxBits, argName, func, maybeNil)
    maxBits = math.min(maxBits or 32, 52) -- lua number (double) can only store integers < 2^53
    local maxVal = 2 ^ (maxBits - 1) -- assuming signed integers, which, for now, are the only ones in use
    assert(type(arg) == "number" and math.floor(arg) == arg and arg < maxVal and arg >= -maxVal or
               (maybeNil and arg == nil),
           string.format("Argument \"%s\" of function \"%s\" has to be a whole number <= %d",
                         argName, func, maxVal))
end

-- function wrappers
function discordRPC.initialize(applicationId, autoRegister, optionalSteamId)
    local func = "discordRPC.Initialize"
    checkStrArg(applicationId, nil, "applicationId", func)
    checkArg(autoRegister, "boolean", "autoRegister", func)
    if optionalSteamId ~= nil then
        checkStrArg(optionalSteamId, nil, "optionalSteamId", func)
    end

    local eventHandlers = ffi.new("struct DiscordEventHandlers")
    eventHandlers.ready = ready_proxy
    eventHandlers.disconnected = disconnected_proxy
    eventHandlers.errored = errored_proxy
    eventHandlers.joinGame = joinGame_proxy
    eventHandlers.spectateGame = spectateGame_proxy
    eventHandlers.joinRequest = joinRequest_proxy

    discordRPClib.Discord_Initialize(applicationId, eventHandlers, autoRegister and 1 or 0,
                                     optionalSteamId)
end

function discordRPC.shutdown()
    discordRPClib.Discord_Shutdown()
end

function discordRPC.runCallbacks()
    discordRPClib.Discord_RunCallbacks()
end
-- http://luajit.org/ext_ffi_semantics.html#callback :
-- It is not allowed, to let an FFI call into a C function (runCallbacks)
-- get JIT-compiled, which in turn calls a callback, calling into Lua again (e.g. discordRPC.ready).
-- Usually this attempt is caught by the interpreter first and the C function
-- is blacklisted for compilation.
-- solution:
-- "Then you'll need to manually turn off JIT-compilation with jit.off() for
-- the surrounding Lua function that invokes such a message polling function."
-- jit.off(discordRPC.runCallbacks)

function discordRPC.updatePresence(presence)
    local func = "discordRPC.updatePresence"
    checkArg(presence, "table", "presence", func)

    -- -1 for string length because of 0-termination
    checkStrArg(presence.state, 127, "presence.state", func, true)
    checkStrArg(presence.details, 127, "presence.details", func, true)

    checkIntArg(presence.startTimestamp, 64, "presence.startTimestamp", func, true)
    checkIntArg(presence.endTimestamp, 64, "presence.endTimestamp", func, true)

    checkStrArg(presence.largeImageKey, 31, "presence.largeImageKey", func, true)
    checkStrArg(presence.largeImageText, 127, "presence.largeImageText", func, true)
    checkStrArg(presence.smallImageKey, 31, "presence.smallImageKey", func, true)
    checkStrArg(presence.smallImageText, 127, "presence.smallImageText", func, true)
    checkStrArg(presence.partyId, 127, "presence.partyId", func, true)

    checkIntArg(presence.partySize, 32, "presence.partySize", func, true)
    checkIntArg(presence.partyMax, 32, "presence.partyMax", func, true)

    checkStrArg(presence.matchSecret, 127, "presence.matchSecret", func, true)
    checkStrArg(presence.joinSecret, 127, "presence.joinSecret", func, true)
    checkStrArg(presence.spectateSecret, 127, "presence.spectateSecret", func, true)

    checkIntArg(presence.instance, 8, "presence.instance", func, true)

    local cpresence = ffi.new("struct DiscordRichPresence")
    cpresence.state = presence.state
    cpresence.details = presence.details
    cpresence.startTimestamp = presence.startTimestamp or 0
    cpresence.endTimestamp = presence.endTimestamp or 0
    cpresence.largeImageKey = presence.largeImageKey
    cpresence.largeImageText = presence.largeImageText
    cpresence.smallImageKey = presence.smallImageKey
    cpresence.smallImageText = presence.smallImageText
    cpresence.partyId = presence.partyId
    cpresence.partySize = presence.partySize or 0
    cpresence.partyMax = presence.partyMax or 0
    cpresence.matchSecret = presence.matchSecret
    cpresence.joinSecret = presence.joinSecret
    cpresence.spectateSecret = presence.spectateSecret
    cpresence.instance = presence.instance or 0

    discordRPClib.Discord_UpdatePresence(cpresence)
end

function discordRPC.clearPresence()
    discordRPClib.Discord_ClearPresence()
end

local replyMap = {no = 0, yes = 1, ignore = 2}

-- maybe let reply take ints too (0, 1, 2) and add constants to the module
function discordRPC.respond(userId, reply)
    checkStrArg(userId, nil, "userId", "discordRPC.respond")
    assert(replyMap[reply],
           "Argument 'reply' to discordRPC.respond has to be one of \"yes\", \"no\" or \"ignore\"")
    discordRPClib.Discord_Respond(userId, replyMap[reply])
end

-- garbage collection callback
getmetatable(discordRPC.gcDummy).__gc = function()
    discordRPC.shutdown()
    ready_proxy:free()
    disconnected_proxy:free()
    errored_proxy:free()
    joinGame_proxy:free()
    spectateGame_proxy:free()
    joinRequest_proxy:free()
end

return discordRPC

end,

----------------------
-- Modules part end --
----------------------
        }
        if files[path] then
            return files[path]
        else
            return origin_seacher(path)
        end
    end
end
---------------------------------------------------------
---------------- Auto Bundled Code Block ----------------
---------------------------------------------------------
local components = require "insurrection.components"
local specialEvents = require "insurrection.specialEvents"
local balltze = Balltze
local engine = Engine
package.preload["luna"] = nil
package.loaded["luna"] = nil
local luna = require "luna"
local chimera = require "insurrection.mods.chimera"
local interface = require "insurrection.interface"
async = require"async".async
local dispatch = require"async".dispatch
require"async".configure("base, table, package, string")
execute_script = engine.hsc.executeScript

---@type Logger
logger = nil
DebugMode = false
APILocalMode = false
IsDebugCustomization = false
IsDebugLocalCustomizationEnabled = false
inspect = require "inspect"
constants = require "insurrection.constants"
api = require "insurrection.api"
discord = require "insurrection.discord"
store = require "insurrection.redux.store"
local isNewMap = true

---@type uiWidgetDefinition?
local editableWidget
---@type tag?
local editableWidgetTagEntry
---@type tag
local lastOpenWidgetTag
---@type tag
local lastClosedWidgetTag
---@type tag
local lastListFocusedWidgetTag
---@type tag?
local lastFocusedWidgetTagEntry
-- Multithread lanes
Lanes = {}

local customBipedPaths = {
    bleed_it_out = {[[keymind\the_flood\characters\unsc\gridharvolur\gridharvolur_mk_ii_[b]_mp]]},
    b30_coop_evolved = {
        [[[shm]\halo_1\characters\mjolnir_gen_1\mjolnir_gen_1_mp]],
        [[[shm]\halo_1\characters\marine\marine_mp]],
        [[[shm]\halo_1\characters\elite\elite_mp]],
        [[[shm]\halo_1\characters\grunt\grunt_mp]]
    }
    -- forge_island_dev = {[ze[[shm]\halo_4\characters\mjolnir_gen2\mjolnir_gen2_mp]]}
}
CustomBipedPaths = customBipedPaths

local customExternalTags = {
    {constants.path.nameplateCollection, engine.tag.classes.tagCollection},
    {constants.path.pauseMenu, engine.tag.classes.uiWidgetDefinition},
    {constants.path.dialog, engine.tag.classes.uiWidgetDefinition},
    {constants.path.customSounds, engine.tag.classes.tagCollection},
    {constants.path.christmasHat, engine.tag.classes.scenery},
    {constants.path.xmasObjects, engine.tag.classes.tagCollection}
}

function log(...)
    if DebugMode then
        local args = {...}
        for i = 1, #args do
            if type(args[i]) == "string" then
                args[i] = args[i]:replace("\n", " ")
            end
        end
        logger:debug(table.unpack(args))
    end
end

function PluginMetadata()
    return {
        name = "Insurrection",
        author = "Shadowmods Team",
        version = "2.6.0",
        targetApi = "1.0.0-rc.1",
        reloadable = true
    }
end

local function initialize()
    log("Initializing Insurrection")
    api.loadUrl()
    components.free()
    constants.get()
    interface.load()
    interface.changeAspectRatio()
    chimera.fontOverride()
end

local commands = {
    debug = {
        description = "Enable Insurrection debug mode",
        help = "<boolean>",
        minArgs = 1,
        maxArgs = 1,
        execute = function(enable)
            DebugMode = luna.bool(enable)
            engine.core.consolePrint("Debug mode: " .. tostring(DebugMode))
        end
    },
    setup_fonts = {
        description = "Setup Insurrection fonts as default game fonts",
        help = "<boolean>",
        minArgs = 1,
        maxArgs = 1,
        execute = function(enable)
            local revert = not luna.bool(enable)
            chimera.setupFonts(revert)
            if not revert then
                interface.dialog("SUCCESS", "Fonts have been setup",
                                 "Please restart the game to see changes.")
                return
            end
            interface.dialog("SUCCESS", "Fonts have been reverted",
                             "Please restart the game to see changes.")
        end
    },
    debug_customization = {
        description = "Enable Insurrection customization debug mode",
        execute = function()
            IsDebugCustomization = true
            interface.blur(false)
            interface.close(true)
            -- execute_script("set_customization_background 1")
            -- execute_script("object_create customization_biped")
        end
    },
    test = {
        description = "Test command",
        execute = function()
            interface.dialog("SUCCESS", "Test command", "This is a test command")
        end
    }
}

local isChimeraLoaded = false

local function loadChimeraCompatibility()
    -- Load Chimera compatibility
    for k, v in pairs(balltze.chimera) do
        if not k:includes "timer" and not k:includes "execute_script" and
            not k:includes "set_callback" then
            _G[k] = v
        end
    end
    server_type = engine.netgame.getServerType()

    -- Replace Chimera functions with Balltze functions
    write_bit = balltze.memory.writeBit
    write_byte = balltze.memory.writeInt8
    write_word = balltze.memory.writeInt16
    write_dword = balltze.memory.writeInt32
    write_int = balltze.memory.writeInt32
    write_float = balltze.memory.writeFloat
    write_string = function(address, value)
        for i = 1, #value do
            write_byte(address + i - 1, string.byte(value, i))
        end
        if #value == 0 then
            write_byte(address, 0)
        end
    end
    execute_script = engine.hsc.executeScript
end

local onMapLoadEvent
local onTickEvent

function PluginLoad()
    logger = balltze.logger.createLogger("Insurrection")
    logger:muteDebug(not DebugMode)

    local function importCustomizableBipeds()
        for mapName, bipeds in pairs(customBipedPaths) do
            for _, bipedPath in pairs(bipeds) do
                balltze.features.importTagFromMap(mapName, bipedPath, engine.tag.classes.biped)
            end
        end
    end

    importCustomizableBipeds()

    if not onMapLoadEvent then
        onMapLoadEvent = balltze.event.mapLoad.subscribe(function(event)
            if event.time == "before" then
                isNewMap = true
                if event.context:mapName() == "ui" then
                    logger:debug("Importing external customizable bipeds...")
                    importCustomizableBipeds()
                    -- elseif api.session.lobbyKey then
                else
                    balltze.features.clearTagImports()
                    for _, tagPath in pairs(customExternalTags) do
                        balltze.features.importTagFromMap("ui", tagPath[1], tagPath[2])
                    end
                end
            else
                balltze.features.clearTagImports()
            end
        end, "lowest")
    end

    if not onTickEvent then
        onTickEvent = balltze.event.tick.subscribe(function(event)
            if event.time == "before" then
                if not isChimeraLoaded and balltze.chimera then
                    logger:debug("Chimera compatibility loaded")
                    loadChimeraCompatibility()
                    isChimeraLoaded = true
                end
                if isChimeraLoaded then
                    if isNewMap then
                        logger:debug("New map loaded, initializing Insurrection data...")
                        initialize()
                        specialEvents.onPostMapLoad()
                        isNewMap = false
                    end
                    interface.onTick()
                    specialEvents.onTick()
                    -- Multithread callback resolve
                    local success, message = pcall(dispatch)
                    if not success then
                        logger:error(tostring(message))
                    end
                end
            end
        end)
    end

    for command, data in pairs(commands) do
        balltze.command.registerCommand(command, command, data.description, data.help, false,
                                        data.minArgs or 0, data.maxArgs or 0, false, true,
                                        function(...)
            local success, result = pcall(data.execute, table.unpack(...))
            if not success then
                logger:error("Error executing command '{}': {}", command, result)
                return false
            end
            return true
        end)
    end

    components.callbacks()

    return true
end

function PluginUnload()
end
